// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
#if NET_4_0
using System.Threading;
#endif

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Security.Authentication.Sasl;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Imap4.Client.Session {
  public static class ImapSessionCreator {
    public static ImapSession CreateSession(IImapSessionProfile profile,
                                            SaslClientMechanism authMechanismSpecified,
                                            UpgradeConnectionStreamCallback createSslStreamCallback)
    {
#if NET_4_0
      return CreateSession(profile,
                           authMechanismSpecified,
                           createSslStreamCallback,
                           CancellationToken.None);
    }

    public static ImapSession CreateSession(IImapSessionProfile profile,
                                            SaslClientMechanism authMechanismSpecified,
                                            UpgradeConnectionStreamCallback createSslStreamCallback,
                                            CancellationToken cancellationToken)
    {
#endif
      ImapSession session;

#if NET_4_0
      var result = CreateSession(profile,
                                 authMechanismSpecified,
                                 createSslStreamCallback,
                                 cancellationToken,
                                 out session);
#else
      var result = CreateSession(profile,
                                 authMechanismSpecified,
                                 createSslStreamCallback,
                                 out session);
#endif

      if (result.Succeeded)
        return session;
      else
        throw new ImapAuthenticationException(result);
    }

    public static ImapCommandResult CreateSession(IImapSessionProfile profile,
                                                  SaslClientMechanism authMechanismSpecified,
                                                  UpgradeConnectionStreamCallback createSslStreamCallback,
                                                  out ImapSession session)
    {
#if NET_4_0
      return CreateSession(profile,
                           authMechanismSpecified,
                           createSslStreamCallback,
                           CancellationToken.None,
                           out session);
    }

    public static ImapCommandResult CreateSession(IImapSessionProfile profile,
                                                  SaslClientMechanism authMechanismSpecified,
                                                  UpgradeConnectionStreamCallback createSslStreamCallback,
                                                  CancellationToken cancellationToken,
                                                  out ImapSession session)
    {
      cancellationToken.ThrowIfCancellationRequested();
#endif

      if (profile == null)
        throw new ArgumentNullException("profile");

      var authority = profile.Authority;
      var securePort = string.Equals(authority.Scheme, ImapUri.UriSchemeImaps, StringComparison.OrdinalIgnoreCase);

      if (securePort && createSslStreamCallback == null)
        throw new ArgumentNullException("createSslStreamCallback");

      session = null;

      ImapSession s = null;
      ImapCommandResult authenticationResult = null;

      try {
        s = new ImapSession(authority.Host,
                            authority.Port,
                            profile.Timeout,
                            profile.SendTimeout,
                            profile.ReceiveTimeout,
                            true,
                            securePort
                              ? createSslStreamCallback
                              : null);

        s.HandlesIncapableAsException = false;
        s.HandlesReferralAsException = false;

        if (s.ServerCapabilities.Count == 0) {
#if NET_4_0
          // cancel before querying server capability
          cancellationToken.ThrowIfCancellationRequested();
#endif

          // try querying server capability (ignore error)
          s.Capability();
        }

        if (!s.ServerCapabilities.Contains(ImapCapability.Imap4Rev1))
          throw new ImapIncapableException(ImapCapability.Imap4Rev1);

        if (profile.UseTlsIfAvailable && s.ServerCapabilities.Contains(ImapCapability.StartTls) && !s.IsSecureConnection) {
#if NET_4_0
          // cancel before upgrading connection
          cancellationToken.ThrowIfCancellationRequested();
#endif

          var r = s.StartTls(createSslStreamCallback, true);

          if (r.Failed)
            throw new ImapSecureConnectionException(r.ResultText);
          else if (!s.ServerCapabilities.Contains(ImapCapability.Imap4Rev1))
            throw new ImapIncapableException(ImapCapability.Imap4Rev1);
        }

        if (authMechanismSpecified == null) {
#if NET_4_0
          authenticationResult = Authenticate(s, profile, cancellationToken);
#else
          authenticationResult = Authenticate(s, profile);
#endif
        }
        else {
#if NET_4_0
          // cancel before authentication
          cancellationToken.ThrowIfCancellationRequested();
#endif

          authenticationResult = s.Authenticate(authMechanismSpecified,
                                                true
#if NET_4_0
                                                , cancellationToken
#endif
                                               );
        }

        if (authenticationResult == null)
          throw new ImapNoAppropriateAuthMechanismException();
      }
      finally {
        try {
          if (s != null && (authenticationResult == null || authenticationResult.Failed))
            s.Disconnect(false);
        }
        catch (ImapConnectionException) {
          // ignore
        }
      }

      try {
        if (profile.UseDeflateIfAvailable && s.ServerCapabilities.Contains(ImapCapability.CompressDeflate))
          // try compress (ignore error)
          s.Compress(ImapCompressionMechanism.Deflate);
      }
      catch {
        try {
          s.Disconnect(false);
        }
        catch (ImapConnectionException) {
          // ignore
        }

        throw;
      }

      session = s;

      return authenticationResult;
    }

    private static ImapCommandResult Authenticate(ImapSession session,
                                                  IImapSessionProfile profile
#if NET_4_0
                                                  , CancellationToken cancellationToken
#endif
                                                 )
    {
      var authority = profile.Authority;
      var username = ImapStyleUriParser.GetUser(authority);
      var authMechanism = ImapStyleUriParser.GetAuthType(authority);

      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    An authentication mechanism (as used by the IMAP AUTHENTICATE
       *    command) can be expressed by adding ";AUTH=<enc-auth-type>" to the
       *    end of the user name in an IMAP URL.  When such an <enc-auth-type> is
       *    indicated, the client SHOULD request appropriate credentials from
       *    that mechanism and use the "AUTHENTICATE" command instead of the
       *    "LOGIN" command.  If no user name is specified, one MUST be obtained
       *    from the mechanism or requested from the user/configuration as
       *    appropriate.
       *
       *    If a user name is included with no authentication mechanism, then
       *    ";AUTH=*" is assumed.
       */

      // TODO: URLAUTH
      /*
      if (!session.ServerCapabilities.Has(ImapCapability.UrlAuth)
        throw new ImapWebException("URLAUTH incapable", WebExceptionStatus.ProtocolError);
      */
      var canFallback = false;

      if (authMechanism == null) {
        if (string.IsNullOrEmpty(username)) {
          authMechanism = ImapAuthenticationMechanism.Anonymous;
          canFallback = true;
        }
        else {
          authMechanism = ImapAuthenticationMechanism.SelectAppropriate;
        }
      }

      if (authMechanism == ImapAuthenticationMechanism.SelectAppropriate) {
        var allowUsingPlainTextMechanism = session.IsSecureConnection || profile.AllowInsecureLogin;

        return AuthenticateWithAppropriateMechanism(session,
                                                    allowUsingPlainTextMechanism,
                                                    profile.Credentials,
                                                    username,
                                                    GetUsingSaslMechanisms(profile.UsingSaslMechanisms ?? EmptyArray<string>.Instance)
#if NET_4_0
                                                    , cancellationToken
#endif
                                                   );
      }
      else if (authMechanism == ImapAuthenticationMechanism.Anonymous) {
        return AuthenticateAsAnonymous(session,
                                       username,
                                       canFallback
#if NET_4_0
                                       , cancellationToken
#endif
                                      );
      }
      else {
        return AuthenticateWithSuppliedMechanism(session,
                                                 profile.Credentials,
                                                 username,
                                                 authMechanism
#if NET_4_0
                                                 , cancellationToken
#endif
                                                );
      }
    }

    private static IEnumerable<string> GetUsingSaslMechanisms(string[] usingSaslMechanisms)
    {
      var availableMechanisms = SaslClientMechanism.GetAvailableMechanisms();

      foreach (var mechanism in usingSaslMechanisms) {
        if (availableMechanisms.Any(delegate(string val) {
          return string.Equals(val, mechanism, StringComparison.OrdinalIgnoreCase);
        }))
          yield return mechanism;
      }
    }

    private static ImapCommandResult AuthenticateWithAppropriateMechanism(ImapSession session,
                                                                          bool allowUsingPlainTextMechanism,
                                                                          ICredentialsByHost credentials,
                                                                          string username,
                                                                          IEnumerable<string> usingSaslMechanisms
#if NET_4_0
                                                                          , CancellationToken cancellationToken
#endif
                                                                         )
    {
      ImapCommandResult result = null;

      foreach (var mechanism in usingSaslMechanisms) {
        if (!allowUsingPlainTextMechanism && SaslClientMechanism.IsMechanismPlainText(mechanism))
          // disallow plain text mechanism
          continue;

        if (string.Equals(mechanism, SaslMechanisms.Anonymous, StringComparison.OrdinalIgnoreCase))
          // disallow 'ANONYMOUS' mechanism
          continue;

        var authMechanism = ImapAuthenticationMechanism.GetKnownOrCreate(mechanism);

        if (session.ServerCapabilities.IsCapable(authMechanism)) {
#if NET_4_0
          cancellationToken.ThrowIfCancellationRequested();
#endif

          result = session.Authenticate(credentials,
                                        username,
                                        authMechanism,
                                        true
#if NET_4_0
                                        , cancellationToken
#endif
                                       );

          if (result.Succeeded)
            break;
        }
      }

      if ((result == null || result.Failed) &&
          allowUsingPlainTextMechanism &&
          !session.ServerCapabilities.Contains(ImapCapability.LoginDisabled)) {
#if NET_4_0
        cancellationToken.ThrowIfCancellationRequested();
#endif

        result = session.Login(credentials, username, true);
      }

      return result;
    }

    private static ImapCommandResult AuthenticateWithSuppliedMechanism(ImapSession session,
                                                                       ICredentialsByHost credentials,
                                                                       string username,
                                                                       ImapAuthenticationMechanism authMechanism
#if NET_4_0
                                                                       , CancellationToken cancellationToken
#endif
                                                                      )
    {
      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    An authentication mechanism (as used by the IMAP AUTHENTICATE
       *    command) can be expressed by adding ";AUTH=<enc-auth-type>" to the
       *    end of the user name in an IMAP URL.  When such an <enc-auth-type> is
       *    indicated, the client SHOULD request appropriate credentials from
       *    that mechanism and use the "AUTHENTICATE" command instead of the
       *    "LOGIN" command.  If no user name is specified, one MUST be obtained
       *    from the mechanism or requested from the user/configuration as
       *    appropriate.
       */
#if NET_4_0
      cancellationToken.ThrowIfCancellationRequested();
#endif

      return session.Authenticate(credentials,
                                  username,
                                  authMechanism,
                                  true
#if NET_4_0
                                  , cancellationToken
#endif
                                 );
    }

    private static ImapCommandResult AuthenticateAsAnonymous(ImapSession session,
                                                             string username,
                                                             bool canFallback
#if NET_4_0
                                                             , CancellationToken cancellationToken
#endif
                                                            )
    {
      /*
       * http://tools.ietf.org/html/rfc5092
       * 3.2. IMAP User Name and Authentication Mechanism
       * 
       *    If no user name and no authentication mechanism are supplied, the
       *    client MUST authenticate as anonymous to the server.  If the server
       *    advertises AUTH=ANONYMOUS IMAP capability, the client MUST use the
       *    AUTHENTICATE command with ANONYMOUS [ANONYMOUS] SASL mechanism.  If
       *    SASL ANONYMOUS is not available, the (case-insensitive) user name
       *    "anonymous" is used with the "LOGIN" command and the Internet email
       *    address of the end user accessing the resource is supplied as the
       *    password.  The latter option is given in order to provide for
       *    interoperability with deployed servers.
       * 
       *    Note that, as described in RFC 3501, the "LOGIN" command MUST NOT be
       *    used when the IMAP server advertises the LOGINDISABLED capability.
       */
      ImapCommandResult result = null;

      if (string.IsNullOrEmpty(username))
        username = "anonymous@";

      if (session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Anonymous)) {
#if NET_4_0
        cancellationToken.ThrowIfCancellationRequested();
#endif

        // try AUTHENTICATE ANONYUMOUS
        result = session.Authenticate(new NetworkCredential(username, string.Empty),
                                      null,
                                      ImapAuthenticationMechanism.Anonymous,
                                      true
#if NET_4_0
                                      , cancellationToken
#endif
                                      );
      }

      if ((result == null || (result.Failed && canFallback)) &&
          !session.ServerCapabilities.Contains(ImapCapability.LoginDisabled)) {
#if NET_4_0
        cancellationToken.ThrowIfCancellationRequested();
#endif

        // try anonymous LOGIN
        result = session.Login(new NetworkCredential("anonymous", username),
                               null,
                               true);
      }

      return result;
    }
  }
}
