// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Collections.Generic;
using System.Linq;
#if ASYNC_OPS
using System.Threading;
using System.Threading.Tasks;
#endif

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  partial class ImapClient {
    internal static readonly bool DefaultSubscription = false;
    internal static readonly bool DefaultSelectReadOnly = false;

    public ImapOpenedMailboxInfo OpenedMailbox {
      get { ThrowIfNotConnected(); return openedMailbox; }
    }

    public bool IsIdling {
      get {
        ThrowIfNotConnected();

        if (openedMailbox == null)
          return false;
        else
          return openedMailbox.IsIdling;
      }
    }

    public ImapCapabilitySet DisabledCapabilities {
      get { return disabledCapabilities; }
    }

    /*
     * LIST/LSUB
     */
    public ImapMailboxInfo GetInbox()
    {
      return GetMailbox(ImapMailbox.Inbox,
                        ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetInbox(ImapMailboxListOptions options)
    {
      return GetMailbox(ImapMailbox.Inbox,
                        options);
    }

#if ASYNC_OPS
    public Task<ImapMailboxInfo> GetInboxAsync()
    {
      return GetMailboxAsync(ImapMailbox.Inbox,
                             ImapMailboxListOptions.Default);
    }

    public Task<ImapMailboxInfo> GetInboxAsync(ImapMailboxListOptions options)
    {
      return GetMailboxAsync(ImapMailbox.Inbox,
                             options);
    }
#endif

    public ImapMailboxInfo GetMailbox(string mailboxName)
    {
      return GetMailbox(mailboxName,
                        ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetMailbox(string mailboxName,
                                      ImapMailboxListOptions options)
    {
      mailboxName = GetValidatedMailboxName(mailboxName, "mailboxName");

      return RunExclusively(() => GetMailboxCore(mailboxName, options, true, false));
    }

#if ASYNC_OPS
    public Task<ImapMailboxInfo> GetMailboxAsync(string mailboxName)
    {
      return GetMailboxAsync(mailboxName,
                             ImapMailboxListOptions.Default);
    }

    public Task<ImapMailboxInfo> GetMailboxAsync(string mailboxName,
                                                 ImapMailboxListOptions options)
    {
      mailboxName = GetValidatedMailboxName(mailboxName, "mailboxName");

      return StartTaskExclusively(() => GetMailboxCore(mailboxName, options, true, false),
                                  CancellationToken.None);
    }
#endif

    /// <param name="mailboxName">must be validated by caller</param>
    private ImapMailboxInfo GetMailboxCore(string mailboxName,
                                           ImapMailboxListOptions options,
                                           bool throwIfNotFound,
                                           bool throwIfNonExistentOrUnselectable)
    {
      var mailbox = GetMailboxesCore(mailboxName,
                                     false,
                                     options,
                                     null,
                                     ListMethod.SelectAppropriate,
                                     (m) => ImapMailbox.NameEquals(m.Name, mailboxName)).FirstOrDefault();

      if (mailbox == null && throwIfNotFound)
        throw new ImapMailboxNotFoundException(mailboxName);

      if (mailbox != null && throwIfNonExistentOrUnselectable)
        mailbox.ThrowIfNonExistentOrUnselectable();

      return mailbox;
    }

    public ImapMailboxInfo GetMailbox(ImapSpecialMailbox specialMailbox)
    {
      return GetMailbox(specialMailbox,
                        ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetMailbox(ImapSpecialMailbox specialMailbox,
                                      ImapMailboxListOptions options)
    {
      return RunExclusively(() => GetMailboxCore(specialMailbox, options, true, false));
    }

#if ASYNC_OPS
    public Task<ImapMailboxInfo> GetMailboxAsync(ImapSpecialMailbox specialMailbox)
    {
      return GetMailboxAsync(specialMailbox,
                             ImapMailboxListOptions.Default);
    }

    public Task<ImapMailboxInfo> GetMailboxAsync(ImapSpecialMailbox specialMailbox,
                                                 ImapMailboxListOptions options)
    {
      return StartTaskExclusively(() => GetMailboxCore(specialMailbox, options, true, false),
                                  CancellationToken.None);
    }
#endif

    private ImapMailboxInfo GetMailboxCore(ImapSpecialMailbox specialMailbox,
                                           ImapMailboxListOptions options,
                                           bool throwIfNotFound,
                                           bool throwIfNonExistentOrUnselectable)
    {
      var specialUseAvailable =
        IsCapabilityAvailable(ImapCapability.SpecialUse) ||
        IsCapabilityAvailable(ImapCapability.GimapXlist); // XXX: Gmail supports SPECIAL-USE but don't list it

      if (!specialUseAvailable)
        throw new ImapIncapableException(ImapCapability.SpecialUse);

      var expectedUseFlag = TranslateSpecialMailboxUseFlag(specialMailbox);
      var selectionOptions = specialUseAvailable ? ImapListSelectionOptions.SpecialUse : null;
      var listMethod = IsCapabilityAvailable(ImapCapability.ListExtended) ? ListMethod.ListExtended : ListMethod.ListOrListRemote;

      var mailbox = GetMailboxesCore(string.Empty,
                                     true,
                                     options,
                                     selectionOptions,
                                     listMethod,
                                     (m) => m.Flags.Contains(expectedUseFlag)).FirstOrDefault();

      if (mailbox == null && throwIfNotFound)
        throw new ImapMailboxNotFoundException(expectedUseFlag.ToString());

      if (mailbox != null && throwIfNonExistentOrUnselectable)
        mailbox.ThrowIfNonExistentOrUnselectable();

      return mailbox;
    }

    private static ImapMailboxFlag TranslateSpecialMailboxUseFlag(ImapSpecialMailbox specialMailbox)
    {
      switch (specialMailbox) {
        case ImapSpecialMailbox.All:      return ImapMailboxFlag.All;
        case ImapSpecialMailbox.Archive:  return ImapMailboxFlag.Archive;
        case ImapSpecialMailbox.Drafts:   return ImapMailboxFlag.Drafts;
        case ImapSpecialMailbox.Flagged:  return ImapMailboxFlag.Flagged;
        case ImapSpecialMailbox.Junk:     return ImapMailboxFlag.Junk;
        case ImapSpecialMailbox.Sent:     return ImapMailboxFlag.Sent;
        case ImapSpecialMailbox.Trash:    return ImapMailboxFlag.Trash;

        default:
          throw ExceptionUtils.CreateArgumentMustBeValidEnumValue("specialMailbox", specialMailbox);
      }
    }

    public ImapMailboxInfo GetOrCreateMailbox(string mailboxName)
    {
      return GetOrCreateMailbox(mailboxName,
                                DefaultSubscription,
                                ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetOrCreateMailbox(string mailboxName,
                                              ImapMailboxListOptions options)
    {
      return GetOrCreateMailbox(mailboxName,
                                DefaultSubscription,
                                options);
    }

    public ImapMailboxInfo GetOrCreateMailbox(string mailboxName,
                                              bool subscribeIfCreated)
    {
      return GetOrCreateMailbox(mailboxName,
                                subscribeIfCreated,
                                ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetOrCreateMailbox(string mailboxName,
                                              bool subscribeIfCreated,
                                              ImapMailboxListOptions options)
    {
      mailboxName = GetValidatedMailboxName(mailboxName, "mailboxName");

      return RunExclusively(() => GetOrCreateMailboxCore(mailboxName, subscribeIfCreated, options));
    }

#if ASYNC_OPS
    public Task<ImapMailboxInfo> GetOrCreateMailboxAsync(string mailboxName)
    {
      return GetOrCreateMailboxAsync(mailboxName,
                                     DefaultSubscription,
                                     ImapMailboxListOptions.Default);
    }

    public Task<ImapMailboxInfo> GetOrCreateMailboxAsync(string mailboxName,
                                                         ImapMailboxListOptions options)
    {
      return GetOrCreateMailboxAsync(mailboxName,
                                     DefaultSubscription,
                                     options);
    }

    public Task<ImapMailboxInfo> GetOrCreateMailboxAsync(string mailboxName,
                                                         bool subscribeIfCreated)
    {
      return GetOrCreateMailboxAsync(mailboxName,
                                     subscribeIfCreated,
                                     ImapMailboxListOptions.Default);
    }

    public Task<ImapMailboxInfo> GetOrCreateMailboxAsync(string mailboxName,
                                                         bool subscribeIfCreated,
                                                         ImapMailboxListOptions options)
    {
      mailboxName = GetValidatedMailboxName(mailboxName, "mailboxName");

      return StartTaskExclusively(() => GetOrCreateMailboxCore(mailboxName, subscribeIfCreated, options),
                                  CancellationToken.None);
    }
#endif

    private ImapMailboxInfo GetOrCreateMailboxCore(string mailboxName,
                                                   bool subscribeIfCreated,
                                                   ImapMailboxListOptions options)
    {
      var mailbox = GetMailboxCore(mailboxName, options, false, false);

      if (mailbox == null)
        return CreateMailboxCore(mailboxName, subscribeIfCreated);
      else
        return mailbox;
    }

#if IMAP_CREATE_SPECAIL_MAILBOX
    public ImapMailboxInfo GetOrCreateMailbox(ImapSpecialMailbox mailbox)
    {
      return GetOrCreateMailbox(mailbox,
                                DefaultSubscription,
                                ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetOrCreateMailbox(ImapSpecialMailbox mailbox,
                                              ImapMailboxListOptions options)
    {
      return GetOrCreateMailbox(mailbox,
                                DefaultSubscription,
                                options);
    }

    public ImapMailboxInfo GetOrCreateMailbox(ImapSpecialMailbox mailbox,
                                              bool subscribeIfCreated)
    {
      return GetOrCreateMailbox(mailbox,
                                subscribeIfCreated,
                                ImapMailboxListOptions.Default);
    }

    public ImapMailboxInfo GetOrCreateMailbox(ImapSpecialMailbox mailbox,
                                              bool subscribeIfCreated,
                                              ImapMailboxListOptions options)
    {
      var m = GetMailboxNoException(mailbox, options);

      if (m == null)
        return CreateMailbox(mailbox, subscribeIfCreated);
      else
        return m;
    }
#endif

#if NET_4_5
    public IReadOnlyList<ImapMailboxInfo> GetMailboxes()
#else
    public IList<ImapMailboxInfo> GetMailboxes()
#endif
    {
      return GetMailboxes(ImapMailboxListOptions.Default);
    }

#if NET_4_5
    public IReadOnlyList<ImapMailboxInfo> GetMailboxes(ImapMailboxListOptions options)
#else
    public IList<ImapMailboxInfo> GetMailboxes(ImapMailboxListOptions options)
#endif
    {
      return RunExclusively(() => {
        return ToReadOnlyList(GetMailboxesCore(string.Empty,
                                               true,
                                               options,
                                               null,
                                               ListMethod.SelectAppropriate,
                                               null));
      });
    }

#if ASYNC_OPS
#if NET_4_5
    public Task<IReadOnlyList<ImapMailboxInfo>> GetMailboxesAsync()
#else
    public Task<IList<ImapMailboxInfo>> GetMailboxesAsync()
#endif
    {
      return GetMailboxesAsync(ImapMailboxListOptions.Default);
    }

#if NET_4_5
    public Task<IReadOnlyList<ImapMailboxInfo>> GetMailboxesAsync(ImapMailboxListOptions options)
#else
    public Task<IList<ImapMailboxInfo>> GetMailboxesAsync(ImapMailboxListOptions options)
#endif
    {
      return StartTaskExclusively(() => {
        return ToReadOnlyList(GetMailboxesCore(string.Empty,
                                               true,
                                               options,
                                               null,
                                               ListMethod.SelectAppropriate,
                                               null));
      }, CancellationToken.None);
    }
#endif

#if NET_4_5
    internal IReadOnlyList<ImapMailboxInfo> GetMailboxesCore
#else
    internal IList<ImapMailboxInfo> GetMailboxesCore
#endif
    (
      ImapMailbox mailbox,
      ImapMailboxListOptions options
    )
    {
      if (mailbox.Flags.Contains(ImapMailboxFlag.NoInferiors))
        // non hierarchical (ignore \HasNoChildren)
        return EmptyArray<ImapMailboxInfo>.Instance;
      else if (string.IsNullOrEmpty(mailbox.HierarchyDelimiter))
        // non hierarchical
        return EmptyArray<ImapMailboxInfo>.Instance;
      else
        return ToReadOnlyList(GetMailboxesCore(mailbox.Name + mailbox.HierarchyDelimiter,
                                               true,
                                               options,
                                               null,
                                               ListMethod.SelectAppropriate,
                                               null));
    }

#if NET_4_5
    private static IReadOnlyList<ImapMailboxInfo> ToReadOnlyList(IEnumerable<ImapMailboxInfo> mailboxes)
#else
    private static IList<ImapMailboxInfo> ToReadOnlyList(IEnumerable<ImapMailboxInfo> mailboxes)
#endif
    {
#if NET_4_5
      return mailboxes.ToList();
#else
      return mailboxes.ToList().AsReadOnly();
#endif
    }

    private enum ListMethod {
      SelectAppropriate,
      ListOrListRemote,
      ListExtended,
      ListRemote,
      List,
    }

    private IEnumerable<ImapMailboxInfo> GetMailboxesCore(string mailboxName,
                                                          bool patternMatching,
                                                          ImapMailboxListOptions options,
                                                          ImapListSelectionOptions selectionOptions,
                                                          ListMethod listMethod,
                                                          Predicate<ImapMailbox> match)
    {
      string referenceName, mailboxNamePattern;

      if (patternMatching) {
        referenceName = mailboxName;
        mailboxNamePattern = ((int)(options & ImapMailboxListOptions.TopLevelOnly) != 0)
          ? "%"
          : "*";
      }
      else {
        referenceName = string.Empty;
        mailboxNamePattern = mailboxName;
      }

      bool subscribedOnly, listRemote, requestStatus;

      TranslateListOptions(options, out subscribedOnly, out listRemote, out requestStatus);

      ImapMailbox[] mailboxes;

      switch (listMethod) {
        case ListMethod.SelectAppropriate:
          if (IsCapabilityAvailable(ImapCapability.ListExtended))
            goto case ListMethod.ListExtended;
          else if (listRemote && IsCapabilityAvailable(ImapCapability.MailboxReferrals))
            goto case ListMethod.ListRemote;
          else
            goto case ListMethod.List;

        case ListMethod.ListOrListRemote:
          if (listRemote)
            goto case ListMethod.ListRemote;
          else
            goto case ListMethod.List;

        case ListMethod.ListExtended: {
          var returnOptions = ImapListReturnOptions.Children;

          if (selectionOptions == null)
            selectionOptions = ImapListSelectionOptions.Empty;

          if (subscribedOnly)
            selectionOptions += ImapListSelectionOptions.Subscribed;

          if (listRemote)
            selectionOptions += ImapListSelectionOptions.Remote;

          if (requestStatus && IsCapabilityAvailable(ImapCapability.ListStatus)) {
            returnOptions += ImapListReturnOptions.StatusDataItems(GetAvailableStatusDataItem());
            requestStatus = false; // no need to request STATUS
          }

          ProcessResult(Session.ListExtended(referenceName,
                                             new[] {mailboxNamePattern},
                                             selectionOptions,
                                             returnOptions,
                                             out mailboxes));

          break;
        }

        case ListMethod.ListRemote: {
          if (subscribedOnly)
            ProcessResult(Session.RLsub(referenceName,
                                        mailboxNamePattern,
                                        out mailboxes));
          else
            ProcessResult(Session.RList(referenceName,
                                        mailboxNamePattern,
                                        out mailboxes));

          break;
        }

        case ListMethod.List: {
          if (subscribedOnly)
            ProcessResult(Session.Lsub(referenceName,
                                       mailboxNamePattern,
                                       out mailboxes));
          else
            ProcessResult(Session.List(referenceName,
                                       mailboxNamePattern,
                                       out mailboxes));

          break;
        }

        default:
          throw ExceptionUtils.CreateArgumentMustBeValidEnumValue("listMethod", listMethod);
      }

      var selectedMailboxName = openedMailbox == null ? null : openedMailbox.FullName;

      foreach (var m in mailboxes) {
        if (match != null && !match(m))
          continue;

        if (requestStatus && !m.IsUnselectable && !m.NameEquals(selectedMailboxName))
          ProcessResult(Session.Status(m, GetAvailableStatusDataItem()));

        yield return new ImapMailboxInfo(this, m);
      }
    }

    private static void TranslateListOptions(ImapMailboxListOptions options,
                                             out bool subscribedOnly,
                                             out bool listRemote,
                                             out bool requestStatus)
    {
      subscribedOnly  = 0 != (int)(options & ImapMailboxListOptions.SubscribedOnly);
      listRemote      = 0 != (int)(options & ImapMailboxListOptions.Remote);
      requestStatus   = 0 != (int)(options & ImapMailboxListOptions.RequestStatus);
    }

    /*
     * CREATE
     */
    public ImapMailboxInfo CreateMailbox(string mailboxName)
    {
      return CreateMailbox(mailboxName, DefaultSubscription);
    }

    public ImapMailboxInfo CreateMailbox(string mailboxName,
                                         bool subscribeIfCreated)
    {
      mailboxName = GetValidatedMailboxName(mailboxName, "mailboxName");

      return RunExclusively(() => CreateMailboxCore(mailboxName, subscribeIfCreated));
    }

#if ASYNC_OPS
    public Task<ImapMailboxInfo> CreateMailboxAsync(string mailboxName)
    {
      return CreateMailboxAsync(mailboxName, DefaultSubscription);
    }

    public Task<ImapMailboxInfo> CreateMailboxAsync(string mailboxName,
                                                    bool subscribeIfCreated)
    {
      mailboxName = GetValidatedMailboxName(mailboxName, "mailboxName");

      return StartTaskExclusively(() => CreateMailboxCore(mailboxName, subscribeIfCreated),
                                  CancellationToken.None);
    }
#endif

    private ImapMailboxInfo CreateMailboxCore(string mailboxName,
                                              bool subscribeIfCreated)
    {
      ImapMailbox createdMailbox;

      ProcessResult(Session.Create(mailboxName, out createdMailbox));

      if (subscribeIfCreated)
        ProcessResult(Session.Subscribe(createdMailbox));

      return CreateFlagsRetrieved(createdMailbox);
    }

    internal ImapMailboxInfo CreateFlagsRetrieved(ImapMailbox mailbox)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      // retrieve mailbox flags
      ProcessResult(Session.List(mailbox));

      return new ImapMailboxInfo(this, mailbox);
    }

#if IMAP_CREATE_SPECAIL_MAILBOX
    public ImapMailboxInfo CreateMailbox(ImapSpecialMailbox mailbox)
    {
      return CreateMailbox(mailbox, DefaultSubscription);
    }

    public ImapMailboxInfo CreateMailbox(ImapSpecialMailbox mailbox,
                                         bool subscribeIfCreated)
    {
      throw new NotImplementedException();
    }
#endif

    /*
     * SELECT/EXAMINE
     */
    public ImapOpenedMailboxInfo OpenInbox()
    {
      return OpenMailbox(ImapMailbox.Inbox,
                         DefaultSelectReadOnly);
    }

    public ImapOpenedMailboxInfo OpenInbox(bool asReadOnly)
    {
      return OpenMailbox(ImapMailbox.Inbox,
                         asReadOnly);
    }

#if ASYNC_OPS
    public Task<ImapOpenedMailboxInfo> OpenInboxAsync()
    {
      return OpenMailboxAsync(ImapMailbox.Inbox,
                              DefaultSelectReadOnly);
    }

    public Task<ImapOpenedMailboxInfo> OpenInboxAsync(bool asReadOnly)
    {
      return OpenMailboxAsync(ImapMailbox.Inbox,
                              asReadOnly);
    }
#endif

    public ImapOpenedMailboxInfo OpenMailbox(ImapMailboxInfo mailbox)
    {
      return OpenMailbox(mailbox,
                         DefaultSelectReadOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(ImapMailboxInfo mailbox,
                                             bool asReadOnly)
    {
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");
      if (mailbox.Client != this)
        throw new ArgumentException("can't open mailbox of different session");

      mailbox.ThrowIfNonExistentOrUnselectable();

      return RunExclusively(() => OpenMailboxCore(mailbox, asReadOnly));
    }

#if ASYNC_OPS
    public Task<ImapOpenedMailboxInfo> OpenMailboxAsync()
    {
      return OpenMailboxAsync(ImapMailbox.Inbox,
                              DefaultSelectReadOnly);
    }

    public Task<ImapOpenedMailboxInfo> OpenMailboxAsync(bool asReadOnly)
    {
      return OpenMailboxAsync(ImapMailbox.Inbox,
                              asReadOnly);
    }

    public Task<ImapOpenedMailboxInfo> OpenMailboxAsync(ImapMailboxInfo mailbox)
    {
      return OpenMailboxAsync(mailbox,
                              DefaultSelectReadOnly);
    }

    public Task<ImapOpenedMailboxInfo> OpenMailboxAsync(ImapMailboxInfo mailbox,
                                             bool asReadOnly)
    {
      if (mailbox == null)
        throw new ArgumentNullException("mailbox");
      if (mailbox.Client != this)
        throw new ArgumentException("can't open mailbox of different session");

      mailbox.ThrowIfNonExistentOrUnselectable();

      return StartTaskExclusively(() => OpenMailboxCore(mailbox, asReadOnly),
                                  CancellationToken.None);
    }
#endif

    /// <param name="mailbox">mailbox must exist and be selectable. must be checked by caller.</param>
    private ImapOpenedMailboxInfo OpenMailboxCore(ImapMailboxInfo mailbox,
                                                  bool asReadOnly)
    {
      if (IsCapabilityAvailable(ImapCapability.CondStore))
        ProcessResult(asReadOnly
                      ? Session.ExamineCondStore(mailbox.FullName)
                      : Session.SelectCondStore(mailbox.FullName));
      else
        ProcessResult(asReadOnly
                      ? Session.Examine(mailbox.FullName)
                      : Session.Select(mailbox.FullName));

      return SetOpenedMailbox(new ImapOpenedMailboxInfo(this, Session.SelectedMailbox));
    }

    public ImapOpenedMailboxInfo OpenMailbox(string mailboxName)
    {
      return OpenMailbox(mailboxName,
                         ImapMailboxListOptions.Default,
                         DefaultSelectReadOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(string mailboxName,
                                             bool asReadOnly)
    {
      return OpenMailbox(mailboxName,
                         ImapMailboxListOptions.Default,
                         asReadOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(string mailboxName,
                                             ImapMailboxListOptions options)
    {
      return OpenMailbox(mailboxName,
                         options,
                         DefaultSelectReadOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(string mailboxName,
                                             ImapMailboxListOptions options,
                                             bool asReadOnly)
    {
      mailboxName = GetValidatedMailboxName(mailboxName, "mailboxName");

      return RunExclusively(() => {
        return OpenMailboxCore(GetMailboxCore(mailboxName, options, true, true),
                               asReadOnly);
      });
    }

#if ASYNC_OPS
    public Task<ImapOpenedMailboxInfo> OpenMailboxAsync(string mailboxName)
    {
      return OpenMailboxAsync(mailboxName,
                              ImapMailboxListOptions.Default,
                              DefaultSelectReadOnly);
    }

    public Task<ImapOpenedMailboxInfo> OpenMailboxAsync(string mailboxName,
                                                        bool asReadOnly)
    {
      return OpenMailboxAsync(mailboxName,
                              ImapMailboxListOptions.Default,
                              asReadOnly);
    }

    public Task<ImapOpenedMailboxInfo> OpenMailboxAsync(string mailboxName,
                                                        ImapMailboxListOptions options)
    {
      return OpenMailboxAsync(mailboxName,
                              options,
                              DefaultSelectReadOnly);
    }

    public Task<ImapOpenedMailboxInfo> OpenMailboxAsync(string mailboxName,
                                                        ImapMailboxListOptions options,
                                                        bool asReadOnly)
    {
      mailboxName = GetValidatedMailboxName(mailboxName, "mailboxName");

      return StartTaskExclusively(() => {
        return OpenMailboxCore(GetMailboxCore(mailboxName, options, true, true),
                               asReadOnly);
      }, CancellationToken.None);
    }
#endif

    public ImapOpenedMailboxInfo OpenMailbox(ImapSpecialMailbox specialMailbox)
    {
      return OpenMailbox(specialMailbox,
                         ImapMailboxListOptions.Default,
                         DefaultSelectReadOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(ImapSpecialMailbox specialMailbox,
                                             bool asReadOnly)
    {
      return OpenMailbox(specialMailbox,
                         ImapMailboxListOptions.Default,
                         asReadOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(ImapSpecialMailbox specialMailbox,
                                             ImapMailboxListOptions options)
    {
      return OpenMailbox(specialMailbox,
                         options,
                         DefaultSelectReadOnly);
    }

    public ImapOpenedMailboxInfo OpenMailbox(ImapSpecialMailbox specialMailbox,
                                             ImapMailboxListOptions options,
                                             bool asReadOnly)
    {
      return RunExclusively(() => {
        return OpenMailboxCore(GetMailboxCore(specialMailbox, options, true, true),
                               asReadOnly);
      });
    }

#if ASYNC_OPS
    public Task<ImapOpenedMailboxInfo> OpenMailboxAsync(ImapSpecialMailbox specialMailbox)
    {
      return OpenMailboxAsync(specialMailbox,
                              ImapMailboxListOptions.Default,
                              DefaultSelectReadOnly);
    }

    public Task<ImapOpenedMailboxInfo> OpenMailboxAsync(ImapSpecialMailbox specialMailbox,
                                                        bool asReadOnly)
    {
      return OpenMailboxAsync(specialMailbox,
                              ImapMailboxListOptions.Default,
                              asReadOnly);
    }

    public Task<ImapOpenedMailboxInfo> OpenMailboxAsync(ImapSpecialMailbox specialMailbox,
                                                        ImapMailboxListOptions options)
    {
      return OpenMailboxAsync(specialMailbox,
                              options,
                              DefaultSelectReadOnly);
    }

    public Task<ImapOpenedMailboxInfo> OpenMailboxAsync(ImapSpecialMailbox specialMailbox,
                                                        ImapMailboxListOptions options,
                                                        bool asReadOnly)
    {
      return StartTaskExclusively(() => {
        return OpenMailboxCore(GetMailboxCore(specialMailbox, options, true, true),
                               asReadOnly);
      }, CancellationToken.None);
    }
#endif

    public ImapOpenedMailboxInfo OpenOrCreateMailbox(string mailboxName)
    {
      return OpenOrCreateMailbox(mailboxName,
                                 DefaultSubscription,
                                 DefaultSelectReadOnly,
                                 ImapMailboxListOptions.Default);
    }

    public ImapOpenedMailboxInfo OpenOrCreateMailbox(string mailboxName,
                                                     ImapMailboxListOptions options)
    {
      return OpenOrCreateMailbox(mailboxName,
                                 DefaultSubscription,
                                 DefaultSelectReadOnly,
                                 options);
    }

    public ImapOpenedMailboxInfo OpenOrCreateMailbox(string mailboxName,
                                                     bool asReadOnly)
    {
      return OpenOrCreateMailbox(mailboxName,
                                 DefaultSubscription,
                                 asReadOnly,
                                 ImapMailboxListOptions.Default);
    }

    public ImapOpenedMailboxInfo OpenOrCreateMailbox(string mailboxName,
                                                     bool subscribeIfCreated,
                                                     bool asReadOnly,
                                                     ImapMailboxListOptions options)
    {
      mailboxName = GetValidatedMailboxName(mailboxName, "mailboxName");

      return RunExclusively(() => OpenOrCreateMailboxCore(mailboxName,
                                                          subscribeIfCreated,
                                                          asReadOnly,
                                                          options));
    }

#if ASYNC_OPS
    public Task<ImapOpenedMailboxInfo> OpenOrCreateMailboxAsync(string mailboxName)
    {
      return OpenOrCreateMailboxAsync(mailboxName,
                                      DefaultSubscription,
                                      DefaultSelectReadOnly,
                                      ImapMailboxListOptions.Default);
    }

    public Task<ImapOpenedMailboxInfo> OpenOrCreateMailboxAsync(string mailboxName,
                                                                ImapMailboxListOptions options)
    {
      return OpenOrCreateMailboxAsync(mailboxName,
                                      DefaultSubscription,
                                      DefaultSelectReadOnly,
                                      options);
    }

    public Task<ImapOpenedMailboxInfo> OpenOrCreateMailboxAsync(string mailboxName,
                                                                bool asReadOnly)
    {
      return OpenOrCreateMailboxAsync(mailboxName,
                                      DefaultSubscription,
                                      asReadOnly,
                                      ImapMailboxListOptions.Default);
    }

    public Task<ImapOpenedMailboxInfo> OpenOrCreateMailboxAsync(string mailboxName,
                                                                bool subscribeIfCreated,
                                                                bool asReadOnly,
                                                                ImapMailboxListOptions options)
    {
      mailboxName = GetValidatedMailboxName(mailboxName, "mailboxName");

      return StartTaskExclusively(() => OpenOrCreateMailboxCore(mailboxName,
                                                                subscribeIfCreated,
                                                                asReadOnly,
                                                                options),
                                  CancellationToken.None);
    }
#endif

    private ImapOpenedMailboxInfo OpenOrCreateMailboxCore(string mailboxName,
                                                          bool subscribeIfCreated,
                                                          bool asReadOnly,
                                                          ImapMailboxListOptions options)
    {
      options &= (~ImapMailboxListOptions.RequestStatus); // no need to request status of mailbox

      var mailbox = GetMailboxCore(mailboxName, options, false, false);

      if (mailbox == null)
        mailbox = CreateMailboxCore(mailboxName, subscribeIfCreated);

      mailbox.ThrowIfNonExistentOrUnselectable();

      return OpenMailboxCore(mailbox, asReadOnly);
    }

#if IMAP_CREATE_SPECAIL_MAILBOX
    public ImapOpenedMailboxInfo OpenOrCreateMailbox(ImapSpecialMailbox mailbox)
    {
      return OpenOrCreateMailbox(mailbox,
                                 DefaultSubscription,
                                 DefaultSelectReadOnly,
                                 ImapMailboxListOptions.Default);
    }

    public ImapOpenedMailboxInfo OpenOrCreateMailbox(ImapSpecialMailbox mailbox,
                                                     ImapMailboxListOptions options)
    {
      return OpenOrCreateMailbox(mailbox,
                                 DefaultSubscription,
                                 DefaultSelectReadOnly,
                                 options);
    }

    public ImapOpenedMailboxInfo OpenOrCreateMailbox(ImapSpecialMailbox mailbox,
                                                     bool asReadOnly)
    {
      return OpenOrCreateMailbox(mailbox,
                                 DefaultSubscription,
                                 asReadOnly,
                                 ImapMailboxListOptions.Default);
    }

    public ImapOpenedMailboxInfo OpenOrCreateMailbox(ImapSpecialMailbox mailbox,
                                                     bool subscribeIfCreated,
                                                     bool asReadOnly,
                                                     ImapMailboxListOptions options)
    {
      options &= (~ImapMailboxListOptions.RequestStatus); // no need to request status of mailbox

      var m = GetMailboxNoException(mailbox, options);

      if (m == null)
        m = CreateMailbox(mailbox, subscribeIfCreated);

      return OpenMailbox(m, asReadOnly);
    }
#endif

    /*
     * CLOSE
     */
    public void CloseMailbox()
    {
      RunExclusively(CloseMailboxCore);
    }

#if ASYNC_OPS
    public Task CloseMailboxAsync()
    {
      return StartTaskExclusively(CloseMailboxCore, CancellationToken.None);
    }
#endif

    private void CloseMailboxCore()
    {
      /*
       *       The CLOSE command permanently removes all messages that have the
       *       \Deleted flag set from the currently selected mailbox, and returns
       *       to the authenticated state from the selected state.  No untagged
       *       EXPUNGE responses are sent.
       * 
       *       No messages are removed, and no error is given, if the mailbox is
       *       selected by an EXAMINE command or is otherwise selected read-only.
       */
      if (Session.SelectedMailbox != null)
        ProcessResult(Session.Close());

      SetOpenedMailbox(null);
    }

    internal void CloseMailboxIfSelected(ImapMailboxInfo mailbox)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      if (Session.SelectedMailbox != null && Session.SelectedMailbox.NameEquals(mailbox.Mailbox)) {
        ProcessResult(Session.Close());

        SetOpenedMailbox(null);
      }
    }

    /*
     * NOOP
     */
    public void Refresh()
    {
      RunExclusively(RefreshCore);
    }

#if ASYNC_OPS
    public Task RefreshAsync()
    {
      return StartTaskExclusively(RefreshCore, CancellationToken.None);
    }
#endif

    private void RefreshCore()
    {
      if (openedMailbox == null)
        ProcessResult(Session.NoOp());
      else
        openedMailbox.ProcessResult(Session.NoOp());
    }

    /*
     * LIST (root)
     */
    /// <remarks><paramref name="rootMailboxName"/> must be validated by caller</remarks>
    private string GetHierarchyDelimiterOf(string rootMailboxName)
    {
      string hierarchyDelimiter;

      if (Session.HierarchyDelimiters.TryGetValue(rootMailboxName, out hierarchyDelimiter)) {
        return hierarchyDelimiter;
      }
      else {
        return RunExclusively(() => {
          ImapMailboxList mailboxList;

          ProcessResult(Session.ListRoot(rootMailboxName, out mailboxList));

          return mailboxList.HierarchyDelimiter;
        });
      }
    }

    public string CombineMailboxName(string name1, string name2)
    {
      GetValidatedMailboxName(name1, "name1");
      GetValidatedMailboxName(name2, "name2");

      return string.Join(GetHierarchyDelimiterOf(name1),
#if NET_4_0
                         name1, name2);
#else
                         new[] {name1, name2});
#endif
    }

    public string CombineMailboxName(string name1, string name2, string name3)
    {
      GetValidatedMailboxName(name1, "name1");
      GetValidatedMailboxName(name2, "name2");
      GetValidatedMailboxName(name3, "name3");

      return string.Join(GetHierarchyDelimiterOf(name1),
#if NET_4_0
                         name1, name2, name3);
#else
                         new[] {name1, name2, name3});
#endif
    }

    public string CombineMailboxName(string name1, string name2, string name3, string name4)
    {
      GetValidatedMailboxName(name1, "name1");
      GetValidatedMailboxName(name2, "name2");
      GetValidatedMailboxName(name3, "name3");
      GetValidatedMailboxName(name4, "name4");

      return string.Join(GetHierarchyDelimiterOf(name1),
#if NET_4_0
                         name1, name2, name3, name4);
#else
                         new[] {name1, name2, name3, name4});
#endif
    }

    public string CombineMailboxName(IEnumerable<string> names)
    {
      if (names == null)
        throw new ArgumentNullException("names");

      return CombineMailboxName((string[])names.ToArray());
    }

    public string CombineMailboxName(params string[] names)
    {
      if (names == null)
        throw new ArgumentNullException("names");
      if (names.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyArray("names");

      Array.ForEach(names, n => GetValidatedMailboxName(n, "names"));

      if (names.Length == 1)
        return names[0];
      else
        return string.Join(GetHierarchyDelimiterOf(names[0]), names);
    }

    /*
     * GETQUOTA
     */
    public ImapQuota GetQuota(string quotaRoot)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException("quotaRoot");

      if (!IsCapabilityAvailable(ImapCapability.Quota))
        return null; // as default

      return RunExclusively(() => {
        ImapQuota quota;

        ProcessResult(Session.GetQuota(quotaRoot, out quota));

        return quota;
      });
    }

#if ASYNC_OPS
    public Task<ImapQuota> GetQuotaAsync(string quotaRoot)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException("quotaRoot");

      if (!IsCapabilityAvailable(ImapCapability.Quota)) {
        var tcs = new TaskCompletionSource<ImapQuota>();

        tcs.SetResult(null); // as default

        return tcs.Task;
      }

      return StartTaskExclusively(() => {
        ImapQuota quota;

        ProcessResult(Session.GetQuota(quotaRoot, out quota));

        return quota;
      }, CancellationToken.None);
    }
#endif

    public double GetQuotaUsage(string quotaRoot, string resourceName)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException("quotaRoot");
      if (resourceName == null)
        throw new ArgumentNullException("resourceName");
      if (resourceName.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("resourceName");

      return GetQuotaUsage(GetQuota(quotaRoot), resourceName);
    }

#if ASYNC_OPS
    public Task<double> GetQuotaUsageAsync(string quotaRoot, string resourceName)
    {
      if (quotaRoot == null)
        throw new ArgumentNullException("quotaRoot");
      if (resourceName == null)
        throw new ArgumentNullException("resourceName");
      if (resourceName.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("resourceName");

      return GetQuotaAsync(quotaRoot).ContinueWith(t => GetQuotaUsage(t.Result, resourceName));
    }
#endif

    private double GetQuotaUsage(ImapQuota quota, string resourceName)
    {
      if (quota == null)
        return 0.0; // as default

      foreach (var resource in quota.Resources) {
        if (string.Equals(resource.Name, resourceName, StringComparison.OrdinalIgnoreCase)) {
          if (resource.Limit == 0L)
            return 0.0;
          else
            return (double)resource.Usage / (double)resource.Limit;
        }
      }

      throw new ImapProtocolViolationException(string.Format("no such resource name: '{0}'", resourceName));
    }

    /*
     * utilities
     */
    private ImapOpenedMailboxInfo SetOpenedMailbox(ImapOpenedMailboxInfo newlyOpenedMailbox)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      if (openedMailbox != null)
        openedMailbox.DetachAllMailboxObject();

      openedMailbox = newlyOpenedMailbox;

      return openedMailbox;
    }

    internal static string GetValidatedMailboxName(string mailboxName, string paramName)
    {
      if (mailboxName == null)
        throw new ArgumentNullException(paramName ?? "mailboxName");
      if (mailboxName.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString(paramName ?? "mailboxName");

      return mailboxName;
    }

    internal protected ImapCommandResult ProcessResult(ImapCommandResult result)
    {
      return ProcessResult(result, null);
    }

    internal protected virtual ImapCommandResult ProcessResult(ImapCommandResult result,
                                                               Func<ImapTaggedStatusResponse, bool> throwIfErrorResponse)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      var alert = result.GetResponseCode(ImapResponseCode.Alert);

      if (alert != null)
        RaiseEvent(EventType.AlertReceived,
                   new ImapAlertReceivedEventArgs(alert.ResponseText.Text, alert.Condition),
                   false);

      if (result.Succeeded)
        return result;

      if (throwIfErrorResponse != null && result.TaggedStatusResponse != null)
        if (!throwIfErrorResponse(result.TaggedStatusResponse))
          return result;

      if (result.Code == ImapCommandResultCode.Bye)
        throw new ImapConnectionException(string.Format("disconnected from server: \"{0}\"",
                                                        result.ResponseText));
      else
        throw new ImapErrorResponseException(result);
    }

    internal bool IsCapabilityAvailable(ImapCapability requiredCapability)
    {
      if (disabledCapabilities.Contains(requiredCapability))
        return false;

      return ServerCapabilities.Contains(requiredCapability);
    }

    internal void ThrowIfCapabilityUnavailable(IImapExtension extension)
    {
      foreach (var requiredCapability in extension.RequiredCapabilities) {
        ThrowIfCapabilityUnavailable(requiredCapability);
      }
    }

    internal void ThrowIfCapabilityUnavailable(ImapCapability requiredCapability)
    {
      if (!IsCapabilityAvailable(requiredCapability))
        throw new ImapIncapableException(requiredCapability);
    }

    internal ImapStatusDataItem GetAvailableStatusDataItem()
    {
      if (IsCapabilityAvailable(ImapCapability.CondStore))
        return ImapStatusDataItem.StandardAll + ImapStatusDataItem.HighestModSeq;
      else
        return ImapStatusDataItem.StandardAll;
    }

    private ImapOpenedMailboxInfo openedMailbox;
    private readonly ImapCapabilitySet disabledCapabilities = new ImapCapabilitySet();
  }
}
