// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Runtime.Remoting.Messaging;
using System.Threading;
#if ASYNC_OPS
using System.Threading.Tasks;
#endif

using Smdn.Net.Imap4.Client.Session;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Imap4.Client {
  public partial class ImapClient : IDisposable {
    public ImapClientProfile Profile {
      get { return profile; }
    }

    public int Timeout {
      get { return (session == null) ? profile.Timeout : session.TransactionTimeout; }
      set
      {
        profile.Timeout = value;

        if (session != null)
          session.TransactionTimeout = value;
      }
    }

    public int SendTimeout {
      get { return (session == null) ? profile.SendTimeout : session.SendTimeout; }
      set
      {
        profile.SendTimeout = value;

        if (session != null)
          session.SendTimeout = value;
      }
    }

    public int ReceiveTimeout {
      get { return (session == null) ? profile.ReceiveTimeout : session.ReceiveTimeout; }
      set
      {
        profile.ReceiveTimeout = value;

        if (session != null)
          session.ReceiveTimeout = value;
      }
    }

    public bool IsConnected {
      get { return (session != null && session.State != ImapSessionState.NotConnected); }
    }

    public bool IsBusy {
      get { return sessionStatus == SessionStatus_Busy; }
    }

    internal ImapSession Session {
      get { ThrowIfNotConnected(); return session; }
    }

    public bool IsSecureSession {
      get { return Session.IsSecureConnection; }
    }

    public ImapCapabilitySet ServerCapabilities {
      get { return Session.ServerCapabilities; }
    }

#if NET_4_5
    public IReadOnlyDictionary<string, string> ServerID {
#else
    public IDictionary<string, string> ServerID {
#endif
      get { return Session.ServerID; }
    }

    public ImapNamespace ServerNamespace {
      get { return Session.Namespaces; }
    }

    /*
     * construction
     */
    public ImapClient()
      : this(new ImapClientProfile())
    {
    }

    public ImapClient(Uri authority)
      : this(new ImapClientProfile(authority))
    {
    }

    public ImapClient(string host)
      : this(new ImapClientProfile(host))
    {
    }

    public ImapClient(string host, string userName)
      : this(new ImapClientProfile(host, userName))
    {
    }

    public ImapClient(string host, bool securePort, string userName)
      : this(new ImapClientProfile(host, securePort, userName))
    {
    }

    public ImapClient(string host, int port)
      : this(new ImapClientProfile(host, port))
    {
    }

    public ImapClient(string host, int port, string userName)
      : this(new ImapClientProfile(host, port, userName))
    {
    }

    public ImapClient(string host, int port, bool securePort)
      : this(new ImapClientProfile(host, port, securePort))
    {
    }

    public ImapClient(string host, int port, bool securePort, string userName)
      : this(new ImapClientProfile(host, port, securePort, userName))
    {
    }

    public ImapClient(string host, int port, string userName, string authType)
      : this(new ImapClientProfile(host, port, userName, authType))
    {
    }

    public ImapClient(string host, int port, bool securePort, string userName, string authType)
      : this(new ImapClientProfile(host, port, securePort, userName, authType))
    {
    }

    public ImapClient(string host, int port, bool securePort, string userName, string authType, int timeout)
      : this(new ImapClientProfile(host, port, securePort, userName, authType, timeout))
    {
    }

    public ImapClient(ImapClientProfile profile)
    {
      if (profile == null)
        throw new ArgumentNullException("profile");

      this.profile = profile;
    }

    /*
     * connect impl
     */
    private static ImapSession CreateSession(ConnectParams @params)
    {
      ImapSession session = null;

      try {
        // XXX: received ALERT while connecting would be ignored
        session = ImapSessionCreator.CreateSession(@params.Profile,
                                                   @params.AuthMechanism,
                                                   @params.CreateSslStreamCallback ?? ImapSslConnection.CreateSslStream
#if ASYNC_OPS
                                                   , @params.CancellationToken
#endif
                                                  );
      }
      finally {
        @params.Profile.SetCredentials(null);
      }

      // update server info
      if (session.ServerCapabilities.Contains(ImapCapability.Namespace))
        session.Namespace();

      if (session.ServerCapabilities.Contains(ImapCapability.ID))
        session.ID(null); // TODO: client ID

      session.UpdateSelectedMailboxSizeAndStatus = false;

      return session;
    }

    internal class ConnectParams {
      public readonly ImapClientProfile Profile;
      public readonly SaslClientMechanism AuthMechanism;
      public readonly UpgradeConnectionStreamCallback CreateSslStreamCallback;
#if ASYNC_OPS
      public readonly CancellationToken CancellationToken;
#endif

      public ConnectParams(ImapClientProfile profile,
                           ICredentialsByHost credentials,
                           UpgradeConnectionStreamCallback createSslStreamCallback
#if ASYNC_OPS
                           , CancellationToken cancellationToken
#endif
                          )
      {
        Profile = profile.Clone();
        Profile.SetCredentials(credentials);

        CreateSslStreamCallback = createSslStreamCallback;
#if ASYNC_OPS
        CancellationToken = cancellationToken;
#endif
      }

      public ConnectParams(ImapClientProfile profile,
                           SaslClientMechanism authMechanism,
                           UpgradeConnectionStreamCallback createSslStreamCallback
#if ASYNC_OPS
                           , CancellationToken cancellationToken
#endif
                          )
      {
        Profile = profile.Clone();
        AuthMechanism = authMechanism;
        CreateSslStreamCallback = createSslStreamCallback;
#if ASYNC_OPS
        CancellationToken = cancellationToken;
#endif
      }
    }

    public class ConnectAsyncResult : IAsyncResult {
      public object AsyncState {
        get { return asyncState; }
      }

      public WaitHandle AsyncWaitHandle {
        get { return innerAsyncResult.AsyncWaitHandle; }
      }

      public bool CompletedSynchronously {
        get { return innerAsyncResult.CompletedSynchronously; }
      }

      public bool IsCompleted {
        get { return innerAsyncResult.IsCompleted; }
      }

      public bool EndConnectCalled {
        get { return endConnectCalled != 0; }
      }

      internal bool OwnerDisposed {
        get; set;
      }

      internal ConnectAsyncResult() {}

      internal void BeginConnect(ConnectParams @params,
                                 AsyncCallback asyncCallback,
                                 object asyncState)
      {
        this.asyncState = asyncState;

        var proc = (Func<ConnectParams, ImapSession>)ImapClient.CreateSession;

        innerAsyncResult = (AsyncResult)proc.BeginInvoke(@params, asyncCallback, asyncState);
      }

      internal ImapSession EndConnect()
      {
        ImapSession createdSession = null;

        try {
          createdSession = ((Func<ConnectParams, ImapSession>)innerAsyncResult.AsyncDelegate).EndInvoke(innerAsyncResult);

          if (OwnerDisposed)
            return null;
          else
            return createdSession;
        }
        finally {
          if (OwnerDisposed && createdSession != null)
            createdSession.Disconnect(false); // ignore exceptions
        }
      }

      private AsyncResult innerAsyncResult;
      private object asyncState;
      internal int endConnectCalled = 0;
    }

    /*
     * BeginConnect()/EndConnect()
     */
    public IAsyncResult BeginConnect(string password)
    {
      return BeginConnect(password, null, null, null);
    }

    public IAsyncResult BeginConnect(string password,
                                     UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      return BeginConnect(password, createSslStreamCallback, null, null);
    }

    public IAsyncResult BeginConnect(string password,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
    {
      return BeginConnect(password, null, asyncCallback, asyncState);
    }

    public IAsyncResult BeginConnect(string password,
                                     UpgradeConnectionStreamCallback createSslStreamCallback,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
    {
      return BeginConnect(new ConnectParams(profile,
                                            new NetworkCredential(profile.UserName, password),
                                            createSslStreamCallback
#if ASYNC_OPS
                                            , CancellationToken.None
#endif
                                           ),
                          asyncCallback,
                          asyncState);
    }

    public IAsyncResult BeginConnect(ICredentialsByHost credentials)
    {
      return BeginConnect(credentials, null, null, null);
    }

    public IAsyncResult BeginConnect(ICredentialsByHost credentials,
                                     UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      return BeginConnect(credentials, createSslStreamCallback, null, null);
    }

    public IAsyncResult BeginConnect(ICredentialsByHost credentials,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
    {
      return BeginConnect(credentials, null, asyncCallback, asyncState);
    }

    public IAsyncResult BeginConnect(ICredentialsByHost credentials,
                                     UpgradeConnectionStreamCallback createSslStreamCallback,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
    {
      return BeginConnect(new ConnectParams(profile,
                                            credentials,
                                            createSslStreamCallback
#if ASYNC_OPS
                                            , CancellationToken.None
#endif
                                           ),
                          asyncCallback,
                          asyncState);
    }

    public IAsyncResult BeginConnect(SaslClientMechanism authMechanism)
    {
      return BeginConnect(authMechanism, null, null, null);
    }

    public IAsyncResult BeginConnect(SaslClientMechanism authMechanism,
                                     UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      return BeginConnect(authMechanism, createSslStreamCallback, null, null);
    }

    public IAsyncResult BeginConnect(SaslClientMechanism authMechanism,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
    {
      return BeginConnect(authMechanism, null, asyncCallback, asyncState);
    }

    public IAsyncResult BeginConnect(SaslClientMechanism authMechanism,
                                     UpgradeConnectionStreamCallback createSslStreamCallback,
                                     AsyncCallback asyncCallback,
                                     object asyncState)
    {
      return BeginConnect(new ConnectParams(profile,
                                            authMechanism,
                                            createSslStreamCallback
#if ASYNC_OPS
                                            , CancellationToken.None
#endif
                                           ),
                          asyncCallback,
                          asyncState);
    }

    private IAsyncResult BeginConnect(ConnectParams @params,
                                      AsyncCallback asyncCallback,
                                      object asyncState)
    {
      ThrowIfAlreadyConnected();

      SetBusy();

      connectAsyncResult = new ConnectAsyncResult();
      connectAsyncResult.BeginConnect(@params, asyncCallback, asyncState);

      return connectAsyncResult;
    }

    public void EndConnect(IAsyncResult asyncResult)
    {
      if (asyncResult == null)
        throw new ArgumentNullException("asyncResult");
      if (asyncResult != connectAsyncResult)
        throw ExceptionUtils.CreateArgumentMustBeValidIAsyncResult("asyncResult");

      if (Interlocked.CompareExchange(ref connectAsyncResult.endConnectCalled, 1, 0) == 1)
        throw new InvalidOperationException("EndConnect already called");

      try {
        session = connectAsyncResult.EndConnect();
      }
      finally {
        connectAsyncResult = null;

        UnsetBusy();
      }
    }

#if ASYNC_OPS
    /*
     * ConnectAsync()
     */
    public Task ConnectAsync(string password)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            new NetworkCredential(profile.UserName, password),
                                            null,
                                            CancellationToken.None));
    }

    public Task ConnectAsync(string password,
                             UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            new NetworkCredential(profile.UserName, password),
                                            createSslStreamCallback,
                                            CancellationToken.None));
    }

    public Task ConnectAsync(string password,
                             UpgradeConnectionStreamCallback createSslStreamCallback,
                             CancellationToken cancellationToken)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            new NetworkCredential(profile.UserName, password),
                                            createSslStreamCallback,
                                            cancellationToken));
    }

    public Task ConnectAsync(ICredentialsByHost credentials)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            credentials,
                                            null,
                                            CancellationToken.None));
    }

    public Task ConnectAsync(ICredentialsByHost credentials,
                             UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            credentials,
                                            createSslStreamCallback,
                                            CancellationToken.None));
    }

    public Task ConnectAsync(ICredentialsByHost credentials,
                             UpgradeConnectionStreamCallback createSslStreamCallback,
                             CancellationToken cancellationToken)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            credentials,
                                            createSslStreamCallback,
                                            cancellationToken));
    }

    public Task ConnectAsync(SaslClientMechanism authMechanism)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            authMechanism,
                                            null,
                                            CancellationToken.None));
    }

    public Task ConnectAsync(SaslClientMechanism authMechanism,
                             UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            authMechanism,
                                            createSslStreamCallback,
                                            CancellationToken.None));
    }

    public Task ConnectAsync(SaslClientMechanism authMechanism,
                             UpgradeConnectionStreamCallback createSslStreamCallback,
                             CancellationToken cancellationToken)
    {
      return ConnectAsync(new ConnectParams(profile,
                                            authMechanism,
                                            createSslStreamCallback,
                                            cancellationToken));
    }

    private Task ConnectAsync(ConnectParams @params)
    {
      return Task.Factory.FromAsync(BeginConnect(@params, null, null), EndConnect);
    }
#endif

    /*
     * Connect()
     */
    public void Connect(string password)
    {
      Connect(new ConnectParams(profile,
                                new NetworkCredential(profile.UserName, password),
                                null
#if ASYNC_OPS
                                , CancellationToken.None
#endif
                               ));
    }

    public void Connect(string password, UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      Connect(new ConnectParams(profile,
                                new NetworkCredential(profile.UserName, password),
                                createSslStreamCallback
#if ASYNC_OPS
                                , CancellationToken.None
#endif
                               ));
    }

#if ASYNC_OPS
    public void Connect(string password,
                        UpgradeConnectionStreamCallback createSslStreamCallback,
                        CancellationToken cancellationToken)
    {
      Connect(new ConnectParams(profile,
                                new NetworkCredential(profile.UserName, password),
                                createSslStreamCallback,
                                cancellationToken));
    }
#endif

    public void Connect(ICredentialsByHost credentials)
    {
      Connect(new ConnectParams(profile,
                                credentials,
                                null
#if ASYNC_OPS
                                , CancellationToken.None
#endif
                               ));
    }

    public void Connect(ICredentialsByHost credentials, UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      Connect(new ConnectParams(profile,
                                credentials,
                                createSslStreamCallback
#if ASYNC_OPS
                                , CancellationToken.None
#endif
                               ));
    }

#if ASYNC_OPS
    public void Connect(ICredentialsByHost credentials,
                        UpgradeConnectionStreamCallback createSslStreamCallback,
                        CancellationToken cancellationToken)
    {
      Connect(new ConnectParams(profile,
                                credentials,
                                createSslStreamCallback,
                                cancellationToken));
    }
#endif

    public void Connect(SaslClientMechanism authMechanism)
    {
      if (authMechanism == null)
        throw new ArgumentNullException("authMechanism");

      Connect(new ConnectParams(profile,
                                authMechanism,
                                null
#if ASYNC_OPS
                                , CancellationToken.None
#endif
                               ));
    }

    public void Connect(SaslClientMechanism authMechanism, UpgradeConnectionStreamCallback createSslStreamCallback)
    {
      if (authMechanism == null)
        throw new ArgumentNullException("authMechanism");

      Connect(new ConnectParams(profile,
                                authMechanism,
                                createSslStreamCallback
#if ASYNC_OPS
                                , CancellationToken.None
#endif
                               ));
    }

#if ASYNC_OPS
    public void Connect(SaslClientMechanism authMechanism,
                        UpgradeConnectionStreamCallback createSslStreamCallback,
                        CancellationToken cancellationToken)
    {
      if (authMechanism == null)
        throw new ArgumentNullException("authMechanism");

      Connect(new ConnectParams(profile,
                                authMechanism,
                                createSslStreamCallback,
                                cancellationToken));
    }
#endif

    private void Connect(ConnectParams @params)
    {
      ThrowIfAlreadyConnected();

      SetBusy();

      try {
        session = CreateSession(@params);
      }
      finally {
        UnsetBusy();
      }
    }

    /*
     * destruction
     */
    void IDisposable.Dispose()
    {
      if (session != null) {
        (session as IDisposable).Dispose();
        session = null;
      }

      if (connectAsyncResult != null)
        connectAsyncResult.OwnerDisposed = true;
    }

    public void Disconnect()
    {
      Disconnect(false);
    }

    public void Logout()
    {
      Disconnect(true);
    }

    public void Disconnect(bool logout)
    {
      SetBusy();

      try {
        if (session == null)
          return;

        session.Disconnect(logout);
        session = null;
      }
      finally {
        UnsetBusy();
      }
    }

    /*
     * methods for exclusive execution
     */
    private const int SessionStatus_Ready = 0;
    private const int SessionStatus_Busy = 1;

    private int sessionStatus = SessionStatus_Ready;

    internal void SetBusy()
    {
      if (SessionStatus_Busy == Interlocked.CompareExchange(ref sessionStatus, SessionStatus_Busy, SessionStatus_Ready))
        throw new InvalidOperationException("client is busy; concurrent operations are not supported");
    }

    internal void UnsetBusy()
    {
      if (SessionStatus_Busy == Interlocked.CompareExchange(ref sessionStatus, SessionStatus_Ready, SessionStatus_Busy))
        RaiseQueuedEvents();
    }

    internal void RunExclusively(Action operation)
    {
      ThrowIfNotConnected();

      SetBusy();

      try {
        operation();
      }
      finally {
        UnsetBusy();
      }
    }

    internal T RunExclusively<T>(Func<T> operation)
    {
      ThrowIfNotConnected();

      SetBusy();

      try {
        return operation();
      }
      finally {
        UnsetBusy();
      }
    }

#if ASYNC_OPS
    internal Task StartTaskExclusively(Action operation,
                                       CancellationToken cancellationToken)
    {
      ThrowIfNotConnected();

      SetBusy();

      var t = Task.Factory.StartNew(() => {
        try {
          operation();
        }
        finally {
          UnsetBusy();
        }
      }, cancellationToken);

      // XXX
      if (cancellationToken.IsCancellationRequested && t.Status == TaskStatus.WaitingToRun)
        UnsetBusy();
      else if (t.IsCanceled)
        UnsetBusy();

      return t;
    }

    internal Task<T> StartTaskExclusively<T>(Func<T> operation,
                                             CancellationToken cancellationToken,
                                             bool unsetBusyIfOperationFinished = true)
    {
      ThrowIfNotConnected();

      SetBusy();

      var t = Task.Factory.StartNew(() => {
        try {
          return operation();
        }
        finally {
          if (unsetBusyIfOperationFinished)
            UnsetBusy();
        }
      }, cancellationToken);

      // XXX
      if (cancellationToken.IsCancellationRequested && t.Status == TaskStatus.WaitingToRun)
        UnsetBusy();
      else if (t.IsCanceled)
        UnsetBusy();

      return t;
    }
#endif

    /*
     * methods for checking and validating internal status
     */
    private void ThrowIfAlreadyConnected()
    {
      if (IsConnected)
        throw new InvalidOperationException("already connected");
    }

    private void ThrowIfNotConnected()
    {
      if (!IsConnected)
        throw new InvalidOperationException("not connected or already disconnected");
    }

#if DEBUG
    internal void ThrowIfNotInExclusiveOperation()
    {
      if (sessionStatus == SessionStatus_Ready)
        throw new InvalidOperationException("not in exclusive operation");
    }
#endif

    public override string ToString()
    {
      return string.Format("{{ImapClient: Authority='{0}', IsConnected={1}, OpenedMailbox={2}}}",
                           profile.Authority,
                           IsConnected,
                           openedMailbox);
    }

    private ImapSession session;
    private ImapClientProfile profile;
    private ConnectAsyncResult connectAsyncResult;
  }
}
