// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
#if NET_4_0
// LazyInitializer
using System.Threading;
#else
using Smdn.Threading;
#endif

using Smdn.Formats;
using Smdn.Formats.Mime;
using Smdn.Net.Imap4.Client.Session;

namespace Smdn.Net.Imap4.Client {
  public sealed partial class ImapMessageInfo : ImapMessageInfoBase, IImapMessageAttribute {
    public static readonly long ExpungedMessageSequenceNumber = 0L;

    public long Sequence {
      get; internal set;
    }

    public long Uid {
      get; private set;
    }

    public Uri Url {
      get { return uriBuilder.Uri; }
    }

    /*
     * IImapMessageDynamicAttribute
     */
    private ImapMessageDynamicAttribute dynamicAttribute;

    internal ImapMessageDynamicAttribute DynamicAttribute {
      get { EnsureDynamicAttributesFetched(); return dynamicAttribute; }
      set { dynamicAttribute = value; }
    }

    internal ImapMessageDynamicAttribute GetDynamicAttribute()
    {
      return dynamicAttribute;
    }

    public IImapMessageFlagSet Flags {
      get { return DynamicAttribute.Flags; }
    }

    [CLSCompliant(false)]
    public ulong ModSeq {
      get { return DynamicAttribute.ModSeq; }
    }

    /*
     * IImapMessageStaticAttribute
     */
    private ImapMessageStaticAttribute staticAttribute;

    internal ImapMessageStaticAttribute StaticAttribute {
      get { EnsureStaticAttributesFetched(); return staticAttribute; }
      set { staticAttribute = value; }
    }

    private ImapMessagePartInfo messagePart = null;

    public ImapMessagePartInfo MessagePart {
      get {
        return LazyInitializer.EnsureInitialized(ref messagePart, delegate {
          return new ImapMessagePartInfo(this);
        });
      }
    }

    [Obsolete("use MessagePart")]
    public IImapBodyStructure BodyStructure {
      get { return StaticAttribute.BodyStructure; }
    }

    public ImapEnvelope Envelope {
      get { return StaticAttribute.Envelope; }
    }

    public DateTimeOffset InternalDate {
      get { return StaticAttribute.InternalDate.Value; }
    }

    DateTimeOffset? IImapMessageStaticAttribute.InternalDate {
      get { return StaticAttribute.InternalDate; }
    }

    public long Length {
      get { return StaticAttribute.Rfc822Size; }
    }

    long IImapMessageStaticAttribute.Rfc822Size {
      get { return StaticAttribute.Rfc822Size; }
    }

    long IImapMessageStaticAttribute.GetBinarySizeOf(string section)
    {
      return StaticAttribute.GetBinarySizeOf(section);
    }

    /*
     * miscellaneous getter properties
     */
    public bool IsMarkedAsDeleted {
      get { return Flags.Contains(ImapMessageFlag.Deleted); }
    }

    public bool IsSeen {
      get { return Flags.Contains(ImapMessageFlag.Seen); }
    }

    public bool IsRecent {
      get { return Flags.Contains(ImapMessageFlag.Recent); }
    }

    public bool IsAnswered {
      get { return Flags.Contains(ImapMessageFlag.Answered); }
    }

    public bool IsDraft {
      get { return Flags.Contains(ImapMessageFlag.Draft); }
    }

    public bool IsFlagged {
      get { return Flags.Contains(ImapMessageFlag.Flagged); }
    }

    public bool IsDeleted {
      get { return Sequence == ExpungedMessageSequenceNumber; }
    }

    public string EnvelopeSubject {
      get { return StaticAttribute.Envelope.GetDecodedSubject(null, false); }
    }

    public DateTimeOffset? EnvelopeDate {
      get { return StaticAttribute.Envelope.GetDecodedDate(false); }
    }

    public MimeType MediaType {
      get { return StaticAttribute.BodyStructure.MediaType; }
    }

    public bool IsMultiPart {
      get { return StaticAttribute.BodyStructure.IsMultiPart; }
    }

    public bool HasEncapsulatedMessage {
      get { return StaticAttribute.BodyStructure.HasEncapsulatedMessage; }
    }

    private ImapSequenceSet uidSet;

    private ImapSequenceSet SequenceOrUidSet {
      get
      {
        if (uidSet == null)
          uidSet = ImapSequenceSet.CreateUidSet(Uid);

        return uidSet;
      }
    }

    internal ImapMessageInfo(ImapOpenedMailboxInfo mailbox, long uid, long sequence)
      : base(mailbox)
    {
      this.Uid = uid;
      this.Sequence = sequence;

      (this as IImapUrl).SetBaseUrl(mailbox.Mailbox.UrlBuilder);
    }

    private ImapUriBuilder uriBuilder;

    protected override ImapSequenceSet GetSequenceOrUidSet()
    {
      return SequenceOrUidSet;
    }

    internal void SetDeleted()
    {
      Sequence = ExpungedMessageSequenceNumber;

      if (IsAvailable)
        Mailbox.Detach(this);
    }

    /*
     * message part
     */
    public ImapMessagePartInfo GetMessagePart(string partSpecifier)
    {
      partSpecifier = ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart(partSpecifier);

      if (string.IsNullOrEmpty(partSpecifier)) {
        return MessagePart;
      }
      else {
        return MessagePart.Find(delegate(ImapMessagePartInfo part) {
          return string.Equals(part.Specifier, partSpecifier, StringComparison.Ordinal);
        });
      }
    }

    public ImapMessagePartInfo GetMessagePart(params int[] partSpecifiers)
    {
      return GetMessagePart((IEnumerable<int>)partSpecifiers);
    }

    public ImapMessagePartInfo GetMessagePart(/*params*/ IEnumerable<int> partSpecifiers)
    {
      var builder = new StringBuilder(0x10);

      foreach (var partSpecifier in partSpecifiers ?? EmptyArray<int>.Instance) {
        if (0 < builder.Length)
          builder.Append('.');

        builder.Append(partSpecifier); // not check whether each value is valid or not here
      }

      return GetMessagePart(builder.ToString());
    }

    public ImapMessagePartInfo FindMessagePart(Predicate<ImapMessagePartInfo> match)
    {
      return MessagePart.Find(match);
    }

    public List<ImapMessagePartInfo> FindAllMessageParts(Predicate<ImapMessagePartInfo> match)
    {
      return MessagePart.FindAll(match);
    }

    // FindMessagePartByMimeType
    public ImapMessagePartInfo FindMessagePart(MimeType mediaType)
    {
      if (mediaType == null)
        throw  new ArgumentNullException("mediaType");

      return MessagePart.Find(delegate(ImapMessagePartInfo part) {
        return mediaType.EqualsIgnoreCase(part.ContentType);
      });
    }

    public List<ImapMessagePartInfo> FindAllMessageParts(MimeType mediaType)
    {
      if (mediaType == null)
        throw  new ArgumentNullException("mediaType");

      return MessagePart.FindAll(delegate(ImapMessagePartInfo part) {
        return mediaType.EqualsIgnoreCase(part.ContentType);
      });
    }

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(ImapUriBuilder baseUrl)
    {
      uriBuilder = baseUrl.Clone();

      uriBuilder.Uid = Uid;
      uriBuilder.UidValidity = UidValidity;
    }

    public override string ToString()
    {
      return string.Format("{{ImapMessageInfo: Authority='{0}', Mailbox='{1}', Sequence={2}, Uid={3}, UidValidity={4}}}",
                           ImapStyleUriParser.GetStrongAuthority(Url),
                           Mailbox.FullName,
                           Sequence,
                           Uid,
                           UidValidity);
    }
  }
}
