// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

namespace Smdn.Net.Imap4.Client {
  internal sealed class ImapMessageInfoCollection :
    IEnumerable<ImapMessageInfo>,
    IImapMessageAttributeCollection<ImapMessageDynamicAttribute>
  {
    IEnumerator<ImapMessageInfo> IEnumerable<ImapMessageInfo>.GetEnumerator()
    {
      return uidMap.Values.GetEnumerator();
    }

    System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
    {
      return (this as IEnumerable<ImapMessageInfo>).GetEnumerator();
    }

    internal static int CompareMessageAttributeSequence(IImapMessageAttribute x, IImapMessageAttribute y)
    {
      if (x.Sequence < y.Sequence)
        return -1;
      else if (x.Sequence > y.Sequence)
        return 1;
      else
        return 0;
    }

    public void Add(ImapMessageInfo item)
    {
      uidMap.Add(item.Uid, item);
      items.Add(item);

      sorted = false;
    }

    public void Remove(ImapMessageInfo item)
    {
      uidMap.Remove(item.Uid);
      items.Remove(item);
    }

    public void Clear()
    {
      uidMap.Clear();
      items.Clear();
    }

    public void ChangeSequenceNumber(ImapMessageInfo item, long newSequenceNumber)
    {
      item.Sequence = newSequenceNumber;

      sorted = false;
    }

    public ImapMessageInfo FindByUid(long uid)
    {
      ImapMessageInfo ret;

      if (uidMap.TryGetValue(uid, out ret))
        return ret;
      else
        return null;
    }

    public ImapMessageInfo FindBySequence(long sequenceNumber)
    {
      if (!sorted) {
        items.Sort(CompareMessageAttributeSequence);
        sorted = true;
      }

      return FindBySequence(items, sequenceNumber);
    }

    /// <param name="list">list of IImapMessageAttribute. must be sorted.</param>
    internal static TImapMessageAttribute FindBySequence<TImapMessageAttribute>(IList<TImapMessageAttribute> list, long sequenceNumber)
      where TImapMessageAttribute : class, IImapMessageAttribute
    {
      // binary search
      var min = 0;
      var max = list.Count - 1;

      while (min <= max) {
        var mid = min + (max - min) / 2;
        var item = list[mid];

        if (item.Sequence == sequenceNumber)
          return item;
        else if (item.Sequence < sequenceNumber)
          min = mid + 1;
        else if (item.Sequence > sequenceNumber)
          max = mid - 1;
      }

      return null;
    }

    ImapMessageDynamicAttribute IImapMessageAttributeCollection<ImapMessageDynamicAttribute>.Find(long sequence)
    {
      var item = FindBySequence(sequence);

      if (item == null)
        return null;
      else
        return item.GetDynamicAttribute();
    }

    void IImapMessageAttributeCollection<ImapMessageDynamicAttribute>.Add(ImapMessageDynamicAttribute @value)
    {
      var item = FindBySequence(@value.Sequence);

      if (item != null)
        item.DynamicAttribute = @value;
    }

    private bool sorted = false;
    private readonly Dictionary<long, ImapMessageInfo> uidMap = new Dictionary<long, ImapMessageInfo>();
    private readonly List<ImapMessageInfo> items = new List<ImapMessageInfo>();
  }
}
