// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Collections.Generic;
using System.Linq;
#if ASYNC_OPS
using System.Threading;
using System.Threading.Tasks;
#endif

using Smdn.Net.Imap4;

namespace Smdn.Net.Imap4.Client {
  internal class ConcreteMessageInfoList : ImapMessageInfoList {
    internal ConcreteMessageInfoList(ImapOpenedMailboxInfo mailbox,
                                     ImapMessageFetchAttributeOptions fetchOptions,
                                     ImapSequenceSet sequenceOrUidSet)
      : base(mailbox, fetchOptions, false)
    {
      this.sequenceOrUidSet = sequenceOrUidSet;
    }

    protected override ImapSequenceSet GetSequenceOrUidSet()
    {
      return sequenceOrUidSet;
    }

    internal ImapMessageInfo ToSingleMessage()
    {
      return Yield(ToSingleMessage);
    }

#if ASYNC_OPS
    internal Task<ImapMessageInfo> ToSingleMessageAsync()
    {
      return YieldAsync(ToSingleMessage, CancellationToken.None);
    }
#endif

    private ImapMessageInfo ToSingleMessage(IEnumerable<ImapMessageInfo> messages)
    {
      var message = messages.FirstOrDefault();

      if (message == null)
        throw new ImapMessageNotFoundException(sequenceOrUidSet);

      return message;
    }

    private readonly ImapSequenceSet sequenceOrUidSet;
  }

  public abstract class ImapMessageInfoList : ImapMessageInfoBase, IEnumerable<ImapMessageInfo> {
    protected ImapMessageInfoList(ImapOpenedMailboxInfo mailbox,
                                  ImapMessageFetchAttributeOptions fetchOptions,
                                  bool orderBySequenceNumber)
      : base(mailbox)
    {
      this.fetchOptions = fetchOptions;
      this.orderBySequenceNumber = orderBySequenceNumber;
    }

    private const int splitCount = 100;

    internal IEnumerable<ImapMessageInfo> YieldMessages()
    {
#if ASYNC_OPS
      return YieldMessages(CancellationToken.None);
    }

    private IEnumerable<ImapMessageInfo> YieldMessages(CancellationToken cancellationToken)
    {
#endif

#if DEBUG
      Client.ThrowIfNotInExclusiveOperation();
#endif

      CheckAttachedToSelectedMailbox();

      var sequenceOrUidSet = GetSequenceOrUidSet();

      if (sequenceOrUidSet.IsEmpty)
        yield break;

      Mailbox.CheckUidValidity(UidValidity, sequenceOrUidSet);

      bool fetchStaticAttr, fetchDynamicAttr;
      var fetchDataItem = ImapMessageInfo.TranslateFetchOption(Mailbox,
                                                               fetchOptions,
                                                               out fetchStaticAttr,
                                                               out fetchDynamicAttr);

      foreach (var fetchSet in sequenceOrUidSet.SplitIntoEach(splitCount)) {
        ImapMessageAttribute[] messages;

#if ASYNC_OPS
        cancellationToken.ThrowIfCancellationRequested();
#endif

        Mailbox.ProcessMessageStatusChangedResult(Client.Session.Fetch(fetchSet,
                                                                       fetchDataItem,
                                                                       out messages),
                                                  messages);

        if (orderBySequenceNumber) {
          Array.Sort(messages, ImapMessageInfoCollection.CompareMessageAttributeSequence);

          foreach (var sequence in fetchSet) {
            var message = ImapMessageInfoCollection.FindBySequence(messages, sequence);

            yield return Mailbox.ToMessageInfo(message, fetchStaticAttr, fetchDynamicAttr);
          }
        }
        else {
          foreach (var message in messages) {
            yield return Mailbox.ToMessageInfo(message, fetchStaticAttr, fetchDynamicAttr);
          }
        }
      }
    }

    public IEnumerator<ImapMessageInfo> GetEnumerator()
    {
      return ToList().GetEnumerator();
    }

    System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
    {
      return ToList().GetEnumerator();
    }

    internal TResult Yield<TResult>(Converter<IEnumerable<ImapMessageInfo>, TResult> converter)
    {
      return Client.RunExclusively(() => {
#if ASYNC_OPS
        return converter(YieldMessages(CancellationToken.None));
#else
        return converter(YieldMessages());
#endif
      });
    }

#if ASYNC_OPS
    internal Task<TResult> YieldAsync<TResult>(Converter<IEnumerable<ImapMessageInfo>, TResult> converter,
                                               CancellationToken cancellationToken)
    {
      return Client.StartTaskExclusively(() => {
        return converter(YieldMessages(cancellationToken));
      }, cancellationToken);
    }
#endif

    public List<ImapMessageInfo> ToList()
    {
      return Yield(Enumerable.ToList);
    }

    public ImapMessageInfo[] ToArray()
    {
      return Yield(Enumerable.ToArray);
    }

#if ASYNC_OPS
    public Task<List<ImapMessageInfo>> ToListAsync()
    {
      return YieldAsync(Enumerable.ToList, CancellationToken.None);
    }

    public Task<List<ImapMessageInfo>> ToListAsync(CancellationToken cancellationToken)
    {
      return YieldAsync(Enumerable.ToList, cancellationToken);
    }

    public Task<ImapMessageInfo[]> ToArrayAsync()
    {
      return YieldAsync(Enumerable.ToArray, CancellationToken.None);
    }

    public Task<ImapMessageInfo[]> ToArrayAsync(CancellationToken cancellationToken)
    {
      return YieldAsync(Enumerable.ToArray, cancellationToken);
    }
#endif

    public override string ToString()
    {
      return string.Format("{{ImapMessageInfoList: Mailbox='{0}', UidValidity={1}}}",
                           Mailbox.FullName,
                           UidValidity);
    }

    private readonly ImapMessageFetchAttributeOptions fetchOptions;
    private readonly bool orderBySequenceNumber;
  }
}
