﻿//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
#if NET_4_0
// LazyInitializer
using System.Threading;
#else
using Smdn.Threading;
#endif

using Smdn.Collections;
using Smdn.Formats;
using Smdn.Formats.Mime;
using Smdn.Net.Imap4;

namespace Smdn.Net.Imap4.Client {
  public sealed partial class ImapMessagePartInfo : IImapUrl {
    public ImapClient Client {
      get { return message.Client; }
    }

    public ImapMessageInfo Message {
      get { return message; }
    }

    /*
     * BODYSTRUCTURE
     */
    [Obsolete]
    public IImapBodyStructure BodyStructure {
      get { return structure; }
    }

    public Uri Url {
      get { return structure.Url; }
    }

    public string Specifier {
      get { return structure.Section; }
    }

    public MimeType ContentType {
      get { return structure.MediaType; }
    }

    public bool IsMultiPart {
      get { return structure.IsMultiPart; }
    }

    public bool HasEncapsulatedMessage {
      get { return structure.HasEncapsulatedMessage; }
    }

    public bool IsEncapsulatedMessagePart {
      get {
        if (parent == null)
          return false;
        else
          return parent.HasEncapsulatedMessage;
      }
    }

    public ImapMessagePartInfo ParentPart {
      get { return parent; }
    }

    public bool IsRootPart {
      get { return parent == null; }
    }

    public int NestLevel {
      get {
        if (parent == null)
          return 0;
        else
          return parent.NestLevel + 1;
      }
    }

    /*
     * BODYSTRUCTURE extensions
     */
    public string ContentDisposition {
      get {
        var extension = structure as IImapBodyStructureExtension;

        if (extension == null)
          return null;
        else if (extension.Disposition == null)
          return null;
        else
          return extension.Disposition.Type;
      }
    }

    public bool IsDispositionAttachment {
      get { return string.Equals(ContentDisposition, "attachment", StringComparison.OrdinalIgnoreCase); }
    }

    public bool IsDispositionInline {
      get { return string.Equals(ContentDisposition, "inline", StringComparison.OrdinalIgnoreCase); }
    }

#if NET_4_5
    public IReadOnlyDictionary<string, string>
#else
    public IDictionary<string, string>
#endif
    ContentDispositionParameters {
      get {
        var extension = structure as IImapBodyStructureExtension;

        if (extension == null)
          return ReadOnlyDictionary<string, string>.Empty;
        else if (extension.Disposition == null)
          return ReadOnlyDictionary<string, string>.Empty;
        else
          return extension.Disposition.Parameters;
      }
    }

#if NET_4_5
    public IReadOnlyList<string>
#else
    public IList<string>
#endif
    ContentLanguages
    {
      get {
        var extension = structure as IImapBodyStructureExtension;

        if (extension == null)
          return EmptyArray<string>.Instance;
        else
          return extension.Languages;
      }
    }

    public Uri ContentLocation {
      get {
        var extension = structure as IImapBodyStructureExtension;

        if (extension == null)
          return null;
        else
          return extension.Location;
      }
    }

    /*
     * single part or message/rfc822
     */
#if NET_4_5
    public IReadOnlyDictionary<string, string>
#else
    public IDictionary<string, string>
#endif
    ContentTypeParameters {
      get {
        if (structure.IsMultiPart) {
          var extendedMultipart = (structure as ImapExtendedMultiPartBodyStructure);

          if (extendedMultipart == null)
            return ReadOnlyDictionary<string, string>.Empty;
          else
            return extendedMultipart.Parameters;
        }
        else {
          return (structure as ImapSinglePartBodyStructure).Parameters;
        }
      }
    }

    public string ContentTypeCharset {
      get {
        string charset;

        if (ContentTypeParameters.TryGetValue("charset", out charset))
          return charset;
        else
          return null;
      }
    }

    public string ContentID {
      get {
        if (structure.IsMultiPart)
          return null;
        else
          return (structure as ImapSinglePartBodyStructure).Id;
      }
    }

    public string ContentDescription {
      get {
        if (structure.IsMultiPart)
          return null;
        else
          return (structure as ImapSinglePartBodyStructure).Description;
      }
    }

    public string ContentTransferEncoding {
      get {
        if (structure.IsMultiPart)
          return null;
        else
          return (structure as ImapSinglePartBodyStructure).Encoding;
      }
    }

    public ContentTransferEncodingMethod ContentTransferEncodingMethod {
      get {
        var encoding = ContentTransferEncoding;

        if (string.IsNullOrEmpty(encoding))
          return ContentTransferEncodingMethod.Unknown;
        else
          return Smdn.Formats.Mime.ContentTransferEncoding.GetEncodingMethod(encoding);
      }
    }

    public string ContentMD5 {
      get {
        if (structure.IsMultiPart) {
          return null;
        }
        else if (structure.HasEncapsulatedMessage) {
          var extended = structure as ImapExtendedMessageRfc822BodyStructure;

          if (extended == null)
            return null;
          else
            return extended.MD5;
        }
        else {
          var extended = structure as ImapExtendedSinglePartBodyStructure;

          if (extended == null)
            return null;
          else
            return extended.MD5;
        }
      }
    }

    public long Length {
      get {
        if (structure.IsMultiPart)
          return 0L;
        else
          return (structure as ImapSinglePartBodyStructure).Size;
      }
    }

    /// <summary>
    /// gets the size of the body in text lines, or the size in text lines of the encapsulated message
    /// </summary>
    public long LineCount {
      get {
        if (structure.IsMultiPart)
          return 0L;
        else
          return (structure as ImapSinglePartBodyStructure).LineCount;
      }
    }

    private ImapMessagePartInfo encapsulatedMessagePart = null;

    public ImapMessagePartInfo EncapsulatedMessagePart {
      get {
        if (structure.HasEncapsulatedMessage) {
          return LazyInitializer.EnsureInitialized(ref encapsulatedMessagePart, delegate {
            return new ImapMessagePartInfo(this, (structure as ImapMessageRfc822BodyStructure).BodyStructure);
          });
        }
        else {
          return null;
        }
      }
    }

    public ImapEnvelope EncapsulatedMessageEnvelope {
      get {
        if (structure.HasEncapsulatedMessage)
          return (structure as ImapMessageRfc822BodyStructure).Envelope;
        else
          return null;
      }
    }

    /*
     * multipart
     */
#if NET_4_5
    private IReadOnlyList<ImapMessagePartInfo> nestedMessageParts = null;

    public IReadOnlyList<ImapMessagePartInfo>
#else
    private IList<ImapMessagePartInfo> nestedMessgaeParts = null;

    public IList<ImapMessagePartInfo>
#endif
    NestedMessageParts
    {
      get {
        if (IsMultiPart) {
#if NET_4_5
          return LazyInitializer.EnsureInitialized(ref nestedMessageParts, delegate {
            return (structure as ImapMultiPartBodyStructure).NestedStructures.Select(s => new ImapMessagePartInfo(this, s)).ToList();
          });
#else
          return LazyInitializer.EnsureInitialized(ref nestedMessgaeParts, delegate {
            var ret = (structure as ImapMultiPartBodyStructure).NestedStructures.Select(s => new ImapMessagePartInfo(this, s)).ToList();

            return ret.AsReadOnly();
          });
#endif
        }
        else {
          return EmptyArray<ImapMessagePartInfo>.Instance;
        }
      }
    }

    private readonly ImapMessageInfo message;
    private readonly IImapBodyStructure structure;
    private readonly ImapMessagePartInfo parent;

    internal ImapMessagePartInfo(ImapMessageInfo message)
    {
#if DEBUG
      if (message == null)
        throw new ArgumentNullException("message");
#endif

      this.message = message;
      this.structure = message.StaticAttribute.BodyStructure;
      this.parent = null;
    }

    private ImapMessagePartInfo(ImapMessagePartInfo parent, IImapBodyStructure structure)
    {
#if DEBUG
      if (parent == null)
        throw new ArgumentNullException("parent");
      if (structure == null)
        throw new ArgumentNullException("structure");
#endif

      this.message = parent.message;
      this.structure = structure;
      this.parent = parent;
    }

    /*
    * Traverse
    */
    public void Traverse(Action<ImapMessagePartInfo> visitPart)
    {
      Traverse(false, visitPart);
    }

    public void Traverse(bool visitItself, Action<ImapMessagePartInfo> visitPart)
    {
      Traverse(visitItself ? null : this, visitPart, null);
    }

    public void Traverse(Action<ImapMessagePartInfo> visitInToPart,
                         Action<ImapMessagePartInfo> visitOutFromPart)
    {
      Traverse(false, visitInToPart, visitOutFromPart);
    }

    public void Traverse(bool visitItself,
                         Action<ImapMessagePartInfo> visitInToPart,
                         Action<ImapMessagePartInfo> visitOutFromPart)
    {
      Traverse(visitItself ? null : this, visitInToPart, visitOutFromPart);
    }

    private void Traverse(ImapMessagePartInfo rootPart,
                          Action<ImapMessagePartInfo> visitInToPart,
                          Action<ImapMessagePartInfo> visitOutFromPart)
    {
      if (visitInToPart != null && this != rootPart)
        visitInToPart(this);

      foreach (var nestedMessagePart in NestedMessageParts) {
        nestedMessagePart.Traverse(rootPart, visitInToPart, visitOutFromPart);
      }

      if (HasEncapsulatedMessage)
        EncapsulatedMessagePart.Traverse(rootPart, visitInToPart, visitOutFromPart);

      if (visitOutFromPart != null && this != rootPart)
        visitOutFromPart(this);
    }

    /*
     * Flatten()
     */
    public List<ImapMessagePartInfo> Flatten()
    {
      return Flatten(false);
    }

    public List<ImapMessagePartInfo> Flatten(bool containItself)
    {
      var ret = new List<ImapMessagePartInfo>();

      Traverse(containItself, delegate(ImapMessagePartInfo part) {
        ret.Add(part);
      });

      return ret;
    }

    /*
     * Find()/FindAll()
     */
    public List<ImapMessagePartInfo> FindAll(Predicate<ImapMessagePartInfo> match)
    {
      if (match == null)
        throw new ArgumentNullException("match");

      return Flatten(false).FindAll(match);
    }

    public ImapMessagePartInfo Find(Predicate<ImapMessagePartInfo> match)
    {
      return FindAll(match).FirstOrDefault();
    }

    public static class Predicates {
      public static bool IsMultiPart(ImapMessagePartInfo part)
      {
        return part.IsMultiPart;
      }

      public static bool HasEncapsulatedMessagePart(ImapMessagePartInfo part)
      {
        return part.HasEncapsulatedMessage;
      }

      public static bool IsEncapsulatedMessagePart(ImapMessagePartInfo part)
      {
        return part.IsEncapsulatedMessagePart;
      }

      public static bool IsTextPart(ImapMessagePartInfo part)
      {
        return part.ContentType.TypeEqualsIgnoreCase("text");
      }

      public static bool IsPlainTextPart(ImapMessagePartInfo part)
      {
        return part.ContentType.EqualsIgnoreCase(MimeType.TextPlain);
      }

      public static bool IsImagePart(ImapMessagePartInfo part)
      {
        return part.ContentType.TypeEqualsIgnoreCase("image");
      }

      public static bool IsAttachmentPart(ImapMessagePartInfo part)
      {
        return part.IsDispositionAttachment;
      }

      public static bool IsInlinePart(ImapMessagePartInfo part)
      {
        return part.IsDispositionInline;
      }

      public static bool SeemsToBeAttachmentPart(ImapMessagePartInfo part)
      {
        if (part.IsDispositionAttachment)
          return true;
        if (part.IsDispositionInline)
          return false;

        if (part.IsMultiPart)
          return false;
        if (part.IsRootPart)
          return false;

        if (MimeType.MessagePartial.EqualsIgnoreCase(part.ContentType))
          // partial message
          return false;

        if (MimeType.MultipartAlternative.EqualsIgnoreCase(part.parent.ContentType))
          // alternative content part
          return false;

        if (part.parent.IsRootPart &&
            MimeType.MultipartMixed.EqualsIgnoreCase(part.parent.ContentType) &&
            (part.ContentType.EqualsIgnoreCase("text/plain") || part.ContentType.EqualsIgnoreCase("text/html")))
          // text part under root part
          return false;

        return true;
      }
    }

    /*
     * utility methods
     */
    public string GetContentFileName()
    {
      return GetContentFileName(null, false);
    }

    public string GetContentFileName(EncodingSelectionCallback selectFallbackEncoding,
                                     bool throwIfMalformedOrUnsupported)
    {
      // try get filename from Content-Disposition filename parameter
      var ret = MimeUtils.GetDecodedParameterValue("filename",
                                                   ContentDispositionParameters,
                                                   selectFallbackEncoding,
                                                   throwIfMalformedOrUnsupported);

      // try get filename from Content-Type name parameter
      return ret ?? MimeUtils.GetDecodedParameterValue("name",
                                                       ContentTypeParameters,
                                                       selectFallbackEncoding,
                                                       throwIfMalformedOrUnsupported);
    }

    /// <summary>Infrastructure. It is not intended to be used directly from your code.</summary>
    void IImapUrl.SetBaseUrl(ImapUriBuilder baseUrl)
    {
      throw new NotSupportedException();
    }

    public override string ToString()
    {
      return string.Format("{{ImapMessagePartInfo: Authority='{0}', Mailbox='{1}', Uid={2}, UidValidity={3}, Specifier={4}}}",
                           ImapStyleUriParser.GetStrongAuthority(Url),
                           Message.Mailbox.FullName,
                           Message.Uid,
                           Message.UidValidity,
                           Specifier);
    }
  }
}

