using System;
using System.Collections.Generic;
using System.Text;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client.Session {
  [TestFixture]
  public class ImapSessionCommandsAnyStateTests : ImapSessionTestsBase {
    [Test]
    public void TestCapability()
    {
      Connect(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsNotNull(session.ServerCapabilities);
        Assert.AreEqual(0, session.ServerCapabilities.Count);

        server.EnqueueResponse("* CAPABILITY IMAP4rev1 CHILDREN\r\n" +
                               "0000 OK done\r\n");

        ImapCapabilitySet capabilities;

        Assert.IsTrue((bool)session.Capability(out capabilities));

        Assert.AreEqual("0000 CAPABILITY\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(capabilities);
        Assert.AreEqual(2, capabilities.Count);
        Assert.IsTrue(capabilities.Contains(ImapCapability.Imap4Rev1));
        Assert.IsTrue(capabilities.Contains(ImapCapability.Children));

        Assert.IsNotNull(session.ServerCapabilities);
        Assert.AreEqual(2, session.ServerCapabilities.Count);
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Children));

        try {
          session.ServerCapabilities.Add(ImapCapability.Imap4);
          Assert.Fail("NotSupportedException not thrown");
        }
        catch (NotSupportedException) {
        }
      });
    }

    [Test]
    public void TestCapability_NotHandleIncapableAsException()
    {
      Connect(delegate(ImapSession session, ImapPseudoServer server) {
        session.HandlesIncapableAsException = true;

        Assert.IsNotNull(session.ServerCapabilities);
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));

        // CAPABILITY transaction 1
        server.EnqueueResponse("* CAPABILITY IMAP4\r\n" +
                               "0000 OK done\r\n");

        Assert.IsTrue((bool)session.Capability());

        Assert.AreEqual("0000 CAPABILITY\r\n",
                        server.DequeueRequest());

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));

        // CAPABILITY transaction 2
        server.EnqueueResponse("* CAPABILITY IMAP4\r\n" +
                               "0001 OK done\r\n");

        Assert.IsTrue((bool)session.Capability());

        Assert.AreEqual("0001 CAPABILITY\r\n",
                        server.DequeueRequest());

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
      });
    }

    [Test]
    public void TestNoOp_InNonAuthenticatedState()
    {
      Connect(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);

        // NOOP transaction
        server.EnqueueResponse("0000 OK done\r\n");

        Assert.IsTrue((bool)session.NoOp());

        Assert.AreEqual("0000 NOOP\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestNoOp_InAuthenticatedState()
    {
      Authenticate(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);

        // NOOP transaction
        server.EnqueueResponse("0002 OK done\r\n");

        Assert.IsTrue((bool)session.NoOp());

        Assert.AreEqual("0002 NOOP\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestNoOp_InSelectedState()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.AreEqual(ImapSessionState.Selected, session.State);

        // NOOP transaction
        server.EnqueueResponse("0004 OK NOOP completed\r\n");

        Assert.IsTrue((bool)session.NoOp());

        Assert.AreEqual("0004 NOOP\r\n",
                        server.DequeueRequest());

        return 1;
      });
    }

    [Test]
    public void TestNoOp_StatusUpdate()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // NOOP transaction
        server.EnqueueResponse("* 22 EXPUNGE\r\n" +
                               "* 18 EXPUNGE\r\n" +
                               "* 23 EXISTS\r\n" +
                               "* 3 RECENT\r\n" +
                               "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                               "* 14 FETCH (FLAGS (\\Seen \\Deleted))\r\n" +
                               "0004 OK NOOP completed\r\n");

        Assert.IsTrue((bool)session.NoOp());

        Assert.AreEqual("0004 NOOP\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(23L, session.SelectedMailbox.ExistsMessage, "selected mailbox exist message count");
        Assert.AreEqual(3L, session.SelectedMailbox.RecentMessage, "selected mailbox recent message count");
        Assert.IsNotNull(session.SelectedMailbox.ApplicableFlags, "selected mailbox applicable flags");
        Assert.AreEqual(5, session.SelectedMailbox.ApplicableFlags.Count);
        Assert.IsTrue(session.SelectedMailbox.ApplicableFlags.Contains(ImapMessageFlag.Answered));
        Assert.IsTrue(session.SelectedMailbox.ApplicableFlags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(session.SelectedMailbox.ApplicableFlags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(session.SelectedMailbox.ApplicableFlags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(session.SelectedMailbox.ApplicableFlags.Contains(ImapMessageFlag.Draft));

        return 1;
      });
    }

    [Test]
    public void TestNoOp_StatusUpdate_ExistsBroken()
    {
      SelectMailbox(1, delegate(ImapSession session, ImapPseudoServer server) {
        Assert.AreEqual(1L, session.SelectedMailbox.ExistsMessage);

        // NOOP transaction
        server.EnqueueResponse("* 1 EXPUNGE\r\n" +
                               "* 1 EXPUNGE\r\n" +
                               "* 1 EXPUNGE\r\n" +
                               "0004 OK NOOP completed\r\n");

        Assert.IsTrue((bool)session.NoOp());

        Assert.AreEqual("0004 NOOP\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(0L, session.SelectedMailbox.ExistsMessage);

        return 1;
      });
    }

    [Test]
    public void TestLogout_InNonAuthenticatedState()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.EnqueueTaggedResponse("LOGOUT\r\n",
                                     "* BYE Logging out\r\n" + 
                                     "$tag OK logged out.\r\n",
                                     true);

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.IsTrue(session.Logout().Code == ImapCommandResultCode.Bye);

          Assert.AreEqual(ImapSessionState.NotConnected, session.State);
          Assert.AreEqual(null, session.Authority);
        }
      }
    }

    [Test]
    public void TestLogout_InAuthenticatedState()
    {
      Authenticate(delegate(ImapSession session, ImapPseudoServer server) {
        server.EnqueueResponse("* BYE Logging out\r\n" + 
                               "0001 OK logged out.\r\n");

        Assert.IsTrue(session.Logout().Code == ImapCommandResultCode.Bye);

        Assert.AreEqual("0002 LOGOUT\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        Assert.AreEqual(null, session.Authority);
      });
    }

    [Test]
    public void TestLogout_InSelectedState()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsNotNull(session.SelectedMailbox);

        server.EnqueueResponse("* BYE Logging out\r\n" + 
                               "0004 OK logged out.\r\n");

        Assert.IsTrue(session.Logout().Code == ImapCommandResultCode.Bye);

        Assert.AreEqual("0004 LOGOUT\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        Assert.AreEqual(null, session.Authority);
        Assert.IsNull(session.SelectedMailbox);

        return -1;
      });
    }

    [Test]
    public void TestID()
    {
      Authenticate(new[] {"ID"}, delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ID));

        // ID transaction
        server.EnqueueResponse("* ID NIL\r\n" +
                               "0002 OK ID completed\r\n");

#if NET_4_5
        IReadOnlyDictionary<string, string> serverParameterList;
#else
        IDictionary<string, string> serverParameterList;
#endif

        Assert.IsTrue((bool)session.ID(new Dictionary<string, string>() {
          {"name", "sodr"},
          {"version", "19.34"},
          {"vendor", "Pink Floyd Music Limited"},
        }, out serverParameterList));

        Assert.AreEqual("0002 ID (\"name\" \"sodr\" \"version\" \"19.34\" \"vendor\" \"Pink Floyd Music Limited\")\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(serverParameterList);
        Assert.AreEqual(0, serverParameterList.Count);
        Assert.AreEqual(0, session.ServerID.Count);

#if !NET_4_5
        Assert.IsTrue(serverParameterList.IsReadOnly, "returns read-only dictionary");
        Assert.Throws<NotSupportedException>(() => session.ServerID.Add("key", "value"));
#endif
      });
    }

    [Test]
    public void TestID_ParameterNil()
    {
      Authenticate(new[] {"ID"}, delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ID));

        // ID transaction
        server.EnqueueResponse("* ID (\"name\" \"Cyrus\" \"version\" \"1.5\")\r\n" +
                               "0002 OK ID completed\r\n");

        Assert.IsTrue((bool)session.ID(new Dictionary<string, string>() {
          {"name", null},
          {"version", null},
          {"vendor", null},
        }));

        Assert.AreEqual("0002 ID (\"name\" NIL \"version\" NIL \"vendor\" NIL)\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(2, session.ServerID.Count);
        Assert.AreEqual("Cyrus", session.ServerID["name"]);
        Assert.AreEqual("Cyrus", session.ServerID["NAME"]);
        Assert.AreEqual("1.5", session.ServerID["version"]);
        Assert.AreEqual("1.5", session.ServerID["VERSION"]);

#if !NET_4_5
        Assert.IsTrue(session.ServerID.IsReadOnly, "dictionary is read-only");
        Assert.Throws<NotSupportedException>(() => session.ServerID.Add("key", "value"));
#endif
      });
    }

    [Test]
    public void TestID_TooMatchFieldValuePair()
    {
      Authenticate(new[] {"ID"}, delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ID));

        var paramList = new Dictionary<string, string>();

        for (var i = 0; i < 31; i++) {
          paramList.Add(string.Format("field{0}", i), string.Format("value{0}", i));
        }

        Assert.Throws<ArgumentException>(() => session.ID(paramList));
      });
    }

    [Test]
    public void TestID_TooLongField()
    {
      Authenticate(new[] {"ID"}, delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ID));

        Assert.Throws<ArgumentException>(() => session.ID(new Dictionary<string, string>() {
          {new string('x', 31), "value"},
        }));
      });
    }

    [Test]
    public void TestID_TooLongValue()
    {
      Authenticate(new[] {"ID"}, delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ID));

        Assert.Throws<ArgumentException>(() => session.ID(new Dictionary<string, string>() {
          {"field", new string('x', 1025)},
        }));
      });
    }

    [Test]
    public void TestID_ContaisNoAsciiChar()
    {
      Authenticate(new[] {"ID"}, delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.ID));

        Assert.Throws<ArgumentException>(() => session.ID(new Dictionary<string, string>() {
          {"オワタ", "＼(＾o＾)／"},
        }));
      });
    }

    [Test]
    public void TestLanguage_Incapable()
    {
      Authenticate(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Language));

        session.HandlesIncapableAsException = true;

        Assert.Throws<ImapIncapableException>(() => session.Language());
      });
    }

    [Test]
    public void TestLanguage_ListSupported()
    {
      Authenticate(new[] {"LANGUAGE"}, delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        // LANGUAGE transaction
        server.EnqueueResponse("* LANGUAGE (EN DE IT i-default)\r\n" +
                               "0002 OK Supported languages have been enumerated\r\n");

        string[] supportedLanguageTags;

        Assert.IsTrue((bool)session.Language(out supportedLanguageTags));

        Assert.AreEqual("0002 LANGUAGE\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(4, supportedLanguageTags.Length);
        Assert.AreEqual("EN", supportedLanguageTags[0]);
        Assert.AreEqual("DE", supportedLanguageTags[1]);
        Assert.AreEqual("IT", supportedLanguageTags[2]);
        Assert.AreEqual("i-default", supportedLanguageTags[3]);
      });
    }

    [Test]
    public void TestLanguage_SelectDefault()
    {
      Authenticate2(new[] {"LANGUAGE"}, delegate(ImapSession session, ImapPseudoServer2 server) {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        server.EnqueueTaggedResponse("LANGUAGE default\r\n",
                                     "* LANGUAGE (DE)\r\n" +
                                     "$tag OK Sprachwechsel durch LANGUAGE-Befehl ausgeführt\r\n",
                                     () => server.SetCharset(Encoding.UTF8),
                                     null);

        string selectedLanguageTag;

        var result = session.Language(out selectedLanguageTag);

        Assert.IsTrue((bool)result);
        Assert.AreEqual("DE", selectedLanguageTag);
        Assert.AreEqual("Sprachwechsel durch LANGUAGE-Befehl ausgeführt", result.TaggedStatusResponse.ResponseText.Text);

        Assert.AreEqual("DE", session.SelectedLanguage);
      });
    }

    [Test]
    public void TestLanguage_SelectSupported()
    {
      Authenticate(new[] {"LANGUAGE"}, delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        // LANGUAGE transaction
        server.EnqueueResponse("* LANGUAGE (EN)\r\n" +
                               "0002 OK Now speaking English\r\n");

        string selectedLanguageTag;

        var result = session.Language(out selectedLanguageTag, "FR-CA", "EN-CA");

        Assert.IsTrue((bool)result);
        Assert.AreEqual("EN", selectedLanguageTag);
        Assert.AreEqual("Now speaking English", result.TaggedStatusResponse.ResponseText.Text);

        Assert.AreEqual("0002 LANGUAGE FR-CA EN-CA\r\n",
                        server.DequeueRequest());

        Assert.AreEqual("EN", session.SelectedLanguage);
      });
    }

    [Test]
    public void TestLanguage_SelectWithCultureInfo()
    {
      Authenticate(new[] {"LANGUAGE"}, delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        // LANGUAGE transaction
        server.EnqueueResponse("0002 NO Unsupported language\r\n");

        var result = session.Language(System.Globalization.CultureInfo.GetCultureInfo("ja-JP"));

        Assert.IsFalse((bool)result);
        Assert.AreEqual("Unsupported language", result.TaggedStatusResponse.ResponseText.Text);

        Assert.AreEqual("0002 LANGUAGE ja-JP\r\n",
                        server.DequeueRequest());

        Assert.IsNull(session.SelectedLanguage);
      });
    }

    [Test]
    public void TestLanguage_NamespaceResponse()
    {
      Authenticate2(new[] {"LANGUAGE"}, delegate(ImapSession session, ImapPseudoServer2 server) {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        server.EnqueueTaggedResponse("LANGUAGE DE",
                                     "* LANGUAGE (DE)\r\n" +
                                     "* NAMESPACE ((\"\" \"/\")) ((\"Other Users/\" \"/\" \"TRANSLATION\"" +
                                     " (\"Andere Ben&APw-tzer/\"))) ((\"Public Folders/\" \"/\" " +
                                     " \"TRANSLATION\" (\"Gemeinsame Postf&AM8-cher/\")))\r\n" +
                                     "$tag OK Sprachwechsel durch LANGUAGE-Befehl ausgeführt\r\n",
                                     () => server.SetCharset(Encoding.UTF8),
                                     null);

        var result = session.Language("DE");

        Assert.IsTrue((bool)result);
        Assert.AreEqual("Sprachwechsel durch LANGUAGE-Befehl ausgeführt", result.TaggedStatusResponse.ResponseText.Text);

        Assert.AreEqual("DE", session.SelectedLanguage);

        Assert.AreEqual(1, session.Namespaces.PersonalNamespaces.Count);
        Assert.AreEqual(string.Empty, session.Namespaces.PersonalNamespaces[0].Prefix);
        Assert.AreEqual("/", session.Namespaces.PersonalNamespaces[0].HierarchyDelimiter);

        Assert.AreEqual(1, session.Namespaces.OtherUsersNamespaces.Count);
        Assert.AreEqual("Other Users/", session.Namespaces.OtherUsersNamespaces[0].Prefix);
        Assert.AreEqual("/", session.Namespaces.OtherUsersNamespaces[0].HierarchyDelimiter);
        Assert.AreEqual("Andere Benützer/", session.Namespaces.OtherUsersNamespaces[0].TranslatedPrefix);

        Assert.AreEqual(1, session.Namespaces.SharedNamespaces.Count);
        Assert.AreEqual("Public Folders/", session.Namespaces.SharedNamespaces[0].Prefix);
        Assert.AreEqual("/", session.Namespaces.SharedNamespaces[0].HierarchyDelimiter);
        // RFC error? (Gemeinsame Postfächer)
        Assert.AreEqual("Gemeinsame PostfÏcher/", session.Namespaces.SharedNamespaces[0].TranslatedPrefix);
      });
    }

    [Test]
    public void TestLanguage_ReceiveResponseAsUTF8()
    {
      Authenticate2(new[] {"LANGUAGE"}, delegate(ImapSession session, ImapPseudoServer2 server) {
        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Language));
        Assert.IsNull(session.SelectedLanguage);

        server.EnqueueTaggedResponse("LANGUAGE JA\r\n",
                                     "* LANGUAGE (JA)\r\n" +
                                     "$tag OK 完了\r\n",
                                     () => server.SetCharset(Encoding.UTF8),
                                     null);

        var result = session.Language("JA");

        Assert.IsTrue((bool)result);
        Assert.AreEqual("完了", result.TaggedStatusResponse.ResponseText.Text);

        // following command/response
        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK 完了\r\n");

        session.NoOp();

        Assert.IsTrue((bool)result);
        Assert.AreEqual("完了", result.TaggedStatusResponse.ResponseText.Text);
      });
    }
  }
}
