#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Collections.Generic;
using System.Linq;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMailboxInfoTests {
    [Test]
    public void TestRefresh()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             "INBOX",
                             Refresh);
    }

    [Test]
    public void TestRefresh_CondStoreDisabled()
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.CondStore},
                             new[] {ImapMailboxFlag.NoInferiors},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.CondStore);

        Refresh(server, mailbox);
      });
    }

    private void Refresh(ImapPseudoServer2 server, ImapMailboxInfo mailbox)
    {
      var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
        server.EnqueueTaggedResponse("STATUS INBOX (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n",
                                     "* STATUS INBOX ()\r\n" +
                                     "$tag OK done\r\n");
      });

      mailbox.Refresh();

      responseTask.Wait();

      Assert.IsFalse(mailbox.Client.IsBusy);
    }

#if ASYNC_OPS
    [Test]
    public void TestRefreshAsync()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("STATUS INBOX (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n",
                                     "* STATUS INBOX ()\r\n" +
                                     "$tag OK done\r\n");
        });

        var t = mailbox.RefreshAsync();

        responseTask.Wait();

        t.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestRefresh_ReRetrieveMailboxFlags()
    {
      Refresh_ReRetrieveMailboxFlags(mailbox => mailbox.Refresh(true));
    }

#if ASYNC_OPS
    [Test]
    public void TestRefreshAsync_ReRetrieveMailboxFlags()
    {
      Refresh_ReRetrieveMailboxFlags(mailbox => mailbox.RefreshAsync(true).Wait());
    }
#endif

    private void Refresh_ReRetrieveMailboxFlags(Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] { ImapMailboxFlag.NoInferiors },
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX",
                                       "* LIST (\\Marked \\NoInferiors) \"/\" \"INBOX\"\r\n" +
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("STATUS INBOX (",
                                       "* STATUS INBOX ()\r\n" +
                                       "$tag OK done\r\n");
        });

        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.NoInferiors);
        CollectionAssert.DoesNotContain(mailbox.Flags, ImapMailboxFlag.Marked);

        testAction(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.NoInferiors);
        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.Marked);
      });
    }

    [Test]
    public void TestRefresh_ReRetrieveMailboxFlags_MailboxDeleted()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             "INBOX.child",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX.child",
                                       "$tag OK done\r\n");
        });

        Assert.IsTrue(mailbox.Exists);

        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.NoInferiors);

        mailbox.Refresh(true);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.Exists);

        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.NoInferiors);
      });
    }

    [Test]
    public void TestRefresh_ReRetrieveMailboxFlags_MailboxReCreated()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             "INBOX.child",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("DELETE INBOX.child",
                                       "$tag OK done\r\n");
        });

        mailbox.Delete();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.Exists);

        server.EnqueueTaggedResponse("LIST \"\" INBOX.child",
                                     "* LIST (\\NoInferiors) \"/\" \"INBOX.child\"\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("STATUS INBOX.child (",
                                     "* STATUS INBOX.child ()\r\n" +
                                     "$tag OK done\r\n");

        mailbox.Refresh(true);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsTrue(mailbox.Exists);

        CollectionAssert.Contains(mailbox.Flags, ImapMailboxFlag.NoInferiors);
      });
    }

    [Test]
    public void TestRefresh_CondStoreCapable()
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.CondStore},
                             new[] {ImapMailboxFlag.NoInferiors},
                             "INBOX.child",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("STATUS INBOX.child (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN HIGHESTMODSEQ)\r\n",
                                       "* STATUS INBOX.child ()\r\n" +
                                       "$tag OK done\r\n");
        });

        mailbox.Refresh();

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestRefresh_NonExistentMailbox()
    {
      Refresh_NonExistentMailbox(mailbox => mailbox.Refresh());
    }

#if ASYNC_OPS
    [Test]
    public void TestRefreshAsync_NonExistentMailbox()
    {
      Refresh_NonExistentMailbox(mailbox => mailbox.RefreshAsync().Wait());
    }
#endif

    private void Refresh_NonExistentMailbox(Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        // do nothing
        testAction(mailbox);

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestRefresh_UnselectableMailbox()
    {
      Refresh_UnselectableMailbox(mailbox => mailbox.Refresh());
    }

#if ASYNC_OPS
    [Test]
    public void TestRefreshAsync_UnselectableMailbox()
    {
      Refresh_UnselectableMailbox(mailbox => mailbox.RefreshAsync().Wait());
    }
#endif

    private void Refresh_UnselectableMailbox(Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoSelect},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsUnselectable);

        // do nothing
        testAction(mailbox);

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestOpen()
    {
      Open(false, mailbox => mailbox.Open());
    }

    [Test]
    public void TestOpen_AsReadOnly()
    {
      Open(true, mailbox => mailbox.Open(true));
    }

#if ASYNC_OPS
    [Test]
    public void TestOpenAsync()
    {
      Open(false, mailbox => mailbox.OpenAsync().Result);
    }

    [Test]
    public void TestOpenAsync_AsReadOnly()
    {
      Open(true, mailbox => mailbox.OpenAsync(true).Result);
    }
#endif

    private void Open(bool asReadOnly, Func<ImapMailboxInfo, ImapOpenedMailboxInfo> testFunc)
    {
      TestUtils2.TestMailbox("INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse(asReadOnly ? "EXAMINE INBOX\r\n" : "SELECT INBOX\r\n",
                                       "* 172 EXISTS\r\n" +
                                       "* 1 RECENT\r\n" +
                                       "* OK [UNSEEN 12] Message 12 is first unseen\r\n" +
                                       "* OK [UIDVALIDITY 3857529045] UIDs valid\r\n" +
                                       "* OK [UIDNEXT 4392] Predicted next UID\r\n" +
                                       "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft)\r\n" +
                                       "* OK [PERMANENTFLAGS (\\Deleted \\Seen \\*)] Limited\r\n" +
                                       string.Format("$tag OK [{0}] done\r\n", asReadOnly ? "READ-ONLY" : "READ-WRITE"));
        });

        Assert.IsNull(mailbox.Client.OpenedMailbox);
        Assert.IsFalse(mailbox.IsOpen);

        using (var opened = testFunc(mailbox)) {
          responseTask.Wait();

          Assert.IsFalse(mailbox.Client.IsBusy);

          try {
            Assert.IsNotNull(opened);
            Assert.IsTrue(opened.IsOpen);

            Assert.AreEqual(172L, opened.ExistMessageCount);
            Assert.AreEqual(1L, opened.RecentMessageCount);
            Assert.AreEqual(12L, opened.FirstUnseenMessageNumber);
            Assert.AreEqual(4392L, opened.NextUid);

            Assert.AreSame(opened, mailbox.Client.OpenedMailbox);
            Assert.IsTrue(mailbox.IsOpen);
            Assert.AreEqual(opened.ExistMessageCount, mailbox.ExistMessageCount);
            Assert.AreEqual(opened.UnseenMessageCount, mailbox.UnseenMessageCount);
            Assert.AreEqual(opened.NextUid, mailbox.NextUid);

            Assert.IsNotNull(opened.ApplicableFlags);
            Assert.AreEqual(5, opened.ApplicableFlags.Count);

            Assert.IsNotNull(opened.PermanentFlags);
            Assert.AreEqual(3, opened.PermanentFlags.Count);

            Assert.IsTrue(opened.IsAllowedToCreateKeywords);
            Assert.IsTrue(opened.IsUidPersistent);

            if (asReadOnly)
              Assert.IsTrue(opened.IsReadOnly);
            else
              Assert.IsFalse(opened.IsReadOnly);
          }
          finally {
            server.EnqueueTaggedResponse("CLOSE\r\n",
                                         "$tag OK done\r\n");
          }
        }

        Assert.IsNull(mailbox.Client.OpenedMailbox);
        Assert.IsFalse(mailbox.IsOpen);
      });
    }

    [Test]
    public void TestOpen_Unselectable()
    {
      Open_Unselectable(mailbox => mailbox.Open());
    }

    [Test]
    public void TestOpen_Unselectable_AsReadOnly()
    {
      Open_Unselectable(mailbox => mailbox.Open(true));
    }

#if ASYNC_OPS
    [Test]
    public void TestOpenAsync_Unselectable()
    {
      Open_Unselectable(mailbox => mailbox.OpenAsync());
    }

    [Test]
    public void TestOpenAsync_Unselectable_AsReadOnly()
    {
      Open_Unselectable(mailbox => mailbox.OpenAsync(true));
    }
#endif

    private void Open_Unselectable(Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoSelect},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => testAction(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNull(mailbox.Client.OpenedMailbox);
        Assert.IsFalse(mailbox.IsOpen);
      });
    }

    [Test, Ignore("not implemented; reopen")]
    public void TestOpen_ReopenItself()
    {
    }

    [Test, Ignore("not implemented; reopen")]
    public void TestOpen_OpenAnotherNewly()
    {
    }

    [Test]
    public void TestDelete()
    {
      Delete(false, mailbox => mailbox.Delete());
    }

    [Test]
    public void TestDelete_Unsubscribe()
    {
      Delete(true, mailbox => mailbox.Delete(true));
    }

#if ASYNC_OPS
    [Test]
    public void TestDeleteAsync()
    {
      Delete(false, mailbox => mailbox.DeleteAsync().Wait());
    }

    [Test]
    public void TestDeleteAsync_Unsubscribe()
    {
      Delete(true, mailbox => mailbox.DeleteAsync(true).Wait());
    }
#endif

    private void Delete(bool unsubscribe, Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("Trash", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("DELETE Trash\r\n",
                                       "$tag OK done\r\n");

          if (unsubscribe)
            server.EnqueueTaggedResponse("UNSUBSCRIBE Trash\r\n",
                                         "$tag OK done\r\n");
        });

        Assert.IsTrue(mailbox.Exists);

        testAction(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.Exists);
      });
    }

    [Test]
    public void TestDelete_DeleteOpened()
    {
      Delete_DeleteOpened(mailbox => mailbox.Delete());
    }

#if ASYNC_OPS
    [Test]
    public void TestDeleteAsync_DeleteOpened()
    {
      Delete_DeleteOpened(mailbox => mailbox.DeleteAsync().Wait());
    }
#endif

    private void Delete_DeleteOpened(Action<ImapOpenedMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("Trash", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SELECT Trash\r\n",
                                       "$tag OK done\r\n");
        });

        var opened = mailbox.Open();

        Assert.AreSame(opened, mailbox.Client.OpenedMailbox);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.IsTrue(opened.IsOpen);

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("DELETE Trash\r\n",
                                     "$tag OK done\r\n");

        Assert.IsTrue(opened.Exists);

        testAction(opened);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNull(mailbox.Client.OpenedMailbox);
        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(opened.IsOpen);
        Assert.IsFalse(opened.Exists);
      });
    }

    [Test]
    public void TestDelete_MailboxNonExistent()
    {
      Delete_MailboxNonExistent(mailbox => mailbox.Delete(true));
    }

#if ASYNC_OPS
    [Test]
    public void TestDeleteAsync_MailboxNonExistent()
    {
      Delete_MailboxNonExistent(mailbox => mailbox.DeleteAsync(true).Wait());
    }
#endif

    private void Delete_MailboxNonExistent(Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "Trash",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("UNSUBSCRIBE Trash\r\n",
                                       "$tag OK done\r\n");
        });

        testAction(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.Exists);
      });
    }

    [Test]
    public void TestDelete_NonExistent()
    {
      DeleteNonExistent(false);
    }

    [Test]
    public void TestDelete_NonExistent_Unsubscribe()
    {
      DeleteNonExistent(true);
    }

    private void DeleteNonExistent(bool unsubscribe)
    {
      TestUtils2.TestMailbox("Trash",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.Exists);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          // someone deletes 'Trash'
          // C2: TAG CREATE "Trash"
          // S2: TAG OK deleted "Trash"

          server.EnqueueTaggedResponse("DELETE Trash\r\n",
                                       "$tag NO [NONEXISTENT] done\r\n");

          if (unsubscribe)
            server.EnqueueTaggedResponse("UNSUBSCRIBE Trash\r\n",
                                         "$tag OK done\r\n");
        });

        mailbox.Delete(unsubscribe);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsFalse(mailbox.Exists);
      });
    }

    [Test]
    public void TestCreate_RecreateDeleted()
    {
      CreateRecreateDeleted(false, mailbox => mailbox.Create());
    }

    [Test]
    public void TestCreate_RecreateDeleted_Subscribe()
    {
      CreateRecreateDeleted(true, mailbox => mailbox.Create(true));
    }

#if ASYNC_OPS
    [Test]
    public void TestCreateAsync_RecreateDeleted()
    {
      CreateRecreateDeleted(false, mailbox => mailbox.CreateAsync().Wait());
    }

    [Test]
    public void TestCreateAsync_RecreateDeleted_Subscribe()
    {
      CreateRecreateDeleted(true, mailbox => mailbox.CreateAsync(true).Wait());
    }
#endif

    private void CreateRecreateDeleted(bool subscribe, Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("Trash", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.Exists);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("DELETE Trash\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UNSUBSCRIBE Trash\r\n",
                                       "$tag OK done\r\n");
        });

        mailbox.Delete(true);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Exists);

        server.EnqueueTaggedResponse("CREATE Trash\r\n",
                                     "$tag OK done\r\n");

        if (subscribe)
          server.EnqueueTaggedResponse("SUBSCRIBE Trash\r\n",
                                       "$tag OK done\r\n");

        testAction(mailbox);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsTrue(mailbox.Exists);
      });
    }

    [Test]
    public void TestCreate_NonExistent()
    {
      CreateNonExistent(false, mailbox => mailbox.Create());
    }

    [Test]
    public void TestCreate_NonExistent_Subscribe()
    {
      CreateNonExistent(true, mailbox => mailbox.Create(true));
    }

#if ASYNC_OPS
    [Test]
    public void TestCreateAsync_NonExistent()
    {
      CreateNonExistent(false, mailbox => mailbox.CreateAsync().Wait());
    }

    [Test]
    public void TestCreateAsync_NonExistent_Subscribe()
    {
      CreateNonExistent(true, mailbox => mailbox.CreateAsync(true).Wait());
    }
#endif

    private void CreateNonExistent(bool subscribe, Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "deleted",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("CREATE deleted\r\n",
                                       "$tag OK done\r\n");

          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE deleted\r\n",
                                         "$tag OK done\r\n");
        });

        testAction(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsTrue(mailbox.Exists);
      });
    }

    [Test]
    public void TestCreate_AlreadyExists()
    {
      CreateAlreadyExists(false, mailbox => mailbox.Create());
    }

    [Test]
    public void TestCreate_AlreadyExists_Subscribe()
    {
      CreateAlreadyExists(true, mailbox => mailbox.Create(true));
    }

#if ASYNC_OPS
    [Test]
    public void TestCreateAsync_AlreadyExists()
    {
      CreateAlreadyExists(false, mailbox => mailbox.CreateAsync().Wait());
    }

    [Test]
    public void TestCreateAsync_AlreadyExists_Subscribe()
    {
      CreateAlreadyExists(true, mailbox => mailbox.CreateAsync(true).Wait());
    }
#endif

    private void CreateAlreadyExists(bool subscribe, Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "Sent",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          // someone creates 'Sent'
          // C2: TAG CREATE "Sent"
          // S2: TAG OK created "Sent"

          server.EnqueueTaggedResponse("CREATE Sent\r\n",
                                       "$tag NO [ALREADYEXISTS] Mailbox 'Sent' already exists\r\n");

          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE Sent\r\n",
                                         "$tag OK done\r\n");
        });

        testAction(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsTrue(mailbox.Exists);
      });
    }

    [Test]
    public void TestMoveTo_MoveToInbox()
    {
      MoveToInbox(mailbox => mailbox.MoveTo("INBOX"));
    }

    [Test]
    public void TestMoveTo_MoveToInbox_Subscribe()
    {
      MoveToInbox(mailbox => mailbox.MoveTo("INBOX", true));
    }

#if ASYNC_OPS
    [Test]
    public void TestMoveToAsync_MoveToInbox()
    {
      MoveToInbox(mailbox => mailbox.MoveToAsync("INBOX"));
    }

    [Test]
    public void TestMoveToAsync_MoveToInbox_Subscribe()
    {
      MoveToInbox(mailbox => mailbox.MoveToAsync("INBOX", true));
    }
#endif

    private void MoveToInbox(Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("/", "Draft", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("Draft", mailbox.Name);
        Assert.AreEqual("Draft", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
        Assert.IsFalse(mailbox.IsInbox);

        Assert.Throws<ImapProtocolViolationException>(() => testAction(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveTo_MoveInbox()
    {
      MoveToMoveInbox(false, (inbox, name) => inbox.MoveTo(name));
    }

    [Test]
    public void TestMoveTo_MoveInbox_Subscribe()
    {
      MoveToMoveInbox(true, (inbox, name) => inbox.MoveTo(name, true));
    }

#if ASYNC_OPS
    [Test]
    public void TestMoveToAsync_MoveInbox()
    {
      MoveToMoveInbox(false, (inbox, name) => inbox.MoveToAsync(name).Wait());
    }

    [Test]
    public void TestMoveToAsync_MoveInbox_Subscribe()
    {
      MoveToMoveInbox(true, (inbox, name) => inbox.MoveToAsync(name, true).Wait());
    }
#endif

    private void MoveToMoveInbox(bool subscribe, Action<ImapMailboxInfo, string> testAction)
    {
      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("INBOX", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
        Assert.IsTrue(mailbox.IsInbox);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("RENAME INBOX INBOX/backup\r\n",
                                       "$tag OK done\r\n");

          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE INBOX/backup\r\n",
                                         "$tag OK done\r\n");
        });

        testAction(mailbox, "INBOX/backup");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("backup", mailbox.Name);
        Assert.AreEqual("INBOX/backup", mailbox.FullName);
        Assert.AreEqual("INBOX", mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox()
    {
      MoveToDestinationMailbox(false, (mailbox, dest) => mailbox.MoveTo(dest));
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_Subscribe()
    {
      MoveToDestinationMailbox(true, (mailbox, dest) => mailbox.MoveTo(dest, true));
    }

#if ASYNC_OPS
    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox()
    {
      MoveToDestinationMailbox(false, (mailbox, dest) => mailbox.MoveToAsync(dest).Wait());
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_Subscribe()
    {
      MoveToDestinationMailbox(true, (mailbox, dest) => mailbox.MoveToAsync(dest, true).Wait());
    }
#endif

    private void MoveToDestinationMailbox(bool subscribe, Action<ImapMailboxInfo, ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("/", "src", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("src", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);

        server.EnqueueTaggedResponse("LIST \"\" dest\r\n",
                                     "* LIST () \"/\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = mailbox.Client.GetMailbox("dest");

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          if (subscribe) {
            server.EnqueueTaggedResponse("LSUB src/ *",
                                         "* LSUB () \"/\" src/sub1\r\n" +
                                         "* LSUB () \"/\" src/sub2/sub3\r\n" +
                                         "* LSUB () \"/\" src/sub4/sub5/sub6\r\n" +
                                         "$tag OK done\r\n");
          }

          server.EnqueueTaggedResponse("RENAME src dest/src\r\n",
                                       "$tag OK done\r\n");

          if (subscribe) {
            server.EnqueueTaggedResponse("UNSUBSCRIBE src\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest/src\r\n",
                                         "$tag OK done\r\n");

            server.EnqueueTaggedResponse("UNSUBSCRIBE src/sub1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest/src/sub1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UNSUBSCRIBE src/sub2/sub3\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest/src/sub2/sub3\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UNSUBSCRIBE src/sub4/sub5/sub6\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest/src/sub4/sub5/sub6\r\n",
                                         "$tag OK done\r\n");
          }
        });

        testAction(mailbox, destMailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("dest/src", mailbox.FullName);
        Assert.AreEqual("dest", mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_DestinationNonHierarchical()
    {
      TestUtils2.TestMailbox(null,
                             null,
                             string.Empty,
                             "dest",
                             MoveToDestinationMailboxDestinationCanNotHaveChild);
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_DestinationNoInferiors()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "dest",
                             MoveToDestinationMailboxDestinationCanNotHaveChild);
    }

    private void MoveToDestinationMailboxDestinationCanNotHaveChild(ImapPseudoServer2 server, ImapMailboxInfo destMailbox)
    {
      server.EnqueueTaggedResponse("LIST \"\" src\r\n",
                                   "* LIST () \"/\" src\r\n" +
                                   "$tag OK done\r\n");

      var sourceMailbox = destMailbox.Client.GetMailbox("src");

      Assert.Throws<ImapProtocolViolationException>(() => sourceMailbox.MoveTo(destMailbox));

      Assert.IsFalse(destMailbox.Client.IsBusy);
    }

#if ASYNC_OPS
    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_DestinationNonHierarchical()
    {
      TestUtils2.TestMailbox(null,
                             null,
                             string.Empty,
                             "dest",
                             MoveToAsyncDestinationMailboxDestinationCanNotHaveChild);
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_DestinationNoInferiors()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "dest",
                             MoveToAsyncDestinationMailboxDestinationCanNotHaveChild);
    }

    private void MoveToAsyncDestinationMailboxDestinationCanNotHaveChild(ImapPseudoServer2 server, ImapMailboxInfo destMailbox)
    {
      server.EnqueueTaggedResponse("LIST \"\" src\r\n",
                                   "* LIST () \"/\" src\r\n" +
                                   "$tag OK done\r\n");

      var sourceMailbox = destMailbox.Client.GetMailbox("src");

      Assert.Throws<ImapProtocolViolationException>(() => sourceMailbox.MoveToAsync(destMailbox));

      Assert.IsFalse(destMailbox.Client.IsBusy);
    }
#endif

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_AlreadyExists()
    {
      MoveToDestinationMailboxAlreadyExists(false, (mailbox, dest) => mailbox.MoveTo(dest));
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_AlreadyExists_Subscribe()
    {
      MoveToDestinationMailboxAlreadyExists(true, (mailbox, dest) => mailbox.MoveTo(dest, true));
    }

#if ASYNC_OPS
    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_AlreadyExists()
    {
      MoveToDestinationMailboxAlreadyExists(false, (mailbox, dest) => mailbox.MoveToAsync(dest).Wait());
    }

    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_AlreadyExists_Subscribe()
    {
      MoveToDestinationMailboxAlreadyExists(true, (mailbox, dest) => mailbox.MoveToAsync(dest, true).Wait());
    }
#endif

    private void MoveToDestinationMailboxAlreadyExists(bool subscribe,
                                                       Action<ImapMailboxInfo, ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("/", "src", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("src", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);

        server.EnqueueTaggedResponse("LIST \"\" dest\r\n",
                                     "* LIST () \"/\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = mailbox.Client.GetMailbox("dest");

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          if (subscribe) {
            server.EnqueueTaggedResponse("LSUB src/ *",
                                         "* LSUB () \"/\" src/sub1\r\n" +
                                         "* LSUB () \"/\" src/sub2/sub3\r\n" +
                                         "$tag OK done\r\n");
          }

          server.EnqueueTaggedResponse("RENAME src dest/src\r\n",
                                       "$tag NO [ALREADYEXISTS] mailbox exists\r\n");
        });

        var ex = TestUtils.Throws<ImapErrorResponseException>(() => testAction(mailbox, destMailbox));

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(ImapResponseCode.AlreadyExists,
                        ex.Result.TaggedStatusResponse.ResponseText.Code);

        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("src", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToDestinationMailbox_MailboxOfDifferentSession()
    {
      MoveTo_MoveToDestinationMailbox_MailboxOfDifferentSession((mailbox, dest) => mailbox.MoveTo(dest));
    }

#if ASYNC_OPS
    [Test]
    public void TestMoveToAsync_MoveToDestinationMailbox_MailboxOfDifferentSession()
    {
      MoveTo_MoveToDestinationMailbox_MailboxOfDifferentSession((mailbox, dest) => mailbox.MoveToAsync(dest).Wait());
    }
#endif

    private void MoveTo_MoveToDestinationMailbox_MailboxOfDifferentSession(Action<ImapMailboxInfo, ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 serverSrc, ImapMailboxInfo mailboxSrc) {
        Assert.AreEqual("INBOX", mailboxSrc.Name);
        Assert.AreEqual("INBOX", mailboxSrc.FullName);
        Assert.AreEqual(string.Empty, mailboxSrc.ParentMailboxName);
        Assert.IsTrue(mailboxSrc.IsInbox);

        TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 serverDest, ImapMailboxInfo mailboxDest) {
          Assert.AreEqual("INBOX", mailboxDest.Name);
          Assert.AreEqual("INBOX", mailboxDest.FullName);
          Assert.AreEqual(string.Empty, mailboxDest.ParentMailboxName);
          Assert.IsTrue(mailboxDest.IsInbox);

          Assert.AreNotSame(mailboxSrc.Client, mailboxDest.Client);

          Assert.Throws<NotSupportedException>(() => testAction(mailboxSrc, mailboxDest));

          Assert.IsFalse(mailboxSrc.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_ContainsWildcard()
    {
      TestUtils2.TestMailbox(".", "Box", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("RENAME Box \"Box*\"\r\n",
                                       "$tag OK done\r\n");
        });

        mailbox.MoveTo("Box*");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("Box*", mailbox.Name);
        Assert.AreEqual("Box*", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Null()
    {
      TestUtils2.TestMailbox(".", "Box", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.MoveTo((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Empty()
    {
      TestUtils2.TestMailbox(".", "Box", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.MoveTo(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestMoveToAsync_MoveToNewName_Null()
    {
      TestUtils2.TestMailbox(".", "Box", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.MoveToAsync((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveToAsync_MoveToNewName_Empty()
    {
      TestUtils2.TestMailbox(".", "Box", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.MoveToAsync(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestMoveTo_MoveToSameName()
    {
      MoveTo_MoveToSameName(mailbox => mailbox.MoveTo(mailbox.FullName));
    }

#if ASYNC_OPS
    [Test]
    public void TestMoveToAsync_MoveToSameName()
    {
      MoveTo_MoveToSameName(mailbox => mailbox.MoveToAsync(mailbox.FullName));
    }
#endif

    private void MoveTo_MoveToSameName(Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("/", "Draft", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("Draft", mailbox.Name);
        Assert.AreEqual("Draft", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
        Assert.IsFalse(mailbox.IsInbox);

        Assert.Throws<ImapProtocolViolationException>(() => testAction(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveTo_MoveToSameName_Inbox()
    {
      MoveTo_MoveToSameName_Inbox(mailbox => mailbox.MoveTo("inbox"));
    }

#if ASYNC_OPS
    [Test]
    public void TestMoveToAsync_MoveToSameName_Inbox()
    {
      MoveTo_MoveToSameName_Inbox(mailbox => mailbox.MoveToAsync("inbox"));
    }
#endif

    private void MoveTo_MoveToSameName_Inbox(Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("INBOX", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
        Assert.IsTrue(mailbox.IsInbox);

        Assert.Throws<ImapProtocolViolationException>(() => testAction(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Deepen()
    {
      MoveToNewNameDeepen(false);
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Deepen_Subscribe()
    {
      MoveToNewNameDeepen(true);
    }

    private void MoveToNewNameDeepen(bool subscribe)
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.ListExtended},
                             null,
                             ".",
                             "src",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          if (subscribe) {
            server.EnqueueTaggedResponse("LIST (SUBSCRIBED) src. *",
                                         "* LIST (\\Subscribed) \".\" src.sub1\r\n" +
                                         "* LIST (\\Subscribed \\NonExistent) \".\" src.sub2\r\n" +
                                         "$tag OK done\r\n");
          }

          server.EnqueueTaggedResponse("RENAME src Trash.src\r\n",
                                       "$tag OK done\r\n");

          if (subscribe) {
            server.EnqueueTaggedResponse("UNSUBSCRIBE src\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE Trash.src\r\n",
                                         "$tag OK done\r\n");

            server.EnqueueTaggedResponse("UNSUBSCRIBE src.sub1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE Trash.src.sub1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UNSUBSCRIBE src.sub2\r\n",
                                         "$tag OK done\r\n");
          }
        });

        if (subscribe)
          mailbox.MoveTo("Trash.src", true);
        else
          mailbox.MoveTo("Trash.src");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("Trash.src", mailbox.FullName);
        Assert.AreEqual("Trash", mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_MakeShallow()
    {
      MoveToNewNameShallow(false);
    }

    [Test]
    public void TestMoveTo_MoveToNewName_MakeShallow_Subscribe()
    {
      MoveToNewNameShallow(true);
    }

    private void MoveToNewNameShallow(bool subscribe)
    {
      TestUtils2.TestMailbox("/",
                             "nested/src",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("nested/src", mailbox.FullName);
        Assert.AreEqual("nested", mailbox.ParentMailboxName);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          if (subscribe) {
            server.EnqueueTaggedResponse("LSUB nested/src/ *",
                                         "* LSUB () \"/\" nested/src/sub1\r\n" +
                                         "* LSUB () \"/\" nested/src/sub2/sub3\r\n" +
                                         "$tag OK done\r\n");
          }

          server.EnqueueTaggedResponse("RENAME nested/src dest\r\n",
                                       "$tag OK done\r\n");

          if (subscribe) {
            server.EnqueueTaggedResponse("UNSUBSCRIBE nested/src\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest\r\n",
                                         "$tag OK done\r\n");

            server.EnqueueTaggedResponse("UNSUBSCRIBE nested/src/sub1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest/sub1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UNSUBSCRIBE nested/src/sub2/sub3\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE dest/sub2/sub3\r\n",
                                         "$tag OK done\r\n");
          }
        });

        if (subscribe)
          mailbox.MoveTo("dest", true);
        else
          mailbox.MoveTo("dest");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("dest", mailbox.Name);
        Assert.AreEqual("dest", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Opened()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK done\r\n");

        var opened = mailbox.Open();

        Assert.AreSame(opened, mailbox.Client.OpenedMailbox);
        Assert.IsTrue(mailbox.IsOpen);
        Assert.IsTrue(opened.IsOpen);
        Assert.AreEqual("INBOX", opened.FullName);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("CLOSE\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("RENAME INBOX Trash.INBOX\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("SELECT Trash.INBOX\r\n",
                                       "$tag OK done\r\n");
        });

        mailbox.MoveTo("Trash.INBOX");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("INBOX", mailbox.Name);
        Assert.AreEqual("Trash.INBOX", mailbox.FullName);
        Assert.AreEqual("Trash", mailbox.ParentMailboxName);

        Assert.IsNull(mailbox.Client.OpenedMailbox);
        Assert.IsFalse(mailbox.IsOpen);
        Assert.IsFalse(opened.IsOpen);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Subscribe_HasNoChildren()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.HasNoChildren},
                             ".",
                             "Draft",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LSUB Draft. *\r\n",
                                       "* LSUB () \".\" Draft.Created\r\n" + // other client created children
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("RENAME Draft Sent\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UNSUBSCRIBE Draft\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("SUBSCRIBE Sent\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("UNSUBSCRIBE Draft.Created\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("SUBSCRIBE Sent.Created\r\n",
                                       "$tag OK done\r\n");
        });

        mailbox.MoveTo("Sent", true);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual("Sent", mailbox.Name);
        Assert.AreEqual("Sent", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Subscribe_NonHierarchical()
    {
      TestUtils2.TestMailbox(null,
                             null,
                             string.Empty,
                             "Sent",
                             MoveToNewNameSubscribeCanNotHaveChild);
    }

    [Test]
    public void TestMoveTo_MoveToNewName_Subscribe_NoInferiors()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "Sent",
                             MoveToNewNameSubscribeCanNotHaveChild);
    }

    private void MoveToNewNameSubscribeCanNotHaveChild(ImapPseudoServer2 server, ImapMailboxInfo mailbox)
    {
      var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
        server.EnqueueTaggedResponse("RENAME Sent OldSent\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("UNSUBSCRIBE Sent\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SUBSCRIBE OldSent\r\n",
                                     "$tag OK done\r\n");
      });

      mailbox.MoveTo("OldSent", true);

      responseTask.Wait();

      Assert.IsFalse(mailbox.Client.IsBusy);
    }

    [Test]
    public void TestMoveTo_MoveToNewName_AlreadyExists()
    {
      MoveToNewNameAlreadyExists(false, (mailbox, name) => mailbox.MoveTo(name));
    }

    [Test]
    public void TestMoveTo_MoveToNewName_AlreadyExists_Subscribe()
    {
      MoveToNewNameAlreadyExists(true, (mailbox, name) => mailbox.MoveTo(name, true));
    }

#if ASYNC_OPS
    [Test]
    public void TestMoveToAsync_MoveToNewName_AlreadyExists()
    {
      MoveToNewNameAlreadyExists(false, (mailbox, name) => mailbox.MoveToAsync(name).Wait());
    }

    [Test]
    public void TestMoveToAsync_MoveToNewName_AlreadyExists_Subscribe()
    {
      MoveToNewNameAlreadyExists(true, (mailbox, name) => mailbox.MoveToAsync(name, true).Wait());
    }
#endif

    private void MoveToNewNameAlreadyExists(bool subscribe, Action<ImapMailboxInfo, string> testAction)
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.ListExtended},
                             null,
                             ".",
                             "src",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("src", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          if (subscribe) {
            server.EnqueueTaggedResponse("LIST (SUBSCRIBED) src. * RETURN (CHILDREN)\r\n",
                                         "* LIST (\\Subscribed) \".\" src.sub1\r\n" +
                                         "* LIST (\\Subscribed \\NonExistent) \".\" src.sub2\r\n" +
                                         "$tag OK done\r\n");
          }

          server.EnqueueTaggedResponse("RENAME src Trash.src\r\n",
                                       "$tag NO [ALREADYEXISTS] mailbox exists\r\n");
        });

        var ex = TestUtils.Throws<ImapErrorResponseException>(() => testAction(mailbox, "Trash.src"));

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(ImapResponseCode.AlreadyExists,
                        ex.Result.TaggedStatusResponse.ResponseText.Code);

        Assert.AreEqual("src", mailbox.Name);
        Assert.AreEqual("src", mailbox.FullName);
        Assert.AreEqual(string.Empty, mailbox.ParentMailboxName);
      });
    }

    [Test]
    public void TestSubscribe()
    {
      Subscribe(false, mailbox => mailbox.Subscribe());
    }

    [Test]
    public void TestSubscribe_Recursive()
    {
      Subscribe(true, mailbox => mailbox.Subscribe(true));
    }

#if ASYNC_OPS
    [Test]
    public void TestSubscribeAsync()
    {
      Subscribe(false, mailbox => mailbox.SubscribeAsync().Wait());
    }

    [Test]
    public void TestSubscribeAsync_Recursive()
    {
      Subscribe(true, mailbox => mailbox.SubscribeAsync(true).Wait());
    }
#endif

    private void Subscribe(bool recursive, Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("SUBSCRIBE INBOX\r\n",
                                       "$tag OK done\r\n");

          if (recursive) {
            server.EnqueueTaggedResponse("LIST INBOX. *\r\n",
                                         "* LIST () \".\" INBOX.Child1\r\n" +
                                         "* LIST () \".\" INBOX.Child2\r\n" +
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE INBOX.Child1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("SUBSCRIBE INBOX.Child2\r\n",
                                         "$tag OK done\r\n");
          }
        });

        testAction(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestUnsubscribe()
    {
      Unsubscribe(false, mailbox => mailbox.Unsubscribe());
    }

    [Test]
    public void TestUnsubscribe_Recursive()
    {
      Unsubscribe(true, mailbox => mailbox.Unsubscribe(true));
    }

#if ASYNC_OPS
    [Test]
    public void TestUnsubscribeAsync()
    {
      Unsubscribe(false, mailbox => mailbox.UnsubscribeAsync().Wait());
    }

    [Test]
    public void TestUnsubscribeAsync_Recursive()
    {
      Unsubscribe(true, mailbox => mailbox.UnsubscribeAsync(true).Wait());
    }
#endif

    private void Unsubscribe(bool recursive, Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("UNSUBSCRIBE INBOX\r\n",
                                       "$tag OK done\r\n");

          if (recursive) {
            server.EnqueueTaggedResponse("LSUB INBOX. *\r\n",
                                         "* LSUB () \".\" INBOX.Child1\r\n" +
                                         "* LSUB () \".\" INBOX.Child2\r\n" +
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UNSUBSCRIBE INBOX.Child1\r\n",
                                         "$tag OK done\r\n");
            server.EnqueueTaggedResponse("UNSUBSCRIBE INBOX.Child2\r\n",
                                         "$tag OK done\r\n");
          }
        });

        testAction(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetChildren()
    {
      GetChildren(new ImapMailboxFlag[0], mailbox => mailbox.GetChildren());
    }

    [Test]
    public void TestGetChildren_HasNoChildren()
    {
      GetChildren(new[] {ImapMailboxFlag.HasNoChildren}, mailbox => mailbox.GetChildren());
    }

#if ASYNC_OPS
    [Test]
    public void TestGetChildrenAsync()
    {
      GetChildren(new ImapMailboxFlag[0], mailbox => mailbox.GetChildrenAsync().Result);
    }

    [Test]
    public void TestGetChildrenAsync_HasNoChildren()
    {
      GetChildren(new[] {ImapMailboxFlag.HasNoChildren}, mailbox => mailbox.GetChildrenAsync().Result);
    }
#endif

#if NET_4_5
    private void GetChildren(ImapMailboxFlag[] flags, Func<ImapMailboxInfo, IReadOnlyList<ImapMailboxInfo>> testFunc)
#else
    private void GetChildren(ImapMailboxFlag[] flags, Func<ImapMailboxInfo, IList<ImapMailboxInfo>> testFunc)
#endif
    {
      TestUtils2.TestMailbox(null, flags, "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST INBOX. *\r\n",
                                       "* LIST () \".\" INBOX.Child1\r\n" +
                                       "* LIST () \".\" INBOX.Child2\r\n" +
                                       "* LIST () \".\" INBOX.Child2.Child3\r\n" +
                                       "$tag OK done\r\n");
        });

        var children = testFunc(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(children);
        Assert.AreEqual(3, children.Count, "Count");
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => children.Clear(), "readonly");
#endif

        using (var enumerator = children.GetEnumerator()) {
          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreEqual("Child1", enumerator.Current.Name);
          Assert.AreEqual("INBOX.Child1", enumerator.Current.FullName);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreEqual("Child2", enumerator.Current.Name);
          Assert.AreEqual("INBOX.Child2", enumerator.Current.FullName);

          Assert.IsTrue(enumerator.MoveNext());
          Assert.AreEqual("Child3", enumerator.Current.Name);
          Assert.AreEqual("INBOX.Child2.Child3", enumerator.Current.FullName);

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetChildren_NonHierarchical()
    {
      TestUtils2.TestMailbox(null,
                             null,
                             string.Empty,
                             "INBOX",
                             GetChildrenNoInferiorsOrNonHierarchical);
    }

    [Test]
    public void TestGetChildren_NoInferiors()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             GetChildrenNoInferiorsOrNonHierarchical);
    }

    private void GetChildrenNoInferiorsOrNonHierarchical(ImapPseudoServer2 server, ImapMailboxInfo mailbox)
    {
      var children = mailbox.GetChildren();

      Assert.IsFalse(mailbox.Client.IsBusy);

      Assert.IsNotNull(children);
      Assert.AreEqual(0, children.Count, "Count");
#if !NET_4_5
      Assert.Throws<NotSupportedException>(() => children.Clear(), "readonly");
#endif

      using (var enumerator = children.GetEnumerator()) {
        Assert.IsFalse(enumerator.MoveNext());
      }
    }

#if ASYNC_OPS
    [Test]
    public void TestGetChildrenAsync_NonHierarchical()
    {
      TestUtils2.TestMailbox(null,
                             null,
                             string.Empty,
                             "INBOX",
                             GetChildrenAsync_NoInferiorsOrNonHierarchical);
    }

    [Test]
    public void TestGetChildrenAsync_NoInferiors()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             GetChildrenAsync_NoInferiorsOrNonHierarchical);
    }

    private void GetChildrenAsync_NoInferiorsOrNonHierarchical(ImapPseudoServer2 server, ImapMailboxInfo mailbox)
    {
      var children = mailbox.GetChildrenAsync().Result;

      Assert.IsFalse(mailbox.Client.IsBusy);

      Assert.IsNotNull(children);
      Assert.AreEqual(0, children.Count, "Count");
#if !NET_4_5
      Assert.Throws<NotSupportedException>(() => children.Clear(), "readonly");
#endif

      using (var enumerator = children.GetEnumerator()) {
        Assert.IsFalse(enumerator.MoveNext());
      }
    }
#endif

    [Test]
    public void TestGetParent()
    {
      GetParent(mailbox => mailbox.GetParent());
    }

#if ASYNC_OPS
    [Test]
    public void TestGetParentAsync()
    {
      GetParent(mailbox => mailbox.GetParentAsync().Result);
    }
#endif

    private void GetParent(Func<ImapMailboxInfo, ImapMailboxInfo> testFunc)
    {
      TestUtils2.TestMailbox(".", "INBOX.Child", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                       "* LIST () \".\" INBOX\r\n" +
                                       "$tag OK done\r\n");
        });

        var parent = testFunc(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(parent);
        Assert.AreEqual("INBOX", parent.Name);
        Assert.AreEqual("INBOX", parent.FullName);
      });
    }

    [Test]
    public void TestGetParent_OfTopLevelMailbox()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetParent());

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestGetParentAsync_OfTopLevelMailbox()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetParentAsync());

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestGetParent_NotFound()
    {
      GetParent_NotFound(mailbox => mailbox.GetParent());
    }

#if ASYNC_OPS
    [Test]
    public void TestGetParentAsync_NotFound()
    {
      GetParent_NotFound(mailbox => mailbox.GetParentAsync().Result);
    }
#endif

    private void GetParent_NotFound(Func<ImapMailboxInfo, ImapMailboxInfo> testFunc)
    {
      TestUtils2.TestMailbox(".", "Sent.Child", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Sent\r\n",
                                       "$tag OK done\r\n");
        });

        var ex = TestUtils.Throws<ImapMailboxNotFoundException>(() => testFunc(mailbox));

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("Sent", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });
      });
    }

    [Test]
    public void TestGetOrCreateParent()
    {
      GetOrCreateParent(false, mailbox => mailbox.GetOrCreateParent());
    }

    [Test]
    public void TestGetOrCreateParent_Subscribe()
    {
      GetOrCreateParent(true, mailbox => mailbox.GetOrCreateParent(true));
    }

#if ASYNC_OPS
    [Test]
    public void TestGetOrCreateParentAsync()
    {
      GetOrCreateParent(false, mailbox => mailbox.GetOrCreateParentAsync().Result);
    }

    [Test]
    public void TestGetOrCreateParentAsync_Subscribe()
    {
      GetOrCreateParent(true, mailbox => mailbox.GetOrCreateParentAsync(true).Result);
    }
#endif

    private void GetOrCreateParent(bool subscribe, Func<ImapMailboxInfo, ImapMailboxInfo> testFunc)
    {
      TestUtils2.TestMailbox(".", "Sent.Child", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Sent\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("CREATE Sent\r\n",
                                       "$tag OK done\r\n");
          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE Sent\r\n",
                                         "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Sent\r\n",
                                       "* LIST () \".\" Sent\r\n" +
                                       "$tag OK done\r\n");
        });

        var parent = testFunc(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(parent);
        Assert.AreEqual("Sent", parent.Name);
        Assert.AreEqual("Sent", parent.FullName);
      });
    }

    [Test]
    public void TestGetOrCreateParent_OfTopLevelMailbox()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetOrCreateParent());

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestGetOrCreateParentAsync_OfTopLevelMailbox()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetOrCreateParentAsync());

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestGetFullNameOf()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX.Child", mailbox.GetFullNameOf("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX/Child", mailbox.GetFullNameOf("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_ArgumentEmpty()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_ArgumentNull()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_CanNotHaveChild()
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_With2Inferiors()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX.1.2", mailbox.GetFullNameOf("1", "2"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX/1/2", mailbox.GetFullNameOf("1", "2"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf("1", "2"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(null, "2"));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf("1", null));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(string.Empty, "2"));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf("1", string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_With3Inferiors()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX.1.2.3", mailbox.GetFullNameOf("1", "2", "3"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX/1/2/3", mailbox.GetFullNameOf("1", "2", "3"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf("1", "2", "3"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(null, "2", "3"));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf("1", null, "3"));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf("1", "2", null));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(string.Empty, "2", "3"));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf("1", string.Empty, "3"));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf("1", "2", string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_WithArrayOfInferiors()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX.1.2.3.4", mailbox.GetFullNameOf(new[] {"1", "2", "3", "4"}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX/1/2/3/4", mailbox.GetFullNameOf(new[] {"1", "2", "3", "4"}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox(null,
                  new[] {ImapMailboxFlag.NoInferiors},
                  ".",
                  "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf(new[] {"1", "2", "3", "4"}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new string[] {null, "2", "3", "4"}));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new string[] {"1", null, "3", "4"}));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new string[] {"1", "2", null, "4"}));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new string[] {"1", "2", "3", null}));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new string[] {string.Empty, "2", "3", "4"}));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new string[] {"1", string.Empty, "3", "4"}));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new string[] {"1", "2", string.Empty, "4"}));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new string[] {"1", "2", "3", string.Empty}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetFullNameOf_WithListOfInferiors()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX.1.2.3.4", mailbox.GetFullNameOf(new List<string>() {"1", "2", "3", "4"}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.AreEqual("INBOX/1/2/3/4", mailbox.GetFullNameOf(new List<string>() {"1", "2", "3", "4"}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf(new List<string>() {"1", "2", "3", "4"}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new List<string>() {null, "2", "3", "4"}));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new List<string>() {"1", null, "3", "4"}));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new List<string>() {"1", "2", null, "4"}));
        Assert.Throws<ArgumentNullException>(() => mailbox.GetFullNameOf(new List<string>() {"1", "2", "3", null}));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new List<string>() {string.Empty, "2", "3", "4"}));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new List<string>() {"1", string.Empty, "3", "4"}));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new List<string>() {"1", "2", string.Empty, "4"}));
        Assert.Throws<ArgumentException>(() => mailbox.GetFullNameOf(new List<string>() {"1", "2", "3", string.Empty}));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestSplitMailboxHierarchy()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        CollectionAssert.AreEqual(new[] {"INBOX"},
                                  mailbox.SplitMailboxHierarchy("INBOX"));
        CollectionAssert.AreEqual(new[] {"INBOX", "1", "2"},
                                  mailbox.SplitMailboxHierarchy("INBOX.1.2"));
        CollectionAssert.AreEqual(new[] {"INBOX/1/2"},
                                  mailbox.SplitMailboxHierarchy("INBOX/1/2"));

        Assert.Throws<ArgumentException>(() => mailbox.SplitMailboxHierarchy(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.Throws<ArgumentNullException>(() => mailbox.SplitMailboxHierarchy(null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox("/", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        CollectionAssert.AreEqual(new[] {"INBOX"},
                                  mailbox.SplitMailboxHierarchy("INBOX"));
        CollectionAssert.AreEqual(new[] {"INBOX.1.2"},
                                  mailbox.SplitMailboxHierarchy("INBOX.1.2"));
        CollectionAssert.AreEqual(new[] {"INBOX", "1", "2"},
                                  mailbox.SplitMailboxHierarchy("INBOX/1/2"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });

      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             ".",
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.SplitMailboxHierarchy("INBOX.1.2"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestCreateChild()
    {
      CreateChild(false, (mailbox, name) => mailbox.CreateChild(name));
    }

    [Test]
    public void TestCreateChild_Subscribe()
    {
      CreateChild(true, (mailbox, name) => mailbox.CreateChild(name, true));
    }

#if ASYNC_OPS
    [Test]
    public void TestCreateChildAsync()
    {
      CreateChild(false, (mailbox, name) => mailbox.CreateChildAsync(name).Result);
    }

    [Test]
    public void TestCreateChildAsync_Subscribe()
    {
      CreateChild(true, (mailbox, name) => mailbox.CreateChildAsync(name, true).Result);
    }
#endif

    private void CreateChild(bool subscribe, Func<ImapMailboxInfo, string, ImapMailboxInfo> testFunc)
    {
      TestUtils2.TestMailbox(".", "Sent", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("CREATE Sent.Child\r\n",
                                       "$tag OK done\r\n");
          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE Sent.Child\r\n",
                                         "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Sent.Child\r\n",
                                       "* LIST () \".\" Sent.Child\r\n" +
                                       "$tag OK done\r\n");
        });

        var child = testFunc(mailbox, "Child");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(child);
        Assert.AreEqual("Child", child.Name);
        Assert.AreEqual("Sent.Child", child.FullName);
      });
    }

    [Test]
    public void TestCreateChild_NameEmpty()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.CreateChild(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestCreateChild_NameNull()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.CreateChild((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestCreateChildAsync_NameEmpty()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.CreateChildAsync(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestCreateChildAsync_NameNull()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.CreateChildAsync((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestCreateChild_AlreadyExists()
    {
      CreateChildAlreadyExists(false, (mailbox, name) => mailbox.CreateChild(name));
    }

    [Test]
    public void TestCreateChild_AlreadyExists_Subscribe()
    {
      CreateChildAlreadyExists(true, (mailbox, name) => mailbox.CreateChild(name, true));
    }

#if ASYNC_OPS
    [Test]
    public void TestCreateChildAsync_AlreadyExists()
    {
      CreateChildAlreadyExists(false, (mailbox, name) => mailbox.CreateChildAsync(name).Result);
    }

    [Test]
    public void TestCreateChildAsync_AlreadyExists_Subscribe()
    {
      CreateChildAlreadyExists(true, (mailbox, name) => mailbox.CreateChildAsync(name, true).Result);
    }
#endif

    private void CreateChildAlreadyExists(bool subscribe, Func<ImapMailboxInfo, string, ImapMailboxInfo> testFunc)
    {
      TestUtils2.TestMailbox(".", "Sent", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("CREATE Sent.Child\r\n",
                                       "$tag NO [ALREADYEXISTS] mailbox already exists\r\n");
        });

        var ex = TestUtils.Throws<ImapErrorResponseException>(() => testFunc(mailbox, "Child"));

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(ImapResponseCode.AlreadyExists,
                        ex.Result.TaggedStatusResponse.ResponseText.Code);
      });
    }

    [Test]
    public void TestGetChild()
    {
      GetChild((mailbox, name) => mailbox.GetChild(name));
    }

#if ASYNC_OPS
    [Test]
    public void TestGetChildAsync()
    {
      GetChild((mailbox, name) => mailbox.GetChildAsync(name).Result);
    }
#endif

    private void GetChild(Func<ImapMailboxInfo, string, ImapMailboxInfo> testFunc)
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" INBOX.Child\r\n",
                                       "* LIST () \".\" INBOX.Child\r\n" +
                                       "$tag OK done\r\n");
        });

        var child = testFunc(mailbox, "Child");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(child);
        Assert.AreEqual("Child", child.Name);
        Assert.AreEqual("INBOX.Child", child.FullName);
      });
    }

    [Test]
    public void TestGetChild_NameEmpty()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.GetChild(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetChild_NameNull()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetChild((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestGetChildAsync_NameEmpty()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.GetChildAsync(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetChildAsync_NameNull()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetChildAsync((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#endif

    [Test]
    public void TestGetChild_NotFound()
    {
      GetChild_NotFound((mailbox, name) => mailbox.GetChild(name));
    }

#if ASYNC_OPS
    [Test]
    public void TestGetChildAsync_NotFound()
    {
      GetChild_NotFound((mailbox, name) => mailbox.GetChildAsync(name).Result);
    }
#endif

    private void GetChild_NotFound(Func<ImapMailboxInfo, string, ImapMailboxInfo> testFunc)
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("$tag OK done\r\n");
        });

        var ex = TestUtils.Throws<ImapMailboxNotFoundException>(() => testFunc(mailbox, "Child"));

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("INBOX.Child", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });
      });
    }

    [Test]
    public void TestGetOrCreateChild_ExistentChild()
    {
      GetOrCreateChildExistent(false, (mailbox, name) => mailbox.GetOrCreateChild(name));
    }

    [Test]
    public void TestGetOrCreateChild_ExistentChild_Subscribe()
    {
      GetOrCreateChildExistent(true, (mailbox, name) => mailbox.GetOrCreateChild(name, true));
    }

#if ASYNC_OPS
    [Test]
    public void TestGetOrCreateChildAsync_ExistentChild()
    {
      GetOrCreateChildExistent(false, (mailbox, name) => mailbox.GetOrCreateChildAsync(name).Result);
    }

    [Test]
    public void TestGetOrCreateChildAsync_ExistentChild_Subscribe()
    {
      GetOrCreateChildExistent(true, (mailbox, name) => mailbox.GetOrCreateChildAsync(name, true).Result);
    }
#endif

    private void GetOrCreateChildExistent(bool subscribe, Func<ImapMailboxInfo, string, ImapMailboxInfo> testFunc)
    {
      TestUtils2.TestMailbox(".", "Sent", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Sent.Child\r\n",
                                       "* LIST () \".\" Sent.Child\r\n" +
                                       "$tag OK done\r\n");
        });

        var child = testFunc(mailbox, "Child");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(child);
        Assert.AreEqual("Child", child.Name);
        Assert.AreEqual("Sent.Child", child.FullName);
      });
    }

    [Test]
    public void TestGetOrCreateChild_NonExistentChild()
    {
      GetOrCreateChildNonExistent(false, (mailbox, name) => mailbox.GetOrCreateChild(name));
    }

    [Test]
    public void TestGetOrCreateChild_NonExistentChild_Subscribe()
    {
      GetOrCreateChildNonExistent(true, (mailbox, name) => mailbox.GetOrCreateChild(name, true));
    }

#if ASYNC_OPS
    [Test]
    public void TestGetOrCreateChildAsync_NonExistentChild()
    {
      GetOrCreateChildNonExistent(false, (mailbox, name) => mailbox.GetOrCreateChildAsync(name).Result);
    }

    [Test]
    public void TestGetOrCreateChildAsync_NonExistentChild_Subscribe()
    {
      GetOrCreateChildNonExistent(true, (mailbox, name) => mailbox.GetOrCreateChildAsync(name, true).Result);
    }
#endif

    private void GetOrCreateChildNonExistent(bool subscribe, Func<ImapMailboxInfo, string, ImapMailboxInfo> testFunc)
    {
      TestUtils2.TestMailbox(".", "Sent", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("LIST \"\" Sent.Child\r\n",
                                       "$tag OK done\r\n");
          server.EnqueueTaggedResponse("CREATE Sent.Child\r\n",
                                       "$tag OK done\r\n");
          if (subscribe)
            server.EnqueueTaggedResponse("SUBSCRIBE Sent.Child\r\n",
                                         "$tag OK done\r\n");
          server.EnqueueTaggedResponse("LIST \"\" Sent.Child\r\n",
                                       "* LIST () \".\" Sent.Child\r\n" +
                                       "$tag OK done\r\n");
        });

        var child = testFunc(mailbox, "Child");

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(child);
        Assert.AreEqual("Child", child.Name);
        Assert.AreEqual("Sent.Child", child.FullName);
      });
    }

    [Test]
    public void TestGetOrCreateChild_NameEmpty()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.GetOrCreateChild(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetOrCreateChild_NameNull()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetOrCreateChild((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestGetOrCreateChildAsync_NameEmpty()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentException>(() => mailbox.GetOrCreateChildAsync(string.Empty));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestGetOrCreateChildAsync_NameNull()
    {
      TestUtils2.TestMailbox(".", "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentNullException>(() => mailbox.GetOrCreateChildAsync((string)null));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestGetChild_NonHierarchical()
    {
      TestNonHierarchicalMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetChild("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestGetChildAsync_NonHierarchical()
    {
      TestNonHierarchicalMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetChildAsync("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestGetOrCreateChild_NonHierarchical()
    {
      TestNonHierarchicalMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetOrCreateChild("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestGetOrCreateChildAsync_NonHierarchical()
    {
      TestNonHierarchicalMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetOrCreateChildAsync("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestCreateChild_NonHierarchical()
    {
      TestNonHierarchicalMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.CreateChild("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestCreateChildAsync_NonHierarchical()
    {
      TestNonHierarchicalMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.CreateChildAsync("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestGetFullNameOf_NonHierarchical()
    {
      TestNonHierarchicalMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    private void TestNonHierarchicalMailbox(Action<ImapMailboxInfo> action)
    {
      TestUtils2.TestMailbox(string.Empty, "INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.CanHaveChild);

        action(mailbox);
      });
    }

    [Test]
    public void TestGetChild_NoInferiors()
    {
      TestNoInferiorsMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetChild("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestGetChildAsync_NoInferiors()
    {
      TestNoInferiorsMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetChildAsync("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestGetOrCreateChild_NoInferiors()
    {
      TestNoInferiorsMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetOrCreateChild("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestGetOrCreateChildAsync_NoInferiors()
    {
      TestNoInferiorsMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetOrCreateChildAsync("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestCreateChild_NoInferiors()
    {
      TestNoInferiorsMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.CreateChild("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestCreateChildAsync_NoInferiors()
    {
      TestNoInferiorsMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.CreateChildAsync("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }
#endif

    [Test]
    public void TestGetFullNameOf_NoInferiors()
    {
      TestNoInferiorsMailbox(delegate(ImapMailboxInfo mailbox) {
        Assert.Throws<ImapProtocolViolationException>(() => mailbox.GetFullNameOf("Child"));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    private void TestNoInferiorsMailbox(Action<ImapMailboxInfo> action)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoInferiors},
                             string.Empty,
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.CanHaveChild);

        action(mailbox);
      });
    }

    [Test]
    public void TestGetQuota()
    {
      GetQuota(mailbox => mailbox.GetQuota());
    }

#if ASYNC_OPS
    [Test]
    public void TestGetQuotaAsync()
    {
      GetQuota(mailbox => mailbox.GetQuotaAsync().Result);
    }
#endif

#if NET_4_5
    private void GetQuota(Func<ImapMailboxInfo, IReadOnlyList<ImapQuota>> testFunc)
#else
    private void GetQuota(Func<ImapMailboxInfo, IList<ImapQuota>> testFunc)
#endif
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.Quota},
                             null,
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("GETQUOTAROOT INBOX\r\n",
                                       "* QUOTAROOT INBOX \"\"\r\n" +
                                       "* QUOTA \"\" (STORAGE 10 512)\r\n" +
                                       "$tag OK done\r\n");
        });

        var quotas = testFunc(mailbox);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(1, quotas.Count);
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => quotas.Clear());
#endif

        var quota = quotas[0];

        Assert.AreEqual(string.Empty, quota.Root);
        Assert.AreEqual(1, quota.Resources.Count);
        Assert.AreEqual("STORAGE", quota.Resources[0].Name);
        Assert.AreEqual(10, quota.Resources[0].Usage);
        Assert.AreEqual(512, quota.Resources[0].Limit);
      });
    }

    [Test]
    public void TestGetQuota_QuotaIncapable()
    {
      GetQuota_QuotaIncapable(mailbox => mailbox.GetQuota());
    }

#if ASYNC_OPS
    [Test]
    public void TestGetQuotaAsync_QuotaIncapable()
    {
      GetQuota_QuotaIncapable(mailbox => mailbox.GetQuotaAsync().Result);
    }
#endif

#if NET_4_5
    private void GetQuota_QuotaIncapable(Func<ImapMailboxInfo, IReadOnlyList<ImapQuota>> testFunc)
#else
    private void GetQuota_QuotaIncapable(Func<ImapMailboxInfo, IList<ImapQuota>> testFunc)
#endif
    {
      TestUtils2.TestMailbox("INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var quotas = testFunc(mailbox);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(0, quotas.Count());
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => quotas.Clear());
#endif
      });
    }

    [Test]
    public void TestGetQuota_QuotaDisabled()
    {
      GetQuota_QuotaDisabled(mailbox => mailbox.GetQuota());
    }

#if ASYNC_OPS
    [Test]
    public void TestGetQuotaAsync_QuotaDisabled()
    {
      GetQuota_QuotaDisabled(mailbox => mailbox.GetQuotaAsync().Result);
    }
#endif

#if NET_4_5
    private void GetQuota_QuotaDisabled(Func<ImapMailboxInfo, IReadOnlyList<ImapQuota>> testFunc)
#else
    private void GetQuota_QuotaDisabled(Func<ImapMailboxInfo, IList<ImapQuota>> testFunc)
#endif
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.Quota},
                             null,
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.Quota);

        var quotas = testFunc(mailbox);

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.AreEqual(0, quotas.Count());
#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => quotas.Clear());
#endif
      });
    }
  }
}
