#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMailboxInfoOperationsAppendTests {
    [Test]
    public void TestAppendMessage_IAppendMessage()
    {
      AppendMessageAsync_IAppendMessage((mailbox, message) => mailbox.AppendMessage(message));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_IAppendMessage()
    {
      AppendMessageAsync_IAppendMessage((mailbox, message) => mailbox.AppendMessageAsync(message).Wait());
    }
#endif

    private void AppendMessageAsync_IAppendMessage(Action<ImapMailboxInfo, IImapAppendMessage> testAction)
    {
      TestUtils2.TestMailbox("INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        const string uploadingMessage = "MIME-Version: 1.0\r\n\r\ntest message";

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage),
                                            new DateTimeOffset(2011, 1, 24, 14, 52, 30, TimeSpan.FromHours(+9.0)),
                                            new ImapMessageFlagSet(new[] {"$label1"}, ImapMessageFlag.Seen));

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse("APPEND INBOX ",
                                       "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");
        });

        testAction(mailbox, message);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.That(server.DequeueRequest(),
                    Is.StringEnding(string.Format("APPEND INBOX (\\Seen $label1) \"24-Jan-2011 14:52:30 +0900\" {{{0}}}\r\n{1}\r\n",
                                                  uploadingMessage.Length,
                                                  uploadingMessage)));
      });
    }

    [Test]
    public void TestAppendMessage_IAppendMessage_MailboxNonExistent()
    {
      AppendMessage_IAppendMessage_MailboxNonExistent((mailbox, stream) => mailbox.AppendMessage(stream));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_IAppendMessage_MailboxNonExistent()
    {
      AppendMessage_IAppendMessage_MailboxNonExistent((mailbox, stream) => mailbox.AppendMessageAsync(stream));
    }
#endif

    private void AppendMessage_IAppendMessage_MailboxNonExistent(Action<ImapMailboxInfo, Stream> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        using (var stream = new MemoryStream(new byte[] {0x40})) {
          Assert.Throws<ImapProtocolViolationException>(() => testAction(mailbox, stream));
        }

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_IAppendMessage_MailboxUnselectable()
    {
      AppendMessage_IAppendMessage_MailboxUnselectable((mailbox, stream) => mailbox.AppendMessage(stream));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_IAppendMessage_MailboxUnselectable()
    {
      AppendMessage_IAppendMessage_MailboxUnselectable((mailbox, stream) => mailbox.AppendMessageAsync(stream));
    }
#endif

    private void AppendMessage_IAppendMessage_MailboxUnselectable(Action<ImapMailboxInfo, Stream> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoSelect},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsUnselectable);

        using (var stream = new MemoryStream(new byte[] {0x40})) {
          Assert.Throws<ImapProtocolViolationException>(() => testAction(mailbox, stream));
        }

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_IAppendMessage_NoTryCreate()
    {
      AppendMessage_IAppendMessage_NoTryCreate((mailbox, stream) => mailbox.AppendMessage(stream));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_IAppendMessage_NoTryCreate()
    {
      AppendMessage_IAppendMessage_NoTryCreate((mailbox, stream) => mailbox.AppendMessageAsync(stream).Wait());
    }
#endif

    private void AppendMessage_IAppendMessage_NoTryCreate(Action<ImapMailboxInfo, Stream> testAction)
    {
      TestUtils2.TestMailbox("NonExistent",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueLiteralContinuationResponse("APPEND NonExistent {1}\r\n",
                                                    "$tag NO [TRYCREATE] mailbox not exist\r\n",
                                                    false);
        });

        var ex = TestUtils.Throws<ImapMailboxNotFoundException>(() => testAction(mailbox, new MemoryStream(new byte[] {0x40})));

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("NonExistent", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });
      });
    }

    [Test]
    public void TestAppendMessages()
    {
      TestUtils2.TestMailbox("INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        AppendMessages(server, mailbox, messages => mailbox.AppendMessages(messages));
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessagesAsync()
    {
      TestUtils2.TestMailbox("INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        AppendMessages(server, mailbox, messages => mailbox.AppendMessagesAsync(messages).Wait());
      });
    }
#endif

    [Test]
    public void TestAppendMessages_MultiappendDisabled()
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.MultiAppend},
                             null,
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.MultiAppend);

        AppendMessages(server, mailbox, messages => mailbox.AppendMessages(messages));
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessagesAsync_MultiappendDisabled()
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.MultiAppend},
                             null,
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.MultiAppend);

        AppendMessages(server, mailbox, messages => mailbox.AppendMessagesAsync(messages).Wait());
      });
    }
#endif

    private void AppendMessages(ImapPseudoServer2 server,
                                ImapMailboxInfo mailbox,
                                Action<IEnumerable<IImapAppendMessage>> testAction)
    {
      const string uploadingMessage1 = "MIME-Version: 1.0\r\n\r\ntest message1";
      const string uploadingMessage2 = "MIME-Version: 1.0\r\n\r\ntest message2";
      const string uploadingMessage3 = "MIME-Version: 1.0\r\n\r\ntest message3";

      var messages = new[] {
        new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage1),
                              new DateTimeOffset(2011, 1, 24, 14, 52, 30, TimeSpan.FromHours(+9.0)),
                              new ImapMessageFlagSet(new[] {"$label1"}, ImapMessageFlag.Seen)),
        new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage2),
                              new DateTimeOffset(2011, 1, 24, 14, 52, 31, TimeSpan.FromHours(+9.0)),
                              new ImapMessageFlagSet(new[] {"$label2"}, ImapMessageFlag.Seen)),
        new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage3),
                              new DateTimeOffset(2011, 1, 24, 14, 52, 32, TimeSpan.FromHours(+9.0)),
                              new ImapMessageFlagSet(new[] {"$label3"}, ImapMessageFlag.Seen)),
      };

      var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
        server.EnqueueTaggedResponse(string.Format("APPEND INBOX (\\Seen $label1) \"24-Jan-2011 14:52:30 +0900\" {{{0}}}\r\n{1}\r\n",
                                                   uploadingMessage1.Length,
                                                   uploadingMessage1),
                                     "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");
        server.EnqueueTaggedResponse(string.Format("APPEND INBOX (\\Seen $label2) \"24-Jan-2011 14:52:31 +0900\" {{{0}}}\r\n{1}\r\n",
                                                   uploadingMessage2.Length,
                                                   uploadingMessage2),
                                     "$tag OK [APPENDUID 38505 3956] APPEND completed\r\n");
        server.EnqueueTaggedResponse(string.Format("APPEND INBOX (\\Seen $label3) \"24-Jan-2011 14:52:32 +0900\" {{{0}}}\r\n{1}\r\n",
                                                   uploadingMessage3.Length,
                                                   uploadingMessage3),
                                     "$tag OK [APPENDUID 38505 3957] APPEND completed\r\n");
      });

      testAction(messages);

      responseTask.Wait();

      Assert.IsFalse(mailbox.Client.IsBusy);
    }

    [Test]
    public void TestAppendMessages_MultiappendCapable()
    {
      AppendMessages_MultiappendCapable((mailbox, messages) => mailbox.AppendMessages(messages));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessagesAsync_MultiappendCapable()
    {
      AppendMessages_MultiappendCapable((mailbox, messages) => mailbox.AppendMessagesAsync(messages).Wait());
    }
#endif

    private void AppendMessages_MultiappendCapable(Action<ImapMailboxInfo, IEnumerable<IImapAppendMessage>> testAction)
    {
      TestUtils2.TestMailbox(new[] {ImapCapability.MultiAppend},
                             null,
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        const string uploadingMessage1 = "MIME-Version: 1.0\r\n\r\ntest message1";
        const string uploadingMessage2 = "MIME-Version: 1.0\r\n\r\ntest message2";
        const string uploadingMessage3 = "MIME-Version: 1.0\r\n\r\ntest message3";

        var messages = new[] {
          new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage1),
                                new DateTimeOffset(2011, 1, 24, 14, 52, 30, TimeSpan.FromHours(+9.0)),
                                new ImapMessageFlagSet(new[] {"$label1"}, ImapMessageFlag.Seen)),
          new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage2),
                                new DateTimeOffset(2011, 1, 24, 14, 52, 31, TimeSpan.FromHours(+9.0)),
                                new ImapMessageFlagSet(new[] {"$label2"}, ImapMessageFlag.Seen)),
          new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage3),
                                new DateTimeOffset(2011, 1, 24, 14, 52, 32, TimeSpan.FromHours(+9.0)),
                                new ImapMessageFlagSet(new[] {"$label3"}, ImapMessageFlag.Seen)),
        };

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          var expectedCommand =
            "APPEND INBOX " +
            string.Format("(\\Seen $label1) \"24-Jan-2011 14:52:30 +0900\" {{{0}}}\r\n{1} ",
                          uploadingMessage1.Length,
                          uploadingMessage1) +
            string.Format("(\\Seen $label2) \"24-Jan-2011 14:52:31 +0900\" {{{0}}}\r\n{1} ",
                          uploadingMessage2.Length,
                          uploadingMessage2) +
            string.Format("(\\Seen $label3) \"24-Jan-2011 14:52:32 +0900\" {{{0}}}\r\n{1}\r\n",
                          uploadingMessage3.Length,
                          uploadingMessage3);

          server.EnqueueTaggedResponse(expectedCommand,
                                       "$tag OK [APPENDUID 38505 3955,3956,3957] APPEND completed\r\n");
        });

        testAction(mailbox, messages);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessages_NonExistentMailbox()
    {
      AppendMessages_NonExistentMailbox((mailbox, messages) => mailbox.AppendMessages(messages));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessagesAsync_NonExistentMailbox()
    {
      AppendMessages_NonExistentMailbox((mailbox, messages) => mailbox.AppendMessagesAsync(messages));
    }
#endif

    private void AppendMessages_NonExistentMailbox(Action<ImapMailboxInfo, IEnumerable<IImapAppendMessage>> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        var messages = new[] {
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
        };

        Assert.Throws<ImapProtocolViolationException>(() => testAction(mailbox, messages));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessages_UnselectableMailbox()
    {
      AppendMessages_UnselectableMailbox((mailbox, messages) => mailbox.AppendMessages(messages));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessagesAsync_UnselectableMailbox()
    {
      AppendMessages_UnselectableMailbox((mailbox, messages) => mailbox.AppendMessagesAsync(messages).Wait());
    }
#endif

    private void AppendMessages_UnselectableMailbox(Action<ImapMailboxInfo, IEnumerable<IImapAppendMessage>> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoSelect},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsUnselectable);

        var messages = new[] {
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
        };

        TestUtils.Throws<ImapProtocolViolationException>(() => testAction(mailbox, messages));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessages_NoTryCreate()
    {
      AppendMessages_NoTryCreate((mailbox, messages) => mailbox.AppendMessages(messages));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessagesAsync_NoTryCreate()
    {
      AppendMessages_NoTryCreate((mailbox, messages) => mailbox.AppendMessagesAsync(messages).Wait());
    }
#endif

    private void AppendMessages_NoTryCreate(Action<ImapMailboxInfo, IEnumerable<IImapAppendMessage>> testAction)
    {
      TestUtils2.TestMailbox("NonExistent",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueLiteralContinuationResponse("APPEND NonExistent {1}\r\n",
                                                    "$tag NO [TRYCREATE] mailbox not exist\r\n",
                                                    false);
        });

        var messages = new[] {
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
        };

        var ex = TestUtils.Throws<ImapMailboxNotFoundException>(() => testAction(mailbox, messages));

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("NonExistent", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });
      });
    }

    [Test]
    public void TestAppendMessage_MessageInfo_SameSession()
    {
      AppendMessage_MessageInfo_SameSession((mailbox, message) => mailbox.AppendMessage(message));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_MessageInfo_SameSession()
    {
      AppendMessage_MessageInfo_SameSession((mailbox, message) => mailbox.AppendMessageAsync(message).Wait());
    }
#endif

    private void AppendMessage_MessageInfo_SameSession(Action<ImapMailboxInfo, ImapMessageInfo> testAction)
    {
      TestUtils2.TestMailbox("INBOX", delegate(ImapPseudoServer2 server, ImapMailboxInfo sourceMailbox) {
        server.EnqueueTaggedResponse("LIST \"\" dest\r\n",
                                     "* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = sourceMailbox.Client.GetMailbox("dest");

        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "* EXISTS 3\r\n" +
                                     "* OK [UIDVALIDITY 23]\r\n" +
                                     "$tag OK [READ-WRITE] done\r\n");

        using (var openedSourceMailbox = sourceMailbox.Open()) {
          server.EnqueueTaggedResponse("FETCH 1 (UID)\r\n",
                                       "* FETCH 1 (UID 3)\r\n" +
                                       "$tag OK done\r\n");

          var message = openedSourceMailbox.GetMessageBySequence(1L);

          var responseTask = TestUtils2.QueueResponseTask(server, sourceMailbox.Client, delegate {
            server.EnqueueTaggedResponse("UID COPY 3 dest\r\n",
                                         "$tag OK done\r\n");
          });

          testAction(destMailbox, message);

          responseTask.Wait();

          Assert.IsFalse(destMailbox.Client.IsBusy);

          server.EnqueueTaggedResponse("CLOSE\r\n",
                                       "$tag OK done\r\n");
        }
      });
    }

    [Test]
    public void TestAppendMessage_MessageInfo_DifferentSession()
    {
      AppendMessage_MessageInfo_DifferentSession((mailbox, message) => mailbox.AppendMessage(message));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_MessageInfo_DifferentSession()
    {
      AppendMessage_MessageInfo_DifferentSession((mailbox, message) => mailbox.AppendMessageAsync(message).Wait());
    }
#endif

    private void AppendMessage_MessageInfo_DifferentSession(Action<ImapMailboxInfo, ImapMessageInfo> testAction)
    {
      TestUtils2.TestMailbox("dest", delegate(ImapPseudoServer2 destServer, ImapMailboxInfo destMailbox) {
        var selectResp = "* EXISTS 3\r\n" +
          "* OK [UIDVALIDITY 23]\r\n" +
            "$tag OK [READ-WRITE] done\r\n";

        TestUtils2.TestOpenedMailbox("INBOX",
                                     selectResp,
                                     delegate(ImapPseudoServer2 sourceServer, ImapOpenedMailboxInfo sourceMailbox) {
          sourceServer.EnqueueTaggedResponse("FETCH 1 (UID)\r\n",
                                             "* FETCH 1 (UID 3)\r\n" +
                                             "$tag OK done\r\n");

          var message = sourceMailbox.GetMessageBySequence(1L);

          sourceServer.EnqueueTaggedResponse("UID FETCH 3 (RFC822.SIZE INTERNALDATE FLAGS BODY.PEEK[]<0.10240>)\r\n",
                                             "* FETCH 1 (" +
                                             "FLAGS (\\Answered $label1) " +
                                             "INTERNALDATE \"25-Jan-2011 15:29:06 +0900\" " +
                                             "RFC822.SIZE 12 " +
                                             "BODY[] {12}\r\ntest message)\r\n" +
                                             "$tag OK done\r\n");

          var responseTask = TestUtils2.QueueResponseTask(destServer, destMailbox.Client, delegate {
            destServer.EnqueueTaggedResponse("APPEND dest (\\Answered $label1) \"25-Jan-2011 15:29:06 +0900\" {12}\r\ntest message\r\n",
                                             "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");
          });

          testAction(destMailbox, message);

          responseTask.Wait();

          Assert.IsFalse(destMailbox.Client.IsBusy);
        });
      });
    }

    [Test]
    public void TestAppendMessage_MessageInfo_MailboxNonExistent()
    {
      AppendMessage_MessageInfo_MailboxNonExistent((mailbox, message) => mailbox.AppendMessage(message));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_MessageInfo_MailboxNonExistent()
    {
      AppendMessage_MessageInfo_MailboxNonExistent((mailbox, message) => mailbox.AppendMessageAsync(message).Wait());
    }
#endif

    private void AppendMessage_MessageInfo_MailboxNonExistent(Action<ImapMailboxInfo, ImapMessageInfo> testAction)
    {
      TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        server.EnqueueTaggedResponse("LIST ",
                                     string.Format("* LIST (\\NonExistent) \".\" INBOX.dest\r\n") +
                                     "$tag OK done\r\n");

        var mailbox = message.Client.GetMailbox("INBOX.dest");

        Assert.IsFalse(mailbox.Exists);

        Assert.Throws<ImapProtocolViolationException>(() => testAction(mailbox, message));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_MessageInfo_MailboxUnselectable()
    {
      AppendMessage_MessageInfo_MailboxUnselectable((mailbox, message) => mailbox.AppendMessage(message));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_MessageInfo_MailboxUnselectable()
    {
      AppendMessage_MessageInfo_MailboxUnselectable((mailbox, message) => mailbox.AppendMessageAsync(message).Wait());
    }
#endif

    private void AppendMessage_MessageInfo_MailboxUnselectable(Action<ImapMailboxInfo, ImapMessageInfo> testAction)
    {
      TestUtils2.TestMessage(delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        server.EnqueueTaggedResponse("LIST ",
                                     string.Format("* LIST (\\Noselect) \".\" INBOX.dest\r\n") +
                                     "$tag OK done\r\n");

        var mailbox = message.Client.GetMailbox("INBOX.dest");

        Assert.IsTrue(mailbox.IsUnselectable);

        Assert.Throws<ImapProtocolViolationException>(() => testAction(mailbox, message));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_Action()
    {
      AppendMessageAction((mailbox, length, dtm, flags, write) => mailbox.AppendMessage(dtm, flags, write));
    }

    [Test]
    public void TestAppendMessage_Action_LengthSpecified()
    {
      AppendMessageAction((mailbox, length, dtm, flags, write) => mailbox.AppendMessage(length, dtm, flags, write));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_Action()
    {
      AppendMessageAction((mailbox, length, dtm, flags, write) => mailbox.AppendMessageAsync(dtm, flags, write).Wait());
    }

    [Test]
    public void TestAppendMessageAsync_Action_LengthSpecified()
    {
      AppendMessageAction((mailbox, length, dtm, flags, write) => mailbox.AppendMessageAsync(length, dtm, flags, write).Wait());
    }
#endif

    private void AppendMessageAction(Action<ImapMailboxInfo, long, DateTimeOffset, IImapMessageFlagSet, Action<Stream>> testAction)
    {
      TestUtils2.TestMailbox("INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        const string uploadingMessage = "MIME-Version: 1.0\r\n\r\ntest message";

        var length = uploadingMessage.Length;
        var flags = new ImapMessageFlagSet(new[] {"$label1"}, ImapMessageFlag.Seen);
        var internalDate = new DateTimeOffset(2011, 1, 24, 14, 52, 30, TimeSpan.FromHours(+9.0));
        var writeUploadMessage = new Action<Stream>(delegate(Stream stream) {
          var writer = new StreamWriter(stream, Encoding.ASCII);

          writer.Write(uploadingMessage);
          writer.Flush();
        });

        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueTaggedResponse(string.Format("APPEND INBOX (\\Seen $label1) \"24-Jan-2011 14:52:30 +0900\" {{{0}}}\r\n{1}\r\n",
                                                     length,
                                                     uploadingMessage),
                                       "$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");
        });

        testAction(mailbox, length, internalDate, flags, writeUploadMessage);

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_Action_MailboxNonExistent()
    {
      AppendMessage_Action_MailboxNonExistent(mailbox => mailbox.AppendMessage(stream => {}));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_Action_MailboxNonExistent()
    {
      AppendMessage_Action_MailboxNonExistent(mailbox => mailbox.AppendMessageAsync(stream => {}));
    }
#endif

    private void AppendMessage_Action_MailboxNonExistent(Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NonExistent},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        Assert.Throws<ImapProtocolViolationException>(() => testAction(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_Action_MailboxUnselectable()
    {
      AppendMessage_Action_MailboxUnselectable(mailbox => mailbox.AppendMessage(stream => {}));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_Action_MailboxUnselectable()
    {
      AppendMessage_Action_MailboxUnselectable(mailbox => mailbox.AppendMessageAsync(stream => {}));
    }
#endif

    private void AppendMessage_Action_MailboxUnselectable(Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox(null,
                             new[] {ImapMailboxFlag.NoSelect},
                             "INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsUnselectable);

        Assert.Throws<ImapProtocolViolationException>(() => testAction(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_Action_LengthSpecified_OutOfRange()
    {
      AppendMessage_Action_LengthSpecified_OutOfRange(mailbox => mailbox.AppendMessage(-1L, stream => {}));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_Action_LengthSpecified_OutOfRange()
    {
      AppendMessage_Action_LengthSpecified_OutOfRange(mailbox => mailbox.AppendMessageAsync(-1L, stream => {}));
    }
#endif

    private void AppendMessage_Action_LengthSpecified_OutOfRange(Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("INBOX",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => testAction(mailbox));

        Assert.IsFalse(mailbox.Client.IsBusy);
      });
    }

    [Test]
    public void TestAppendMessage_Action_NoTryCreate()
    {
      AppendMessage_Action_NoTryCreate(mailbox => mailbox.AppendMessage(1L, stream => stream.WriteByte(0x40)));
    }

#if ASYNC_OPS
    [Test]
    public void TestAppendMessageAsync_Action_NoTryCreate()
    {
      AppendMessage_Action_NoTryCreate(mailbox => mailbox.AppendMessageAsync(1L, stream => stream.WriteByte(0x40)).Wait());
    }
#endif

    private void AppendMessage_Action_NoTryCreate(Action<ImapMailboxInfo> testAction)
    {
      TestUtils2.TestMailbox("NonExistent",
                             delegate(ImapPseudoServer2 server, ImapMailboxInfo mailbox) {
        var responseTask = TestUtils2.QueueResponseTask(server, mailbox.Client, delegate {
          server.EnqueueLiteralContinuationResponse("APPEND NonExistent {1}\r\n",
                                                    "$tag NO [TRYCREATE] mailbox not exist\r\n",
                                                    false);
        });

        var ex = TestUtils.Throws<ImapMailboxNotFoundException>(() => testAction(mailbox));

        responseTask.Wait();

        Assert.IsFalse(mailbox.Client.IsBusy);

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("NonExistent", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });
      });
    }
  }
}