﻿using System;
using System.Text;
using NUnit.Framework;

using Smdn.Formats;
using Smdn.Formats.Mime;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMessagePartInfoTests {
    private void TestMessage(string fetchResponseBodyStructure, Action<ImapMessageInfo> testAction)
    {
      var envelope =
        "(" +
        "\"Tue, 23 Mar 2010 09:46:52 +0900\" " +
        "\"test message\" " + 
        "((\"from address\" NIL \"from\" \"mail.example.com\")) " + 
        "((\"sender address\" NIL \"sender\" \"mail.example.com\")) " + 
        "((\"reply-to address\" NIL \"reply-to\" \"mail.example.com\")) " + 
        "((\"to address\" NIL \"to\" \"mail.example.com\")) " +
        "((\"cc address\" NIL \"cc\" \"mail.example.com\")) " +
        "((\"bcc address\" NIL \"bcc\" \"mail.example.com\")) " +
        "\"<in.reply.to@mail.example.com>\" " + 
        "\"<message.id@mail.example.com>\" " + 
        ")";
      var fetchResp =
        "* FETCH 1 (UID 1 " +
        "BODYSTRUCTURE " + fetchResponseBodyStructure +
        " ENVELOPE " + envelope +
        " INTERNALDATE \"23-Mar-2010 11:22:33 +0900\" RFC822.SIZE 1024)\r\n" +
        "$tag OK done\r\n";

      TestUtils2.TestMessage(fetchResp,
                             ImapMessageFetchAttributeOptions.StaticAttributes,
                             delegate(ImapPseudoServer2 server, ImapMessageInfo message) {
        Assert.IsFalse(message.Client.IsBusy);

        Assert.IsNotNull(message.MessagePart);

        testAction(message);
      });
    }

    [Test]
    public void TestProperties_SinglePart_WithNoExtensionData()
    {
      const string respBodyStructure =
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") \"content-id\" \"content-description\" \"7bit\" 2109 54)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.IsEmpty(message.MessagePart.Specifier, "Specifier");
        Assert.IsFalse(message.MessagePart.IsMultiPart, "IsMultiPart");
        Assert.IsFalse(message.MessagePart.HasEncapsulatedMessage, "HasEncapsulatedMessage");
        Assert.IsFalse(message.MessagePart.IsEncapsulatedMessagePart, "IsEncapsulatedMessagePart");
        Assert.IsNull(message.MessagePart.ParentPart, "ParentPart");
        Assert.IsTrue(message.MessagePart.IsRootPart, "IsRoot");
        Assert.AreEqual(0, message.MessagePart.NestLevel, "NestLevel");
        Assert.AreEqual(message.Url.AbsoluteUri, message.MessagePart.Url.AbsoluteUri, "Url");

        Assert.IsTrue(message.MessagePart.ContentType.EqualsIgnoreCase("text/plain"), "Content-Type");
        Assert.AreEqual(1, message.MessagePart.ContentTypeParameters.Count, "ContentTypeParameters.Count");
        Assert.AreEqual("us-ascii", message.MessagePart.ContentTypeParameters["charset"], "ContentTypeParameters[charset]");
        Assert.AreEqual("us-ascii", message.MessagePart.ContentTypeParameters["CHARSET"], "ContentTypeParameters[CHARSET]");
        Assert.AreEqual("us-ascii", message.MessagePart.ContentTypeCharset);

        Assert.AreEqual("content-id", message.MessagePart.ContentID, "ContentID");
        Assert.AreEqual("content-description", message.MessagePart.ContentDescription, "ContentDescription");
        Assert.AreEqual("7bit", message.MessagePart.ContentTransferEncoding, "ContentTransferEncoding");
        Assert.AreEqual(ContentTransferEncodingMethod.SevenBit,
                        message.MessagePart.ContentTransferEncodingMethod, "ContentTransferEncodingMethod");
        Assert.AreEqual(2109, message.MessagePart.Length, "Length");
        Assert.AreEqual(54, message.MessagePart.LineCount, "LineCount");

        Assert.IsNull(message.MessagePart.ContentMD5, "ContentMD5");
        Assert.IsNull(message.MessagePart.ContentDisposition, "ContentDisposition");
        Assert.IsFalse(message.MessagePart.IsDispositionAttachment, "IsDispositionAttachment");
        Assert.IsFalse(message.MessagePart.IsDispositionInline, "IsDispositionInline");
        Assert.IsEmpty(message.MessagePart.ContentDispositionParameters, "ContentDispositionParameters");
        Assert.IsEmpty(message.MessagePart.ContentLanguages, "ContentLanguages");
        Assert.IsNull(message.MessagePart.ContentLocation, "ContentLocation");

        Assert.IsNull(message.MessagePart.EncapsulatedMessagePart, "EncapsulatedMessagePart");
        Assert.IsNull(message.MessagePart.EncapsulatedMessageEnvelope, "EncapsulatedMessageEnvelope");
        Assert.IsEmpty(message.MessagePart.NestedMessageParts, "NestedMessageParts");

        Assert.AreSame(message.MessagePart.NestedMessageParts, message.MessagePart.NestedMessageParts, "NestedMessageParts equality");
      });
    }

    [Test]
    public void TestProperties_SinglePart_WithExtensionData()
    {
      const string respBodyStructure =
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") \"content-id\" \"content-description\" \"7bit\" 2109 54 \"content-md5\" (\"attachment\" (\"filename\" \"foo.txt\")) (ja en) \"http://localhost/\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("content-md5", message.MessagePart.ContentMD5, "ContentMD5");
        Assert.AreEqual("attachment", message.MessagePart.ContentDisposition, "ContentDisposition");
        Assert.IsTrue(message.MessagePart.IsDispositionAttachment, "IsDispositionAttachment");
        Assert.IsFalse(message.MessagePart.IsDispositionInline, "IsDispositionInline");
        Assert.AreEqual(1, message.MessagePart.ContentDispositionParameters.Count, "ContentDispositionParameters.Count");
        Assert.AreEqual("foo.txt", message.MessagePart.ContentDispositionParameters["filename"], "ContentDispositionParameters[filename]");
        Assert.AreEqual("foo.txt", message.MessagePart.ContentDispositionParameters["FILENAME"], "ContentDispositionParameters[FILENAME]");
        Assert.AreEqual(2, message.MessagePart.ContentLanguages.Count, "ContentLanguages.Count");
        CollectionAssert.AreEquivalent(new[] {"ja", "en"},
                                       message.MessagePart.ContentLanguages,
                                       "ContentLanguages");
        Assert.AreEqual("http://localhost/",
                        message.MessagePart.ContentLocation.AbsoluteUri,
                        "ContentLocation");
      });
    }

    [Test]
    public void TestProperties_SinglePart_WithExtensionData_Nil()
    {
      const string respBodyStructure =
        "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 1 1 NIL NIL NIL NIL)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.IsNull(message.MessagePart.ContentMD5, "ContentMD5");
        Assert.IsNull(message.MessagePart.ContentDisposition, "ContentDisposition");
        Assert.IsFalse(message.MessagePart.IsDispositionAttachment, "IsDispositionAttachment");
        Assert.IsFalse(message.MessagePart.IsDispositionInline, "IsDispositionInline");
        Assert.IsEmpty(message.MessagePart.ContentDispositionParameters, "ContentDispositionParameters");
        Assert.IsEmpty(message.MessagePart.ContentLanguages, "ContentLanguages");
        Assert.IsNull(message.MessagePart.ContentLocation, "ContentLocation");
      });
    }

    [Test]
    public void TestProperties_Encapsulating_WithNoExtensionData()
    {
      const string envelope = "(NIL \"test mail\" NIL NIL NIL NIL NIL NIL NIL NIL)";
      const string respBodyStructure = 
        "(" +
        "\"message\" \"rfc822\" (\"foo\" \"bar\") \"content-id\" \"content-description\" \"7bit\" 188 " +
        envelope + " " +
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") NIL NIL \"7bit\" 121 6 NIL NIL NIL NIL) 10" +
        ")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.IsEmpty(message.MessagePart.Specifier, "Specifier");
        Assert.IsFalse(message.MessagePart.IsMultiPart, "IsMultiPart");
        Assert.IsTrue(message.MessagePart.HasEncapsulatedMessage, "HasEncapsulatedMessage");
        Assert.IsFalse(message.MessagePart.IsEncapsulatedMessagePart, "IsEncapsulatedMessagePart");
        Assert.IsNull(message.MessagePart.ParentPart, "ParentPart");
        Assert.IsTrue(message.MessagePart.IsRootPart, "IsRoot");
        Assert.AreEqual(0, message.MessagePart.NestLevel, "NestLevel");
        Assert.AreEqual(message.Url.AbsoluteUri, message.MessagePart.Url.AbsoluteUri, "Url");

        Assert.IsTrue(message.MessagePart.ContentType.EqualsIgnoreCase("message/rfc822"), "Content-Type");
        Assert.AreEqual(1, message.MessagePart.ContentTypeParameters.Count, "ContentTypeParameters.Count");
        Assert.AreEqual("bar", message.MessagePart.ContentTypeParameters["foo"], "ContentTypeParameters[foo]");
        Assert.AreEqual("bar", message.MessagePart.ContentTypeParameters["FOO"], "ContentTypeParameters[FOO]");
        Assert.IsNull(message.MessagePart.ContentTypeCharset);

        Assert.AreEqual("content-id", message.MessagePart.ContentID, "ContentID");
        Assert.AreEqual("content-description", message.MessagePart.ContentDescription, "ContentDescription");
        Assert.AreEqual("7bit", message.MessagePart.ContentTransferEncoding, "ContentTransferEncoding");
        Assert.AreEqual(ContentTransferEncodingMethod.SevenBit,
                        message.MessagePart.ContentTransferEncodingMethod, "ContentTransferEncodingMethod");
        Assert.AreEqual(188, message.MessagePart.Length, "Length");
        Assert.AreEqual(10, message.MessagePart.LineCount, "LineCount");

        Assert.IsNull(message.MessagePart.ContentMD5, "ContentMD5");
        Assert.IsNull(message.MessagePart.ContentDisposition, "ContentDisposition");
        Assert.IsFalse(message.MessagePart.IsDispositionAttachment, "IsDispositionAttachment");
        Assert.IsFalse(message.MessagePart.IsDispositionInline, "IsDispositionInline");
        Assert.IsEmpty(message.MessagePart.ContentDispositionParameters, "ContentDispositionParameters");
        Assert.IsEmpty(message.MessagePart.ContentLanguages, "ContentLanguages");
        Assert.IsNull(message.MessagePart.ContentLocation, "ContentLocation");

        Assert.IsEmpty(message.MessagePart.NestedMessageParts, "NestedMessageParts");

        var encapsulatedPart = message.MessagePart.EncapsulatedMessagePart;

        Assert.IsNotNull(encapsulatedPart, "EncapsulatedMessagePart");
        Assert.AreSame(encapsulatedPart, message.MessagePart.EncapsulatedMessagePart, "EncapsulatedMessagePart equality");
        Assert.IsFalse(encapsulatedPart.HasEncapsulatedMessage, "EncapsulatedMessagePart.HasEncapsulatedMessage");
        Assert.IsTrue(encapsulatedPart.IsEncapsulatedMessagePart, "EncapsulatedMessagePart.IsEncapsulatedMessagePart");
        Assert.AreEqual(1, encapsulatedPart.NestLevel, "EncapsulatedMessagePart.NestLevel");
        Assert.IsTrue(message.MessagePart.EncapsulatedMessagePart.ContentType.EqualsIgnoreCase("text/plain"), "EncapsulatedMessagePart.ContentType");
        Assert.AreEqual("us-ascii", message.MessagePart.EncapsulatedMessagePart.ContentTypeCharset, "EncapusulatedMessagePart.ContentTypeCharset");
        Assert.AreEqual(121, message.MessagePart.EncapsulatedMessagePart.Length, "EncapsulatedMessagePart.Length");
        Assert.AreEqual(6, message.MessagePart.EncapsulatedMessagePart.LineCount, "EncapsulatedMessagePart.LineCount");
        Assert.AreSame(message.MessagePart, encapsulatedPart.ParentPart, "EncapsulatedMessagePart.ParentPart #1");
        Assert.AreSame(encapsulatedPart.ParentPart, encapsulatedPart.ParentPart, "EncapsulatedMessagePart.ParentPart #2");

        Assert.IsNotNull(message.MessagePart.EncapsulatedMessageEnvelope, "EncapsulatedMessageEnvelope");
        Assert.AreEqual("test mail", message.MessagePart.EncapsulatedMessageEnvelope.Subject, "EncapsulatedMessageEnvelope.Subject");
      });
    }

    [Test]
    public void TestProperties_Encapsulating_WithExtensionData()
    {
      const string envelope = "(NIL \"test mail\" NIL NIL NIL NIL NIL NIL NIL NIL)";
      const string respBodyStructure = 
        "(" +
        "\"message\" \"rfc822\" (\"foo\" \"bar\") \"content-id\" \"content-description\" \"7bit\" 188 " +
        envelope + " " +
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") NIL NIL \"7bit\" 121 6 NIL NIL NIL NIL) 10 \"content-md5\" (\"attachment\" (\"filename\" \"foo.txt\")) (ja en) \"http://localhost/\"" +
        ")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("content-md5", message.MessagePart.ContentMD5, "ContentMD5");
        Assert.AreEqual("attachment", message.MessagePart.ContentDisposition, "ContentDisposition");
        Assert.IsTrue(message.MessagePart.IsDispositionAttachment, "IsDispositionAttachment");
        Assert.IsFalse(message.MessagePart.IsDispositionInline, "IsDispositionInline");
        Assert.AreEqual(1, message.MessagePart.ContentDispositionParameters.Count, "ContentDispositionParameters.Count");
        Assert.AreEqual("foo.txt", message.MessagePart.ContentDispositionParameters["filename"], "ContentDispositionParameters[filename]");
        Assert.AreEqual("foo.txt", message.MessagePart.ContentDispositionParameters["FILENAME"], "ContentDispositionParameters[FILENAME]");
        Assert.AreEqual(2, message.MessagePart.ContentLanguages.Count, "ContentLanguages.Count");
        CollectionAssert.AreEquivalent(new[] {"ja", "en"},
                                       message.MessagePart.ContentLanguages,
                                       "ContentLanguages");
        Assert.AreEqual("http://localhost/",
                        message.MessagePart.ContentLocation.AbsoluteUri,
                        "ContentLocation");
      });
    }
      
    [Test]
    public void TestProperties_MultiPart_WithNoExtensionData()
    {
      const string respBodyStructure =
        "(" + 
        "(\"TEXT\" \"PLAIN\" (\"CHARSET\" \"US-ASCII\") NIL NIL \"7BIT\" 1152 23)" + 
        "(\"TEXT\" \"PLAIN\" (\"CHARSET\" \"US-ASCII\" \"NAME\" \"cc.diff\") \"<960723163407.20117h@cac.washington.edu>\" \"Compiler diff\" \"BASE64\" 4554 73)" +
        " \"MIXED\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.IsEmpty(message.MessagePart.Specifier, "Specifier");
        Assert.IsTrue(message.MessagePart.IsMultiPart, "IsMultiPart");
        Assert.IsFalse(message.MessagePart.HasEncapsulatedMessage, "HasEncapsulatedMessage");
        Assert.IsFalse(message.MessagePart.IsEncapsulatedMessagePart, "IsEncapsulatedMessagePart");
        Assert.IsNull(message.MessagePart.ParentPart, "ParentPart");
        Assert.IsTrue(message.MessagePart.IsRootPart, "IsRoot");
        Assert.AreEqual(0, message.MessagePart.NestLevel, "NestLevel");
        Assert.AreEqual(message.Url.AbsoluteUri, message.MessagePart.Url.AbsoluteUri, "Url");

        Assert.IsTrue(message.MessagePart.ContentType.EqualsIgnoreCase("multipart/mixed"), "Content-Type");
        Assert.IsEmpty(message.MessagePart.ContentTypeParameters, "ContentTypeParameters");
        Assert.IsNull(message.MessagePart.ContentTypeCharset);

        Assert.IsNull(message.MessagePart.ContentMD5, "ContentMD5");
        Assert.IsNull(message.MessagePart.ContentDisposition, "ContentDisposition");
        Assert.IsFalse(message.MessagePart.IsDispositionAttachment, "IsDispositionAttachment");
        Assert.IsFalse(message.MessagePart.IsDispositionInline, "IsDispositionInline");
        Assert.IsEmpty(message.MessagePart.ContentDispositionParameters, "ContentDispositionParameters");
        Assert.IsEmpty(message.MessagePart.ContentLanguages, "ContentLanguages");
        Assert.IsNull(message.MessagePart.ContentLocation, "ContentLocation");

        Assert.IsNull(message.MessagePart.EncapsulatedMessagePart, "EncapsulatedMessagePart");
        Assert.IsNull(message.MessagePart.EncapsulatedMessageEnvelope, "EncapsulatedMessageEnvelope");

        Assert.AreSame(message.MessagePart.NestedMessageParts, message.MessagePart.NestedMessageParts, "NestedMessageParts equality");
        Assert.AreEqual(2, message.MessagePart.NestedMessageParts.Count, "NestedMessageParts.Count");

        var part1 = message.MessagePart.NestedMessageParts[0];

        Assert.AreSame(part1, message.MessagePart.NestedMessageParts[0]);
        Assert.AreSame(part1, message.GetMessagePart(1));
        Assert.AreEqual("1", part1.Specifier, "part1.Specifier");
        Assert.IsFalse(part1.IsMultiPart, "part1.IsMultiPart");
        Assert.IsFalse(part1.HasEncapsulatedMessage, "part1.HasEncapsulatedMessage");
        Assert.IsFalse(part1.IsEncapsulatedMessagePart, "part1.IsEncapsulatedMessagePart");
        Assert.AreSame(message.MessagePart, part1.ParentPart, "part1.ParentPart #1");
         Assert.AreSame(part1.ParentPart, part1.ParentPart, "part1.ParentPart #2");
        Assert.IsFalse(part1.IsRootPart, "part1.IsRoot");
        Assert.AreEqual(1, part1.NestLevel, "part1.NestLevel");
        Assert.AreEqual(part1.Url.AbsoluteUri, message.MessagePart.Url.AbsoluteUri + "/;SECTION=1", "part1.Url");

        var part2 = message.MessagePart.NestedMessageParts[1];

        Assert.AreSame(part2, message.MessagePart.NestedMessageParts[1]);
        Assert.AreSame(part2, message.GetMessagePart(2));
        Assert.AreEqual("2", part2.Specifier, "part2.Specifier");
        Assert.IsFalse(part2.IsMultiPart, "part2.IsMultiPart");
        Assert.IsFalse(part2.HasEncapsulatedMessage, "part2.HasEncapsulatedMessage");
        Assert.IsFalse(part2.IsEncapsulatedMessagePart, "part2.IsEncapsulatedMessagePart");
        Assert.AreSame(message.MessagePart, part2.ParentPart, "part2.ParentPart #1");
        Assert.AreSame(part2.ParentPart, part2.ParentPart, "part2.ParentPart #2");
        Assert.IsFalse(part2.IsRootPart, "part2.IsRoot");
        Assert.AreEqual(1, part2.NestLevel, "part2.NestLevel");
        Assert.AreEqual(part2.Url.AbsoluteUri, message.MessagePart.Url.AbsoluteUri + "/;SECTION=2", "part2.Url");
      });
    }

    [Test]
    public void TestProperties_MultiPart_WithExtensionData()
    {
      const string respBodyStructure =
        "(" + 
        "(\"TEXT\" \"PLAIN\" (\"CHARSET\" \"US-ASCII\") NIL NIL \"7BIT\" 1152 23)" + 
        "(\"TEXT\" \"PLAIN\" (\"CHARSET\" \"US-ASCII\" \"NAME\" \"cc.diff\") \"<960723163407.20117h@cac.washington.edu>\" \"Compiler diff\" \"BASE64\" 4554 73)" +
        " \"MIXED\" (\"foo\" \"bar\") (\"inline\" (\"filename\" \"foo.txt\")) (ja en) \"http://localhost/\"" +
        ")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.IsTrue(message.MessagePart.ContentType.EqualsIgnoreCase("multipart/mixed"), "Content-Type");
        Assert.AreEqual(1, message.MessagePart.ContentTypeParameters.Count, "ContentTypeParameters.Count");
        Assert.AreEqual("bar", message.MessagePart.ContentTypeParameters["foo"], "ContentTypeParameters[foo]");
        Assert.AreEqual("bar", message.MessagePart.ContentTypeParameters["FOO"], "ContentTypeParameters[FOO]");

        Assert.IsNull(message.MessagePart.ContentMD5, "ContentMD5");
        Assert.AreEqual("inline", message.MessagePart.ContentDisposition, "ContentDisposition");
        Assert.IsFalse(message.MessagePart.IsDispositionAttachment, "IsDispositionAttachment");
        Assert.IsTrue(message.MessagePart.IsDispositionInline, "IsDispositionInline");
        Assert.AreEqual(1, message.MessagePart.ContentDispositionParameters.Count, "ContentDispositionParameters.Count");
        Assert.AreEqual("foo.txt", message.MessagePart.ContentDispositionParameters["filename"], "ContentDispositionParameters[filename]");
        Assert.AreEqual("foo.txt", message.MessagePart.ContentDispositionParameters["FILENAME"], "ContentDispositionParameters[FILENAME]");
        Assert.AreEqual(2, message.MessagePart.ContentLanguages.Count, "ContentLanguages.Count");
        CollectionAssert.AreEquivalent(new[] {"ja", "en"},
                                       message.MessagePart.ContentLanguages,
                                       "ContentLanguages");
        Assert.AreEqual("http://localhost/",
                        message.MessagePart.ContentLocation.AbsoluteUri,
                        "ContentLocation");
      });
    }

    [Test]
    public void TestProperties_MultiPart_SingleContent()
    {
      const string respBodyStructure =
        "(" + 
        "(\"TEXT\" \"PLAIN\" (\"CHARSET\" \"US-ASCII\") NIL NIL \"7BIT\" 1152 23)" + 
        " \"MIXED\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.IsEmpty(message.MessagePart.Specifier, "Specifier");
        Assert.IsTrue(message.MessagePart.IsMultiPart, "IsMultiPart");
        Assert.IsFalse(message.MessagePart.HasEncapsulatedMessage, "HasEncapsulatedMessage");
        Assert.IsFalse(message.MessagePart.IsEncapsulatedMessagePart, "IsEncapsulatedMessagePart");
        Assert.IsNull(message.MessagePart.ParentPart, "ParentPart");
        Assert.IsTrue(message.MessagePart.IsRootPart, "IsRoot");
        Assert.AreEqual(0, message.MessagePart.NestLevel, "NestLevel");
        Assert.AreEqual(message.Url.AbsoluteUri, message.MessagePart.Url.AbsoluteUri, "Url");

        Assert.IsNull(message.MessagePart.EncapsulatedMessagePart, "EncapsulatedMessagePart");
        Assert.IsNull(message.MessagePart.EncapsulatedMessageEnvelope, "EncapsulatedMessageEnvelope");

        Assert.AreEqual(1, message.MessagePart.NestedMessageParts.Count, "NestedMessageParts.Count");

        var part1 = message.MessagePart.NestedMessageParts[0];

        Assert.AreSame(part1, message.GetMessagePart(1));
        Assert.AreEqual("1", part1.Specifier, "part1.Specifier");
        Assert.IsFalse(part1.IsMultiPart, "part1.IsMultiPart");
        Assert.IsFalse(part1.HasEncapsulatedMessage, "part1.HasEncapsulatedMessage");
        Assert.IsFalse(part1.IsEncapsulatedMessagePart, "part1.IsEncapsulatedMessagePart");
        Assert.AreSame(message.MessagePart, part1.ParentPart, "part1.ParentPart #1");
        Assert.AreSame(part1.ParentPart, part1.ParentPart, "part1.ParentPart #2");
        Assert.IsFalse(part1.IsRootPart, "part1.IsRoot");
        Assert.AreEqual(1, part1.NestLevel, "part1.NestLevel");
        Assert.AreEqual(part1.Url.AbsoluteUri, message.MessagePart.Url.AbsoluteUri + "/;SECTION=1", "part1.Url");
      });
    }

    [Test]
    public void TestTraverse_SinglePart()
    {
      const string respBodyStructure =
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") \"content-id\" \"content-description\" \"7bit\" 2109 54)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var visitInToCount = 0;
        var visitOutFromCount = 0;
        var visitInToMessagePartAssersions = new Action<ImapMessagePartInfo>[] {
          p => {
            Assert.AreEqual(string.Empty, p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("text/plain"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
          },
        };
        var visitOutFromMessagePartAssersions = visitInToMessagePartAssersions;

        message.MessagePart.Traverse(true,
                                     partInTo => visitInToMessagePartAssersions[visitInToCount++](partInTo),
                                     partOutTo => visitOutFromMessagePartAssersions[visitOutFromCount++](partOutTo));

        Assert.AreEqual(visitInToMessagePartAssersions.Length,
                        visitInToCount,
                        "visitInToCount, visitItself");
        Assert.AreEqual(visitOutFromMessagePartAssersions.Length,
                        visitOutFromCount,
                        "visitOutFromCount, visitItself");

        visitInToCount = 0;
        visitOutFromCount = 0;

        message.MessagePart.Traverse(false,
                                     partInTo => visitInToCount++,
                                     partOutTo => visitOutFromCount++);

        Assert.AreEqual(0, visitInToCount, "visitInToCount, not visitItself");
        Assert.AreEqual(0, visitOutFromCount, "visitOutFromCount, not visitItself");
      });
    }

    [Test]
    public void TestTraverse_Encapsulating()
    {
      const string envelope = "(NIL \"test mail\" NIL NIL NIL NIL NIL NIL NIL NIL)";
      const string respBodyStructure = 
        "(" +
        "\"message\" \"rfc822\" (\"foo\" \"bar\") \"content-id\" \"content-description\" \"7bit\" 188 " +
        envelope + " " +
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") NIL NIL \"7bit\" 121 6 NIL NIL NIL NIL) 10" +
        ")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var visitInToCount = 0;
        var visitOutFromCount = 0;
        var visitInToMessagePartAssersions = new Action<ImapMessagePartInfo>[] {
          p => {
            Assert.AreEqual(string.Empty, p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("message/rfc822"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsTrue(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
          },
          p => {
            Assert.AreEqual("1", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("text/plain"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsTrue(p.IsEncapsulatedMessagePart);
          },
        };
        var visitOutFromMessagePartAssersions = (Action<ImapMessagePartInfo>[])visitInToMessagePartAssersions.Clone();

        Array.Reverse(visitOutFromMessagePartAssersions);

        message.MessagePart.Traverse(true,
                                     partInTo => visitInToMessagePartAssersions[visitInToCount++](partInTo),
                                     partOutTo => visitOutFromMessagePartAssersions[visitOutFromCount++](partOutTo));

        Assert.AreEqual(visitInToMessagePartAssersions.Length,
                        visitInToCount,
                        "visitInToCount, visitItself");
        Assert.AreEqual(visitOutFromMessagePartAssersions.Length,
                        visitOutFromCount,
                        "visitOutFromCount, visitItself");

        visitInToCount = 0;
        visitOutFromCount = 0;

        message.MessagePart.Traverse(false,
                                     partInTo => {visitInToCount++; Assert.AreEqual("1", partInTo.Specifier); },
                                     partOutTo => {visitOutFromCount++; Assert.AreEqual("1", partOutTo.Specifier); });

        Assert.AreEqual(1, visitInToCount, "visitInToCount, not visitItself");
        Assert.AreEqual(1, visitOutFromCount, "visitOutFromCount, not visitItself");
      });
    }

    [Test]
    public void TestTraverse_MultiPart()
    {
      const string bodyStructurePart1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 1 1)";
      const string bodyStructurePart2_1 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 21)";
      const string bodyStructurePart2_2_1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 22 1)";
      const string bodyStructurePart2_2_2 = "(\"text\" \"richtext\" NIL NIL NIL \"base64\" 22 2)";
      const string bodyStructurePart2_2_3 = "(\"text\" \"html\" NIL NIL NIL \"quoted-printable\" 22 3)";
      const string bodyStructurePart2_2 = "(" + bodyStructurePart2_2_1 + bodyStructurePart2_2_2 + bodyStructurePart2_2_3 + " \"alternative\")";
      const string bodyStructurePart2 = "(" + bodyStructurePart2_1 + bodyStructurePart2_2 + " \"mixed\")";
      const string bodyStructurePart3_1 = "(\"image\" \"gif\" NIL NIL NIL \"base64\" 31)";
      const string bodyStructurePart3_2 = "(\"image\" \"png\" NIL NIL NIL \"base64\" 32)";
      const string bodyStructurePart3 = "(" + bodyStructurePart3_1 + bodyStructurePart3_2 + " \"alternative\")";
      const string bodyStructurePart4 = "(\"application\" \"octet-stream\" NIL NIL NIL \"base64\" 4)";

      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + bodyStructurePart3 + bodyStructurePart4 + " \"mixed\")";

      // multipart/mixed
      //   1 text/plain
      //   2 multipart/mixed
      //     2.1 image/jpeg
      //     2.2 multipart/alternative
      //       2.2.1 text/plain
      //       2.2.2 text/richtext
      //       2.2.3 text/html
      //   3 multipart/alternative
      //     3.1 image/gif
      //     3.2 image/png
      //   4 application/octet-stream

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var visitInToCount = 0;
        var visitOutFromCount = 0;
        var visitInToMessagePartAssersions = new Action<ImapMessagePartInfo>[] {
          p => {
            Assert.AreEqual(string.Empty, p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("multipart/mixed"));
            Assert.IsTrue(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
            Assert.AreEqual(0, p.NestLevel);
          },
          p => {
            Assert.AreEqual("1", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("text/plain"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
            Assert.AreEqual(1, p.NestLevel);
          },
          p => {
            Assert.AreEqual("2", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("multipart/mixed"));
            Assert.IsTrue(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
            Assert.AreEqual(1, p.NestLevel);
          },
          p => {
            Assert.AreEqual("2.1", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("image/jpeg"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
            Assert.AreEqual(2, p.NestLevel);
          },
          p => {
            Assert.AreEqual("2.2", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("multipart/alternative"));
            Assert.IsTrue(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
            Assert.AreEqual(2, p.NestLevel);
          },
          p => {
            Assert.AreEqual("2.2.1", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("text/plain"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
            Assert.AreEqual(3, p.NestLevel);
          },
          p => {
            Assert.AreEqual("2.2.2", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("text/richtext"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
            Assert.AreEqual(3, p.NestLevel);
          },
          p => {
            Assert.AreEqual("2.2.3", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("text/html"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
            Assert.AreEqual(3, p.NestLevel);
          },
          p => {
            Assert.AreEqual("3", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("multipart/alternative"));
            Assert.IsTrue(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
            Assert.AreEqual(1, p.NestLevel);
          },
          p => {
            Assert.AreEqual("3.1", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("image/gif"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
            Assert.AreEqual(2, p.NestLevel);
          },
          p => {
            Assert.AreEqual("3.2", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("image/png"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
            Assert.AreEqual(2, p.NestLevel);
          },
          p => {
            Assert.AreEqual("4", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("application/octet-stream"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
            Assert.AreEqual(1, p.NestLevel);
          },
        };
        var visitOutFromMessagePartAssersions = new Action<ImapMessagePartInfo>[] {
          p => {
            Assert.AreEqual("1", p.Specifier);
          },
          p => {
            Assert.AreEqual("2.1", p.Specifier);
          },
          p => {
            Assert.AreEqual("2.2.1", p.Specifier);
          },
          p => {
            Assert.AreEqual("2.2.2", p.Specifier);
          },
          p => {
            Assert.AreEqual("2.2.3", p.Specifier);
          },
          p => {
            Assert.AreEqual("2.2", p.Specifier);
          },
          p => {
            Assert.AreEqual("2", p.Specifier);
          },
          p => {
            Assert.AreEqual("3.1", p.Specifier);
          },
          p => {
            Assert.AreEqual("3.2", p.Specifier);
          },
          p => {
            Assert.AreEqual("3", p.Specifier);
          },
          p => {
            Assert.AreEqual("4", p.Specifier);
          },
          p => {
            Assert.AreEqual(string.Empty, p.Specifier);
          },
        };

        message.MessagePart.Traverse(true,
                                     partInTo => visitInToMessagePartAssersions[visitInToCount++](partInTo),
                                     partOutTo => visitOutFromMessagePartAssersions[visitOutFromCount++](partOutTo));

        Assert.AreEqual(visitInToMessagePartAssersions.Length,
                        visitInToCount,
                        "visitInToCount, visitItself");
        Assert.AreEqual(visitOutFromMessagePartAssersions.Length,
                        visitOutFromCount,
                        "visitOutFromCount, visitItself");

        visitInToCount = 0;
        visitOutFromCount = 0;

        visitInToMessagePartAssersions = visitInToMessagePartAssersions.Slice(1);
        visitOutFromMessagePartAssersions = visitOutFromMessagePartAssersions.Slice(0, visitOutFromMessagePartAssersions.Length - 1);

        message.MessagePart.Traverse(false,
                                     partInTo => visitInToMessagePartAssersions[visitInToCount++](partInTo),
                                     partOutTo => visitOutFromMessagePartAssersions[visitOutFromCount++](partOutTo));

        Assert.AreEqual(visitInToMessagePartAssersions.Length,
                        visitInToCount,
                        "visitInToCount, not visitItself");
        Assert.AreEqual(visitOutFromMessagePartAssersions.Length,
                        visitOutFromCount,
                        "visitOutFromCount, not visitItself");
      });
    }

    [Test]
    public void TestTraverse_MultiPart_TraverseFromNestedPart()
    {
      const string bodyStructurePart1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 1 1)";
      const string bodyStructurePart2_1 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 21)";
      const string bodyStructurePart2_2_1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 22 1)";
      const string bodyStructurePart2_2_2 = "(\"text\" \"richtext\" NIL NIL NIL \"base64\" 22 2)";
      const string bodyStructurePart2_2_3 = "(\"text\" \"html\" NIL NIL NIL \"quoted-printable\" 22 3)";
      const string bodyStructurePart2_2 = "(" + bodyStructurePart2_2_1 + bodyStructurePart2_2_2 + bodyStructurePart2_2_3 + " \"alternative\")";
      const string bodyStructurePart2 = "(" + bodyStructurePart2_1 + bodyStructurePart2_2 + " \"mixed\")";
      const string bodyStructurePart3_1 = "(\"image\" \"gif\" NIL NIL NIL \"base64\" 31)";
      const string bodyStructurePart3_2 = "(\"image\" \"png\" NIL NIL NIL \"base64\" 32)";
      const string bodyStructurePart3 = "(" + bodyStructurePart3_1 + bodyStructurePart3_2 + " \"alternative\")";
      const string bodyStructurePart4 = "(\"application\" \"octet-stream\" NIL NIL NIL \"base64\" 4)";

      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + bodyStructurePart3 + bodyStructurePart4 + " \"mixed\")";

      // multipart/mixed
      //   1 text/plain
      //   2 multipart/mixed
      //     2.1 image/jpeg
      //     2.2 multipart/alternative
      //       2.2.1 text/plain
      //       2.2.2 text/richtext
      //       2.2.3 text/html
      //   3 multipart/alternative
      //     3.1 image/gif
      //     3.2 image/png
      //   4 application/octet-stream

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var visitInToCount = 0;
        var visitOutFromCount = 0;
        var visitInToMessagePartAssersions = new Action<ImapMessagePartInfo>[] {
          p => {
            Assert.AreEqual("2", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("multipart/mixed"));
            Assert.IsTrue(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
          },
          p => {
            Assert.AreEqual("2.1", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("image/jpeg"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
          },
          p => {
            Assert.AreEqual("2.2", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("multipart/alternative"));
            Assert.IsTrue(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
          },
          p => {
            Assert.AreEqual("2.2.1", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("text/plain"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
          },
          p => {
            Assert.AreEqual("2.2.2", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("text/richtext"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
          },
          p => {
            Assert.AreEqual("2.2.3", p.Specifier);
            Assert.IsTrue(p.ContentType.EqualsIgnoreCase("text/html"));
            Assert.IsFalse(p.IsMultiPart);
            Assert.IsFalse(p.HasEncapsulatedMessage);
            Assert.IsFalse(p.IsEncapsulatedMessagePart);
          },
        };
        var visitOutFromMessagePartAssersions = new Action<ImapMessagePartInfo>[] {
          p => {
            Assert.AreEqual("2.1", p.Specifier);
          },
          p => {
            Assert.AreEqual("2.2.1", p.Specifier);
          },
          p => {
            Assert.AreEqual("2.2.2", p.Specifier);
          },
          p => {
            Assert.AreEqual("2.2.3", p.Specifier);
          },
          p => {
            Assert.AreEqual("2.2", p.Specifier);
          },
          p => {
            Assert.AreEqual("2", p.Specifier);
          },
        };

        var nestedPart = message.GetMessagePart("2");

        nestedPart.Traverse(true,
                            partInTo => visitInToMessagePartAssersions[visitInToCount++](partInTo),
                            partOutTo => visitOutFromMessagePartAssersions[visitOutFromCount++](partOutTo));

        Assert.AreEqual(visitInToMessagePartAssersions.Length,
                        visitInToCount,
                        "visitInToCount, visitItself");
        Assert.AreEqual(visitOutFromMessagePartAssersions.Length,
                        visitOutFromCount,
                        "visitOutFromCount, visitItself");

        visitInToCount = 0;
        visitOutFromCount = 0;

        visitInToMessagePartAssersions = visitInToMessagePartAssersions.Slice(1);
        visitOutFromMessagePartAssersions = visitOutFromMessagePartAssersions.Slice(0, visitOutFromMessagePartAssersions.Length - 1);

        nestedPart.Traverse(false,
                            partInTo => visitInToMessagePartAssersions[visitInToCount++](partInTo),
                            partOutTo => visitOutFromMessagePartAssersions[visitOutFromCount++](partOutTo));

        Assert.AreEqual(visitInToMessagePartAssersions.Length,
                        visitInToCount,
                        "visitInToCount, not visitItself");
        Assert.AreEqual(visitOutFromMessagePartAssersions.Length,
                        visitOutFromCount,
                        "visitOutFromCount, not visitItself");
      });
    }


    [Test]
    public void TestTraverse_ArgumentNull()
    {
      const string respBodyStructure =
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") \"content-id\" \"content-description\" \"7bit\" 2109 54)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.DoesNotThrow(() => message.MessagePart.Traverse((Action<ImapMessagePartInfo>)null));
        Assert.DoesNotThrow(() => message.MessagePart.Traverse((Action<ImapMessagePartInfo>)null, (Action<ImapMessagePartInfo>)null));
      });
    }

    [Test]
    public void TestFlatten_SinglePart()
    {
      const string respBodyStructure =
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") \"content-id\" \"content-description\" \"7bit\" 2109 54)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var flatten = message.MessagePart.Flatten();

        Assert.IsEmpty(flatten);

        flatten = message.MessagePart.Flatten(true);

        Assert.AreEqual(1, flatten.Count);
        Assert.AreSame(message.MessagePart, flatten[0]);
      });
    }

    [Test]
    public void TestFlatten_Encapsulating()
    {
      const string envelope = "(NIL \"test mail\" NIL NIL NIL NIL NIL NIL NIL NIL)";
      const string respBodyStructure = 
        "(" +
        "\"message\" \"rfc822\" (\"foo\" \"bar\") \"content-id\" \"content-description\" \"7bit\" 188 " +
        envelope + " " +
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") NIL NIL \"7bit\" 121 6 NIL NIL NIL NIL) 10" +
        ")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var flatten = message.MessagePart.Flatten();

        Assert.AreEqual(1, flatten.Count);
        Assert.AreEqual("1", flatten[0].Specifier);
        Assert.IsTrue(flatten[0].ContentType.EqualsIgnoreCase("text/plain"));

        flatten = message.MessagePart.Flatten(true);

        Assert.AreEqual(2, flatten.Count);
        Assert.AreSame(message.MessagePart, flatten[0]);

        Assert.AreEqual("1", flatten[1].Specifier);
        Assert.IsTrue(flatten[1].ContentType.EqualsIgnoreCase("text/plain"));

        Assert.IsEmpty(flatten[1].Flatten());
      });
    }

    [Test]
    public void TestFlatten_MultiPart()
    {
      const string bodyStructurePart1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 1 1)";
      const string bodyStructurePart2_1 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 21)";
      const string bodyStructurePart2_2_1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 22 1)";
      const string bodyStructurePart2_2_2 = "(\"text\" \"richtext\" NIL NIL NIL \"base64\" 22 2)";
      const string bodyStructurePart2_2_3 = "(\"text\" \"html\" NIL NIL NIL \"quoted-printable\" 22 3)";
      const string bodyStructurePart2_2 = "(" + bodyStructurePart2_2_1 + bodyStructurePart2_2_2 + bodyStructurePart2_2_3 + " \"alternative\")";
      const string bodyStructurePart2 = "(" + bodyStructurePart2_1 + bodyStructurePart2_2 + " \"mixed\")";
      const string bodyStructurePart3_1 = "(\"image\" \"gif\" NIL NIL NIL \"base64\" 31)";
      const string bodyStructurePart3_2 = "(\"image\" \"png\" NIL NIL NIL \"base64\" 32)";
      const string bodyStructurePart3 = "(" + bodyStructurePart3_1 + bodyStructurePart3_2 + " \"alternative\")";
      const string bodyStructurePart4 = "(\"application\" \"octet-stream\" NIL NIL NIL \"base64\" 4)";

      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + bodyStructurePart3 + bodyStructurePart4 + " \"mixed\")";

      // multipart/mixed
      //   1 text/plain
      //   2 multipart/mixed
      //     2.1 image/jpeg
      //     2.2 multipart/alternative
      //       2.2.1 text/plain
      //       2.2.2 text/richtext
      //       2.2.3 text/html
      //   3 multipart/alternative
      //     3.1 image/gif
      //     3.2 image/png
      //   4 application/octet-stream

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var flatten = message.MessagePart.Flatten();

        Assert.AreEqual(11, flatten.Count);

        var index = 0;

        foreach (var pair in new[] {
          new {Specifier = "1", ContentType = "text/plain"},
          new {Specifier = "2", ContentType = "multipart/mixed"},
          new {Specifier = "2.1", ContentType = "image/jpeg"},
          new {Specifier = "2.2", ContentType = "multipart/alternative"},
          new {Specifier = "2.2.1", ContentType = "text/plain"},
          new {Specifier = "2.2.2", ContentType = "text/richtext"},
          new {Specifier = "2.2.3", ContentType = "text/html"},
          new {Specifier = "3", ContentType = "multipart/alternative"},
          new {Specifier = "3.1", ContentType = "image/gif"},
          new {Specifier = "3.2", ContentType = "image/png"},
          new {Specifier = "4", ContentType = "application/octet-stream"},
        }) {
          Assert.AreEqual(pair.Specifier, flatten[index].Specifier, "Specifier #{0}", index);
          Assert.IsTrue(flatten[index].ContentType.EqualsIgnoreCase(pair.ContentType), "Specifier #{0}", index);

          index++;
        }

        var subpart = message.GetMessagePart("3");

        flatten = subpart.Flatten(true);

        Assert.AreEqual(3, flatten.Count);

        index = 0;

        foreach (var pair in new[] {
          new {Specifier = "3", ContentType = "multipart/alternative"},
          new {Specifier = "3.1", ContentType = "image/gif"},
          new {Specifier = "3.2", ContentType = "image/png"},
        }) {
          Assert.AreEqual(pair.Specifier, flatten[index].Specifier, "Specifier #{0}", index);
          Assert.IsTrue(flatten[index].ContentType.EqualsIgnoreCase(pair.ContentType), "Specifier #{0}", index);

          index++;
        }
      });
    }

    [Test]
    public void TestFind()
    {
      // multipart/mixed
      //   1 text/plain
      //   2 multipart/mixed
      //     2.1 image/jpeg
      //     2.2 multipart/alternative
      //       2.2.1 text/plain
      //       2.2.2 text/richtext
      //       2.2.3 text/html
      //   3 multipart/alternative
      //     3.1 image/gif
      //     3.2 image/png
      //   4 application/octet-stream
      const string bodyStructurePart1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 1 1)";
      const string bodyStructurePart2_1 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 21)";
      const string bodyStructurePart2_2_1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 22 1)";
      const string bodyStructurePart2_2_2 = "(\"text\" \"richtext\" NIL NIL NIL \"base64\" 22 2)";
      const string bodyStructurePart2_2_3 = "(\"text\" \"html\" NIL NIL NIL \"quoted-printable\" 22 3)";
      const string bodyStructurePart2_2 = "(" + bodyStructurePart2_2_1 + bodyStructurePart2_2_2 + bodyStructurePart2_2_3 + " \"alternative\")";
      const string bodyStructurePart2 = "(" + bodyStructurePart2_1 + bodyStructurePart2_2 + " \"mixed\")";
      const string bodyStructurePart3_1 = "(\"image\" \"gif\" NIL NIL NIL \"base64\" 31)";
      const string bodyStructurePart3_2 = "(\"image\" \"png\" NIL NIL NIL \"base64\" 32)";
      const string bodyStructurePart3 = "(" + bodyStructurePart3_1 + bodyStructurePart3_2 + " \"alternative\")";
      const string bodyStructurePart4 = "(\"application\" \"octet-stream\" NIL NIL NIL \"base64\" 4)";

      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + bodyStructurePart3 + bodyStructurePart4 + " \"mixed\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("2", message.MessagePart.Find(p => p.ContentType.EqualsIgnoreCase("multipart/mixed")).Specifier, "#1");
        Assert.AreEqual("2.2.3", message.MessagePart.Find(p => p.ContentTransferEncoding == "quoted-printable").Specifier, "#2");
        Assert.AreEqual("1", message.MessagePart.Find(p => p.ContentType.EqualsIgnoreCase("TEXT/PLAIN")).Specifier, "#3");
        Assert.AreEqual("3.1", message.MessagePart.Find(p => p.ParentPart.Specifier == "3").Specifier, "#4");
        Assert.AreEqual("4", message.MessagePart.Find(p => p.Length == 4).Specifier, "#5");
      });
    }

    [Test]
    public void TestFind_NothingMatched()
    {
      const string respBodyStructure =
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") \"content-id\" \"content-description\" \"7bit\" 2109 54)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.IsNull(message.MessagePart.Find(p => p.IsMultiPart));
      });
    }

    [Test]
    public void TestFind_ArgumentNull()
    {
      const string respBodyStructure =
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") \"content-id\" \"content-description\" \"7bit\" 2109 54)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.MessagePart.Find((Predicate<ImapMessagePartInfo>)null));
      });
    }

    [Test]
    public void TestFindAll()
    {
      // multipart/mixed
      //   1 text/plain
      //   2 multipart/mixed
      //     2.1 image/jpeg
      //     2.2 multipart/alternative
      //       2.2.1 text/plain
      //       2.2.2 text/richtext
      //       2.2.3 text/html
      //   3 multipart/alternative
      //     3.1 image/gif
      //     3.2 image/png
      //   4 application/octet-stream
      const string bodyStructurePart1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 1 1)";
      const string bodyStructurePart2_1 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 21)";
      const string bodyStructurePart2_2_1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 22 1)";
      const string bodyStructurePart2_2_2 = "(\"text\" \"richtext\" NIL NIL NIL \"base64\" 22 2)";
      const string bodyStructurePart2_2_3 = "(\"text\" \"html\" NIL NIL NIL \"quoted-printable\" 22 3)";
      const string bodyStructurePart2_2 = "(" + bodyStructurePart2_2_1 + bodyStructurePart2_2_2 + bodyStructurePart2_2_3 + " \"alternative\")";
      const string bodyStructurePart2 = "(" + bodyStructurePart2_1 + bodyStructurePart2_2 + " \"mixed\")";
      const string bodyStructurePart3_1 = "(\"image\" \"gif\" NIL NIL NIL \"base64\" 31)";
      const string bodyStructurePart3_2 = "(\"image\" \"png\" NIL NIL NIL \"base64\" 32)";
      const string bodyStructurePart3 = "(" + bodyStructurePart3_1 + bodyStructurePart3_2 + " \"alternative\")";
      const string bodyStructurePart4 = "(\"application\" \"octet-stream\" NIL NIL NIL \"base64\" 4)";

      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + bodyStructurePart3 + bodyStructurePart4 + " \"mixed\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var foundPartCount = 0;
        var expectedPlainTextPartSpecifiers = new[] {
          "1", "2.2.1",
        };

        foreach (var foundPart in message.MessagePart.FindAll(p => p.ContentType.EqualsIgnoreCase("text/plain"))) {
          Assert.AreEqual(expectedPlainTextPartSpecifiers[foundPartCount++],
                          foundPart.Specifier,
                          "FindAll(text/plain) #{0}", foundPartCount);
        }

        Assert.AreEqual(expectedPlainTextPartSpecifiers.Length,
                        foundPartCount,
                        "FindAll(text/plain) result count");

        foundPartCount = 0;

        var expectedBase64PartSpecifiers = new[] {
          "2.1", "2.2.2", "3.1", "3.2", "4"
        };

        foreach (var foundPart in message.MessagePart.FindAll(p => p.ContentTransferEncodingMethod == ContentTransferEncodingMethod.Base64)) {
          Assert.AreEqual(expectedBase64PartSpecifiers[foundPartCount++],
                          foundPart.Specifier,
                          "FindAll(base64) #{0}", foundPartCount);
        }

        Assert.AreEqual(expectedBase64PartSpecifiers.Length,
                        foundPartCount,
                        "FindAll(text/plain) result count");
      });
    }

    [Test]
    public void TestFindAll_NothingMatched()
    {
      const string respBodyStructure =
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") \"content-id\" \"content-description\" \"7bit\" 2109 54)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.IsEmpty(message.MessagePart.FindAll(p => p.IsMultiPart));
      });
    }

    [Test]
    public void TestFindAll_ArgumentNull()
    {
      const string respBodyStructure =
        "(\"text\" \"plain\" (\"charset\" \"us-ascii\") \"content-id\" \"content-description\" \"7bit\" 2109 54)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.MessagePart.FindAll((Predicate<ImapMessagePartInfo>)null));
      });
    }

    [Test]
    public void TestFindAll_WithPredicate_IsMultiPart()
    {
      const string bodyStructurePart1_1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 11 11)";
      const string bodyStructurePart1_2 = "(\"text\" \"richtext\" NIL NIL NIL \"base64\" 12 12)";
      const string bodyStructurePart1 = "(" + bodyStructurePart1_1 + bodyStructurePart1_2 + " \"alternative\")";
      const string bodyStructurePart2_1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 21 11)";
      const string bodyStructurePart2_2 = "(\"text\" \"richtext\" NIL NIL NIL \"base64\" 22 22)";
      const string bodyStructurePart2 = "(" + bodyStructurePart1_1 + bodyStructurePart1_2 + " \"mixed\")";
      const string bodyStructurePart3 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 3 3)";

      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + bodyStructurePart3 + " \"mixed\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var match = message.MessagePart.FindAll(ImapMessagePartInfo.Predicates.IsMultiPart);

        Assert.AreEqual(2, match.Count);

        Assert.AreEqual("1", match[0].Specifier);
        Assert.IsTrue(match[0].IsMultiPart);

        Assert.AreEqual("2", match[1].Specifier);
        Assert.IsTrue(match[1].IsMultiPart);
      });
    }

    [Test]
    public void TestFindAll_WithPredicate_HasEncapsulatedMessagePart()
    {
      const string envelope = "(NIL \"test mail\" NIL NIL NIL NIL NIL NIL NIL NIL)";
      const string bodyStructurePart1 = 
        "(" +
        "\"message\" \"rfc822\" NIL NIL NIL \"7bit\" 1 " +
        envelope + " " +
        "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 11 11) 1" +
        ")";
      const string bodyStructurePart2 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 2 2)";
      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + " \"mixed\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var match = message.MessagePart.FindAll(ImapMessagePartInfo.Predicates.HasEncapsulatedMessagePart);

        Assert.AreEqual(1, match.Count);

        Assert.AreEqual("1", match[0].Specifier);
        Assert.IsTrue(match[0].HasEncapsulatedMessage);
      });
    }

    [Test]
    public void TestFindAll_WithPredicate_IsEncapsulatedMessagePart()
    {
      const string envelope = "(NIL \"test mail\" NIL NIL NIL NIL NIL NIL NIL NIL)";
      const string bodyStructurePart1 = 
        "(" +
        "\"message\" \"rfc822\" NIL NIL NIL \"7bit\" 1 " +
        envelope + " " +
        "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 11 11) 1" +
        ")";
      const string bodyStructurePart2 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 2 2)";
      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + " \"mixed\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var match = message.MessagePart.FindAll(ImapMessagePartInfo.Predicates.IsEncapsulatedMessagePart);

        Assert.AreEqual(1, match.Count);

        Assert.AreEqual("1.1", match[0].Specifier);
        Assert.IsTrue(match[0].IsEncapsulatedMessagePart);
      });
    }

    [Test]
    public void TestFindAll_WithPredicate_IsTextPart()
    {
      const string bodyStructurePart1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 1 1)";
      const string bodyStructurePart2 = "(\"text\" \"richtext\" NIL NIL NIL \"base64\" 2 2)";
      const string bodyStructurePart3 = "(\"text\" \"html\" NIL NIL NIL \"quoted-printable\" 3 3)";
      const string bodyStructurePart4 = "(\"image\" \"gif\" NIL NIL NIL \"base64\" 4 4)";

      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + bodyStructurePart3 + bodyStructurePart4 + " \"mixed\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var match = message.MessagePart.FindAll(ImapMessagePartInfo.Predicates.IsTextPart);

        Assert.AreEqual(3, match.Count);
        Assert.AreEqual("1", match[0].Specifier);
        Assert.AreEqual("2", match[1].Specifier);
        Assert.AreEqual("3", match[2].Specifier);
      });
    }

    [Test]
    public void TestFindAll_WithPredicate_IsPlainTextPart()
    {
      const string bodyStructurePart1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 1 1)";
      const string bodyStructurePart2 = "(\"text\" \"richtext\" NIL NIL NIL \"base64\" 2 2)";
      const string bodyStructurePart3 = "(\"text\" \"html\" NIL NIL NIL \"quoted-printable\" 3 3)";
      const string bodyStructurePart4 = "(\"image\" \"gif\" NIL NIL NIL \"base64\" 4 4)";

      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + bodyStructurePart3 + bodyStructurePart4 + " \"mixed\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var match = message.MessagePart.FindAll(ImapMessagePartInfo.Predicates.IsPlainTextPart);

        Assert.AreEqual(1, match.Count);
        Assert.AreEqual("1", match[0].Specifier);
      });
    }

    [Test]
    public void TestFindAll_WithPredicate_IsImagePart()
    {
      const string bodyStructurePart1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 1 1)";
      const string bodyStructurePart2 = "(\"text\" \"richtext\" NIL NIL NIL \"base64\" 2 2)";
      const string bodyStructurePart3 = "(\"text\" \"html\" NIL NIL NIL \"quoted-printable\" 3 3)";
      const string bodyStructurePart4 = "(\"image\" \"gif\" NIL NIL NIL \"base64\" 4 4)";

      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + bodyStructurePart3 + bodyStructurePart4 + " \"mixed\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var match = message.MessagePart.FindAll(ImapMessagePartInfo.Predicates.IsImagePart);

        Assert.AreEqual(1, match.Count);
        Assert.AreEqual("4", match[0].Specifier);
      });
    }

    [Test]
    public void TestFindAll_WithPredicate_IsAttachmentPart()
    {
      const string bodyStructurePart1 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 1 NIL (\"attachment\" NIL) NIL NIL)";
      const string bodyStructurePart2 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 2 NIL (\"inline\" NIL) NIL NIL)";
      const string bodyStructurePart3 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 3)";

      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + bodyStructurePart3 + " \"mixed\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var match = message.MessagePart.FindAll(ImapMessagePartInfo.Predicates.IsAttachmentPart);

        Assert.AreEqual(1, match.Count);
        Assert.AreEqual("1", match[0].Specifier);
      });
    }

    [Test]
    public void TestFindAll_WithPredicate_IsInlinePart()
    {
      const string bodyStructurePart1 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 1 NIL (\"attachment\" NIL) NIL NIL)";
      const string bodyStructurePart2 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 2 NIL (\"inline\" NIL) NIL NIL)";
      const string bodyStructurePart3 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 3)";

      const string respBodyStructure =
        "(" + bodyStructurePart1 + bodyStructurePart2 + bodyStructurePart3 + " \"mixed\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var match = message.MessagePart.FindAll(ImapMessagePartInfo.Predicates.IsInlinePart);

        Assert.AreEqual(1, match.Count);
        Assert.AreEqual("2", match[0].Specifier);
      });
    }

    [Test]
    public void TestFindAll_WithPredicate_SeemsToBeAttachmentPart_1()
    {
      // (* = attachment part)
      // multipart/mixed
      //   1 text/plain; attachment (*)
      //   2 text/plain; inline
      //   3 text/plain
      //   4 multipart/alternative
      //     4.1 text/plain
      //     4.2 text/html
      //     4.3 text/html; attachment (*)
      //     4.4 image/jpeg
      //     4.5 application/octet-stream
      //   5 image/jpeg (*)
      //   6 image/jpeg; inline
      //   7 application/octet-stream (*)
      //   8 application/octet-stream; inline
      //   9 message/partial
      //  10 message/partial; attachment (*)
      //  11 text/html
      //  12 text/css (*)

      const string bodyStructurePart1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 1 1 NIL (\"attachment\" NIL) NIL NIL)";
      const string bodyStructurePart2 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 2 2 NIL (\"inline\" NIL) NIL NIL)";
      const string bodyStructurePart3 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 3 3)";
      const string bodyStructurePart4_1 = "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 41 41)";
      const string bodyStructurePart4_2 = "(\"text\" \"html\" NIL NIL NIL \"7bit\" 42 42)";
      const string bodyStructurePart4_3 = "(\"text\" \"html\" NIL NIL NIL \"7bit\" 43 43 NIL (\"attachment\" NIL) NIL NIL)";
      const string bodyStructurePart4_4 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 44)";
      const string bodyStructurePart4_5 = "(\"application\" \"octet-stream\" NIL NIL NIL \"base64\" 45)";
      const string bodyStructurePart4 =
        "(" +
        bodyStructurePart4_1 +
        bodyStructurePart4_2 +
        bodyStructurePart4_3 +
        bodyStructurePart4_4 +
        bodyStructurePart4_5 +
        " \"alternative\")";
      const string bodyStructurePart5 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 5)";
      const string bodyStructurePart6 = "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 6 NIL (\"inline\" NIL) NIL NIL)";
      const string bodyStructurePart7 = "(\"application\" \"octet-stream\" NIL NIL NIL \"base64\" 7)";
      const string bodyStructurePart8 = "(\"application\" \"octet-stream\" NIL NIL NIL \"base64\" 8 NIL (\"inline\" NIL) NIL NIL)";
      const string bodyStructurePart9 = "(\"message\" \"partial\" NIL NIL NIL \"7bit\" 9)";
      const string bodyStructurePart10 = "(\"message\" \"partial\" NIL NIL NIL \"7bit\" 10 NIL (\"attachment\" NIL) NIL NIL)";
      const string bodyStructurePart11 = "(\"text\" \"html\" NIL NIL NIL \"7bit\" 11 11)";
      const string bodyStructurePart12 = "(\"text\" \"css\" NIL NIL NIL \"7bit\" 12 12)";

      const string respBodyStructure =
        "(" +
        bodyStructurePart1 +
        bodyStructurePart2 +
        bodyStructurePart3 +
        bodyStructurePart4 +
        bodyStructurePart5 +
        bodyStructurePart6 +
        bodyStructurePart7 +
        bodyStructurePart8 +
        bodyStructurePart9 +
        bodyStructurePart10 +
        bodyStructurePart11 +
        bodyStructurePart12 +
        " \"mixed\")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        var match = message.MessagePart.FindAll(ImapMessagePartInfo.Predicates.SeemsToBeAttachmentPart);

        Assert.AreEqual(6, match.Count);

        Assert.AreEqual("1", match[0].Specifier);
        Assert.AreEqual("4.3", match[1].Specifier);
        Assert.AreEqual("5", match[2].Specifier);
        Assert.AreEqual("7", match[3].Specifier);
        Assert.AreEqual("10", match[4].Specifier);
        Assert.AreEqual("12", match[5].Specifier);
      });
    }

    [Test]
    public void TestFindAll_WithPredicate_SeemsToBeAttachmentPart_2()
    {
      const string respBodyStructure =
        "(\"text\" \"plain\" NIL NIL NIL \"7bit\" 1 1)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.IsEmpty(message.MessagePart.FindAll(ImapMessagePartInfo.Predicates.SeemsToBeAttachmentPart));
      });
    }

    [Test]
    public void TestGetContentFileName_FromContentTypeNameParameter()
    {
      const string respBodyStructure =
        "(\"image\" \"jpeg\" (\"name\" \"attached.jpg\") NIL NIL \"base64\" 1)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("attached.jpg", message.MessagePart.GetContentFileName());
      });
    }

    [Test]
    public void TestGetContentFileName_FromContentTypeNameParameter_MimeEncoded()
    {
      const string respBodyStructure =
        "(\"image\" \"jpeg\" (\"name\" \"=?ISO-2022-JP?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?=\") NIL NIL \"base64\" 1)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("添付ファイル.jpg", message.MessagePart.GetContentFileName());
      });
    }

    [Test]
    public void TestGetContentFileName_FromContentTypeExtendedNameParameter()
    {
      const string respBodyStructure =
        "(\"image\" \"jpeg\" (\"name*\" \"ISO-2022-JP''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67\") NIL NIL \"base64\" 1)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("添付ファイル.jpg", message.MessagePart.GetContentFileName());
      });
    }

    [Test]
    public void TestGetContentFileName_FromContentDispositionFilenameParameter()
    {
      const string respBodyStructure =
        "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 1 NIL (\"attachment\" (\"filename\" \"attached.jpg\")) NIL NIL)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("attached.jpg", message.MessagePart.GetContentFileName());
      });
    }

    [Test]
    public void TestGetContentFileName_FromContentDispositionExtendedFilenameParameter()
    {
      const string respBodyStructure =
        "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 1 NIL (\"attachment\" (\"filename*\" \"ISO-2022-JP''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67\")) NIL NIL)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("添付ファイル.jpg", message.MessagePart.GetContentFileName());
      });
    }

    [Test]
    public void TestGetContentFileName_NoNameOrDisposition()
    {
      const string respBodyStructure =
        "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 1 NIL NIL NIL NIL)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.IsNull(message.MessagePart.GetContentFileName());
      });
    }

    [Test]
    public void TestGetContentFileName_NoNameOrFilenameParameter()
    {
      const string respBodyStructure =
        "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 1 NIL (\"attachment\" NIL) NIL NIL)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.IsNull(message.MessagePart.GetContentFileName());
      });
    }

    [Test]
    public void TestGetContentFileName_NameParameter_InvalidMimeEncoding()
    {
      const string respBodyStructure =
        "(\"image\" \"jpeg\" (\"name\" \"=?ISO-2022-JP?X?xxxxxx?=\") NIL NIL \"base64\" 1)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", message.MessagePart.GetContentFileName());
        Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", message.MessagePart.GetContentFileName(null, false));

        Assert.Throws<FormatException>(() => message.MessagePart.GetContentFileName(null, true));
      });
    }

    [Test]
    public void TestGetContentFileName_NameParameter_InvalidCharset()
    {
      const string respBodyStructure =
        "(\"image\" \"jpeg\" (\"name\" \"=?X-UNKNOWN?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?=\") NIL NIL \"base64\" 1)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("=?X-UNKNOWN?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?=", message.MessagePart.GetContentFileName());
        Assert.AreEqual("=?X-UNKNOWN?B?GyRCRTpJVSVVJSElJCVrGyhCLmpwZw==?=", message.MessagePart.GetContentFileName(null, false));

        Assert.Throws<EncodingNotSupportedException>(() => message.MessagePart.GetContentFileName(null, true));

        Assert.AreEqual("添付ファイル.jpg", message.MessagePart.GetContentFileName(delegate(string encoding) {
          Assert.AreEqual("X-UNKNOWN", encoding);

          return Encoding.GetEncoding("iso-2022-jp");
        }, true));
      });
    }

    [Test]
    public void TestGetContentFileName_FilenameParameter_InvalidCharset()
    {
      const string respBodyStructure =
        "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 1 NIL (\"attachment\" (\"filename*\" \"X-UNKNOWN''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67\")) NIL NIL)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("X-UNKNOWN''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67", message.MessagePart.GetContentFileName());
        Assert.AreEqual("X-UNKNOWN''%1B%24%42%45%3A%49%55%25%55%25%21%25%24%25%6B%1B%28%42%2E%6A%70%67", message.MessagePart.GetContentFileName(null, false));

        Assert.Throws<EncodingNotSupportedException>(() => message.MessagePart.GetContentFileName(null, true));

        Assert.AreEqual("添付ファイル.jpg", message.MessagePart.GetContentFileName(delegate(string encoding) {
          Assert.AreEqual("X-UNKNOWN", encoding);

          return Encoding.GetEncoding("iso-2022-jp");
        }, true));
      });
    }

    [Test]
    public void TestGetContentFileName_MultiPartWithExtension()
    {
      const string respBodyStructure =
        "((\"image\" \"jpeg\" NIL NIL NIL \"base64\" 1) \"mixed\" NIL (\"attached\" (\"filename\" \"attached.eml\")) NIL NIL)";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("attached.eml", message.MessagePart.GetContentFileName());
      });
    }

    [Test]
    public void TestGetContentFileName_EncapsulatingWithExtension()
    {
      const string envelope = "(NIL \"test mail\" NIL NIL NIL NIL NIL NIL NIL NIL)";
      const string respBodyStructure = 
        "(" +
        "\"message\" \"rfc822\" NIL NIL NIL \"7bit\" 10 " +
        envelope + " " +
        "(\"image\" \"jpeg\" NIL NIL NIL \"base64\" 1) 10 NIL (\"attachment\" (\"filename\" \"attached.eml\")) NIL NIL" +
        ")";

      TestMessage(respBodyStructure, delegate(ImapMessageInfo message) {
        Assert.AreEqual("attached.eml", message.MessagePart.GetContentFileName());
      });
    }
  }
}

