using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Runtime.Remoting.Messaging;
using System.Threading;
using System.Text;
using System.Text.RegularExpressions;

using Smdn.Net;

namespace Smdn.Net.Imap4 {
  public class ImapPseudoServer2 : IDisposable {
    private class Response {
      public string ExpectedCommand {
        get { return expectedCommand; }
      }

      public string ResponseText {
        get { return responseText; }
      }

      public Action ActionBeforeRespond {
        get { return actionBeforeRespond; }
      }

      public Action ActionAfterRespond {
        get { return actionAfterRespond; }
      }

      public bool IsByeResponse {
        get; set;
      }

      public bool IsPushResponse {
        get; set;
      }

      public bool IsLiteralContinuationResponse {
        get; set;
      }

      public bool AcceptLiteral {
        get; set;
      }

      public Response(string expectedCommand,
                      string responseText,
                      Action actionBeforeRespond,
                      Action actionAfterRespond)
      {
        this.expectedCommand = expectedCommand;
        this.responseText = responseText;
        this.actionBeforeRespond = actionBeforeRespond;
        this.actionAfterRespond = actionAfterRespond;

        IsByeResponse = false;
        IsPushResponse = false;
        IsLiteralContinuationResponse = false;
        AcceptLiteral = false;
      }

      private readonly string expectedCommand;
      private readonly string responseText;
      private readonly Action actionBeforeRespond;
      private readonly Action actionAfterRespond;
    }

    public IPEndPoint ServerEndPoint {
      get { return serverEndPoint; }
    }

    public string Host {
      get { return ServerEndPoint.Address.ToString(); }
    }

    public int Port {
      get { return ServerEndPoint.Port; }
    }

    public string HostPort {
      get { return string.Format("{0}:{1}", Host, Port); }
    }

    public bool ThrowIfServerNotStopped {
      get; set;
    }

    public ImapPseudoServer2()
    {
      ThrowIfServerNotStopped = true;

      Initialize();
    }

    private void Initialize()
    {
      server = new TcpListener(new IPEndPoint(IPAddress.Loopback, 0));
      server.Start();

      serverEndPoint = server.LocalEndpoint as IPEndPoint;

      responseTag = 0;

      SetGreetingResponse("* OK ready\r\n");

      clientRequestWaitHandle = new ManualResetEvent(false);
    }

    void IDisposable.Dispose()
    {
      if (acceptClientAsyncResult != null) {
        var acceptClient = (Action)(acceptClientAsyncResult as AsyncResult).AsyncDelegate;

        if (acceptClientAsyncResult.AsyncWaitHandle.WaitOne(500))
          acceptClient.EndInvoke(acceptClientAsyncResult);
        else if (ThrowIfServerNotStopped)
          throw new ApplicationException("server not stopped");

        acceptClientAsyncResult = null;
      }

      if (server != null) {
        server.Stop();
        server = null;
      }

      if (clientRequestWaitHandle != null) {
        clientRequestWaitHandle.Close();
        clientRequestWaitHandle = null;
      }

      if (serverProcException != null) {
        Console.Error.WriteLine("unhandled server exception:{0}{1}", Environment.NewLine, serverProcException);

        throw new ApplicationException("exception not handled", serverProcException);
      }
    }

    public void Start()
    {
      if (acceptClientAsyncResult != null)
        throw new InvalidOperationException("already started");

      var acceptClient = new Action(AcceptClient);

      acceptClientAsyncResult = acceptClient.BeginInvoke(null, null);
    }

    public void Restart()
    {
      (this as IDisposable).Dispose();

      Initialize();

      Start();
    }

    public static ImapPseudoServer2 StartNew()
    {
      var server = new ImapPseudoServer2();

      server.Start();

      return server;
    }

    public static ImapPseudoServer2 StartNotAccepting()
    {
      var server = new ImapPseudoServer2();

      server.server.Start();

      return server;
    }

    public void SetCharset(Encoding newCharset)
    {
      if (newCharset == null)
        throw new ArgumentNullException("newCharset");

      transferCharset = newCharset;
    }

    public void SetGreetingResponse(string greeting)
    {
      SetGreetingResponse(greeting, null, null, false);
    }

    public void SetGreetingResponse(string greeting,
                                    Action actionBeforeRespond,
                                    Action actionAfterRespond)
    {
      SetGreetingResponse(greeting, actionBeforeRespond, actionAfterRespond, false);
    }

    public void SetGreetingResponse(string greeting,
                                    Action actionBeforeRespond,
                                    Action actionAfterRespond,
                                    bool isByeResponse)
    {
      greetingResponse = new Response(null, greeting, actionBeforeRespond, actionAfterRespond);

      greetingResponse.IsByeResponse = isByeResponse;
    }

    public void EnqueueTaggedResponse(string responseText)
    {
      EnqueueTaggedResponse(null, responseText, null, null);
    }

    public void EnqueueTaggedResponse(string expectedCommand,
                                      string responseText)
    {
      EnqueueTaggedResponse(expectedCommand,
                            responseText,
                            null,
                            null,
                            false);
    }

    public void EnqueueTaggedResponse(string expectedCommand,
                                      string responseText,
                                      bool isByeResponse)
    {
      EnqueueTaggedResponse(expectedCommand,
                            responseText,
                            null,
                            null,
                            isByeResponse);
    }

    public void EnqueueTaggedResponse(string expectedCommand,
                                      string responseText,
                                      Action actionBeforeRespond,
                                      Action actionAfterRespond)
    {
      EnqueueTaggedResponse(expectedCommand,
                            responseText,
                            actionBeforeRespond,
                            actionAfterRespond,
                            false);
    }

    public void EnqueueTaggedResponse(string expectedCommand,
                                      string responseText,
                                      Action actionBeforeRespond,
                                      Action actionAfterRespond,
                                      bool isByeResponse)
    {
#if DEBUG
      if (!responseText.Contains("$tag"))
        throw new ArgumentException("is not tagged response", "responseText");
#endif

      responseText = responseText.Replace("$tag", (responseTag++).ToString("x4"));

      if (0x10000 == responseTag)
        responseTag = 0;

      var response = new Response(expectedCommand,
                                  responseText,
                                  actionBeforeRespond,
                                  actionAfterRespond);

      response.IsByeResponse = isByeResponse;

      responseQueue.Enqueue(response);
    }

    public void EnqueueUntaggedByeResponse(string responseText)
    {
      EnqueueUntaggedByeResponse(responseText, false);
    }

    public void EnqueueUntaggedByeResponse(string responseText, bool asPushResponse)
    {
#if DEBUG
      if (!(responseText.StartsWith("* BYE") || responseText.Contains("\r\n* BYE")))
        throw new ArgumentException("is not BYE response", "responseText");
#endif

      var response = new Response(null,
                                  responseText,
                                  null,
                                  null);

      response.IsByeResponse = true;
      response.IsPushResponse = asPushResponse;

      responseQueue.Enqueue(response);
    }

    public void EnqueueResponse(string responseText)
    {
      EnqueueResponse(null,
                      responseText,
                      null,
                      null);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText)
    {
      EnqueueResponse(expectedCommand,
                      responseText,
                      null,
                      null);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText,
                                Action actionBeforeRespond,
                                Action actionAfterRespond)
    {
      var response = new Response(expectedCommand,
                                  responseText,
                                  actionBeforeRespond,
                                  actionAfterRespond);

      responseQueue.Enqueue(response);
    }

    public void EnqueuePushResponse(string responseText)
    {
      EnqueuePushResponse(responseText,
                          false);
    }

    public void EnqueuePushResponse(string responseText,
                                    bool disconnectAfterSend)
    {
      var response = new Response(null,
                                  responseText,
                                  null,
                                  null);

      response.IsPushResponse = true;
      response.IsByeResponse = disconnectAfterSend;

      responseQueue.Enqueue(response);
    }

    public void EnqueueLiteralContinuationResponse(string responseText,
                                                   bool acceptLiteral)
    {
      EnqueueLiteralContinuationResponse(null,
                                         responseText,
                                         acceptLiteral);
    }

    public void EnqueueLiteralContinuationResponse(string expectedCommand,
                                                   string responseText,
                                                   bool acceptLiteral)
    {
      if (responseText.Contains("$tag"))
        responseText = responseText.Replace("$tag", (responseTag++).ToString("x4"));

      var response = new Response(expectedCommand,
                                  responseText,
                                  null,
                                  null);

      response.IsLiteralContinuationResponse = true;
      response.AcceptLiteral = acceptLiteral;

      responseQueue.Enqueue(response);
    }

    // TODO: rename
    public string DequeueRequest()
    {
      return DequeueRequest(null);
    }

    public string DequeueRequest(Encoding encoding)
    {
      string ret;

      lock (commandQueue) {
        if (0 < commandQueue.Count)
          ret = commandQueue.Dequeue();
        else
          return null;
      }

      if (encoding == null)
        return ret;
      else
        return encoding.GetString(transferCharset.GetBytes(ret));
    }

    public void WaitForClientRequest()
    {
      WaitForClientRequest(TimeSpan.FromSeconds(5));
    }

    public bool WaitForClientRequest(TimeSpan timeout)
    {
      clientRequestWaitHandle.Reset();

      return clientRequestWaitHandle.WaitOne(timeout);
    }

    private void AcceptClient()
    {
      var serverClock = System.Diagnostics.Stopwatch.StartNew();

      WriteLogLine(serverClock, "listening...");

      var client = server.AcceptTcpClient();

      WriteLogLine(serverClock, "client accepted");

      client.ReceiveTimeout = 100;

      try {
        WriteLogLine(serverClock, "server process started");

        ServerProc(client, serverClock);

        WriteLogLine(serverClock, "server process successfully finished");
      }
      catch (Exception ex) {
        serverProcException = ex;

        WriteLogLine(serverClock, "server process stopped by {0}", ex.GetType().FullName);
      }

      client.Close();

      WriteLogLine(serverClock, "connection closed");
    }

    private static void WriteLogLine(Stopwatch serverClock, string format, params object[] args)
    {
      Console.Write("({0:D2}.{1:D4}) ", serverClock.Elapsed.Seconds, serverClock.Elapsed.Milliseconds);
      Console.WriteLine(format, args);
    }

    private static void WriteLog(Stopwatch serverClock, string line)
    {
      Console.Write("({0:D2}.{1:D4}) ", serverClock.Elapsed.Seconds, serverClock.Elapsed.Milliseconds);
      Console.Write(line);
    }

    private void ServerProc(TcpClient client, Stopwatch serverClock)
    {
      var commandStream = client.GetStream();

      Send(greetingResponse, commandStream, serverClock);

      if (greetingResponse.IsByeResponse)
        return;

      var commandReceiveBuffer = new byte[0x1000];
      var commandBuffer = new StringBuilder(0x1000);
      const long endOfLiteral = -1L;
      long literalOctetsToBeReceive = endOfLiteral;

      for (;;) {
        if (!IsConnected(client, serverClock))
          return;

        if (0 < responseQueue.Count && responseQueue.Peek().IsPushResponse) {
          var r = responseQueue.Dequeue();

          Send(r, commandStream, serverClock);

          if (r.IsByeResponse)
            return;
          else
            continue;
        }

        /*
         * wait for command sent
         */
        try {
          // XXX: NetworkStream.Read does not timeout on netfx
          if (Runtime.IsRunningOnNetFx && !commandStream.DataAvailable)
            continue;

          var commandLength = commandStream.Read(commandReceiveBuffer, 0, commandReceiveBuffer.Length);

          if (commandLength == 0 && !IsConnected(client, serverClock))
            return;

          var received = transferCharset.GetString(commandReceiveBuffer, 0, commandLength);

          WriteLog(serverClock, "C:\t" + received);

          commandBuffer.Append(received);

          literalOctetsToBeReceive -= commandLength;

          if (0 < literalOctetsToBeReceive) {
            continue; // continue to receive literal
          }
          else if (literalOctetsToBeReceive == 0L) {
            literalOctetsToBeReceive = endOfLiteral;
            continue; // receive next of literal
          }
        }
        catch (IOException ex) {
          if (!client.Connected)
            return; // disconnected

          var exSocket = ex.InnerException as SocketException;

          if (exSocket != null &&
              (exSocket.SocketErrorCode == SocketError.WouldBlock || exSocket.SocketErrorCode == SocketError.TimedOut)) {
            continue;
          }

          throw;
        }

        /*
         * process received command
         */
        var command = commandBuffer.ToString();

        var literalSizeMatches = literalSizeRegex.Matches(command);

        if (0 < literalSizeMatches.Count) {
          // command ends with literal
          var literalSizeMatch = literalSizeMatches[literalSizeMatches.Count - 1]; // last match

          literalOctetsToBeReceive = long.Parse(literalSizeMatch.Groups[2].Value);

          if ("+".Equals(literalSizeMatch.Groups[3].Value, StringComparison.Ordinal)) {
            // non-sync literal
            continue; // continue to receive literal
          }
          else {
            // sync literal
            if (0 < responseQueue.Count && responseQueue.Peek().IsLiteralContinuationResponse) {
              if (responseQueue.Peek().AcceptLiteral) {
                Send(responseQueue.Dequeue(), commandStream, serverClock);

                continue; // continue to receive literal
              }
              else {
                literalOctetsToBeReceive = endOfLiteral;
              }
            }
            else {
              var continuationResponse = new Response(null, "+ continue\r\n", null, null);

              Send(continuationResponse, commandStream, serverClock);

              continue; // continue to receive literal
            }
          }
        }

        if (!command.EndsWith("\r\n", StringComparison.Ordinal))
          // incomplete command
          continue;

        lock (commandQueue) {
          commandQueue.Enqueue(command);
        }

        /*
         * dequeue response
         */
        while (responseQueue.Count == 0) {
          clientRequestWaitHandle.Set();

          if (!IsConnected(client, serverClock))
            return;

          Thread.Sleep(10);
        }

        var response = responseQueue.Dequeue();

        /*
         * send response
         */
        if (string.IsNullOrEmpty(response.ExpectedCommand) ||
            1 /*SP*/ <= command.IndexOf(response.ExpectedCommand, StringComparison.OrdinalIgnoreCase))
          Send(response, commandStream, serverClock);
        else
          throw new InvalidOperationException(string.Format("unexpected command{0}expected '{1}' but was '{2}'",
                                                            Environment.NewLine,
                                                            response.ExpectedCommand,
                                                            command.Replace("\r", "\\r").Replace("\n", "\\n")));

        commandBuffer.Length = 0;

        if (response.IsByeResponse)
          return;
      }
    }

    private void Send(Response response, Stream requestStream, Stopwatch serverClock)
    {
      if (response.ActionBeforeRespond != null)
        response.ActionBeforeRespond();

      var bytes = transferCharset.GetBytes(response.ResponseText);

      requestStream.Write(bytes, 0, bytes.Length);

      WriteLog(serverClock, "S:\t" + response.ResponseText);

      if (response.ActionAfterRespond != null)
        response.ActionAfterRespond();
    }

    private static readonly byte[] peekClientBuffer = new byte[1];

    private static bool IsConnected(TcpClient client, Stopwatch serverClock)
    {
      try {
        if (client.Client.Poll(0, SelectMode.SelectRead)) {
          if (client.Client.Receive(peekClientBuffer, SocketFlags.Peek) == 0) {
            WriteLogLine(serverClock, "client disconnected");
            return false;
          }
        }
      }
      catch (SocketException ex) {
        switch (ex.SocketErrorCode) {
          case SocketError.ConnectionAborted:
          case SocketError.ConnectionReset:
            return false;

          default:
            Console.Error.WriteLine("unexpected socket error: {0}", ex.SocketErrorCode);
            throw;
        }
      }

      return true;
    }

    private TcpListener server;
    private IPEndPoint serverEndPoint;
    private IAsyncResult acceptClientAsyncResult;
    private readonly Queue<Response> responseQueue = new Queue<Response>();
    private readonly Queue<string> commandQueue = new Queue<string>();
    private ManualResetEvent clientRequestWaitHandle;
    private Response greetingResponse;
    private int responseTag = 0;
    private Exception serverProcException = null;
    private Encoding transferCharset = NetworkTransferEncoding.Transfer8Bit;

    private static readonly Regex literalSizeRegex = new Regex(@"(~?)\{(\d+)(\+?)\}\r\n$", RegexOptions.Singleline | RegexOptions.Compiled | RegexOptions.CultureInvariant);
  }
}

