// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

using Smdn.Collections;

namespace Smdn.Net.Imap4 {
  // http://tools.ietf.org/html/rfc5256
  public sealed class ImapThreadList : IEnumerable<ImapThreadList> {
    /// <summary>sequence number or UID</summary>
    public long Number {
      get; private set;
    }

#if NET_4_5
    public IReadOnlyList<ImapThreadList>
#else
    public IList<ImapThreadList>
#endif
    Children {
      get; private set;
    }

    public bool IsUidList {
      get; private set;
    }

    public bool IsRoot {
      get { return Number == 0; }
    }

    internal static ImapThreadList CreateRootedList
    (
      bool isUidList,
#if NET_4_5
      IReadOnlyList<ImapThreadList> children
#else
      IList<ImapThreadList> children
#endif
    )
    {
      return new ImapThreadList(true, isUidList, 0, children);
    }

    internal static ImapThreadList CreateNested(bool isUidList, long number, ImapThreadList child)
    {
#if DEBUG
      if (child == null)
        throw new ArgumentNullException("child");
#endif

      return new ImapThreadList(false, isUidList, number, new[] { child });
    }

    internal static ImapThreadList CreateNested
    (
      bool isUidList,
      long number,
#if NET_4_5
      IReadOnlyList<ImapThreadList> children
#else
      IList<ImapThreadList> children
#endif
    )
    {
#if DEBUG
      if (children == null)
        throw new ArgumentNullException("children");
#endif

      return new ImapThreadList(false, isUidList, number, children);
    }

    internal static ImapThreadList Create(bool isUidList, long number)
    {
      return new ImapThreadList(false, isUidList, number, EmptyArray<ImapThreadList>.Instance);
    }

    private ImapThreadList
    (
      bool root,
      bool isUidList,
      long number,
#if NET_4_5
      IReadOnlyList<ImapThreadList> children
#else
      IList<ImapThreadList> children
#endif
    )
    {
      if (root)
        number = 0;
      else if (number <= 0)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("number", number);

      this.IsUidList = isUidList;
      this.Number = number;
      this.Children = CollectionUtils.ToNonNullReadOnly(children);
    }

    public IEnumerator<ImapThreadList> GetEnumerator()
    {
      return (Children as IEnumerable<ImapThreadList>).GetEnumerator();
    }

    System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
    {
      return Children.GetEnumerator();
    }

    public ImapSequenceSet ToSequenceSet()
    {
      var sequenceOrUidSet = IEnumerableExtensions.EnumerateDepthFirst(this).Select(delegate(ImapThreadList list) {
        return list.Number;
      }).ToArray();

      return ImapSequenceSet.CreateSetUncheckedDirect(IsUidList, sequenceOrUidSet);
    }

    public void Traverse(Action<ImapThreadList> action)
    {
      if (action == null)
        throw new ArgumentNullException("action");

      foreach (var child in IEnumerableExtensions.EnumerateDepthFirst(this)) {
        action(child);
      }
    }

    public override string ToString ()
    {
      return string.Format("{{IsUidList={0}, Number={1}, Children={{{2}}}}}",
                           IsUidList,
                           Number,
                           CollectionUtils.Join(", ", Children));
    }
  }
}