//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
using System;
using System.Collections.Generic;
using System.Text;

using Smdn.Formats;
using Smdn.Formats.Mime;

namespace Smdn.Net.Imap4 {
  public static class MimeUtils {
    private static readonly char[] extendedInitialValueSplitter = new[] {'\''};

    public static string GetDecodedParameterValue(string paramName,
#if NET_4_5
                                                  IReadOnlyDictionary<string, string> parameters,
#else
                                                  IDictionary<string, string> parameters,
#endif
                                                  EncodingSelectionCallback selectFallbackEncoding,
                                                  bool throwIfMalformedOrUnsupported)
    {
      if (parameters == null)
        throw new ArgumentNullException("parameters");

      string rawValue = null;

      // try get extended parameter value
      if (parameters.TryGetValue(paramName + "*", out rawValue)) {
        /*
         * RFC 2231 MIME Parameter Value and Encoded Word Extensions: Character Sets, Languages, and Continuation
         * https://tools.ietf.org/html/rfc2231
         * 
         * 6. IMAP4 Handling of Parameter Values
         *    IMAP4 [RFC-2060] servers SHOULD decode parameter value continuations
         *    when generating the BODY and BODYSTRUCTURE fetch attributes.
         * 
         * 7. Modifications to MIME ABNF
         *    extended-initial-value := [charset] "'" [language] "'" extended-other-values
         *    extended-other-values := *(ext-octet / attribute-char)
         */
        var vals = rawValue.Split(extendedInitialValueSplitter, StringSplitOptions.None);

        if (vals.Length == 3) {
          Encoding charset = null;

          if (0 < vals[0].Length) {
            if (throwIfMalformedOrUnsupported)
              charset = EncodingUtils.GetEncodingThrowException(vals[0], selectFallbackEncoding);
            else
              charset = EncodingUtils.GetEncoding(vals[0], selectFallbackEncoding);
          }

          if (charset == null)
            return rawValue;
          else
            return PercentEncoding.GetDecodedString(vals[2], charset, false);
        }
        else {
          if (throwIfMalformedOrUnsupported)
            throw new FormatException("malformed parameter value continuations");
          else
            return rawValue;
        }
      }
      else if (parameters.TryGetValue(paramName, out rawValue)) {
        return GetDecodedValue(rawValue, selectFallbackEncoding, throwIfMalformedOrUnsupported);
      }
      else {
        return null;
      }
    }

    /// <param name="rawValue">raw text, can be null.</param>
    /// <returns>decoded or raw value</returns>
    public static string GetDecodedValue(string rawValue,
                                          EncodingSelectionCallback selectFallbackEncoding,
                                          bool throwIfMalformedOrUnsupported)
    {
      Encoding charset;

      return GetDecodedValue(rawValue,
                             selectFallbackEncoding,
                             throwIfMalformedOrUnsupported,
                             out charset);
    }

    /// <param name="rawValue">raw text, can be null.</param>
    /// <returns>decoded or raw value</returns>
    public static string GetDecodedValue(string rawValue,
                                          EncodingSelectionCallback selectFallbackEncoding,
                                          bool throwIfMalformedOrUnsupported,
                                          out Encoding charset)
    {
      charset = null;

      if (rawValue == null)
        return null;

      try {
        MimeEncodingMethod encoding;

        return MimeEncoding.Decode(rawValue,
                                    selectFallbackEncoding,
                                    null,
                                    out encoding,
                                    out charset);
      }
      catch (EncodingNotSupportedException) {
        if (throwIfMalformedOrUnsupported)
          throw;
        else
          return rawValue;
      }
      catch (FormatException) {
        if (throwIfMalformedOrUnsupported)
          throw;
        else
          return rawValue;
      }
    }
  }
}

