using System;
using System.Text;
using NUnit.Framework;

using Smdn.Formats;

namespace Smdn.Net.Imap4 {
  [TestFixture]
  public class ImapAddressTests {
    [Test]
    public void TestGetDecodedName()
    {
      var address = new ImapAddress("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", null, "test", "example.com");

      Assert.AreEqual("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", address.Name);
      Assert.AreEqual("テスト", address.GetDecodedName());
    }

    [Test]
    public void TestGetDecodedName_InvalidMimeEncoding()
    {
      var address = new ImapAddress("=?ISO-2022-JP?X?xxxxxx?=", null, "test", "example.com");

      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", address.Name);
      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", address.GetDecodedName());
      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", address.GetDecodedName(null ,false));

      Assert.Throws<FormatException>(() => address.GetDecodedName(null, true));
    }

    [Test]
    public void TestGetDecodedName_InvalidCharset()
    {
      var address = new ImapAddress("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", null, "test", "example.com");

      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", address.Name);
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", address.GetDecodedName());
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", address.GetDecodedName(null ,false));

      Assert.Throws<EncodingNotSupportedException>(() => address.GetDecodedName(null, true));

      Assert.AreEqual("テスト", address.GetDecodedName(delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return Encoding.GetEncoding("iso-2022-jp");
      }, true));
    }

    [Test]
    public void TestToMailAddress_WithDisplayName()
    {
      var address = (new ImapAddress("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", null, "test", "example.com")).ToMailAddress();

      Assert.AreEqual("テスト", address.DisplayName);
      Assert.AreEqual("test", address.User);
      Assert.AreEqual("example.com", address.Host);

      address = (new ImapAddress("test", null, "test", "example.com")).ToMailAddress();

      Assert.AreEqual("test", address.DisplayName);
      Assert.AreEqual("test", address.User);
      Assert.AreEqual("example.com", address.Host);
    }

    [Test]
    public void TestToMailAddress_WithoutDisplayName()
    {
      var address = (new ImapAddress(null, null, "test", "example.com")).ToMailAddress();

      Assert.AreEqual("test", address.User);
      Assert.AreEqual("example.com", address.Host);
    }

    [Test]
    public void TestToMailAddress_InvalidMimeEncoding()
    {
      var a = new ImapAddress("=?ISO-2022-JP?X?xxxxxx?=", null, "test", "example.com");

      var address = a.ToMailAddress();

      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", address.DisplayName);
      Assert.AreEqual("test", address.User);
      Assert.AreEqual("example.com", address.Host);

      Assert.DoesNotThrow(() => a.ToMailAddress(null, false));
      Assert.Throws<FormatException>(() => a.ToMailAddress(null, true));
    }

    [Test]
    public void TestToMailAddress_InvalidCharset()
    {
      var a = new ImapAddress("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", null, "test", "example.com");
      var address = a.ToMailAddress();

      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", address.DisplayName);
      Assert.AreEqual("test", address.User);
      Assert.AreEqual("example.com", address.Host);

      Assert.DoesNotThrow(() => a.ToMailAddress(null, false));
      Assert.Throws<EncodingNotSupportedException>(() => a.ToMailAddress(null, true));

      Assert.AreEqual("テスト", a.ToMailAddress(delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return Encoding.GetEncoding("iso-2022-jp");
      }, true).DisplayName);
    }

    [Test]
    public void TestToMailAddressCollection()
    {
      var col = ImapAddress.ToMailAddressCollection(new[] {
        new ImapAddress("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", null, "test", "example.net"),
        new ImapAddress("=?ISO-2022-JP?X?xxxxxx?=", null, "test", "example.com"),
        new ImapAddress(null, null, "test", "example.com"),
      });

      Assert.IsNotNull(col);
      Assert.AreEqual(3, col.Count);

      Assert.AreEqual("テスト", col[0].DisplayName);
      Assert.AreEqual("test@example.net", col[0].Address);

      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=", col[1].DisplayName);
      Assert.AreEqual("test@example.com", col[1].Address);

      Assert.IsEmpty(col[2].DisplayName);
      Assert.AreEqual("test@example.com", col[2].Address);
    }

    [Test]
    public void TestToMailAddressCollection_InvalidCharset()
    {
      var addresses = new[] {
        new ImapAddress("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", null, "test", "example.com"),
      };

      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=",
                      ImapAddress.ToMailAddressCollection(addresses)[0].DisplayName);
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=",
                      ImapAddress.ToMailAddressCollection(addresses, null, false)[0].DisplayName);

      Assert.Throws<EncodingNotSupportedException>(() => ImapAddress.ToMailAddressCollection(addresses, null, true));

      Assert.AreEqual("テスト", ImapAddress.ToMailAddressCollection(addresses, delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return Encoding.GetEncoding("iso-2022-jp");
      }, true)[0].DisplayName);
    }

    [Test]
    public void TestToMailAddressCollection_InvalidMimeEncoding()
    {
      var addresses = new[] {
        new ImapAddress("=?ISO-2022-JP?X?xxxxxx?=", null, "test", "example.com"),
      };

      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=",
                      ImapAddress.ToMailAddressCollection(addresses)[0].DisplayName);
      Assert.AreEqual("=?ISO-2022-JP?X?xxxxxx?=",
                      ImapAddress.ToMailAddressCollection(addresses, null, false)[0].DisplayName);

      Assert.Throws<FormatException>(() => ImapAddress.ToMailAddressCollection(addresses, null, true));
    }

    [Test]
    public void TestToMailAddressCollection_Null()
    {
      Assert.Throws<ArgumentNullException>(() => ImapAddress.ToMailAddressCollection(null));
    }

    [Test]
    public void TestToString()
    {
      var address = new ImapAddress("test", "test", "test", "example.com");

      Assert.DoesNotThrow(() => address.ToString());
    }
  }
}
