using System;
using System.Text;
using NUnit.Framework;

using Smdn.Formats;

namespace Smdn.Net.Imap4 {
  [TestFixture]
  public class ImapEnvelopeTests {
    [Test]
    public void TestAddressListProperties_ReadOnly()
    {
#if !NET_4_5
      var addresses = new[] {
        new ImapAddress("test", "test", "test", "test"),
      };
      var envelope = new ImapEnvelope(null, null, addresses, addresses, addresses, addresses, addresses, addresses, null, null);

      Assert.IsTrue(envelope.From.IsReadOnly, "From");
      Assert.IsTrue(envelope.Sender.IsReadOnly, "Sender");
      Assert.IsTrue(envelope.ReplyTo.IsReadOnly, "ReplyTo");
      Assert.IsTrue(envelope.To.IsReadOnly, "To");
      Assert.IsTrue(envelope.Cc.IsReadOnly, "Cc");
      Assert.IsTrue(envelope.Bcc.IsReadOnly, "Bcc");

      Assert.Throws<NotSupportedException>(() => envelope.From.Clear(), "From");
      Assert.Throws<NotSupportedException>(() => envelope.Sender.Clear(), "Sender");
      Assert.Throws<NotSupportedException>(() => envelope.ReplyTo.Clear(), "ReplyTo");
      Assert.Throws<NotSupportedException>(() => envelope.To.Clear(), "To");
      Assert.Throws<NotSupportedException>(() => envelope.Cc.Clear(), "Cc");
      Assert.Throws<NotSupportedException>(() => envelope.Bcc.Clear(), "Bcc");
#endif
    }

    [Test]
    public void TestAddressListProperties_NonNull()
    {
      var envelope = new ImapEnvelope(null, null, null, null, null, null, null, null, null, null);

      CollectionAssert.IsEmpty(envelope.From, "From");
      CollectionAssert.IsEmpty(envelope.Sender, "Sender");
      CollectionAssert.IsEmpty(envelope.ReplyTo, "ReplyTo");
      CollectionAssert.IsEmpty(envelope.To, "To");
      CollectionAssert.IsEmpty(envelope.Cc, "Cc");
      CollectionAssert.IsEmpty(envelope.Bcc, "Bcc");
    }

    private static ImapEnvelope CreateEnvelopeWithDate(string date)
    {
      return new ImapEnvelope(date, null, null, null, null, null, null, null, null, null);
    }

    [Test]
    public void TestGetDecodedDate()
    {
      var inputs = new[] {
        "Fri, 15 Mar 2002 12:32:23 +0900",
        "Fri, 15 Mar 2002 12:32:23 +0900 (JST)",
        "Fri (= Friday), 15 (th) Mar (March = 3rd month of year) 2002 12 (hour):32 (minute):23 (second) (timezone =) +0900 (JST)",
        "Fri (= Friday), 15 (th) Mar (March = 3rd month of year) 2002\r\n12 (hour):32 (minute):23 (second) (timezone =) +0900 (JST)",
      };
      var expected = new DateTimeOffset(2002, 3, 15, 12, 32, 23, new TimeSpan(+9, 0, 0));

      foreach (var input in inputs) {
        var envelope = CreateEnvelopeWithDate(input);

        Assert.AreEqual(expected, envelope.GetDecodedDate(), input);
      }
    }

    [Test]
    public void TestGetDecodedDate_NilOrEmpty()
    {
      var envelope1 = CreateEnvelopeWithDate(null);

      Assert.IsNull(envelope1.GetDecodedDate(), "#1");
      Assert.IsNull(envelope1.GetDecodedDate(true), "#2");

      var envelope2 = CreateEnvelopeWithDate(string.Empty);

      Assert.IsNull(envelope2.GetDecodedDate(), "#3");
      Assert.IsNull(envelope2.GetDecodedDate(true), "#4");
    }

    [Test]
    public void TestGetDecodedDate_Malformed()
    {
      var envelope = CreateEnvelopeWithDate("malformed date and time");

      Assert.IsNull(envelope.GetDecodedDate());
      Assert.IsNull(envelope.GetDecodedDate(false));
      Assert.Throws<FormatException>(() => envelope.GetDecodedDate(true));
    }

    private static ImapEnvelope CreateEnvelopeWithSubject(string subject)
    {
      return new ImapEnvelope(null, subject, null, null, null, null, null, null, null, null);
    }

    [Test]
    public void TestGetDecodedSubject()
    {
      var envelope1 = CreateEnvelopeWithSubject("not encoded");

      Assert.AreEqual("not encoded", envelope1.GetDecodedSubject());
      Assert.AreEqual("not encoded", envelope1.Subject);

      var envelope2 = CreateEnvelopeWithSubject("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=");

      Assert.AreEqual("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", envelope2.Subject);
      Assert.AreEqual("テスト", envelope2.GetDecodedSubject());
    }

    [Test]
    public void TestGetDecodedSubject_NilOrEmpty()
    {
      var envelope1 = CreateEnvelopeWithSubject(null);

      Assert.IsNull(envelope1.Subject);
      Assert.IsNull(envelope1.GetDecodedSubject());

      var envelope2 = CreateEnvelopeWithSubject(string.Empty);

      Assert.IsEmpty(envelope2.Subject);
      Assert.IsNull(envelope2.GetDecodedSubject());
    }

    [Test]
    public void TestGetDecodedSubject_InvalidMimeEncoding()
    {
      var envelope = CreateEnvelopeWithSubject("=?ISO-2022-JP?X?GyRCJUYlOSVIGyhC?=");

      Assert.AreEqual("=?ISO-2022-JP?X?GyRCJUYlOSVIGyhC?=", envelope.Subject);
      Assert.AreEqual("=?ISO-2022-JP?X?GyRCJUYlOSVIGyhC?=", envelope.GetDecodedSubject());
      Assert.AreEqual("=?ISO-2022-JP?X?GyRCJUYlOSVIGyhC?=", envelope.GetDecodedSubject(null, false));

      Assert.Throws<FormatException>(() => envelope.GetDecodedSubject(null, true));
    }

    [Test]
    public void TestGetDecodedSubject_InvalidCharset()
    {
      var envelope = CreateEnvelopeWithSubject("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=");

      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", envelope.Subject);
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", envelope.GetDecodedSubject());
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", envelope.GetDecodedSubject(null ,false));

      Assert.Throws<EncodingNotSupportedException>(() => envelope.GetDecodedSubject(null, true));

      Assert.AreEqual("テスト", envelope.GetDecodedSubject(delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return Encoding.GetEncoding("iso-2022-jp");
      }, true));
    }

    [Test]
    public void TestToString()
    {
      var addresses = new[] {
        new ImapAddress("test1", "test1", "test1", "test1"),
        new ImapAddress("test2", "test2", "test2", "test2"),
      };
      var envelope = new ImapEnvelope("date", "subject", addresses, addresses, addresses, addresses, addresses, addresses, "in-reply-to", "message-id");

      Assert.DoesNotThrow(() => envelope.ToString());
    }
  }
}