using System;
using System.Collections.Generic;
using NUnit.Framework;

namespace Smdn.Net.Imap4 {
  [TestFixture]
  public class ImapSequenceSetTest {
    [Test]
    public void TestSet()
    {
      ImapSequenceSet s;

      s = ImapSequenceSet.CreateSet(2);

      Assert.AreEqual("2", s.ToString());
      Assert.IsTrue(s.IsSingle);
      Assert.AreEqual(1L, s.ConcreteCount);

      s = ImapSequenceSet.CreateSet(2, 4, 5, 7);

      Assert.AreEqual("2,4,5,7", s.ToString());
      Assert.IsFalse(s.IsEmpty);
      Assert.IsFalse(s.IsSingle);
      Assert.AreEqual(4L, s.ConcreteCount);

      s = ImapSequenceSet.CreateSet(new long[] {1, 2, 3, 4, 5, 6, 7, 8});

      Assert.AreEqual("1,2,3,4,5,6,7,8", s.ToString());
      Assert.IsFalse(s.IsEmpty);
      Assert.IsFalse(s.IsSingle);
      Assert.AreEqual(8L, s.ConcreteCount);

      s = ImapSequenceSet.CreateSet(new long[] {1, 2, 3, 4, 5, 6, 7, 8}, 3, 4);

      Assert.AreEqual("4,5,6,7", s.ToString());
      Assert.IsFalse(s.IsEmpty);
      Assert.IsFalse(s.IsSingle);
      Assert.AreEqual(4L, s.ConcreteCount);

      s = ImapSequenceSet.CreateSet(new long[] {1, 2, 3, 4, 5, 6, 7, 8}, 7, 1);

      Assert.AreEqual("8", s.ToString());
      Assert.IsFalse(s.IsEmpty);
      Assert.IsTrue(s.IsSingle);
      Assert.AreEqual(1L, s.ConcreteCount);
    }

    [Test]
    public void TestSet_Serializable()
    {
      var s = ImapSequenceSet.CreateSet(2, 4, 5, 7);

      TestUtils.SerializeBinary(s, delegate(ImapSequenceSet d) {
        Assert.AreEqual("2,4,5,7", d.ToString());
        Assert.IsFalse(d.IsUidSet);
      });

      s = ImapSequenceSet.CreateUidSet(2, 4, 5, 7);

      TestUtils.SerializeBinary(s, delegate(ImapSequenceSet d) {
        Assert.AreEqual("2,4,5,7", d.ToString());
        Assert.IsTrue(d.IsUidSet);
      });
    }

    [Test]
    public void TestSet_FromArray()
    {
      var arr = new long[] {1, 2, 3, 4, 5, 6, 7, 8};

      foreach (var s in new[] {
        ImapSequenceSet.CreateSet(arr),
        ImapSequenceSet.CreateUidSet(arr),
        ImapSequenceSet.CreateSet(true, arr),
        ImapSequenceSet.CreateSet(false, arr),
      }) {
        Assert.AreEqual("1,2,3,4,5,6,7,8", s.ToString());
        Assert.IsFalse(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.AreEqual(8L, s.ConcreteCount);

        Array.Clear(arr, 0, arr.Length);

        Assert.AreEqual("1,2,3,4,5,6,7,8", s.ToString());
        Assert.IsFalse(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.AreEqual(8L, s.ConcreteCount);

        arr = new long[] {1, 2, 3, 4, 5, 6, 7, 8};
      }
    }

    private IEnumerable<long> GenerateSequenceNumbers()
    {
      yield return 1;
      yield return 2;
      yield return 3;
      yield return 4;
      yield return 5;
      yield return 6;
      yield return 7;
      yield return 8;
    }

    [Test]
    public void TestSet_FromIEnumerable()
    {
      foreach (var s in new[] {
        ImapSequenceSet.CreateSet(GenerateSequenceNumbers()),
        ImapSequenceSet.CreateUidSet(GenerateSequenceNumbers()),
        ImapSequenceSet.CreateSet(true, GenerateSequenceNumbers()),
        ImapSequenceSet.CreateSet(false, GenerateSequenceNumbers()),
      }) {
        Assert.AreEqual("1,2,3,4,5,6,7,8", s.ToString());
        Assert.IsFalse(s.IsEmpty);
        Assert.IsFalse(s.IsSingle);
        Assert.AreEqual(8L, s.ConcreteCount);
      }
    }

    [Test]
    public void TestSet_ToArray()
    {
      var s = ImapSequenceSet.CreateSet(2, 4, 5, 7).ToArray();

      Assert.AreEqual(4, s.Length);
      Assert.AreEqual(2, s[0]);
      Assert.AreEqual(4, s[1]);
      Assert.AreEqual(5, s[2]);
      Assert.AreEqual(7, s[3]);
    }

    [Test]
    public void TestSet_GetEnumerator()
    {
      var expected = new[] {2L, 4L, 5L, 7L};
      var index = 0;

      foreach (var num in ImapSequenceSet.CreateSet(2, 4, 5, 7)) {
        Assert.AreEqual(expected[index++], num);
      }
    }

    [Test]
    public void TestSet_ToNumber_Single()
    {
      Assert.AreEqual(3L, ImapSequenceSet.CreateSet(3).ToNumber());
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestSet_ToNumber_NotSingle()
    {
      ImapSequenceSet.CreateSet(2, 4, 5, 7).ToNumber();
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestSet_ToNumber_Empty()
    {
      ImapSequenceSet.CreateSet(new long[0]).ToNumber();
    }

    [Test]
    public void TestRange()
    {
      ImapSequenceSet s;

      s = ImapSequenceSet.CreateRangeSet(15, 18);

      Assert.AreEqual("15:18", s.ToString());
      Assert.IsFalse(s.IsEmpty);
      Assert.IsFalse(s.IsSingle);
      Assert.AreEqual(4L, s.ConcreteCount);

      s = ImapSequenceSet.CreateRangeSet(15L, 15L);

      Assert.AreEqual("15", s.ToString());
      Assert.IsFalse(s.IsEmpty);
      Assert.IsTrue(s.IsSingle);
      Assert.AreEqual(1L, s.ConcreteCount);
    }

    [Test]
    public void TestRange_Serializable()
    {
      var s = ImapSequenceSet.CreateRangeSet(15, 18);

      TestUtils.SerializeBinary(s, delegate(ImapSequenceSet d) {
        Assert.AreEqual(s.GetType(), d.GetType());
        Assert.AreEqual("15:18", d.ToString());
        Assert.IsFalse(d.IsUidSet);
      });

      s = ImapSequenceSet.CreateUidRangeSet(15, 18);

      TestUtils.SerializeBinary(s, delegate(ImapSequenceSet d) {
        Assert.AreEqual(s.GetType(), d.GetType());
        Assert.AreEqual("15:18", d.ToString());
        Assert.IsTrue(d.IsUidSet);
      });
    }

    [Test]
    public void TestRange_ToArray()
    {
      var s = ImapSequenceSet.CreateRangeSet(15, 18).ToArray();

      Assert.AreEqual(4, s.Length);
      Assert.AreEqual(15, s[0]);
      Assert.AreEqual(16, s[1]);
      Assert.AreEqual(17, s[2]);
      Assert.AreEqual(18, s[3]);

      s = ImapSequenceSet.CreateRangeSet(15L, 15L).ToArray();

      Assert.AreEqual(1, s.Length);
      Assert.AreEqual(15L, s[0]);
    }

    [Test]
    public void TestRange_GetEnumerator()
    {
      var expected = new[] {15L, 16L, 17L, 18L};
      var index = 0;

      foreach (var num in ImapSequenceSet.CreateRangeSet(15, 18)) {
        Assert.AreEqual(expected[index++], num);
      }
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestRange_ToNumber()
    {
      ImapSequenceSet.CreateRangeSet(15, 18).ToNumber();
    }

    [Test]
    public void TestFrom()
    {
      var s = ImapSequenceSet.CreateFromSet(72);

      Assert.AreEqual("72:*", s.ToString());
      Assert.IsFalse(s.IsEmpty);
      Assert.IsFalse(s.IsSingle);
      Assert.Less(s.ConcreteCount, 0L);
    }

    [Test]
    public void TestFrom_Serializable()
    {
      var s = ImapSequenceSet.CreateFromSet(72);

      TestUtils.SerializeBinary(s, delegate(ImapSequenceSet d) {
        Assert.AreEqual(s.GetType(), d.GetType());
        Assert.AreEqual("72:*", d.ToString());
        Assert.IsFalse(d.IsUidSet);
      });

      s = ImapSequenceSet.CreateUidFromSet(72);

      TestUtils.SerializeBinary(s, delegate(ImapSequenceSet d) {
        Assert.AreEqual(s.GetType(), d.GetType());
        Assert.AreEqual("72:*", d.ToString());
        Assert.IsTrue(d.IsUidSet);
      });
    }

    [Test]
    [ExpectedException(typeof(NotSupportedException))]
    public void TestFrom_ToArray()
    {
      ImapSequenceSet.CreateFromSet(72).ToArray();
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestFrom_GetEnumerator()
    {
      foreach (var num in ImapSequenceSet.CreateFromSet(72)) {
        Assert.Fail("enumerated: {0}", num);
      }
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestFrom_ToNumber()
    {
      ImapSequenceSet.CreateFromSet(72).ToNumber();
    }

    [Test]
    public void TestTo()
    {
      var s = ImapSequenceSet.CreateToSet(25);

      Assert.AreEqual("*:25", s.ToString());
      Assert.IsFalse(s.IsEmpty);
      Assert.IsFalse(s.IsSingle);
      Assert.AreEqual(25L, s.ConcreteCount);
    }

    [Test]
    public void TestTo_Serializable()
    {
      var s = ImapSequenceSet.CreateToSet(25);

      TestUtils.SerializeBinary(s, delegate(ImapSequenceSet d) {
        Assert.AreEqual(s.GetType(), d.GetType());
        Assert.AreEqual("*:25", d.ToString());
        Assert.IsFalse(d.IsUidSet);
      });

      s = ImapSequenceSet.CreateUidToSet(25);

      TestUtils.SerializeBinary(s, delegate(ImapSequenceSet d) {
        Assert.AreEqual(s.GetType(), d.GetType());
        Assert.AreEqual("*:25", d.ToString());
        Assert.IsTrue(d.IsUidSet);
      });
    }

    [Test]
    [ExpectedException(typeof(NotSupportedException))]
    public void TestTo_ToArray()
    {
      ImapSequenceSet.CreateToSet(25).ToArray();
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestTo_GetEnumerator()
    {
      foreach (var num in ImapSequenceSet.CreateToSet(25)) {
        Assert.Fail("enumerated: {0}", num);
      }
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestTo_ToNumber()
    {
      ImapSequenceSet.CreateToSet(25).ToNumber();
    }

    [Test]
    public void TestAll()
    {
      var s = ImapSequenceSet.CreateAllSet();

      Assert.AreEqual("*", s.ToString());
      Assert.IsFalse(s.IsEmpty);
      Assert.IsFalse(s.IsSingle);
      Assert.Less(s.ConcreteCount, 0L);
    }

    [Test]
    public void TestAll_Serializable()
    {
      var s = ImapSequenceSet.CreateAllSet();

      TestUtils.SerializeBinary(s, delegate(ImapSequenceSet d) {
        Assert.AreEqual(s.GetType(), d.GetType());
        Assert.AreEqual("*", d.ToString());
        Assert.IsFalse(d.IsUidSet);
      });

      s = ImapSequenceSet.CreateUidAllSet();

      TestUtils.SerializeBinary(s, delegate(ImapSequenceSet d) {
        Assert.AreEqual(s.GetType(), d.GetType());
        Assert.AreEqual("*", d.ToString());
        Assert.IsTrue(d.IsUidSet);
      });
    }

    [Test]
    [ExpectedException(typeof(NotSupportedException))]
    public void TestAll_ToArray()
    {
      ImapSequenceSet.CreateAllSet().ToArray();
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestAll_GetEnumerator()
    {
      foreach (var num in ImapSequenceSet.CreateAllSet()) {
        Assert.Fail("enumerated: {0}", num);
      }
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestAll_ToNumber()
    {
      ImapSequenceSet.CreateAllSet().ToNumber();
    }

    [Test]
    public void TestCombine()
    {
      var set1 = ImapSequenceSet.CreateRangeSet(1, 3);
      var set2 = ImapSequenceSet.CreateSet(5, 6, 8, 9);

      ImapSequenceSet c;

      c = ImapSequenceSet.Combine(set1, set2);

      Assert.AreEqual("1:3,5,6,8,9", c.ToString());
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      Assert.AreEqual(7L, c.ConcreteCount);

      TestUtils.SerializeBinary(c, delegate(ImapSequenceSet d) {
        Assert.AreEqual(c.GetType(), d.GetType());
        Assert.AreEqual("1:3,5,6,8,9", d.ToString());
      });

      c = set1 + set2;

      Assert.AreEqual("1:3,5,6,8,9", c.ToString());
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      Assert.AreEqual(7L, c.ConcreteCount);

      c = set1.CombineWith(set2);

      try {
        c.ToNumber();
        Assert.Fail("NotSupportedException not thrown");
      }
      catch (NotSupportedException) {
      }

      Assert.AreEqual("1:3,5,6,8,9", c.ToString());
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      Assert.AreEqual(7L, c.ConcreteCount);

      c = set2 + set1;

      Assert.AreEqual("5,6,8,9,1:3", c.ToString());
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      Assert.AreEqual(7L, c.ConcreteCount);

      TestUtils.SerializeBinary(c, delegate(ImapSequenceSet d) {
        Assert.AreEqual(c.GetType(), d.GetType());
        Assert.AreEqual("5,6,8,9,1:3", d.ToString());
      });

      try {
        c.ToNumber();
        Assert.Fail("NotSupportedException not thrown");
      }
      catch (NotSupportedException) {
      }
    }

    [Test]
    public void TestCombine_WithTo()
    {
      var set1 = ImapSequenceSet.CreateToSet(7);
      var set2 = ImapSequenceSet.CreateSet(5, 6, 8, 9);

      foreach (var c in new[] {
        ImapSequenceSet.Combine(set1, set2),
        set1 + set2,
        set1.CombineWith(set2),
      }) {
        //TODO: Assert.AreEqual("*:7,8,9", c.ToString());
        Assert.AreEqual("*:7,5,6,8,9", c.ToString());
        Assert.IsFalse(c.IsEmpty);
        Assert.IsFalse(c.IsSingle);
        //TODO: Assert.AreEqual(9L, c.ConcreteCount);
        Assert.AreEqual(11L, c.ConcreteCount);

        Assert.Throws<NotSupportedException>(() => c.ToNumber());

        TestUtils.SerializeBinary(c, delegate(ImapSequenceSet d) {
          Assert.AreEqual(c.GetType(), d.GetType());
          Assert.AreEqual("*:7,5,6,8,9", d.ToString());
        });
      }
    }

    [Test]
    public void TestCombine_WithFrom()
    {
      var set1 = ImapSequenceSet.CreateFromSet(7);
      var set2 = ImapSequenceSet.CreateSet(5, 6, 8, 9);

      foreach (var c in new[] {
        ImapSequenceSet.Combine(set1, set2),
        set1 + set2,
        set1.CombineWith(set2),
      }) {
        //TODO: Assert.AreEqual("5,6,7:*", c.ToString());
        Assert.AreEqual("7:*,5,6,8,9", c.ToString());
        Assert.IsFalse(c.IsEmpty);
        Assert.IsFalse(c.IsSingle);
        Assert.Less(c.ConcreteCount, 0L);

        Assert.Throws<NotSupportedException>(() => c.ToNumber());

        TestUtils.SerializeBinary(c, delegate(ImapSequenceSet d) {
          Assert.AreEqual(c.GetType(), d.GetType());
          Assert.AreEqual("7:*,5,6,8,9", d.ToString());
        });
      }
    }

    [Test]
    public void TestCombine_NonEmpty_WithEmpty()
    {
      var set1 = ImapSequenceSet.CreateRangeSet(2, 5);
      var set2 = ImapSequenceSet.CreateSet(new long[0]);

      Assert.IsFalse(set1.IsEmpty);
      Assert.IsFalse(set1.IsSingle);
      Assert.IsTrue(set2.IsEmpty);
      Assert.IsFalse(set2.IsSingle);

      ImapSequenceSet c;

      c = set1 + set2;

      Assert.AreEqual("2:5", c.ToString());
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      Assert.AreEqual(4L, c.ConcreteCount);

      TestUtils.SerializeBinary(c, delegate(ImapSequenceSet d) {
        Assert.AreEqual(c.GetType(), d.GetType());
        Assert.AreEqual("2:5", d.ToString());
      });

      try {
        c.ToNumber();
        Assert.Fail("NotSupportedException not thrown");
      }
      catch (NotSupportedException) {
      }

      c = set2 + set1;

      Assert.AreEqual("2:5", c.ToString());
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      Assert.AreEqual(4L, c.ConcreteCount);

      try {
        c.ToNumber();
        Assert.Fail("NotSupportedException not thrown");
      }
      catch (NotSupportedException) {
      }
    }

    [Test]
    public void TestCombine_Single_WithEmpty()
    {
      var set1 = ImapSequenceSet.CreateSet(3);
      var set2 = ImapSequenceSet.CreateSet(new long[0]);

      Assert.IsFalse(set1.IsEmpty);
      Assert.IsTrue(set1.IsSingle);
      Assert.IsTrue(set2.IsEmpty);
      Assert.IsFalse(set2.IsSingle);

      ImapSequenceSet c;

      c = set1 + set2;

      Assert.AreEqual("3", c.ToString());
      Assert.IsFalse(c.IsEmpty);
      Assert.IsTrue(c.IsSingle);
      Assert.AreEqual(1L, c.ConcreteCount);

      Assert.AreEqual(3L, c.ToNumber());

      TestUtils.SerializeBinary(c, delegate(ImapSequenceSet d) {
        Assert.AreEqual(c.GetType(), d.GetType());
        Assert.AreEqual("3", d.ToString());
      });
    }

    [Test]
    public void TestCombine_WithAll()
    {
      var set1 = ImapSequenceSet.CreateRangeSet(1, 3);
      var set2 = ImapSequenceSet.CreateAllSet();

      var c = set1 + set2;

      Assert.AreEqual("*", c.ToString());
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      Assert.Less(c.ConcreteCount, 0L);

      TestUtils.SerializeBinary(c, delegate(ImapSequenceSet d) {
        Assert.AreEqual(c.GetType(), d.GetType());
        Assert.AreEqual("*", d.ToString());
      });

      try {
        c.ToNumber();
        Assert.Fail("NotSupportedException not thrown");
      }
      catch (NotSupportedException) {
      }
    }

    [Test]
    public void TestCombine_WithCombinedSet()
    {
      var set1 = ImapSequenceSet.CreateSet(1, 2, 3) + ImapSequenceSet.CreateRangeSet(5, 7);
      var set2 = ImapSequenceSet.CreateRangeSet(9, 10) + ImapSequenceSet.CreateSet(15, 16, 18);

      var c = set1 + set2;

      Assert.AreEqual("1,2,3,5:7,9:10,15,16,18", c.ToString());
      Assert.IsFalse(c.IsEmpty);
      Assert.IsFalse(c.IsSingle);
      Assert.AreEqual(11, c.ConcreteCount);

      TestUtils.SerializeBinary(c, delegate(ImapSequenceSet d) {
        Assert.AreEqual(c.GetType(), d.GetType());
        Assert.AreEqual("1,2,3,5:7,9:10,15,16,18", d.ToString());
      });

      try {
        c.ToNumber();
        Assert.Fail("NotSupportedException not thrown");
      }
      catch (NotSupportedException) {
      }
    }

    [Test]
    public void TestCombine_WithSequenceSetAndUidSet()
    {
      foreach (var test in new[] {
        new {Set1 = ImapSequenceSet.CreateAllSet(),         Set2 = ImapSequenceSet.CreateUidAllSet()},
        new {Set1 = ImapSequenceSet.CreateAllSet(),         Set2 = ImapSequenceSet.CreateUidFromSet(2)},
        new {Set1 = ImapSequenceSet.CreateFromSet(3),       Set2 = ImapSequenceSet.CreateUidToSet(5)},
        new {Set1 = ImapSequenceSet.CreateRangeSet(1, 3),   Set2 = ImapSequenceSet.CreateUidRangeSet(5, 7)},
        new {Set1 = ImapSequenceSet.CreateToSet(10),        Set2 = ImapSequenceSet.CreateUidSet(3)},
      }) {
        try {
          ImapSequenceSet.Combine(test.Set1, test.Set2);
          Assert.Fail("ArgumentException not thrown: {0} {1}", test.Set1, test.Set2);
        }
        catch (ArgumentException) {
        }

        try {
          ImapSequenceSet.Combine(test.Set2, test.Set1);
          Assert.Fail("ArgumentException not thrown: {0} {1}", test.Set1, test.Set2);
        }
        catch (ArgumentException) {
        }
      }
    }

    [Test]
    public void TestCombined_ToArray()
    {
      var set1 = ImapSequenceSet.CreateRangeSet(1, 3);
      var set2 = ImapSequenceSet.CreateSet(5, 6, 8, 9);

      var combined = set1 + set2;

      var arr = combined.ToArray();

      Assert.AreEqual(7, arr.Length);
      Assert.AreEqual(1, arr[0]);
      Assert.AreEqual(2, arr[1]);
      Assert.AreEqual(3, arr[2]);
      Assert.AreEqual(5, arr[3]);
      Assert.AreEqual(6, arr[4]);
      Assert.AreEqual(8, arr[5]);
      Assert.AreEqual(9, arr[6]);
    }

    [Test]
    public void TestCombined_GetEnumerator()
    {
      var set1 = ImapSequenceSet.CreateRangeSet(1, 3);
      var set2 = ImapSequenceSet.CreateSet(5, 6, 8, 9);

      var combined = set1 + set2;

      var expected = new[] {1L, 2L, 3L, 5L, 6L, 8L, 9L};
      var index = 0;

      foreach (var num in combined) {
        Assert.AreEqual(expected[index++], num);
      }
    }

    [Test]
    [ExpectedException(typeof(NotSupportedException))]
    public void TestCombined_WithFromSet_ToArray()
    {
      var set1 = ImapSequenceSet.CreateFromSet(10);
      var set2 = ImapSequenceSet.CreateSet(5, 6, 8, 9);

      (set1 + set2).ToArray();
    }

    [Test, ExpectedException(typeof(NotSupportedException))]
    public void TestCombined_WithFromSet_GetEnumerator()
    {
      var set1 = ImapSequenceSet.CreateFromSet(10);
      var set2 = ImapSequenceSet.CreateSet(5, 6, 8, 9);

      foreach (var num in (set1 + set2)) {
        Assert.Fail("enumerated: {0}", num);
      }
    }

    [Test]
    public void TestIsEmpty()
    {
      Assert.IsTrue(ImapSequenceSet.CreateSet(new long[] {}).IsEmpty);

      Assert.IsFalse(ImapSequenceSet.CreateFromSet(1).IsEmpty);
      Assert.IsFalse(ImapSequenceSet.CreateUidFromSet(1).IsEmpty);

      Assert.IsFalse(ImapSequenceSet.CreateToSet(3).IsEmpty);
      Assert.IsFalse(ImapSequenceSet.CreateUidToSet(3).IsEmpty);

      Assert.IsFalse(ImapSequenceSet.CreateRangeSet(1, 3).IsEmpty);
      Assert.IsFalse(ImapSequenceSet.CreateUidRangeSet(1, 3).IsEmpty);

      Assert.IsFalse(ImapSequenceSet.CreateAllSet().IsEmpty);
      Assert.IsFalse(ImapSequenceSet.CreateUidAllSet().IsEmpty);

      Assert.IsFalse(ImapSequenceSet.Combine(ImapSequenceSet.CreateSet(new long[] {}), ImapSequenceSet.CreateSet(new long[] {1})).IsEmpty);
    }

    [Test]
    public void TestSplitIntoEach_Sequential1()
    {
      var range = ImapSequenceSet.CreateRangeSet(1L, 345L);
      var expectedString = new[] {
        "1:100",
        "101:200",
        "201:300",
        "301:345",
      };
      var expectedConcreteCount = new[] {
        100L,
        100L,
        100L,
        45L,
      };
      var index = 0;

      Assert.AreEqual(345L, range.ConcreteCount);

      foreach (var splitted in range.SplitIntoEach(100)) {
        Assert.AreEqual(expectedString[index], splitted.ToString());

        Assert.IsFalse(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.SerializeBinary(splitted, delegate(ImapSequenceSet d) {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedString[index], d.ToString());
          Assert.IsFalse(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(4, index);
    }

    [Test]
    public void TestSplitIntoEach_Sequential2()
    {
      var range = ImapSequenceSet.CreateRangeSet(1L, 400L);
      var expectedString = new[] {
        "1:100",
        "101:200",
        "201:300",
        "301:400",
      };
      var expectedConcreteCount = new[] {
        100L,
        100L,
        100L,
        100L,
      };
      var index = 0;

      Assert.AreEqual(400L, range.ConcreteCount);

      foreach (var splitted in range.SplitIntoEach(100)) {
        Assert.AreEqual(expectedString[index], splitted.ToString());

        Assert.IsFalse(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.SerializeBinary(splitted, delegate(ImapSequenceSet d) {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedString[index], d.ToString());
          Assert.IsFalse(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(4, index);
    }

    [Test]
    public void TestSplitIntoEach_Sequential3()
    {
      var range = ImapSequenceSet.CreateRangeSet(1L, 100L);
      var expectedString = new[] {
        "1:100",
      };
      var expectedConcreteCount = new[] {
        100L,
      };
      var index = 0;

      Assert.AreEqual(100L, range.ConcreteCount);

      foreach (var splitted in range.SplitIntoEach(100)) {
        Assert.AreEqual(expectedString[index], splitted.ToString());

        Assert.IsFalse(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.SerializeBinary(splitted, delegate(ImapSequenceSet d) {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedString[index], d.ToString());
          Assert.IsFalse(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(1, index);
    }

    [Test]
    public void TestSplitIntoEach_Sequential4()
    {
      var range = ImapSequenceSet.CreateRangeSet(3L, 12L);
      var expectedString = new[] {
        "3:6",
        "7:10",
        "11:12",
      };
      var expectedConcreteCount = new[] {
        4L,
        4L,
        2L,
      };
      var index = 0;

      foreach (var splitted in range.SplitIntoEach(4)) {
        Assert.AreEqual(expectedString[index], splitted.ToString());

        Assert.IsFalse(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.SerializeBinary(splitted, delegate(ImapSequenceSet d) {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedString[index], d.ToString());
          Assert.IsFalse(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(3, index);
    }

    [Test]
    public void TestSplitIntoEach_Sequential5()
    {
      var set1 = ImapSequenceSet.CreateUidRangeSet(3L, 8L);
      var set2 = ImapSequenceSet.CreateUidRangeSet(9L, 15L);
      var combined = set1 + set2;
      var expectedString = new[] {
        "3:7",
        "8:12",
        "13:15",
      };
      var expectedConcreteCount = new[] {
        5L,
        5L,
        3L,
      };
      var index = 0;

      foreach (var splitted in combined.SplitIntoEach(5)) {
        Assert.AreEqual(expectedString[index], splitted.ToString());

        Assert.IsTrue(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.SerializeBinary(splitted, delegate(ImapSequenceSet d) {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedString[index], d.ToString());
          Assert.IsTrue(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(3, index);
    }

    [Test]
    public void TestSplitIntoEach_NonSequential1()
    {
      var set1 = ImapSequenceSet.CreateRangeSet(7L, 10L);
      var set2 = ImapSequenceSet.CreateSet(1L, 3L, 5L);
      var set3 = ImapSequenceSet.CreateRangeSet(20L, 25L);
      var set4 = ImapSequenceSet.CreateSet(30L);
      var combined = set1 + set2 + set3 + set4;
      var expectedString = new[] {
        "7:9",
        "10,1,3",
        "5,20,21",
        "22:24",
        "25,30",
      };
      var expectedConcreteCount = new[] {
        3L,
        3L,
        3L,
        3L,
        2L,
      };
      var index = 0;

      foreach (var splitted in combined.SplitIntoEach(3)) {
        Assert.AreEqual(expectedString[index], splitted.ToString());

        Assert.IsFalse(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.SerializeBinary(splitted, delegate(ImapSequenceSet d) {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedString[index], d.ToString());
          Assert.IsFalse(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(5, index);
    }

    [Test]
    public void TestSplitIntoEach_NonSequential2()
    {
      var set1 = ImapSequenceSet.CreateRangeSet(7L, 10L);
      var set2 = ImapSequenceSet.CreateSet(1L, 3L, 5L);
      var set3 = ImapSequenceSet.CreateRangeSet(20L, 25L);
      var set4 = ImapSequenceSet.CreateSet(30L);
      var combined = set1 + set2 + set3 + set4;
      var expectedString = new[] {
        "7,8,9,10,1",
        "3,5,20,21,22",
        "23,24,25,30",
      };
      var expectedConcreteCount = new[] {
        5L,
        5L,
        4L,
      };
      var index = 0;

      foreach (var splitted in combined.SplitIntoEach(5)) {
        Assert.AreEqual(expectedString[index], splitted.ToString());

        Assert.IsFalse(splitted.IsUidSet);
        Assert.IsFalse(splitted.IsSingle);
        Assert.IsFalse(splitted.IsEmpty);
        Assert.AreEqual(expectedConcreteCount[index], splitted.ConcreteCount);

        TestUtils.SerializeBinary(splitted, delegate(ImapSequenceSet d) {
          Assert.AreEqual(splitted.GetType(), d.GetType());
          Assert.AreEqual(expectedString[index], d.ToString());
          Assert.IsFalse(d.IsUidSet);
        });

        index++;
      }

      Assert.AreEqual(3, index);
    }

    [Test]
    public void TestSplitIntoEach_Single()
    {
      var set = ImapSequenceSet.CreateSet(1L);
      var index = 0;

      foreach (var splitted in set.SplitIntoEach(5)) {
        Assert.AreSame(splitted, set);

        index++;
      }

      Assert.AreEqual(1, index);
    }

    [Test]
    public void TestFromUri()
    {
      var arr = ImapSequenceSet.FromUri(new Uri("imap://localhost/INBOX/;UID=1")).ToArray();

      Assert.AreEqual(1, arr.Length);
      Assert.AreEqual(1L, arr[0]);
    }

    [Test, ExpectedException(typeof(ArgumentException))]
    public void TestFromUri_InvalidForm()
    {
      ImapSequenceSet.FromUri(new Uri("imap://localhost/INBOX?UID 1"));
    }
  }
}