// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2010-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.Client;

namespace Smdn.Net.Pop3.Client {
  partial class PopClient {
    private const bool defaultGetUniqueId = false;

    public bool DeleteAfterRetrieve {
      get; set;
    }

    /// <exception cref="InvalidOperationException">throws if IsBusy is true</exception>
    public long MessageCount {
      get { ThrowIfNotConnected(); return EnsureDropListUpdated().MessageCount; }
    }

    /// <exception cref="InvalidOperationException">throws if IsBusy is true</exception>
    public long TotalSize {
      get { ThrowIfNotConnected(); return EnsureDropListUpdated().SizeInOctets; }
    }

    public void UpdateStatus()
    {
      RunExclusively(UpdateStatusCore);
    }

    private void UpdateStatusCore()
    {
      PopDropListing dropList;

      ThrowIfError(session.Stat(out dropList));

      dropListing = dropList;
    }

    private PopDropListing EnsureDropListUpdated()
    {
      if (dropListing.HasValue)
        return dropListing.Value;

      RunExclusively(UpdateStatusCore);

      return dropListing.Value;
    }

    private long GetMessageCount()
    {
      if (!dropListing.HasValue)
        UpdateStatusCore();

      return dropListing.Value.MessageCount;
    }

    public IEnumerable<PopMessageInfo> GetMessages()
    {
      return GetMessages(defaultGetUniqueId);
    }

    public IEnumerable<PopMessageInfo> GetMessages(bool getUniqueId)
    {
      return RunExclusively(delegate {
        return GetMessagesCore(getUniqueId);
      });
    }

    // TODO: CancellationToken
#if NET_4_5
    private IReadOnlyList<PopMessageInfo> GetMessagesCore(bool getUniqueId)
#else
    private IList<PopMessageInfo> GetMessagesCore(bool getUniqueId)
#endif
    {
      List<PopScanListing> scanLists;
      var expectedResultCount = dropListing.HasValue ? (int)dropListing.Value.MessageCount : 0;

      ThrowIfError(session.List(expectedResultCount, out scanLists));

      if (scanLists.Count == 0)
        return EmptyArray<PopMessageInfo>.Instance;

      var ret = new List<PopMessageInfo>(scanLists.Count);

      if (getUniqueId) {
        Dictionary<long, PopUniqueIdListing> uidLists = null;
        List<PopUniqueIdListing> uids = null;

        ThrowIfError(session.Uidl(scanLists.Count, out uids));

        uidLists = new Dictionary<long, PopUniqueIdListing>(scanLists.Count);

        foreach (var uid in uids) {
          uidLists[uid.MessageNumber] = uid;
        }

        foreach (var scanList in scanLists) {
          PopUniqueIdListing uniqueIdList;

          if (uidLists.TryGetValue(scanList.MessageNumber, out uniqueIdList)) {
            ret.Add(ToMessage(scanList, uniqueIdList));
          }
          else {
            var message = ToMessage(scanList, null);

            message.UniqueIdList = null; // XXX: set isUniqueIdListed true

            ret.Add(message);
          }
        }
      }
      else {
        foreach (var scanList in scanLists) {
          ret.Add(ToMessage(scanList, null));
        }
      }

#if NET_4_5
      return ret;
#else
      return ret.AsReadOnly();
#endif
    }

    public PopMessageInfo GetMessage(long messageNumber)
    {
      if (messageNumber <= 0L)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("messageNumber", messageNumber);

      return RunExclusively(delegate {
        return GetMessageByMessageNumber(messageNumber, defaultGetUniqueId);
      });
    }

    public PopMessageInfo GetMessage(long messageNumber, bool getUniqueId)
    {
      if (messageNumber <= 0L)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("messageNumber", messageNumber);

      return RunExclusively(delegate {
        return GetMessageByMessageNumber(messageNumber, getUniqueId);
      });
    }

    // TODO: CancellationToken
    private PopMessageInfo GetMessageByMessageNumber(long messageNumber, bool getUniqueId)
    {
      /*
       * find exist
       */
      var message = FindMessageByMessageNumber(messageNumber);

      if (message != null) {
        message.ThrowIfMarkedAsDeleted();

        if (!getUniqueId || message.UniqueIdList.HasValue)
          return message;
      }

      /*
       * list newly
       */
      if (GetMessageCount() < messageNumber)
        throw new PopMessageNotFoundException(messageNumber);

      PopScanListing scanList;

      var result = session.List(messageNumber, out scanList);

      if (result.Code == PopCommandResultCode.Error)
        throw new PopMessageNotFoundException(messageNumber);
      else
        ThrowIfError(result);

      if (getUniqueId) {
        PopUniqueIdListing uniqueIdList;

        ThrowIfError(session.Uidl(messageNumber, out uniqueIdList));

        return ToMessage(scanList, uniqueIdList);
      }
      else {
        return ToMessage(scanList, null);
      }
    }

    public PopMessageInfo GetFirstMessage()
    {
      return GetFirstMessage(defaultGetUniqueId);
    }

    public PopMessageInfo GetFirstMessage(bool getUniqueId)
    {
      return RunExclusively(delegate {
        return GetMessageByMessageNumber(1L, getUniqueId);
      });
    }

    public PopMessageInfo GetLastMessage()
    {
      return GetLastMessage(defaultGetUniqueId);
    }

    public PopMessageInfo GetLastMessage(bool getUniqueId)
    {
      return RunExclusively(delegate {
        return GetLastMessageCore(getUniqueId);
      });
    }

    private PopMessageInfo GetLastMessageCore(bool getUniqueId)
    {
      var lastMessageNumber = GetMessageCount();

      if (lastMessageNumber == 0L)
        throw new PopMessageNotFoundException(1L);
      else
        return GetMessageByMessageNumber(lastMessageNumber, getUniqueId);
    }

    public PopMessageInfo GetMessage(string uniqueId)
    {
      uniqueId = GetValidatedUniqueId(uniqueId);

      return RunExclusively(delegate {
        return GetMessageByUniqueId(uniqueId);
      });
    }

    // TODO: CancellationToken
    private PopMessageInfo GetMessageByUniqueId(string validatedUniqueId)
    {
      /*
       * find exist
       */
      var message = FindMessageByUniqueId(validatedUniqueId);

      if (message != null) {
        message.ThrowIfMarkedAsDeleted();

        return message;
      }

      /*
       * list newly
       */
      List<PopUniqueIdListing> uniqueIdLists;

      ThrowIfError(session.Uidl(out uniqueIdLists));

      foreach (var uniqueIdList in uniqueIdLists) {
        if (!uniqueIdList.Equals(validatedUniqueId))
          continue;

        PopScanListing scanList;

        ThrowIfError(session.List(uniqueIdList.MessageNumber, out scanList));

        return ToMessage(scanList, uniqueIdList);
      }

      throw new PopMessageNotFoundException(validatedUniqueId);
    }

    private PopMessageInfo FindMessageByMessageNumber(long messageNumber)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      PopMessageInfo message;

      if (messages.TryGetValue(messageNumber, out message))
        return message;
      else
        return null;
    }

    private PopMessageInfo FindMessageByUniqueId(string uniqueId)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      foreach (var message in messages.Values) {
        if (message.UniqueIdList.HasValue && message.UniqueIdList.Value.Equals(uniqueId))
          return message;
      }

      return null;
    }

    private PopMessageInfo ToMessage(PopScanListing scanList, PopUniqueIdListing? uniqueIdList)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      PopMessageInfo message;

      if (messages.TryGetValue(scanList.MessageNumber, out message)) {
        // update values
        message.ScanList = scanList;
        message.UniqueIdList = uniqueIdList ?? message.UniqueIdList;
      }
      else {
        message = new PopMessageInfo(this, scanList, uniqueIdList);

        messages[message.MessageNumber] = message;
      }

      return message;
    }

    public void CancelDelete()
    {
      RunExclusively(CancelDeleteCore);
    }

    private void CancelDeleteCore()
    {
      ThrowIfError(session.Rset());

      dropListing = null; // must be updated

      foreach (var message in messages.Values) {
        message.IsMarkedAsDeleted = false;
      }
    }

    public void KeepAlive()
    {
      RunExclusively(KeepAliveCore);
    }

    private void KeepAliveCore()
    {
      ThrowIfError(session.NoOp());
    }

    internal void Delete(PopMessageInfo message)
    {
#if DEBUG
      ThrowIfNotInExclusiveOperation();
#endif

      ThrowIfError(session.Dele(message.MessageNumber));

      dropListing = null; // must be updated
    }

    /*
     * utilities
     */
    internal static string GetValidatedUniqueId(string uniqueId)
    {
      if (uniqueId == null)
        throw new ArgumentNullException("uniqueId");
      if (uniqueId.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("uniqueId");

      return uniqueId;
    }

    internal static void ThrowIfError(PopCommandResult result)
    {
      if (result.Failed)
        throw new PopErrorResponseException(result);
    }

    private PopDropListing? dropListing = null;
    private Dictionary<long, PopMessageInfo> messages = new Dictionary<long, PopMessageInfo>();
  }
}
