//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2010-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

#if NET_4_0
#define ASYNC_OPS
#endif

#if ASYNC_OPS
using System;
using System.IO;
using System.Threading;
using System.Threading.Tasks;
using System.Text;

using Smdn.IO;

using ReportRetrieveProgress = Smdn.Net.Pop3.Client.Transaction.BuiltIn.RetrTopTransactionBase.ReportRetrieveProgress;

namespace Smdn.Net.Pop3.Client {
  partial class PopMessageInfo {
    /*
     * MarkAsDeletedAsync
     */
    public Task MarkAsDeletedAsync()
    {
      ThrowIfSessionClosed();

      return client.StartTaskExclusively(MarkAsDeletedCore,
                                         CancellationToken.None);
    }

    /*
     * GetUniqueIdAsync
     */
    public Task<string> GetUniqueIdAsync()
    {
      ThrowIfSessionClosed();

      if (isUniqueIdListed) {
#if NET_4_5
        return Task<string>.FromResult(ThrowIfNull(ListedUniqueId));
#else
        var tcs = new TaskCompletionSource<string>();

        tcs.SetResult(ThrowIfNull(ListedUniqueId));

        return tcs.Task;
#endif
      }
      else {
        ThrowIfMarkedAsDeleted();

        return client.StartTaskExclusively(() => ThrowIfNull(ListUniqueId()),
                                           CancellationToken.None);
      }
    }

    private static string ThrowIfNull(string uniqueId)
    {
      if (uniqueId == null)
        throw new PopIncapableException(PopCapability.Uidl);
      else
        return uniqueId;
    }

    /*
     * OpenReadAsync
     */
    public Task<Stream> OpenReadAsync()
    {
      return GetMessageStreamAsync(EntireMessage,
                                   CancellationToken.None,
                                   null,
                                   s => s);
    }

    public Task<Stream> OpenReadAsync(CancellationToken cancellationToken)
    {
      return GetMessageStreamAsync(EntireMessage,
                                   cancellationToken,
                                   null,
                                   s => s);
    }

#if NET_4_5
    public Task<Stream> OpenReadAsync(IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return GetMessageStreamAsync(EntireMessage,
                                   CancellationToken.None,
                                   progress,
                                   s => s);
    }

    public Task<Stream> OpenReadAsync(CancellationToken cancellationToken,
                                      IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return GetMessageStreamAsync(EntireMessage,
                                   cancellationToken,
                                   progress,
                                   s => s);
    }
#endif

    public Task<Stream> OpenReadAsync(int maxLines)
    {
      return GetMessageStreamAsync(maxLines,
                                   CancellationToken.None,
                                   null,
                                   s => s);
    }

    public Task<Stream> OpenReadAsync(int maxLines,
                                      CancellationToken cancellationToken)
    {
      return GetMessageStreamAsync(maxLines,
                                   cancellationToken,
                                   null,
                                   s => s);
    }

#if NET_4_5
    public Task<Stream> OpenReadAsync(int maxLines,
                                      IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return GetMessageStreamAsync(maxLines,
                                   CancellationToken.None,
                                   progress,
                                   s => s);
    }

    public Task<Stream> OpenReadAsync(int maxLines,
                                      CancellationToken cancellationToken,
                                      IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return GetMessageStreamAsync(maxLines,
                                   cancellationToken,
                                   progress,
                                   s => s);
    }
#endif

    private Task<TResult> GetMessageStreamAsync<TResult>(int maxLines,
                                                         CancellationToken cancellationToken,
#if NET_4_5
                                                         IProgress<PopMessageRetrieveProgressInfo> progress,
#else
                                                         object progress,
#endif
                                                         Converter<Stream, TResult> converter)
    {
      PrepareRetrieve(maxLines);

      return client.StartTaskExclusively(() => {
        var messageStream = Retrieve(maxLines, null, cancellationToken, progress);

        // convert stream and yield result
        return converter(messageStream);
      }, cancellationToken);
    }

    /*
     * ReadAsAsync(Converter<Stream>)
     */
    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter)
    {
      return ReadAsAsync(EntireMessage,
                         converter,
                         CancellationToken.None,
                         null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return ReadAsAsync(EntireMessage,
                         converter,
                         cancellationToken,
                         null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return ReadAsAsync(EntireMessage,
                         converter,
                         CancellationToken.None,
                         progress);
    }

    public Task<TResult> ReadAsAsync<TResult>(Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return ReadAsAsync(EntireMessage,
                         converter,
                         cancellationToken,
                         progress);
    }
#endif

    public Task<TResult> ReadAsAsync<TResult>(int maxLines,
                                              Converter<Stream, TResult> converter)
    {
      return ReadAsAsync(maxLines,
                         converter,
                         CancellationToken.None,
                         null);
    }

    public Task<TResult> ReadAsAsync<TResult>(int maxLines,
                                              Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return ReadAsAsync(maxLines,
                         converter,
                         cancellationToken,
                         null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(int maxLines,
                                              Converter<Stream, TResult> converter,
                                              IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return ReadAsAsync(maxLines,
                         converter,
                         CancellationToken.None,
                         progress);
    }
#endif

    // implementation of ReadAsAsync(Converter<Stream>)
#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(int maxLines,
                                              Converter<Stream, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<PopMessageRetrieveProgressInfo> progress)
#else
    private Task<TResult> ReadAsAsync<TResult>(int maxLines,
                                               Converter<Stream, TResult> converter,
                                               CancellationToken cancellationToken,
                                               object progress)
#endif
    {
      if (converter == null)
        throw new ArgumentNullException("converter");

      return GetMessageStreamAsync(maxLines,
                                   cancellationToken,
                                   progress,
                                   delegate(Stream s) {
        using (s) {
          return converter(s);
        }
      });
    }

    /*
     * ReadAsAsync<TResult>(Func<Stream, ...>)
     */

    /*
     * ReadAllBytesAsync()
     */
    public Task<byte[]> ReadAllBytesAsync()
    {
      return ReadAllBytesAsync(EntireMessage,
                               CancellationToken.None,
                               null);
    }

    public Task<byte[]> ReadAllBytesAsync(CancellationToken cancellationToken)
    {
      return ReadAllBytesAsync(EntireMessage,
                               cancellationToken,
                               null);
    }

#if NET_4_5
    public Task<byte[]> ReadAllBytesAsync(IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return ReadAllBytesAsync(EntireMessage,
                               CancellationToken.None,
                               progress);
    }

    public Task<byte[]> ReadAllBytesAsync(CancellationToken cancellationToken,
                                          IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return ReadAllBytesAsync(EntireMessage,
                               cancellationToken,
                               progress);
    }
#endif

    public Task<byte[]> ReadAllBytesAsync(int maxLines)
    {
      return ReadAllBytesAsync(maxLines,
                               CancellationToken.None,
                               null);
    }

    public Task<byte[]> ReadAllBytesAsync(int maxLines,
                                          CancellationToken cancellationToken)
    {
      return ReadAllBytesAsync(maxLines,
                               cancellationToken,
                               null);
    }

#if NET_4_5
    public Task<byte[]> ReadAllBytesAsync(int maxLines,
                                          IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return ReadAllBytesAsync(maxLines,
                               CancellationToken.None,
                               progress);
    }
#endif

    // implementation of ReadAllBytesAsync()
#if NET_4_5
    public Task<byte[]> ReadAllBytesAsync(int maxLines,
                                          CancellationToken cancellationToken,
                                          IProgress<PopMessageRetrieveProgressInfo> progress)
#else
    private Task<byte[]> ReadAllBytesAsync(int maxLines,
                                           CancellationToken cancellationToken,
                                           object progress)
#endif
    {
      return ReadAsAsync(maxLines,
                         s => s.ReadToEnd(),
                         cancellationToken,
                         progress);
    }

    /*
     * WriteToAsync(Stream)
     */
    public Task WriteToAsync(Stream stream)
    {
      return WriteToAsync(stream,
                          EntireMessage,
                          CancellationToken.None,
                          null);
    }

    public Task WriteToAsync(Stream stream,
                             CancellationToken cancellationToken)
    {
      return WriteToAsync(stream,
                          EntireMessage,
                          cancellationToken,
                          null);
    }

#if NET_4_5
    public Task WriteToAsync(Stream stream,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return WriteToAsync(stream,
                          EntireMessage,
                          CancellationToken.None,
                          progress);
    }

    public Task WriteToAsync(Stream stream,
                             CancellationToken cancellationToken,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return WriteToAsync(stream,
                          EntireMessage,
                          cancellationToken,
                          progress);
    }
#endif

    public Task WriteToAsync(Stream stream,
                             int maxLines)
    {
      return WriteToAsync(stream,
                          maxLines,
                          CancellationToken.None,
                          null);
    }

    public Task WriteToAsync(Stream stream,
                             int maxLines,
                             CancellationToken cancellationToken)
    {
      return WriteToAsync(stream,
                          maxLines,
                          cancellationToken,
                          null);
    }

#if NET_4_5
    public Task WriteToAsync(Stream stream,
                             int maxLines,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return WriteToAsync(stream,
                          maxLines,
                          CancellationToken.None,
                          progress);
    }
#endif

    // implementation of WriteToAsync(Stream)
#if NET_4_5
    public Task WriteToAsync(Stream stream,
                             int maxLines,
                             CancellationToken cancellationToken,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
#else
    private Task WriteToAsync(Stream stream,
                              int maxLines,
                              CancellationToken cancellationToken,
                              object progress)
#endif
    {
      if (stream == null)
        throw new ArgumentNullException("stream");
      if (!stream.CanWrite)
        throw ExceptionUtils.CreateArgumentMustBeWritableStream("stream");

      PrepareRetrieve(maxLines);

      return client.StartTaskExclusively(() => {
        Retrieve(maxLines, stream, cancellationToken, progress);
      }, cancellationToken);
    }

    /*
     * WriteToAsync(BinaryWriter)
     */
    public Task WriteToAsync(BinaryWriter writer)
    {
      return WriteToAsync(writer,
                          EntireMessage,
                          CancellationToken.None,
                          null);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             CancellationToken cancellationToken)
    {
      return WriteToAsync(writer,
                          EntireMessage,
                          cancellationToken,
                          null);
    }

#if NET_4_5
    public Task WriteToAsync(BinaryWriter writer,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return WriteToAsync(writer,
                          EntireMessage,
                          CancellationToken.None,
                          progress);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             CancellationToken cancellationToken,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return WriteToAsync(writer,
                          EntireMessage,
                          cancellationToken,
                          progress);
    }
#endif

    public Task WriteToAsync(BinaryWriter writer,
                             int maxLines)
    {
      return WriteToAsync(writer,
                          maxLines,
                          CancellationToken.None,
                          null);
    }

    public Task WriteToAsync(BinaryWriter writer,
                             int maxLines,
                             CancellationToken cancellationToken)
    {
      return WriteToAsync(writer,
                          maxLines,
                          cancellationToken,
                          null);
    }

#if NET_4_5
    public Task WriteToAsync(BinaryWriter writer,
                             int maxLines,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return WriteToAsync(writer,
                          maxLines,
                          CancellationToken.None,
                          progress);
    }
#endif

    // implementation of WriteToAsync(BinaryWriter)
#if NET_4_5
    public Task WriteToAsync(BinaryWriter writer,
                             int maxLines,
                             CancellationToken cancellationToken,
                             IProgress<PopMessageRetrieveProgressInfo> progress)
#else
    private Task WriteToAsync(BinaryWriter writer,
                              int maxLines,
                              CancellationToken cancellationToken,
                              object progress)
#endif
    {
      if (writer == null)
        throw new ArgumentNullException("writer");

      return GetMessageStreamAsync(maxLines,
                                   cancellationToken,
                                   progress,
                                   s => s).ContinueWith(t => {
        using (var stream = t.Result) {
          const int defaultBufferSize = 8 * 1024;

          var chunkedStream = stream as ChunkedMemoryStream;

          stream.CopyTo(writer, (chunkedStream == null) ? defaultBufferSize : chunkedStream.ChunkSize);
        }
      });
    }

    /*
     * SaveAsync()
     */
    public Task SaveAsync(string path)
    {
      return SaveAsync(path,
                       EntireMessage,
                       CancellationToken.None,
                       null);
    }

    public Task SaveAsync(string path,
                          CancellationToken cancellationToken)
    {
      return SaveAsync(path,
                       EntireMessage,
                       cancellationToken,
                       null);
    }

#if NET_4_5
    public Task SaveAsync(string path,
                          IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return SaveAsync(path,
                       EntireMessage,
                       CancellationToken.None,
                       progress);
    }

    public Task SaveAsync(string path,
                          CancellationToken cancellationToken,
                          IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return SaveAsync(path,
                       EntireMessage,
                       cancellationToken,
                       progress);
    }
#endif

    public Task SaveAsync(string path,
                          int maxLines)
    {
      return SaveAsync(path,
                       maxLines,
                       CancellationToken.None,
                       null);
    }

    public Task SaveAsync(string path,
                          int maxLines,
                          CancellationToken cancellationToken)
    {
      return SaveAsync(path,
                       maxLines,
                       cancellationToken,
                       null);
    }

#if NET_4_5
    public Task SaveAsync(string path,
                          int maxLines,
                          IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return SaveAsync(path,
                       maxLines,
                       CancellationToken.None,
                       progress);
    }
#endif

    // implementation of SaveAsync()
#if NET_4_5
    public Task SaveAsync(string path,
                          int maxLines,
                          CancellationToken cancellationToken,
                          IProgress<PopMessageRetrieveProgressInfo> progress)
#else
    private Task SaveAsync(string path,
                           int maxLines,
                           CancellationToken cancellationToken,
                           object progress)
#endif
    {
      if (path == null)
        throw new ArgumentNullException("path");
      if (path.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("path");

      PrepareRetrieve(maxLines);

      return client.StartTaskExclusively(() => {
        using (var fileStream = new FileStream(path, FileMode.Create, FileAccess.Write)) {
          Retrieve(maxLines, fileStream, cancellationToken, progress);
        }
      }, cancellationToken);
    }

    /*
     * OpenTextAsync()
     */
    public Task<StreamReader> OpenTextAsync()
    {
      return OpenTextAsync(iso8859_1,
                           EntireMessage,
                           CancellationToken.None,
                           null);
    }

    public Task<StreamReader> OpenTextAsync(Encoding charset)
    {
      return OpenTextAsync(charset,
                           EntireMessage,
                           CancellationToken.None,
                           null);
    }

    public Task<StreamReader> OpenTextAsync(int maxLines)
    {
      return OpenTextAsync(iso8859_1,
                           maxLines,
                           CancellationToken.None,
                           null);
    }

    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            int maxLines)
    {
      return OpenTextAsync(charset,
                           maxLines,
                           CancellationToken.None,
                           null);
    }

    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            int maxLines,
                                            CancellationToken cancellationToken)
    {
      return OpenTextAsync(charset,
                           maxLines,
                           cancellationToken,
                           null);
    }

#if NET_4_5
    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            int maxLines,
                                            IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return OpenTextAsync(charset,
                           maxLines,
                           CancellationToken.None,
                           progress);
    }
#endif

    // implementation of OpenTextAsync()
#if NET_4_5
    public Task<StreamReader> OpenTextAsync(Encoding charset,
                                            int maxLines,
                                            CancellationToken cancellationToken,
                                            IProgress<PopMessageRetrieveProgressInfo> progress)
#else
    private Task<StreamReader> OpenTextAsync(Encoding charset,
                                             int maxLines,
                                             CancellationToken cancellationToken,
                                             object progress)
#endif
    {
      if (charset == null)
        throw new ArgumentNullException("charset");

      return GetMessageStreamAsync(maxLines,
                                   cancellationToken,
                                   progress,
                                   s => new StreamReader(s, charset));
    }

    /*
     * ReadAsAsync<TResult>(Converter<StreamReader>)
     */
    public Task<TResult> ReadAsAsync<TResult>(Converter<StreamReader, TResult> converter)
    {
      return ReadAsAsync(iso8859_1,
                         EntireMessage,
                         converter,
                         CancellationToken.None,
                         null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              Converter<StreamReader, TResult> converter)
    {
      return ReadAsAsync(charset,
                         EntireMessage,
                         converter,
                         CancellationToken.None,
                         null);
    }

    public Task<TResult> ReadAsAsync<TResult>(int maxLines,
                                              Converter<StreamReader, TResult> converter)
    {
      return ReadAsAsync(iso8859_1,
                         maxLines,
                         converter,
                         CancellationToken.None,
                         null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              int maxLines,
                                              Converter<StreamReader, TResult> converter)
    {
      return ReadAsAsync(charset,
                         maxLines,
                         converter,
                         CancellationToken.None,
                         null);
    }

    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              int maxLines,
                                              Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken)
    {
      return ReadAsAsync(charset,
                         maxLines,
                         converter,
                         cancellationToken,
                         null);
    }

#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              int maxLines,
                                              Converter<StreamReader, TResult> converter,
                                              IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return ReadAsAsync(charset,
                         maxLines,
                         converter,
                         CancellationToken.None,
                         progress);
    }
#endif

    // implementation of ReadAsAsync<TResult>(Converter<StreamReader>)
#if NET_4_5
    public Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                              int maxLines,
                                              Converter<StreamReader, TResult> converter,
                                              CancellationToken cancellationToken,
                                              IProgress<PopMessageRetrieveProgressInfo> progress)
#else
    private Task<TResult> ReadAsAsync<TResult>(Encoding charset,
                                               int maxLines,
                                               Converter<StreamReader, TResult> converter,
                                               CancellationToken cancellationToken,
                                               object progress)
#endif
    {
      if (charset == null)
        throw new ArgumentNullException("charset");
      if (converter == null)
        throw new ArgumentNullException("converter");

      return GetMessageStreamAsync(maxLines,
                                   cancellationToken,
                                   progress,
                                   delegate (Stream s) {
        using (var reader = new StreamReader(s, charset)) {
          return converter(reader);
        }
      });
    }

    /*
     * ReadLinesAsync()
     */

    /*
    * ReadAllLinesAsync()
    */
    public Task<string[]> ReadAllLinesAsync()
    {
      return ReadAllLinesAsync(iso8859_1,
                               EntireMessage,
                               CancellationToken.None,
                               null);
    }

    public Task<string[]> ReadAllLinesAsync(Encoding charset)
    {
      return ReadAllLinesAsync(charset,
                               EntireMessage,
                               CancellationToken.None,
                               null);
    }

    public Task<string[]> ReadAllLinesAsync(int maxLines)
    {
      return ReadAllLinesAsync(iso8859_1,
                               maxLines,
                               CancellationToken.None,
                               null);
    }

    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            int maxLines)
    {
      return ReadAllLinesAsync(charset,
                               maxLines,
                               CancellationToken.None,
                               null);
    }

    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            int maxLines,
                                            CancellationToken cancellationToken)
    {
      return ReadAllLinesAsync(charset,
                               maxLines,
                               cancellationToken,
                               null);
    }

#if NET_4_5
    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            int maxLines,
                                            IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return ReadAllLinesAsync(charset,
                               maxLines,
                               CancellationToken.None,
                               progress);
    }
#endif

    // implementation of ReadAllLinesAsync()
#if NET_4_5
    public Task<string[]> ReadAllLinesAsync(Encoding charset,
                                            int maxLines,
                                            CancellationToken cancellationToken,
                                            IProgress<PopMessageRetrieveProgressInfo> progress)
#else
    private Task<string[]> ReadAllLinesAsync(Encoding charset,
                                             int maxLines,
                                             CancellationToken cancellationToken,
                                             object progress)
#endif
    {
      return ReadAsAsync(charset,
                         maxLines,
                         reader => reader.ReadAllLines(),
                         cancellationToken,
                         progress);
    }

    /*
     * ReadAllTextAsync()
     */
    public Task<string> ReadAllTextAsync()
    {
      return ReadAllTextAsync(iso8859_1,
                              EntireMessage,
                              CancellationToken.None,
                              null);
    }

    public Task<string> ReadAllTextAsync(Encoding charset)
    {
      return ReadAllTextAsync(charset,
                              EntireMessage,
                              CancellationToken.None,
                              null);
    }

    public Task<string> ReadAllTextAsync(int maxLines)
    {
      return ReadAllTextAsync(iso8859_1,
                              maxLines,
                              CancellationToken.None,
                              null);
    }

    public Task<string> ReadAllTextAsync(Encoding charset,
                                         int maxLines)
    {
      return ReadAllTextAsync(charset,
                              maxLines,
                              CancellationToken.None,
                              null);
    }

    public Task<string> ReadAllTextAsync(Encoding charset,
                                         int maxLines,
                                         CancellationToken cancellationToken)
    {
      return ReadAllTextAsync(charset,
                              maxLines,
                              cancellationToken,
                              null);
    }

#if NET_4_5
    public Task<string> ReadAllTextAsync(Encoding charset,
                                         int maxLines,
                                         IProgress<PopMessageRetrieveProgressInfo> progress)
    {
      return ReadAllTextAsync(charset,
                              maxLines,
                              CancellationToken.None,
                              progress);
    }
#endif

    // implementation of ReadAllTextAsync()
#if NET_4_5
    public Task<string> ReadAllTextAsync(Encoding charset,
                                         int maxLines,
                                         CancellationToken cancellationToken,
                                         IProgress<PopMessageRetrieveProgressInfo> progress)
#else
    private Task<string> ReadAllTextAsync(Encoding charset,
                                          int maxLines,
                                          CancellationToken cancellationToken,
                                          object progress)
#endif
    {
      return ReadAsAsync(charset,
                         maxLines,
                         reader => reader.ReadToEnd(),
                         cancellationToken,
                         progress);
    }
  }
}
#endif // ASYNC_OPS
