// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2010-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
#if NET_4_0
using System.Threading;
#endif

using Smdn.IO;
using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.Client;

using ReportRetrieveProgress = Smdn.Net.Pop3.Client.Transaction.BuiltIn.RetrTopTransactionBase.ReportRetrieveProgress;

namespace Smdn.Net.Pop3.Client {
  public sealed partial class PopMessageInfo {
    public PopClient Client {
      get { ThrowIfSessionClosed(); return client; }
      internal set { client = value; }
    }

    public long MessageNumber {
      get { ThrowIfSessionClosed(); return scanList.MessageNumber; }
    }

    public long Length {
      get { ThrowIfSessionClosed(); return scanList.SizeInOctets; }
    }

    public string UniqueId {
      get
      {
        string uniqueId;

        if (TryGetUniqueId(out uniqueId))
          return uniqueId;
        else
          return null;
      }
    }

    internal PopScanListing ScanList {
      get { return scanList; }
      set { scanList = value; }
    }

    internal PopUniqueIdListing? UniqueIdList {
      get { return uniqueIdList; }
      set
      {
        uniqueIdList = value;
        isUniqueIdListed = true;
      }
    }

    internal string ListedUniqueId {
      get
      {
        if (uniqueIdList.HasValue && uniqueIdList.Value.UniqueId != null)
          return uniqueIdList.Value.UniqueId;
        else
          return null;
      }
    }

    public bool IsMarkedAsDeleted {
      get { ThrowIfSessionClosed(); return isMarkedAsDeleted; }
      internal set { isMarkedAsDeleted = value; }
    }

    internal PopMessageInfo(PopClient client, PopScanListing scanList, PopUniqueIdListing? uniqueIdList)
    {
      this.client = client;
      this.scanList = scanList;
      this.uniqueIdList = uniqueIdList;
      this.isMarkedAsDeleted = false;
      this.isUniqueIdListed = uniqueIdList.HasValue;
    }

    private PopClient client;
    private PopScanListing scanList;
    private PopUniqueIdListing? uniqueIdList;
    private bool isMarkedAsDeleted;
    private bool isUniqueIdListed;

    /*
     * operation
     */
    public void MarkAsDeleted()
    {
      ThrowIfSessionClosed();

      client.RunExclusively(MarkAsDeletedCore);
    }

    private void MarkAsDeletedCore()
    {
      if (isMarkedAsDeleted)
        return;

      client.Delete(this);

      isMarkedAsDeleted = true;
    }

    public string GetUniqueId()
    {
      string uniqueId;

      if (TryGetUniqueId(out uniqueId))
        return uniqueId;
      else
        throw new PopIncapableException(PopCapability.Uidl);
    }

    public bool TryGetUniqueId(out string uniqueId)
    {
      ThrowIfSessionClosed();

      if (isUniqueIdListed) {
        uniqueId = ListedUniqueId;
      }
      else {
        ThrowIfMarkedAsDeleted();

        uniqueId = client.RunExclusively<string>(ListUniqueId);
      }

      return (uniqueId != null);
    }

    private string ListUniqueId()
    {
#if DEBUG
      client.ThrowIfNotInExclusiveOperation();
#endif

      PopUniqueIdListing uid;

      var ret = client.Session.Uidl(MessageNumber, out uid);

      switch (ret.Code) {
        case PopCommandResultCode.Ok:
          isUniqueIdListed = true;
          uniqueIdList = uid;
          return uid.UniqueId;

        case PopCommandResultCode.Error:
          isUniqueIdListed = true;
          uniqueIdList = null;
          return null;

        default:
          PopClient.ThrowIfError(ret);
          return null; // never return
      }
    }

    internal const int EntireMessage = -1;

    public Stream OpenRead()
    {
      return OpenRead(EntireMessage);
    }

    /// <param name="maxLines">
    /// A maximum line count to retrieve.
    /// <list type="number">
    ///   <listheader><term>value</term><description>behavior</description></listheader>
    ///   <item><term>-1</term><description>entire message</description></item>
    ///   <item><term>0</term><description>message without body</description></item>
    ///   <item><term>1 or greater</term><description>message with specified lines of body</description></item>
    /// </list>
    /// </param>
    public Stream OpenRead(int maxLines)
    {
      PrepareRetrieve(maxLines);

      return client.RunExclusively(delegate {
#if NET_4_0
        return Retrieve(maxLines, null, CancellationToken.None, null);
#else
        return Retrieve(maxLines, null);
#endif
      });
    }

    private void PrepareRetrieve(int maxLines)
    {
      ThrowIfSessionClosed();
      ThrowIfMarkedAsDeleted();
      ThrowIfMaxLinesOutOfRange(maxLines);
    }

    private Stream Retrieve
    (
#if NET_4_5
      int maxLines,
      Stream retrieveToStream,
      CancellationToken cancellationToken,
      IProgress<PopMessageRetrieveProgressInfo> progress
#elif NET_4_0
      int maxLines,
      Stream retrieveToStream,
      CancellationToken cancellationToken,
      object progress
#else
      int maxLines,
      Stream retrieveToStream
#endif
    )
    {
      var closeStreamIfCancellationRequested = (retrieveToStream == null);

#if NET_4_5
      var reportProgress = (progress == null)
                             ? (ReportRetrieveProgress)null
                             : (ReportRetrieveProgress)((octets, lines) => progress.Report(new PopMessageRetrieveProgressInfo(octets, lines, scanList.SizeInOctets, cancellationToken)));

      if (maxLines == EntireMessage)
        PopClient.ThrowIfError(client.Session.Retr(MessageNumber, reportProgress, ref retrieveToStream));
      else
        PopClient.ThrowIfError(client.Session.Top(MessageNumber, maxLines, reportProgress, ref retrieveToStream));
#else
      if (maxLines == EntireMessage)
        PopClient.ThrowIfError(client.Session.Retr(MessageNumber, null, ref retrieveToStream));
      else
        PopClient.ThrowIfError(client.Session.Top(MessageNumber, maxLines, null, ref retrieveToStream));
#endif

      if (client.DeleteAfterRetrieve) {
#if NET_4_0
        // check cancellation requested before marking as deleted
        if (cancellationToken.IsCancellationRequested && closeStreamIfCancellationRequested)
          retrieveToStream.Close();

        cancellationToken.ThrowIfCancellationRequested();
#endif

        if (!isMarkedAsDeleted)
          MarkAsDeletedCore();
      }

#if NET_4_0
      if (cancellationToken.IsCancellationRequested && closeStreamIfCancellationRequested)
        // check cancellation requested before yielding result
        retrieveToStream.Close();

      cancellationToken.ThrowIfCancellationRequested();
#endif

      return retrieveToStream;
    }

    /*
     * ReadAs<TResult>(Converter<Stream>)
     */
    public TResult ReadAs<TResult>(Converter<Stream, TResult> converter)
    {
      return ReadAs(EntireMessage, converter);
    }

    public TResult ReadAs<TResult>(int maxLines, Converter<Stream, TResult> converter)
    {
      if (converter == null)
        throw new ArgumentNullException("converter");

      using (var stream = OpenRead(maxLines)) {
        return converter(stream);
      }
    }

    /*
     * ReadAs<TResult>(Func<Stream, ...>)
     */
    public TResult ReadAs<T, TResult>
      (Func<Stream, T, TResult> read, T arg)
    {
      return ReadAs(EntireMessage, read, arg);
    }

    public TResult ReadAs<T1, T2, TResult>
      (Func<Stream, T1, T2, TResult> read, T1 arg1, T2 arg2)
    {
      return ReadAs(EntireMessage, read, arg1, arg2);
    }

    public TResult ReadAs<T1, T2, T3, TResult>
      (Func<Stream, T1, T2, T3, TResult> read, T1 arg1, T2 arg2, T3 arg3)
    {
      return ReadAs(EntireMessage, read, arg1, arg2, arg3);
    }

    public TResult ReadAs<T1, T2, T3, T4, TResult>
      (Func<Stream, T1, T2, T3, T4, TResult> read, T1 arg1, T2 arg2, T3 arg3, T4 arg4)
    {
      return ReadAs(EntireMessage, read, arg1, arg2, arg3, arg4);
    }

    public TResult ReadAs<T, TResult>
      (int maxLines,
       Func<Stream, T, TResult> read, T arg)
    {
      if (read == null)
        throw new ArgumentNullException("read");

      using (var stream = OpenRead(maxLines)) {
        return read(stream, arg);
      }
    }

    public TResult ReadAs<T1, T2, TResult>
      (int maxLines,
       Func<Stream, T1, T2, TResult> read, T1 arg1, T2 arg2)
    {
      if (read == null)
        throw new ArgumentNullException("read");

      using (var stream = OpenRead(maxLines)) {
        return read(stream, arg1, arg2);
      }
    }

    public TResult ReadAs<T1, T2, T3, TResult>
      (int maxLines,
       Func<Stream, T1, T2, T3, TResult> read, T1 arg1, T2 arg2, T3 arg3)
    {
      if (read == null)
        throw new ArgumentNullException("read");

      using (var stream = OpenRead(maxLines)) {
        return read(stream, arg1, arg2, arg3);
      }
    }

    public TResult ReadAs<T1, T2, T3, T4, TResult>
      (int maxLines,
       Func<Stream, T1, T2, T3, T4, TResult> read, T1 arg1, T2 arg2, T3 arg3, T4 arg4)
    {
      if (read == null)
        throw new ArgumentNullException("read");

      using (var stream = OpenRead(maxLines)) {
        return read(stream, arg1, arg2, arg3, arg4);
      }
    }

    /*
     * ReadAllBytes()
     */
    public byte[] ReadAllBytes()
    {
      return ReadAs<byte[]>(ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(int maxLines)
    {
      return ReadAs<byte[]>(maxLines, ReadAllBytesProc);
    }

    private static byte[] ReadAllBytesProc(Stream stream)
    {
      return stream.ReadToEnd();
    }

    /*
     * WriteTo(Stream)
     */
    public void WriteTo(Stream stream)
    {
      WriteTo(stream, EntireMessage);
    }

    public void WriteTo(Stream stream, int maxLines)
    {
      if (stream == null)
        throw new ArgumentNullException("stream");
      if (!stream.CanWrite)
        throw ExceptionUtils.CreateArgumentMustBeWritableStream("stream");

      PrepareRetrieve(maxLines);

      client.RunExclusively(delegate {
#if NET_4_0
        Retrieve(maxLines, stream, CancellationToken.None, null);
#else
        Retrieve(maxLines, stream);
#endif
      });
    }

    /*
     * WriteTo(BinaryWriter)
     */
    public void WriteTo(BinaryWriter writer)
    {
      WriteTo(writer, EntireMessage);
    }

    public void WriteTo(BinaryWriter writer, int maxLines)
    {
      if (writer == null)
        throw new ArgumentNullException("writer");

      using (var messageStream = OpenRead(maxLines)) {
        var chunked = messageStream as ChunkedMemoryStream;

        if (chunked == null)
          messageStream.CopyTo(writer);
        else
          messageStream.CopyTo(writer, chunked.ChunkSize);
      }
    }

    /*
     * Save()
     */
    public void Save(string path)
    {
      Save(new FileInfo(path), EntireMessage);
    }

    public void Save(string path, int maxLines)
    {
      Save(new FileInfo(path), maxLines);
    }

    internal void Save(FileInfo file, int maxLines)
    {
      PrepareRetrieve(maxLines);

      client.RunExclusively(delegate {
        using (var fileStream = file.Open(FileMode.Create, FileAccess.Write)) {
#if NET_4_0
          Retrieve(maxLines, fileStream, CancellationToken.None, null);
#else
          Retrieve(maxLines, fileStream);
#endif
        }
      });
    }

    /*
     * OpenText()
     */
    private static readonly Encoding iso8859_1 = Encoding.GetEncoding("ISO-8859-1");

    public StreamReader OpenText()
    {
      return OpenText(iso8859_1, EntireMessage);
    }

    public StreamReader OpenText(Encoding encoding)
    {
      return OpenText(encoding, EntireMessage);
    }

    public StreamReader OpenText(int maxLines)
    {
      return OpenText(iso8859_1, maxLines);
    }

    public StreamReader OpenText(Encoding encoding, int maxLines)
    {
      if (encoding == null)
        throw new ArgumentNullException("encoding");

      return new StreamReader(OpenRead(maxLines), encoding, false);
    }

    /*
     * ReadAs<TResult>(Converter<StreamReader>)
     */
    public TResult ReadAs<TResult>(Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(), converter);
    }

    public TResult ReadAs<TResult>(Encoding encoding, Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(encoding), converter);
    }

    public TResult ReadAs<TResult>(int maxLines, Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(maxLines), converter);
    }

    public TResult ReadAs<TResult>(Encoding encoding, int maxLines, Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(encoding, maxLines), converter);
    }

    private static TResult ReadAsCore<TResult>(StreamReader reader, Converter<StreamReader, TResult> converter)
    {
      using (reader) {
        if (converter == null)
          throw new ArgumentNullException("converter");

        return converter(reader);
      }
    }

    /*
     * ReadLines()
     */
    public IEnumerable<string> ReadLines()
    {
      return ReadLines(iso8859_1, EntireMessage);
    }

    public IEnumerable<string> ReadLines(Encoding encoding)
    {
      return ReadLines(encoding, EntireMessage);
    }

    public IEnumerable<string> ReadLines(int maxLines)
    {
      return ReadLines(iso8859_1, maxLines);
    }

    public IEnumerable<string> ReadLines(Encoding encoding, int maxLines)
    {
      using (var reader = OpenText(encoding, maxLines)) {
        for (;;) {
          var line = reader.ReadLine();

          if (line == null)
            break;
          else
            yield return line;
        }
      }
    }

    /*
     * ReadAllLines()
     */
    public string[] ReadAllLines()
    {
      return ReadLines().ToArray();
    }

    public string[] ReadAllLines(Encoding encoding)
    {
      return ReadLines(encoding).ToArray();
    }

    public string[] ReadAllLines(int maxLines)
    {
      return ReadLines(maxLines).ToArray();
    }

    public string[] ReadAllLines(Encoding encoding, int maxLines)
    {
      return ReadLines(encoding, maxLines).ToArray();
    }

    /*
     * ReadAllText()
     */
    public string ReadAllText()
    {
      return ReadAs<string>(ReadAllTextProc);
    }

    public string ReadAllText(Encoding encoding)
    {
      return ReadAs<string>(encoding, ReadAllTextProc);
    }

    public string ReadAllText(int maxLines)
    {
      return ReadAs<string>(maxLines, ReadAllTextProc);
    }

    public string ReadAllText(Encoding encoding, int maxLines)
    {
      return ReadAs<string>(encoding, maxLines, ReadAllTextProc);
    }

    private static string ReadAllTextProc(StreamReader reader)
    {
      return reader.ReadToEnd();
    }

    /*
     * utility methods
     */
    internal void ThrowIfMarkedAsDeleted()
    {
      if (IsMarkedAsDeleted)
        throw new PopMessageDeletedException(MessageNumber);
    }

    private void ThrowIfSessionClosed()
    {
      if (client == null)
        throw new InvalidOperationException("session has been closed");
    }

    private void ThrowIfMaxLinesOutOfRange(int maxLines)
    {
      if (maxLines < -1)
        throw new ArgumentOutOfRangeException("maxLines", maxLines, "must be -1, 0 or positive number");
    }

    public override string ToString()
    {
      return string.Format("{{PopMessageInfo: Authority='{0}', MessageNumber={1}, Length={2}}}",
                           client.Profile.Authority,
                           MessageNumber,
                           Length);
    }
  }
}
