using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Runtime.Remoting.Messaging;
using System.Threading;

using Smdn.Net;

namespace Smdn.Net.Pop3.Client.Session {
  public class PopPseudoServer : IDisposable {
    private class Response {
      public string ExpectedCommand {
        get { return expectedCommand; }
      }

      public string ResponseText {
        get { return responseText; }
      }

      public Action ActionBeforeRespond {
        get { return actionBeforeRespond; }
      }

      public Action ActionAfterRespond {
        get { return actionAfterRespond; }
      }

      public bool IsQuitResponse {
        get; set;
      }

      public Response(string expectedCommand,
                      string responseText,
                      Action actionBeforeRespond,
                      Action actionAfterRespond)
      {
        this.expectedCommand = expectedCommand;
        this.responseText = responseText;
        this.actionBeforeRespond = actionBeforeRespond;
        this.actionAfterRespond = actionAfterRespond;
      }

      private readonly string expectedCommand;
      private readonly string responseText;
      private readonly Action actionBeforeRespond;
      private readonly Action actionAfterRespond;
    }

    public IPEndPoint ServerEndPoint {
      get { return serverEndPoint; }
    }

    public string Host {
      get { return ServerEndPoint.Address.ToString(); }
    }

    public int Port {
      get { return ServerEndPoint.Port; }
    }

    public string HostPort {
      get { return string.Format("{0}:{1}", Host, Port); }
    }

    public bool ThrowIfServerNotStopped {
      get; set;
    }

    public PopPseudoServer()
    {
      ThrowIfServerNotStopped = true;

      Initialize();
    }

    private void Initialize()
    {
      server = new TcpListener(new IPEndPoint(IPAddress.Loopback, 0));
      server.Start();

      serverEndPoint = server.LocalEndpoint as IPEndPoint;

      SetGreetingResponse("+OK\r\n");

      clientRequestWaitHandle = new ManualResetEvent(false);
    }

    void IDisposable.Dispose()
    {
      if (acceptClientAsyncResult != null) {
        var acceptClient = (Action)(acceptClientAsyncResult as AsyncResult).AsyncDelegate;

        if (acceptClientAsyncResult.AsyncWaitHandle.WaitOne(500))
          acceptClient.EndInvoke(acceptClientAsyncResult);
        else if (ThrowIfServerNotStopped)
          throw new ApplicationException("server not stopped");

        acceptClientAsyncResult = null;
      }

      if (server != null) {
        server.Stop();
        server = null;
      }

      if (clientRequestWaitHandle != null) {
        clientRequestWaitHandle.Close();
        clientRequestWaitHandle = null;
      }

      if (serverProcException != null) {
        Console.Error.WriteLine("unhandled server exception:{0}{1}", Environment.NewLine, serverProcException);

        throw new ApplicationException("exception not handled", serverProcException);
      }
    }

    public void Start()
    {
      if (acceptClientAsyncResult != null)
        throw new InvalidOperationException("already started");

      var acceptClient = new Action(AcceptClient);

      acceptClientAsyncResult = acceptClient.BeginInvoke(null, null);
    }

    public void Restart()
    {
      (this as IDisposable).Dispose();

      Initialize();

      Start();
    }

    public static PopPseudoServer StartNew()
    {
      var server = new PopPseudoServer();

      server.Start();

      return server;
    }

    public static PopPseudoServer StartNotAccepting()
    {
      var server = new PopPseudoServer();

      server.server.Start();

      return server;
    }

    public void SetGreetingResponse(string greeting)
    {
      SetGreetingResponse(greeting, null, null, false);
    }

    public void SetGreetingResponse(string greeting,
                                    Action actionBeforeRespond,
                                    Action actionAfterRespond)
    {
      SetGreetingResponse(greeting, actionBeforeRespond, actionAfterRespond, false);
    }

    public void SetGreetingResponse(string greeting,
                                    Action actionBeforeRespond,
                                    Action actionAfterRespond,
                                    bool isQuitResponse)
    {
      greetingResponse = new Response(null, greeting, actionBeforeRespond, actionAfterRespond);

      greetingResponse.IsQuitResponse = isQuitResponse;
    }

    public void EnqueueResponse(string responseText)
    {
      EnqueueResponse(null, responseText, null, null);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText)
    {
      EnqueueResponse(expectedCommand,
                      responseText,
                      null,
                      null,
                      false);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText,
                                bool isQuitResponse)
    {
      EnqueueResponse(expectedCommand,
                      responseText,
                      null,
                      null,
                      isQuitResponse);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText,
                                Action actionBeforeRespond,
                                Action actionAfterRespond)
    {
      EnqueueResponse(expectedCommand,
                      responseText,
                      actionBeforeRespond,
                      actionAfterRespond,
                      false);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText,
                                Action actionBeforeRespond,
                                Action actionAfterRespond,
                                bool isQuitResponse)
    {
      var response = new Response(expectedCommand,
                                  responseText,
                                  actionBeforeRespond,
                                  actionAfterRespond);

      response.IsQuitResponse = isQuitResponse;

      responseQueue.Enqueue(response);
    }

    // TODO: rename
    public string DequeueRequest()
    {
      lock (commandQueue) {
        if (0 < commandQueue.Count)
          return commandQueue.Dequeue();
        else
          return null;
      }
    }

    public void WaitForClientRequest()
    {
      WaitForClientRequest(TimeSpan.FromSeconds(5));
    }

    public void WaitForClientRequest(TimeSpan timeout)
    {
      clientRequestWaitHandle.Reset();

      clientRequestWaitHandle.WaitOne(timeout);
    }

    private void AcceptClient()
    {
      Console.WriteLine("listening...");

      var client = server.AcceptTcpClient();

      Console.WriteLine("client accepted");

      client.ReceiveTimeout = 100;

      try {
        Console.WriteLine("server process started");

        ServerProc(client);

        Console.WriteLine("server process successfully finished");
      }
      catch (Exception ex) {
        serverProcException = ex;

        Console.WriteLine("server process stopped by {0}", ex.GetType().FullName);
      }

      client.Close();

      Console.WriteLine("connection closed");
    }

    private void ServerProc(TcpClient client)
    {
      var commandStream = client.GetStream();

      Send(greetingResponse, commandStream);

      if (greetingResponse.IsQuitResponse)
        return;

      var commandBuffer = new byte[0x1000];

      for (;;) {
        if (!IsConnected(client))
          return;

        /*        
         * wait for command sent
         */
        string commandText = null;

        try {
          var commandLength = commandStream.Read(commandBuffer, 0, commandBuffer.Length);

          if (commandLength == 0 && !IsConnected(client))
            return;

          commandText = NetworkTransferEncoding.Transfer8Bit.GetString(commandBuffer, 0, commandLength);

          Console.Write("C:\t");
          Console.Write(commandText);

          lock (commandQueue) {
            commandQueue.Enqueue(commandText);
          }
        }
        catch (IOException ex) {
          if (!client.Connected)
            return; // disconnected

          var exSocket = ex.InnerException as SocketException;

          if (exSocket != null &&
              (exSocket.SocketErrorCode == SocketError.WouldBlock || exSocket.SocketErrorCode == SocketError.TimedOut)) {
            continue;
          }

          throw;
        }

        /*
         * send response
         */
        while (responseQueue.Count == 0) {
          clientRequestWaitHandle.Set();

          if (!IsConnected(client))
            return;

          Thread.Sleep(10);
        }

        var response = responseQueue.Dequeue();

        if (string.IsNullOrEmpty(response.ExpectedCommand) ||
            0 <= commandText.IndexOf(response.ExpectedCommand, StringComparison.OrdinalIgnoreCase))
          Send(response, commandStream);
        else
          throw new InvalidOperationException(string.Format("unexpected command{0}expected '{1}' but was '{2}'",
                                                            Environment.NewLine,
                                                            response.ExpectedCommand,
                                                            commandText.Replace("\r", "\\r").Replace("\n", "\\n")));

        if (response.IsQuitResponse)
          return;
      }
    }

    private void Send(Response response, Stream requestStream)
    {
      if (response.ActionBeforeRespond != null)
        response.ActionBeforeRespond();

      var bytes = NetworkTransferEncoding.Transfer8Bit.GetBytes(response.ResponseText);

      requestStream.Write(bytes, 0, bytes.Length);

      Console.Write("S:\t");
      Console.Write(response.ResponseText);

      if (response.ActionAfterRespond != null)
        response.ActionAfterRespond();
    }

    private static readonly byte[] peekClientBuffer = new byte[1];

    private static bool IsConnected(TcpClient client)
    {
      try {
        if (client.Client.Poll(0, SelectMode.SelectRead)) {
          if (client.Client.Receive(peekClientBuffer, SocketFlags.Peek) == 0) {
            Console.WriteLine("client disconnected");
            return false;
          }
        }
      }
      catch (SocketException ex) {
        switch (ex.SocketErrorCode) {
          case SocketError.ConnectionAborted:
          case SocketError.ConnectionReset:
            return false;

          default:
            Console.Error.WriteLine("unexpected socket error: {0}", ex.SocketErrorCode);
            throw;
        }
      }
      
      return true;
    }

    private TcpListener server;
    private IPEndPoint serverEndPoint;
    private IAsyncResult acceptClientAsyncResult;
    private readonly Queue<Response> responseQueue = new Queue<Response>();
    private readonly Queue<string> commandQueue = new Queue<string>();
    private ManualResetEvent clientRequestWaitHandle;
    private Response greetingResponse;
    private Exception serverProcException = null;
  }
}



