using System;
using System.IO;
using System.Net;
using System.Collections.Generic;
#if NET_4_0
using System.Threading;
using System.Threading.Tasks;
#endif
using NUnit.Framework;

using Smdn.Net.Pop3.Protocol;

namespace Smdn.Net.Pop3.Client.Session {
  [TestFixture]
  public class PopSessionCreatorTests {
    private class Profile : IPopSessionProfile {
      public Uri Authority {
        get; set;
      }

      public int Timeout {
        get; set;
      }

      public int SendTimeout {
        get; set;
      }

      public int ReceiveTimeout {
        get; set;
      }

      public ICredentialsByHost Credentials {
        get; set;
      }

      public string[] UsingSaslMechanisms {
        get; set;
      }

      public bool UseTlsIfAvailable {
        get; set;
      }

      public bool AllowInsecureLogin {
        get; set;
      }

      public bool AllowStandardLogin {
        get; set;
      }

      public bool AllowApopLogin {
        get; set;
      }

      public Profile(NetworkCredential cred, string userinfo, string hostPort)
        : this(cred, PopUri.UriSchemePop, userinfo, hostPort)
      {
      }

      public Profile(NetworkCredential cred, string scheme, string userinfo, string hostPort)
      {
        this.UseTlsIfAvailable = true;
        this.AllowInsecureLogin = false;
        this.AllowStandardLogin = true;
        this.AllowApopLogin = true;
        this.Timeout = 1000;
        this.SendTimeout = -1;
        this.ReceiveTimeout = -1;
        this.Credentials = cred;

        if (userinfo == null)
          this.Authority = new Uri(string.Format("{0}://{1}/", scheme, hostPort));
        else
          this.Authority = new Uri(string.Format("{0}://{1}@{2}/", scheme, userinfo, hostPort));
      }
    }

    [Test]
    public void TestCreateSession_CreateSslStream_CallbackNull()
    {
      var prof = new Profile(new NetworkCredential("user", "pass"), "POPS", "user", "pop.invalid");

      Assert.Throws<ArgumentNullException>(() => PopSessionCreator.CreateSession(prof, null, null));
    }

    [Test]
    public void TestCreateSession_ConnectRefusedByTimeout()
    {
      using (var server = PopPseudoServer.StartNotAccepting()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.Timeout = 500;

        try {
          PopSessionCreator.CreateSession(prof, null, null);

          Assert.Fail("PopConnectionException not thrown");
        }
        catch (PopConnectionException ex) {
          var timeoutException = ex.InnerException as TimeoutException;

          Assert.IsNotNull(timeoutException);
        }
      }
    }

    [Test]
    public void TestCreateSession_ConnectRefusedByDnsError()
    {
      var prof = new Profile(new NetworkCredential("user", "pass"), "user", "pop.invalid");

      prof.Timeout = -1;

      try {
        PopSessionCreator.CreateSession(prof, null, null);

        Assert.Fail("PopConnectionException not thrown");
      }
      catch (PopConnectionException ex) {
        var socketException = ex.InnerException as System.Net.Sockets.SocketException;

        Assert.IsNotNull(socketException);
      }
    }

    [Test]
    public void TestCreateSession_ConnectToSecurePort()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "pops", "user", server.HostPort);

        prof.AllowInsecureLogin = false;

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("USER user\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("PASS pass\r\n",
                               "+OK\r\n");

        var streamUpgraded = false;

        using (var session = PopSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SSL stream
        })) {
          Assert.IsTrue(streamUpgraded, "stream upgraded");

          Assert.AreEqual(PopSessionState.Transaction, session.State);
          StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
          Assert.IsTrue(session.IsSecureConnection);
        }
      }
    }

    [Test]
    public void TestCreateSession_ConnectToSecurePort_Exception()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "pops", "user", server.HostPort);

        try {
          using (var session = PopSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
            throw new System.Security.Authentication.AuthenticationException();
          }));
          Assert.Fail("PopUpgradeConnectionException not thrown");
        }
        catch (PopUpgradeConnectionException ex) {
          Assert.IsNotNull(ex.InnerException);
          Assert.IsInstanceOf<System.Security.Authentication.AuthenticationException>(ex.InnerException);
        }
      }
    }

    [Test]
    public void TestCreateSession_Stls_UseTlsIfAvailable()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.AllowInsecureLogin = false;
        prof.UseTlsIfAvailable = true;

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("STLS\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");
        server.EnqueueResponse("USER\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("PASS\r\n",
                               "+OK\r\n");

        var streamUpgraded = false;

        using (var session = PopSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SSL stream
        })) {
          Assert.IsTrue(streamUpgraded, "stream upgraded");

          Assert.AreEqual(PopSessionState.Transaction, session.State);
          StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
          Assert.IsTrue(session.IsSecureConnection);
        }
      }
    }

    [Test]
    public void TestCreateSession_Stls_DontUseTlsWhetherAvailable()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.AllowInsecureLogin = true;
        prof.UseTlsIfAvailable = false;

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("USER\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("PASS\r\n",
                               "+OK\r\n");

        var streamUpgraded = false;

        using (var session = PopSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SSL stream
        })) {
          Assert.IsFalse(streamUpgraded, "stream upgraded");

          Assert.AreEqual(PopSessionState.Transaction, session.State);
          StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
          Assert.IsFalse(session.IsSecureConnection);
        }
      }
    }

    [Test]
    public void TestCreateSession_Stls_Failure()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("STLS\r\n",
                               "-ERR\r\n");

        var streamUpgraded = false;

        try {
          using (var session = PopSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
            streamUpgraded = true;
            return baseStream; // TODO: return SSL stream
          }));
          Assert.Fail("PopUpgradeConnectionException not thrown");
        }
        catch (PopUpgradeConnectionException ex) {
          Assert.IsNull(ex.InnerException);
        }

        Assert.IsFalse(streamUpgraded, "stream upgraded");
      }
    }

    [Test]
    public void TestCreateSession_Stls_Exception()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("STLS\r\n",
                               "+OK\r\n");

        try {
          using (var session = PopSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
            throw new System.Security.Authentication.AuthenticationException();
          }));
          Assert.Fail("PopUpgradeConnectionException not thrown");
        }
        catch (PopUpgradeConnectionException ex) {
          Assert.IsNotNull(ex.InnerException);
          Assert.IsInstanceOf<System.Security.Authentication.AuthenticationException>(ex.InnerException);
        }
      }
    }

    [Test]
    public void TestCreateSession_ProfileTimeoutMustBeApplied()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=plain", server.HostPort);

        prof.AllowInsecureLogin = true;
        prof.Timeout        = 100 * 1000;
        prof.SendTimeout    = 300 * 1000;
        prof.ReceiveTimeout = 500 * 1000;

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);
          StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
          Assert.IsFalse(session.IsSecureConnection);

          Assert.AreEqual(100 * 1000, session.TransactionTimeout);
          Assert.AreEqual(300 * 1000, session.SendTimeout);
          Assert.AreEqual(500 * 1000, session.ReceiveTimeout);
        }
      }
    }

#if NET_4_0
    [Test]
    public void TestCreateSession_CancellationRequested_BeforeCreateSession()
    {
      var prof = new Profile(new NetworkCredential("user", "pass"), "pop", "user", "pop.invalid");

      prof.AllowInsecureLogin = true;

      var cts = new CancellationTokenSource();

      cts.Cancel();

      var t = Task.Factory.StartNew(() => PopSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

      Assert.Throws<AggregateException>(() => t.Wait());

      Assert.IsTrue(t.IsCanceled, "IsCancelled");
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeQueryingServerCapability()
    {
      // connect
      //   <--cancellation
      // CAPA

      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.AllowInsecureLogin = false;
        prof.UseTlsIfAvailable = true;

        var cts = new CancellationTokenSource();

        server.SetGreetingResponse("+OK\r\n",
                                   () => cts.Cancel(),
                                   null);

        var t = Task.Factory.StartNew(() => PopSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        Assert.IsNull(server.DequeueRequest()); // CAPA not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeUpgradeStream()
    {
      // connect
      // CAPA
      //   <--cancellation
      // STLS

      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.AllowInsecureLogin = false;
        prof.UseTlsIfAvailable = true;

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => PopSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n",
                               () => cts.Cancel(),
                               null);
        server.EnqueueResponse("STLS\r\n",
                               "+OK\r\n");

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains("CAPA", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // STLS not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeSendSaslClientResponse()
    {
      // connect
      // CAPA
      // AUTH(SaslMechanism)
      // <--cancellation
      // SASL server challenge

      using (var server = PopPseudoServer.StartNew()) {
        var credential = new NetworkCredential("pop-user", (string)null);
        var prof = new Profile(null, "pop", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5"};
        prof.AllowInsecureLogin = true;

        var cts = new CancellationTokenSource();

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH X-PSEUDO-MECHANISM\r\n",
                               "+ \r\n",
                               () => cts.Cancel(),
                               null);
        server.EnqueueResponse("-ERR\r\n");

        using (var authMechanism = new SaslPseudoMechanism(false, 1)) {
          authMechanism.Credential = credential;

          var t = Task.Factory.StartNew(() => PopSessionCreator.CreateSession(prof, authMechanism, null, cts.Token), cts.Token);

          var ex = Assert.Throws<AggregateException>(() => t.Wait());

          Assert.IsInstanceOf<PopAuthenticationException>(ex.Flatten().InnerException);

          // Assert.IsTrue(t.IsCanceled, "IsCancelled");
        }

        StringAssert.Contains("CAPA", server.DequeueRequest());
        StringAssert.Contains("AUTH X-PSEUDO-MECHANISM", server.DequeueRequest());
        Assert.AreEqual("*\r\n", server.DequeueRequest()); // client response not sent
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_WithSpecificMechanism()
    {
      // connect
      // CAPA
      //   <--cancellation
      // AUTH(SaslMechanism)

      using (var server = PopPseudoServer.StartNew()) {
        var credential = new NetworkCredential("pop-user", (string)null);
        var prof = new Profile(null, "pop", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5"};
        prof.AllowInsecureLogin = true;

        var cts = new CancellationTokenSource();

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n",
                               () => cts.Cancel(),
                               null);

        using (var authMechanism = new SaslPseudoMechanism(false, 1)) {
          authMechanism.Credential = credential;

          var t = Task.Factory.StartNew(() => PopSessionCreator.CreateSession(prof, authMechanism, null, cts.Token), cts.Token);

          Assert.Throws<AggregateException>(() => t.Wait());

          Assert.IsTrue(t.IsCanceled, "IsCancelled");
        }

        StringAssert.Contains("CAPA", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTH not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_SelectAppropriateMechanism()
    {
      // connect
      // CAPA
      //   <--cancellation
      // AUTH(AUTH=*)

      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = true;

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => PopSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n",
                               () => cts.Cancel(),
                               null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains("CAPA", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTH not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_SelectAppropriateMechanism_FallbackToNext()
    {
      // connect
      // CAPA
      // AUTH(AUTH=*)
      //   <--cancellation
      // AUTH(AUTH=*)

      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = true;

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => PopSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH LOGIN ",
                               "-ERR\r\n",
                               () => cts.Cancel(),
                               null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains("CAPA", server.DequeueRequest());
        StringAssert.Contains("AUTH LOGIN ", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTH not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_SelectAppropriateMechanism_FallbackToApop()
    {
      // connect
      // CAPA
      // AUTH LOGIN
      //   <--cancellation
      // APOP

      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN"};
        prof.AllowInsecureLogin = true;
        prof.AllowApopLogin = true;

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => PopSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH LOGIN ",
                               "-ERR\r\n",
                               () => cts.Cancel(),
                               null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains("CAPA", server.DequeueRequest());
        StringAssert.Contains("AUTH LOGIN ", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // APOP not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_SelectAppropriateMechanism_FallbackToUserPass()
    {
      // connect
      // CAPA
      // AUTH LOGIN
      // APOP
      //   <--cancellation
      // USER/PASS

      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN"};
        prof.AllowInsecureLogin = true;
        prof.AllowApopLogin = true;
        prof.AllowStandardLogin = true;

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => PopSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH LOGIN ",
                               "-ERR\r\n");
        server.EnqueueResponse("APOP user ",
                               "-ERR\r\n",
                               () => cts.Cancel(),
                               null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains("CAPA", server.DequeueRequest());
        StringAssert.Contains("AUTH LOGIN ", server.DequeueRequest());
        StringAssert.Contains("APOP user ", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // USER not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_MechanismSpecified()
    {
      // connect
      // CAPA
      //   <--cancellation
      // AUTH(AUTH=XXX)

      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=LOGIN", server.HostPort);

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => PopSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n",
                               () => cts.Cancel(),
                               null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains("CAPA", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTH not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_Anonymous()
    {
      // connect
      // CAPA
      //   <--cancellation
      // AUTH(AUTH=ANONYMOUS)

      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(null, "user%40pop.example.net;auth=anonymous", server.HostPort);

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => PopSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL ANONYMOUS\r\n" +
                               ".\r\n",
                               () => cts.Cancel(),
                               null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains("CAPA", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTH not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_Anonymous_FallbackToUserPass()
    {
      // connect
      // CAPA
      // AUTH(AUTH=ANONYMOUS)
      //   <--cancellation
      // USER/PASS

      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(null, null, server.HostPort);

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => PopSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH ANONYMOUS ",
                               "-ERR\r\n",
                               () => cts.Cancel(),
                               null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains("CAPA", server.DequeueRequest());
        StringAssert.Contains("AUTH ANONYMOUS ", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // USER not requested
      }
    }
#endif

    [Test]
    public void TestAuthenticate_UserAndAuthMechanismSpecified()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=plain", server.HostPort);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);
          StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAuthMechanismSpecified_AuthenticationFailure()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=plain", server.HostPort);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "-ERR\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopAuthenticationException not thrown");
        }
        catch (PopAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.PopCommandResultCode.Error, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_NoUserAndAuthMechanismSpecified()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), ";auth=plain", server.HostPort);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);

          var b = new PopUriBuilder(prof.Authority);

          b.UserName = "user";

          StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
        }
      }
    }

    [Test]
    public void TestAuthenticate_NoUserAndAuthMechanismSpecified_AuthenticationFailure()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), ";auth=plain", server.HostPort);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "-ERR\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopAuthenticationException not thrown");
        }
        catch (PopAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.PopCommandResultCode.Error, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndApopMechanismSpecified()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=+apop", server.HostPort);

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        server.EnqueueResponse("APOP user ",
                               "+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);
          StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndApopMechanismSpecified_AuthenticationFailure()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=+apop", server.HostPort);

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        server.EnqueueResponse("APOP user ",
                               "-ERR\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopAuthenticationException not thrown");
        }
        catch (PopAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.PopCommandResultCode.Error, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndApopMechanismSpecified_ApopIncapable()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=+apop", server.HostPort);

        server.SetGreetingResponse("+OK\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopIncapableException not thrown");
        }
        catch (PopIncapableException ex) {
          Assert.IsNull(ex.InnerException);
        }
      }
    }

    [Test]
    public void TestAuthenticate_NoUserAndApopMechanismSpecified()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), ";auth=+apop", server.HostPort);

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        server.EnqueueResponse("APOP user ",
                               "+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);

          var b = new PopUriBuilder(prof.Authority);

          b.UserName = "user";

          StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
        }
      }
    }

    [Test]
    public void TestAuthenticate_NoUserAndApopMechanismSpecified_AuthenticationFailure()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), ";auth=+apop", server.HostPort);

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        server.EnqueueResponse("APOP user ",
                               "-ERR\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopAuthenticationException not thrown");
        }
        catch (PopAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.PopCommandResultCode.Error, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_NoUserAndApopMechanismSpecified_ApopIncapable()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), ";auth=+apop", server.HostPort);

        server.SetGreetingResponse("+OK\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopIncapableException not thrown");
        }
        catch (PopIncapableException ex) {
          Assert.IsNull(ex.InnerException);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified()
    {
      AuthenticateWithAppropriateMechanism("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified()
    {
      AuthenticateWithAppropriateMechanism("user");
    }

    private void AuthenticateWithAppropriateMechanism(string userinfo)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"ANONYMOUS", "LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = true;

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH LOGIN ",
                               "+ \r\n");
        server.EnqueueResponse("+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);

          var b = new PopUriBuilder(prof.Authority);

          b.AuthType = PopAuthenticationMechanism.Login;

          StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_ExceptionWhileAuthentication()
    {
      AuthenticateExceptionWhileAuthentication("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_ExceptionWhileAuthentication()
    {
      AuthenticateExceptionWhileAuthentication("user");
    }

    private void AuthenticateExceptionWhileAuthentication(string userinfo)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"DIGEST-MD5", "PLAIN"};

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH DIGEST-MD5\r\n",
                               "+ xxx-invalid-response-xxx\r\n");

        try {
          using (var session = PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopException not thrown");
        }
        catch (PopException ex) {
          Assert.IsNotNull(ex.InnerException);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToNextAppropriateMechanism()
    {
      AuthenticateWithAppropriateMechanism_FallbackToNextAppropriateMechanism("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToNextAppropriateMechanism()
    {
      AuthenticateWithAppropriateMechanism_FallbackToNextAppropriateMechanism("user");
    }

    private void AuthenticateWithAppropriateMechanism_FallbackToNextAppropriateMechanism(string userinfo)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"ANONYMOUS", "LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = true;

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH LOGIN ",
                               "+ \r\n");
        server.EnqueueResponse("-ERR\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);

          var b = new PopUriBuilder(prof.Authority);

          b.AuthType = PopAuthenticationMechanism.Plain;

          StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToApop()
    {
      AuthenticateWithAppropriateMechanism_FallbackToApop("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToApop()
    {
      AuthenticateWithAppropriateMechanism_FallbackToApop("user");
    }

    private void AuthenticateWithAppropriateMechanism_FallbackToApop(string userinfo)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"ANONYMOUS", "PLAIN", "LOGIN"};
        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "-ERR\r\n");
        server.EnqueueResponse("AUTH LOGIN ",
                               "+ \r\n");
        server.EnqueueResponse("-ERR\r\n");
        server.EnqueueResponse("APOP user ",
                               "+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);

          var b = new PopUriBuilder(prof.Authority);

          b.AuthType = PopAuthenticationMechanism.Apop;

          StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToApop_AuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanism_FallbackToApop_AuthenticationFailure("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToApop_AuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanism_FallbackToApop_AuthenticationFailure("user");
    }

    private void AuthenticateWithAppropriateMechanism_FallbackToApop_AuthenticationFailure(string userinfo)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"PLAIN"};
        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "-ERR\r\n");
        server.EnqueueResponse("APOP user ",
                               "-ERR\r\n");
        server.EnqueueResponse("USER user\r\n",
                               "-ERR\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopAuthenticationException not thrown");
        }
        catch (PopAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.PopCommandResultCode.Error, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToApop_ApopDisallowed()
    {
      AuthenticateWithAppropriateMechanism_FallbackToApop_ApopDisallowed("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToApop_ApopDisallowed()
    {
      AuthenticateWithAppropriateMechanism_FallbackToApop_ApopDisallowed("user");
    }

    private void AuthenticateWithAppropriateMechanism_FallbackToApop_ApopDisallowed(string userinfo)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"PLAIN"};
        prof.AllowInsecureLogin = true;
        prof.AllowApopLogin = false;

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "-ERR\r\n");
        server.EnqueueResponse("USER user\r\n",
                               "-ERR\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopAuthenticationException not thrown");
        }
        catch (PopAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.PopCommandResultCode.Error, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToUserPass()
    {
      AuthenticateWithAppropriateMechanism_FallbackToUserPass("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToUserPass()
    {
      AuthenticateWithAppropriateMechanism_FallbackToUserPass("user");
    }

    private void AuthenticateWithAppropriateMechanism_FallbackToUserPass(string userinfo)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("+OK\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");
        server.EnqueueResponse("USER user\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("PASS pass\r\n",
                               "+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);

          var b = new PopUriBuilder(prof.Authority);

          b.AuthType = null;

          StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToUserPass_AuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanism_FallbackToUserPass_AuthenticationFailure("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToUserPass_AuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanism_FallbackToUserPass_AuthenticationFailure("user");
    }

    private void AuthenticateWithAppropriateMechanism_FallbackToUserPass_AuthenticationFailure(string userinfo)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("+OK\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");
        server.EnqueueResponse("USER user\r\n",
                               "-ERR\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopAuthenticationException not thrown");
        }
        catch (PopAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.PopCommandResultCode.Error, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToUserPass_UserPassDisallowed()
    {
      AuthenticateWithAppropriateMechanism_FallbackToUserPass_UserPassDisallowed("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToUserPass_UserPassDisallowed()
    {
      AuthenticateWithAppropriateMechanism_FallbackToUserPass_UserPassDisallowed("user");
    }

    private void AuthenticateWithAppropriateMechanism_FallbackToUserPass_UserPassDisallowed(string userinfo)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"PLAIN"};
        prof.AllowInsecureLogin = true;
        prof.AllowStandardLogin = false;

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "-ERR\r\n");
        server.EnqueueResponse("APOP user ",
                               "-ERR\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopAuthenticationException not thrown");
        }
        catch (PopAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.PopCommandResultCode.Error, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_AppropriateMechanismNotFound()
    {
      AuthenticateWithAppropriateMechanism_AppropriateMechanismNotFound("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_AppropriateMechanismNotFound()
    {
      AuthenticateWithAppropriateMechanism_AppropriateMechanismNotFound("user");
    }

    private void AuthenticateWithAppropriateMechanism_AppropriateMechanismNotFound(string userinfo)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.AllowInsecureLogin = false; // LOGINDISABLE equiv

        server.SetGreetingResponse("+OK\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");

        Assert.Throws<PopNoAppropriateAuthMechanismException>(() => PopSessionCreator.CreateSession(prof, null, null));
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_ApopAndUserPassDisallowed()
    {
      AuthenticateWithAppropriateMechanism_ApopAndUserPassDisallowed("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_ApopAndUserPassDisallowed()
    {
      AuthenticateWithAppropriateMechanism_ApopAndUserPassDisallowed("user");
    }

    private void AuthenticateWithAppropriateMechanism_ApopAndUserPassDisallowed(string userinfo)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.AllowInsecureLogin = true;
        prof.AllowApopLogin = false;
        prof.AllowStandardLogin = false;

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");

        Assert.Throws<PopNoAppropriateAuthMechanismException>(() => PopSessionCreator.CreateSession(prof, null, null));
      }
    }

    [Test]
    public void TestAuthenticate_AnonymousMechanismSpecified()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(null, "user%40pop.example.net;auth=anonymous", server.HostPort);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH ANONYMOUS dXNlckBwb3AuZXhhbXBsZS5uZXQ=\r\n",
                               "+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);
          StringAssert.AreEqualIgnoringCase(prof.Authority.AbsoluteUri, session.Authority.AbsoluteUri);
        }
      }
    }

    [Test]
    public void TestAuthenticate_AnonymousMechanismSpecified_AuthenticationFailure()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(null, "user%40pop.example.net;auth=anonymous", server.HostPort);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH ANONYMOUS dXNlckBwb3AuZXhhbXBsZS5uZXQ=\r\n",
                               "-ERR\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopAuthenticationException not thrown");
        }
        catch (PopAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.PopCommandResultCode.Error, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_NoUserAndNoAuthMechanismSpecified()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(null, null, server.HostPort);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH ANONYMOUS ",
                               "-ERR\r\n");
        server.EnqueueResponse("USER anonymous\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("PASS anonymous@\r\n",
                               "+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);

          var b = new PopUriBuilder(prof.Authority);

          b.UserName = "anonymous";

          StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
        }
      }
    }

    [Test]
    public void TestAuthenticate_NoUserAndNoAuthMechanismSpecified_AuthenticationFailure()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(null, null, server.HostPort);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");
        server.EnqueueResponse("USER anonymous\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("PASS anonymous@\r\n",
                               "-ERR\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopAuthenticationException not thrown");
        }
        catch (PopAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.PopCommandResultCode.Error, ex.Result.Code);
        }
      }
    }

    [Test, Ignore("not work; AUTH ANONYMOUS not supported && LOGINDISABLE")]
    public void TestAuthenticate_NoUserAndNoAuthMechanismSpecified_AppropriateMechanismNotFound()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(null, null, server.HostPort);

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");

        Assert.Throws<PopNoAppropriateAuthMechanismException>(() => PopSessionCreator.CreateSession(prof, null, null));
      }
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN", "ANONYMOUS", "CRAM-MD5"};
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL X-UNKNOWN DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH CRAM-MD5\r\n",
                               "+ PDQwMDEzNDQxMTIxNDM1OTQuMTI3MjQ5OTU1MEBsb2NhbGhvc3Q+\r\n");
        server.EnqueueResponse("+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);

          var b = new PopUriBuilder(prof.Authority);

          b.AuthType = PopAuthenticationMechanism.CRAMMD5;

          StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
        }
      }
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_SecureSession_ApopAvailable()
    {
      Authenticate_InsecureLoginDisallowed_SecureSession(true);
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_SecureSession_ApopNotAvailable()
    {
      Authenticate_InsecureLoginDisallowed_SecureSession(false);
    }

    private void Authenticate_InsecureLoginDisallowed_SecureSession(bool apopAvailable)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"ANONYMOUS", "LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse(apopAvailable ? "+OK <timestamp@localhost>\r\n" : "+OK\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("STLS\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH LOGIN ",
                               "-ERR\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "-ERR\r\n");

        if (apopAvailable) {
          server.EnqueueResponse("APOP user ",
                                 "+OK\r\n");
        }
        else {
          server.EnqueueResponse("USER user\r\n",
                                 "+OK\r\n");
          server.EnqueueResponse("PASS pass\r\n",
                                 "+OK\r\n");
        }

        using (var session = PopSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
          return baseStream; // TODO: return SSL stream
        })) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);

          var b = new PopUriBuilder(prof.Authority);

          if (apopAvailable)
            b.AuthType = PopAuthenticationMechanism.Apop;
          else
            b.AuthType = null;

          StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);

          Assert.IsTrue(session.IsSecureConnection);
        }
      }
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_SecureSession_ApopDisallowed()
    {
      Authenticate_InsecureLoginDisallowed_SecureSession_ApopOrUserPassDisallowed(false, true);
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_SecureSession_UserPassDisallowed()
    {
      Authenticate_InsecureLoginDisallowed_SecureSession_ApopOrUserPassDisallowed(true, false);
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_SecureSession_ApopAndUserPassDisallowed()
    {
      Authenticate_InsecureLoginDisallowed_SecureSession_ApopOrUserPassDisallowed(false, false);
    }

    private void Authenticate_InsecureLoginDisallowed_SecureSession_ApopOrUserPassDisallowed(bool allowApopLogin, bool allowStandardLogin)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"ANONYMOUS", "LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = false;
        prof.AllowStandardLogin = allowStandardLogin;
        prof.AllowApopLogin = allowApopLogin;

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("STLS\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH LOGIN ",
                               "-ERR\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "-ERR\r\n");

        if (allowApopLogin) {
          server.EnqueueResponse("+OK\r\n");
        }

        if (allowStandardLogin) {
          server.EnqueueResponse("USER user\r\n",
                                 "+OK\r\n");
          server.EnqueueResponse("PASS pass\r\n",
                                 "+OK\r\n");
        }

        try {
          using (var session = PopSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
            return baseStream; // TODO: return SSL stream
          })) {
            Assert.AreEqual(PopSessionState.Transaction, session.State);

            var b = new PopUriBuilder(prof.Authority);

            if (allowApopLogin)
              b.AuthType = PopAuthenticationMechanism.Apop;
            else
              b.AuthType = null;

            StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);

            Assert.IsTrue(session.IsSecureConnection);
          }
        }
        catch (Exception ex) {
          if (allowApopLogin || allowStandardLogin)
            Assert.Fail("unexpected exception: {0}", ex);
          else
            Assert.IsInstanceOf<PopAuthenticationException>(ex);
        }
      }
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_FallbackToNextAppropriateMechanism()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5"};
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL X-UNKNOWN DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH DIGEST-MD5\r\n",
                               "+ cmVhbG09ImVsd29vZC5pbm5vc29mdC5jb20iLG5vbmNlPSJPQTZNRzl0" +
                               "RVFHbTJoaCIscW9wPSJhdXRoIixhbGdvcml0aG09bWQ1LXNlc3MsY2hh" +
                               "cnNldD11dGYtOA==\r\n");
        server.EnqueueResponse("-ERR\r\n");
        server.EnqueueResponse("AUTH CRAM-MD5\r\n",
                               "+ PDQwMDEzNDQxMTIxNDM1OTQuMTI3MjQ5OTU1MEBsb2NhbGhvc3Q+\r\n");
        server.EnqueueResponse("+OK\r\n");

        using (var session = PopSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(PopSessionState.Transaction, session.State);

          var b = new PopUriBuilder(prof.Authority);

          b.AuthType = PopAuthenticationMechanism.CRAMMD5;

          StringAssert.AreEqualIgnoringCase(b.Uri.AbsoluteUri, session.Authority.AbsoluteUri);
        }
      }
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_FallbackToNextAppropriateMechanism_AuthenticationFailure()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5"};
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL X-UNKNOWN DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH DIGEST-MD5\r\n",
                               "+ cmVhbG09ImVsd29vZC5pbm5vc29mdC5jb20iLG5vbmNlPSJPQTZNRzl0" +
                               "RVFHbTJoaCIscW9wPSJhdXRoIixhbGdvcml0aG09bWQ1LXNlc3MsY2hh" +
                               "cnNldD11dGYtOA==\r\n");
        server.EnqueueResponse("-ERR\r\n");
        server.EnqueueResponse("AUTH CRAM-MD5\r\n",
                               "+ PDQwMDEzNDQxMTIxNDM1OTQuMTI3MjQ5OTU1MEBsb2NhbGhvc3Q+\r\n");
        server.EnqueueResponse("-ERR\r\n");

        try {
          Assert.IsNull(PopSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("PopAuthenticationException not thrown");
        }
        catch (PopAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.PopCommandResultCode.Error, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_AppropriateMechanismNotFound()
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"ANONYMOUS", "LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("+OK <timestamp@localhost>\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL X-UNKNOWN DIGEST-MD5 CRAM-MD5 NTLM PLAIN LOGIN ANONYMOUS\r\n" +
                               ".\r\n");

        Assert.Throws<PopNoAppropriateAuthMechanismException>(() => PopSessionCreator.CreateSession(prof, null, null));
      }
    }

    [Test]
    public void TestAuthenticate_SpecificMechanism()
    {
      AuthenticateSpecificMechanism(new NetworkCredential("sasl-user", (string)null));
    }

    [Test]
    public void TestAuthenticate_SpecificMechanism_CredentialNotSet()
    {
      AuthenticateSpecificMechanism(null);
    }

    private void AuthenticateSpecificMechanism(NetworkCredential credential)
    {
      using (var server = PopPseudoServer.StartNew()) {
        var prof = new Profile(new NetworkCredential("pop-user", (string)null), "pop", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5"};
        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("+OK ready\r\n");
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH X-PSEUDO-MECHANISM\r\n",
                               "+\r\n");
        server.EnqueueResponse("+OK\r\n");

        using (var authMechanism = new SaslPseudoMechanism(false, 1)) {
          authMechanism.Credential = credential;

          using (var session = PopSessionCreator.CreateSession(prof, authMechanism, null)) {
            Assert.AreEqual(PopSessionState.Transaction, session.State);

            if (credential == null)
              Assert.AreEqual(string.Format("pop://;AUTH=X-PSEUDO-MECHANISM@{0}/", server.HostPort),
                              session.Authority.AbsoluteUri);
            else
              Assert.AreEqual(string.Format("pop://sasl-user;AUTH=X-PSEUDO-MECHANISM@{0}/", server.HostPort),
                              session.Authority.AbsoluteUri);
          }

          Assert.AreSame(credential,
                         authMechanism.Credential,
                         "credential must be kept");

          Assert.AreEqual(Smdn.Security.Authentication.Sasl.SaslExchangeStatus.Succeeded,
                          authMechanism.ExchangeStatus);
        }
      }
    }
  }
}
