#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Diagnostics;
using System.Linq;
using System.Net;
using System.Threading;
using NUnit.Framework;

using Smdn.Net.Pop3.Client.Session;
using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.Client;
using Smdn.Security.Authentication.Sasl;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Pop3.Client {
  [TestFixture]
  public class PopClientTests {
    [Test]
    public void TestConstruct()
    {
      Assert.AreEqual("pop://localhost/",
                      DefaultPropertyAssertion(new PopClient()).Profile.Authority.AbsoluteUri,
                      "#1 authority");

      Assert.AreEqual("pop://user;AUTH=+APOP@pop.example.net:10110/",
                      DefaultPropertyAssertion(new PopClient(new Uri("pop://user;AUTH=+APOP@pop.example.net:10110/"))).Profile.Authority.AbsoluteUri,
                      "#2 authority");

      Assert.AreEqual("pop://pop.example.net/",
                      DefaultPropertyAssertion(new PopClient("pop.example.net")).Profile.Authority.AbsoluteUri,
                      "#3 authority");

      Assert.AreEqual("pop://user@pop.example.net/",
                      DefaultPropertyAssertion(new PopClient("pop.example.net", "user")).Profile.Authority.AbsoluteUri,
                      "#4 authority");

      Assert.AreEqual("pops://user@pop.example.net/",
                      DefaultPropertyAssertion(new PopClient("pop.example.net", true, "user")).Profile.Authority.AbsoluteUri,
                      "#5 authority");

      Assert.AreEqual("pop://pop.example.net:10110/",
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110)).Profile.Authority.AbsoluteUri,
                      "#6 authority");

      Assert.AreEqual("pop://user@pop.example.net:10110/", 
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110, "user")).Profile.Authority.AbsoluteUri,
                      "#7 authority");

      Assert.AreEqual("pop://user@pop.example.net:10110/",
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110, "user")).Profile.Authority.AbsoluteUri,
                      "#8 authority");

      Assert.AreEqual("pops://user@pop.example.net:10110/",
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110, true, "user")).Profile.Authority.AbsoluteUri,
                      "#9 authority");

      Assert.AreEqual("pop://user;AUTH=+APOP@pop.example.net:10110/",
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110, "user", "+APOP")).Profile.Authority.AbsoluteUri,
                      "#10 authority");

      Assert.AreEqual("pops://user;AUTH=DIGEST-MD5@pop.example.net:10110/",
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110, true, "user", "DIGEST-MD5")).Profile.Authority.AbsoluteUri,
                      "#11 authority");

      Assert.AreEqual("pops://user@pop.example.net:10110/",
                      DefaultPropertyAssertion(new PopClient("pop.example.net", 10110, true, "user", null, 1000), 1000).Profile.Authority.AbsoluteUri,
                      "#12 authority");

      var profile = new PopClientProfile(new Uri("pops://user@pop.example.net/"));

      Assert.AreEqual("pops://user@pop.example.net/",
                      DefaultPropertyAssertion(new PopClient(profile)).Profile.Authority.AbsoluteUri,
                      "#13 authority");
    }

    private PopClient DefaultPropertyAssertion(PopClient client)
    {
      return DefaultPropertyAssertion(client, System.Threading.Timeout.Infinite);
    }

    private PopClient DefaultPropertyAssertion(PopClient client, int timeoutMilliseconds)
    {
      Assert.IsNotNull(client.Profile);
      Assert.IsTrue(client.Profile.UseTlsIfAvailable);
      Assert.IsFalse(client.Profile.AllowInsecureLogin);
      Assert.IsFalse(client.DeleteAfterRetrieve);
      Assert.AreEqual(timeoutMilliseconds, client.Timeout);
      Assert.AreEqual(System.Threading.Timeout.Infinite, client.ReceiveTimeout);
      Assert.AreEqual(System.Threading.Timeout.Infinite, client.SendTimeout);
      Assert.IsFalse(client.IsBusy);

      Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsSecureSession));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerCapabilities));
      Assert.Throws<InvalidOperationException>(() => Assert.AreNotEqual(0L, client.TotalSize));
      Assert.Throws<InvalidOperationException>(() => Assert.AreNotEqual(0L, client.MessageCount));

      return client;
    }

    [Test]
    public void TestPropertyGetIsSecureSessionNotConnected()
    {
      using (var client = new PopClient()) {
        Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsSecureSession));
      }
    }

    [Test]
    public void TestPropertyGetMessageCountNotConnected()
    {
      using (var client = new PopClient()) {
        Assert.Throws<InvalidOperationException>(() => Assert.AreNotEqual(0L, client.MessageCount));
      }
    }

    [Test]
    public void TestPropertyGetTotalSizeNotConnected()
    {
      using (var client = new PopClient()) {
        Assert.Throws<InvalidOperationException>(() => Assert.AreNotEqual(0L, client.TotalSize));
      }
    }

    [Test]
    public void TestPropertyGetServerCapabilitiesNotConnected()
    {
      using (var client = new PopClient()) {
        Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerCapabilities));
      }
    }

    [Test]
    public void TestConnectWithPassword()
    {
      Connect(delegate(PopPseudoServer server, PopClient client) {
        client.Profile.AllowInsecureLogin = true;

        client.Connect("pass");

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNull((client.Profile as IPopSessionProfile).Credentials);
      });
    }

    [Test]
    public void TestConnectWithCredentials()
    {
      Connect(delegate(PopPseudoServer server, PopClient client) {
        client.Profile.AllowInsecureLogin = true;

        client.Connect(new NetworkCredential("user", "pass"));

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNull((client.Profile as IPopSessionProfile).Credentials);
      });
    }

    [Test]
    public void TestConnectWithSpecifiedSaslMechanism()
    {
      using (var server = PopPseudoServer.StartNew()) {
        using (var client = new PopClient(server.Host, server.Port)) {
          server.EnqueueResponse("CAPA\r\n",
                                 "+OK\r\n" +
                                 ".\r\n");
          server.EnqueueResponse("AUTH X-PSEUDO-MECHANISM\r\n",
                                 "+ \r\n");
          server.EnqueueResponse("+OK\r\n");

          using (var authMechanism = new SaslPseudoMechanism(false, 1)) {
            client.Connect(authMechanism);

            Assert.AreEqual(Smdn.Security.Authentication.Sasl.SaslExchangeStatus.Succeeded,
                            authMechanism.ExchangeStatus);
          }

          Assert.IsTrue(client.IsConnected);
          Assert.IsFalse(client.IsSecureSession);
          Assert.IsNotNull(client.ServerCapabilities);
          Assert.IsNull((client.Profile as IPopSessionProfile).Credentials);
        }
      }
    }

    private class SaslXOAuth2 : XOAuth2Mechanism {
      public bool InitializeCalled = false;

      public SaslXOAuth2(string user, string accessToken)
        : base(user, accessToken)
      {
      }

      public override void Initialize()
      {
        base.Initialize();

        InitializeCalled = true;
      }
    }

    [Test]
    public void TestConnectWithXOAuth2()
    {
      using (var server = PopPseudoServer.StartNew()) {
        using (var client = new PopClient(server.Host, server.Port)) {
          server.EnqueueResponse("CAPA\r\n",
                                 "+OK\r\n" + // AUTH=XOAUTH2 not listed
                                 ".\r\n");
          server.EnqueueResponse("AUTH XOAUTH2",
                                 "+OK\r\n");

          using (var oauth2 = new SaslXOAuth2("someuser@example.com", "vF9dft4qmTc2Nvb3RlckBhdHRhdmlzdGEuY29tCg==")) {
            client.Connect(oauth2);

            Assert.IsTrue(oauth2.InitializeCalled);
            Assert.AreEqual(SaslExchangeStatus.Continuing, oauth2.ExchangeStatus);
            Assert.IsNull(oauth2.LastErrorResults);
          }

          Assert.IsTrue(client.IsConnected);
          Assert.IsFalse(client.IsSecureSession);
          Assert.IsNotNull(client.ServerCapabilities);
        }
      }
    }

    [Test]
    public void TestConnectWithXOAuth2Error()
    {
      using (var server = PopPseudoServer.StartNew()) {
        using (var client = new PopClient(server.Host, server.Port)) {
          server.EnqueueResponse("CAPA\r\n",
                                 "+OK\r\n" + // AUTH=XOAUTH2 not listed
                                 ".\r\n");
          server.EnqueueResponse("AUTH XOAUTH2",
                                 "+ eyJzdGF0dXMiOiI0MDEiLCJzY2hlbWVzIjoiYmVhcmVyIG1hYyIsInNjb\n3BlIjoiaHR0cHM6Ly9tYWlsLmdvb2dsZS5jb20vIn0K\r\n");
          server.EnqueueResponse("-ERR\r\n");

          using (var oauth2 = new SaslXOAuth2("someuser@example.com", "vF9dft4qmTc2Nvb3RlckBhdHRhdmlzdGEuY29tCg==")) {
            Assert.Throws<PopAuthenticationException>(() => client.Connect(oauth2));

            Assert.IsTrue(oauth2.InitializeCalled);
            Assert.AreEqual(SaslExchangeStatus.Failed, oauth2.ExchangeStatus);
            Assert.AreEqual("{\"status\":\"401\",\"schemes\":\"bearer mac\",\"scope\":\"https://mail.google.com/\"}\n", 
                            oauth2.LastErrorResults);
          }

          Assert.IsFalse(client.IsConnected);
        }
      }
    }

    [Test]
    public void TestConnectAlreadyConnected()
    {
      Connect(delegate(PopPseudoServer server, PopClient client) {
        client.Profile.AllowInsecureLogin = true;

        client.Connect(new NetworkCredential("user", "pass"));

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);

        try {
          client.Connect("pass");
          Assert.Fail("InvalidOperationException not thrown");
        }
        catch (InvalidOperationException) {
          Assert.IsNull((client.Profile as IPopSessionProfile).Credentials);
        }
      });
    }

    private void Connect(Action<PopPseudoServer, PopClient> action)
    {
      using (var server = PopPseudoServer.StartNew()) {
        using (var client = new PopClient(server.Host, server.Port, "user")) {
          Assert.IsFalse(client.IsConnected);
          Assert.IsFalse(client.IsBusy);
          Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsSecureSession));
          Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerCapabilities));
          Assert.Throws<InvalidOperationException>(() => Assert.AreNotEqual(0L, client.TotalSize));
          Assert.Throws<InvalidOperationException>(() => Assert.AreNotEqual(0L, client.MessageCount));

          server.EnqueueResponse("CAPA\r\n",
                                 "+OK\r\n" +
                                 ".\r\n");
          server.EnqueueResponse("USER user\r\n",
                                 "+OK\r\n");
          server.EnqueueResponse("PASS pass\r\n",
                                 "+OK\r\n");

          action(server, client);

          Assert.IsFalse(client.IsBusy);
        }
      }
    }

    [Test]
    public void TestBeginConnectWithPassword()
    {
      BeginConnect(true);
    }

    [Test]
    public void TestBeginConnectWithCredentials()
    {
      BeginConnect(false);
    }

    private void BeginConnect(bool usePassword)
    {
      using (var server = PopPseudoServer.StartNew()) {
        using (var client = new PopClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 5000;
          client.Profile.AllowInsecureLogin = true;

          Assert.IsFalse(client.IsBusy, "IsBusy before connect");

          var callbacked = false;
          var asyncStateObject = new object();

          var asyncCallback = (AsyncCallback)delegate(IAsyncResult ar) {
            Assert.IsNotNull(ar);
            Assert.AreSame(ar.AsyncState, asyncStateObject);
            callbacked = true;
          };

          var asyncResult = usePassword
                            ? (PopClient.ConnectAsyncResult)client.BeginConnect("pass", asyncCallback, asyncStateObject)
                            : (PopClient.ConnectAsyncResult)client.BeginConnect(new NetworkCredential("user", "pass"), asyncCallback, asyncStateObject);

          Assert.IsFalse(client.IsConnected);
          Assert.IsNotNull(asyncResult);
          Assert.IsFalse(asyncResult.IsCompleted);
          Assert.IsFalse(asyncResult.EndConnectCalled);
          Assert.IsFalse(callbacked);

          try {
            client.BeginConnect("pass");
            Assert.Fail("InvalidOperationException not thrown");
          }
          catch (InvalidOperationException) {
          }

          Assert.IsTrue(client.IsBusy, "IsBusy while connecting");

          var interval = TimeSpan.FromMilliseconds(250);

          Smdn.Net.TestUtils.Elapses(interval, delegate {
            ThreadPool.QueueUserWorkItem(delegate(object state) {
              var s = state as PopPseudoServer;

              Thread.Sleep(interval);

              s.EnqueueResponse("CAPA\r\n",
                                "+OK\r\n" +
                                ".\r\n");
              s.EnqueueResponse("USER user\r\n",
                                "+OK\r\n");
              s.EnqueueResponse("PASS pass\r\n",
                                "+OK\r\n");
            }, server);

            Assert.IsTrue(asyncResult.AsyncWaitHandle.WaitOne(3000, false));
          });

          Assert.IsFalse(client.IsConnected);

          Thread.Sleep(50); // wait for callback
          Assert.IsTrue(callbacked);
          Assert.IsFalse(asyncResult.EndConnectCalled);
          Assert.IsTrue(client.IsBusy, "IsBusy before EndConnect");

          client.EndConnect(asyncResult);

          Assert.IsFalse(client.IsBusy, "IsBusy after connected");

          Assert.IsTrue(asyncResult.EndConnectCalled);
          Assert.IsTrue(client.IsConnected);
          Assert.IsFalse(client.IsSecureSession);

          Assert.IsNull((client.Profile as IPopSessionProfile).Credentials);

          try {
            client.EndConnect(asyncResult);
            Assert.Fail("InvalidOperationException not thrown");
          }
          catch (ArgumentException) {
          }
        }
      }
    }

    [Test]
    public void TestBeginConnectAlreadyConnected()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);

        client.Profile.AllowInsecureLogin = true;

        try {
          client.BeginConnect("pass");
          Assert.Fail("InvalidOperationException not thrown");
        }
        catch (InvalidOperationException) {
        }

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
      });
    }

    [Test]
    public void TestBeginConnectException()
    {
      using (var server = PopPseudoServer.StartNotAccepting()) {
        server.ThrowIfServerNotStopped = false;

        using (var client = new PopClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 500;

          Assert.IsFalse(client.IsConnected);

          var asyncResult = client.BeginConnect("pass");

          /*          
           * no response
           */

          Assert.Throws<PopConnectionException>(() => client.EndConnect(asyncResult));

          Assert.IsNull((client.Profile as IPopSessionProfile).Credentials);

          Assert.IsFalse(client.IsConnected);
        }
      }
    }

#if ASYNC_OPS
    [Test]
    public void TestConnectAsyncWithPassword()
    {
      ConnectAsync(true);
    }

    [Test]
    public void TestConnectAsyncWithCredential()
    {
      ConnectAsync(false);
    }

    private void ConnectAsync(bool usePassword)
    {
      using (var server = PopPseudoServer.StartNew()) {
        using (var client = new PopClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 5000;
          client.Profile.AllowInsecureLogin = true;

          var t = usePassword
                  ? client.ConnectAsync("pass")
                  : client.ConnectAsync(new NetworkCredential("user", "pass"));

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            var s = state as PopPseudoServer;

            s.WaitForClientRequest();

            s.EnqueueResponse("CAPA\r\n",
                              "+OK\r\n" +
                              ".\r\n");
            s.EnqueueResponse("USER user\r\n",
                              "+OK\r\n");
            s.EnqueueResponse("PASS pass\r\n",
                              "+OK\r\n");
          }, server);

          Assert.IsTrue(t.Wait(10000), "Task.Wait");
          Assert.IsTrue(client.IsConnected);
        }
      }
    }

    [Test]
    public void TestConnectAsyncAlreadyConnected()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);

        client.Profile.AllowInsecureLogin = true;

        try {
          client.ConnectAsync("pass");
          Assert.Fail("InvalidOperationException not thrown");
        }
        catch (InvalidOperationException) {
        }

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
      });
    }

    [Test]
    public void TestConnectAsyncException()
    {
      using (var server = PopPseudoServer.StartNotAccepting()) {
        using (var client = new PopClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 250;

          Assert.IsFalse(client.IsConnected);

          var t = client.ConnectAsync("pass");

          try {
            t.Wait(10000);
            Assert.Fail("AggregateException not thrown");
          }
          catch (AggregateException ex) {
            Assert.IsInstanceOf(typeof(PopConnectionException), ex.InnerException);
            Assert.IsNull((client.Profile as IPopSessionProfile).Credentials);
          }

          Assert.IsFalse(client.IsConnected);
        }
      }
    }

    [Test]
    public void TestConnect_CancellationRequested()
    {
      using (var server = PopPseudoServer.StartNew()) {
        using (var client = new PopClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 1000;
          client.Profile.AllowInsecureLogin = true;

          Assert.IsFalse(client.IsBusy, "IsBusy before connect");

          var cts = new CancellationTokenSource();

          server.SetGreetingResponse("+OK\r\n",
                                     () => cts.Cancel(),
                                     null);

          Assert.Throws<OperationCanceledException>(() => client.Connect("pass", null, cts.Token));

          Assert.IsFalse(client.IsBusy, "IsBusy after cancel");
          Assert.IsFalse(client.IsConnected);
          Assert.IsNull((client.Profile as IPopSessionProfile).Credentials);
        }
      }
    }

    [Test]
    public void TestConnectAsync_CancellationRequested()
    {
      using (var server = PopPseudoServer.StartNew()) {
        using (var client = new PopClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 1000;
          client.Profile.AllowInsecureLogin = true;

          Assert.IsFalse(client.IsBusy, "IsBusy before connect");

          using (var e = new ManualResetEventSlim(false)) {
            var cts = new CancellationTokenSource();

            server.SetGreetingResponse("+OK\r\n",
                                       () => { e.Wait(); cts.Cancel(); },
                                       null);

            var t = client.ConnectAsync("pass", null, cts.Token);

            Assert.IsTrue(client.IsBusy, "IsBusy while connecting");

            e.Set();

            Assert.Throws<AggregateException>(() => t.Wait());

            Assert.IsTrue(t.IsCanceled, "IsCanceled");
          }

          Assert.IsFalse(client.IsBusy, "IsBusy after cancel");
          Assert.IsFalse(client.IsConnected);
          Assert.IsNull((client.Profile as IPopSessionProfile).Credentials);
        }
      }
    }
#endif

    [Test]
    public void TestLogoutAsyncConnectRunning()
    {
      DisconnectAsyncConnectRunning(true);
    }

    [Test]
    public void TestDisconnectAsyncConnectRunning()
    {
      DisconnectAsyncConnectRunning(false);
    }

    private void DisconnectAsyncConnectRunning(bool logout)
    {
      using (var server = PopPseudoServer.StartNew()) {
        server.ThrowIfServerNotStopped = false;

        using (var client = new PopClient(server.Host, server.Port, "user")) {
          client.BeginConnect("pass");

          Assert.IsTrue(client.IsBusy);

          try {
            if (logout)
              client.Logout();
            else
              client.Disconnect();

            Assert.Fail("InvalidOperationException not thrown");
          }
          catch (InvalidOperationException) {
          }
        }
      }
    }

    [Test]
    public void TestDisposeAsyncConnectRunning()
    {
      using (var server = PopPseudoServer.StartNew()) {
        using (var client = new PopClient(server.Host, server.Port, "user")) {
          client.Profile.AllowInsecureLogin = true;
          client.Profile.Timeout = 5000;

          Assert.IsFalse(client.IsConnected);

          var asyncResult = (PopClient.ConnectAsyncResult)client.BeginConnect("pass");

          server.EnqueueResponse("CAPA\r\n",
                                 "+OK\r\n" +
                                 ".\r\n");
          server.EnqueueResponse("USER user\r\n",
                                 "+OK\r\n");
          server.EnqueueResponse("PASS pass\r\n",
                                 "+OK\r\n");

          Assert.IsFalse(asyncResult.EndConnectCalled);

          ((IDisposable)client).Dispose();

          Assert.IsFalse(client.IsConnected);

          client.EndConnect(asyncResult);

          Assert.IsTrue(asyncResult.EndConnectCalled);

          Assert.IsFalse(client.IsConnected);
        }
      }
    }

    [Test]
    public void TestEndConnectInvalidAsyncResult()
    {
      using (PopPseudoServer
             server1 = PopPseudoServer.StartNew(),
             server2 = PopPseudoServer.StartNew()) {
        server1.ThrowIfServerNotStopped = false;
        server2.ThrowIfServerNotStopped = false;

        using (PopClient
               client1 = new PopClient(server1.Host, server1.Port, "user1"),
               client2 = new PopClient(server2.Host, server2.Port, "user2")) {
          client1.Profile.ReceiveTimeout = 100;
          client2.Profile.ReceiveTimeout = 100;

          var asyncResult1 = client1.BeginConnect("pass1");
          var asyncResult2 = client2.BeginConnect(new NetworkCredential("user2", "pass2"));

          Assert.Throws<ArgumentException>(() => client1.EndConnect(asyncResult2));
          Assert.Throws<ArgumentException>(() => client2.EndConnect(asyncResult1));

          try {
            client1.EndConnect(asyncResult1);
          }
          catch (TimeoutException) {
            // possible exception
          }

          try {
            client2.EndConnect(asyncResult2);
          }
          catch (TimeoutException) {
            // possible exception
          }
        }
      }
    }

    [Test]
    public void TestConnectBeginConnectRunning()
    {
      using (var server = PopPseudoServer.StartNew()) {
        server.ThrowIfServerNotStopped = false;

        using (var client = new PopClient(server.Host, server.Port, "user")) {
          client.Profile.AllowInsecureLogin = true;

          client.BeginConnect("pass");

          try {
            client.Connect("pass");
            Assert.Fail("InvalidOperationException not thrown");
          }
          catch (InvalidOperationException) {
          }

          try {
            client.Connect(new NetworkCredential("user", "pass"));
            Assert.Fail("InvalidOperationException not thrown");
          }
          catch (InvalidOperationException) {
          }
        }
      }
    }

    [Test]
    public void TestDisconnect()
    {
      DisposeOrDisconnect(true);
    }

    [Test]
    public void TestDispose()
    {
      DisposeOrDisconnect(false);
    }

    private void DisposeOrDisconnect(bool disconnect)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        if (disconnect)
          client.Disconnect();
        else
          ((IDisposable)client).Dispose();

        Assert.IsFalse(client.IsConnected);

        Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsSecureSession));
        Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerCapabilities));
        Assert.Throws<InvalidOperationException>(() => Assert.AreNotEqual(0L, client.TotalSize));
        Assert.Throws<InvalidOperationException>(() => Assert.AreNotEqual(0L, client.MessageCount));

        if (disconnect)
          Assert.DoesNotThrow(() => client.Disconnect(), "Disconnect again");
        else
          Assert.DoesNotThrow(() => ((IDisposable)client).Dispose(), "Dispose again");
      });
    }

    [Test]
    public void TestCheckConnected()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        client.Disconnect();

        Assert.IsFalse(client.IsConnected);

        Assert.Throws<InvalidOperationException>(delegate { client.KeepAlive(); });
      });
    }

    [Test]
    public void TestOperationAfterDisconnected()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        client.Disconnect();

        OperationAfterException(client);
      });
    }

    [Test]
    public void TestOperationAfterTimeout()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        client.Timeout = 100;

        try {
          client.CancelDelete();
        }
        catch (TimeoutException) {
        }

        OperationAfterException(client);
      });
    }

    [Test]
    public void TestOperationAfterInternalError()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("+OK\r\n" +
                               "\r\n" + // invalid response
                               ".\r\n");

        try {
          client.GetMessages().ToArray();
        }
        catch (PopException) {
        }

        OperationAfterException(client);
      });
    }

    private void OperationAfterException(PopClient client)
    {
      try {
        client.CancelDelete();
        Assert.Fail("InvalidOperationException not thrown");
      }
      catch (InvalidOperationException) {
      }
    }

    [Test]
    public void TestLogout()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n");

        client.Logout();

        Assert.IsFalse(client.IsConnected);

        Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsSecureSession));
        Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerCapabilities));
        Assert.Throws<InvalidOperationException>(() => Assert.AreNotEqual(0L, client.TotalSize));
        Assert.Throws<InvalidOperationException>(() => Assert.AreNotEqual(0L, client.MessageCount));

        Assert.DoesNotThrow(() => client.Logout(), "Logout again");
      });
    }

    [Test]
    public void TestReconnect()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n",
                               true);

        client.Logout();

        Assert.IsFalse(client.IsConnected);

        // reconnect
        server.Restart();

        client.Profile.Host = server.Host;
        client.Profile.Port = server.Port;

        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");
        server.EnqueueResponse("USER user\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("PASS pass\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("STAT\r\n",
                               "+OK 0 0\r\n");

        client.Connect("pass");

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
      });
    }

    [Test]
    public void TestSerializeBinaryErrorResponseException()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        // NOOP
        server.EnqueueResponse("-ERR server error\r\n");

        try {
          client.KeepAlive();
          Assert.Fail("PopErrorResponseException not thrown");
        }
        catch (PopErrorResponseException ex) {
          Smdn.Net.TestUtils.SerializeBinary(ex, delegate(PopErrorResponseException deserialized) {
            Assert.AreEqual(ex.Message, deserialized.Message);
            Assert.IsNotNull(deserialized.Result);
            Assert.IsNotNull(deserialized.Result.StatusResponse);
            Assert.AreEqual(PopStatusIndicator.Negative, deserialized.Result.StatusResponse.Status);
            Assert.AreEqual("server error", deserialized.Result.ResponseText);

            var responses = deserialized.Result.ReceivedResponses.ToArray();

            Assert.AreEqual(1, responses.Length);
            Assert.IsInstanceOf<PopStatusResponse>(responses[0]);
          });
        }
      });
    }
  }
}
