#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Linq;
using System.Threading;
using NUnit.Framework;

using Smdn.Net.Pop3.Client.Session;

namespace Smdn.Net.Pop3.Client {
  [TestFixture]
  public class PopClientOperationsTests {
    [Test]
    public void TestGetMessage()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");

        var message = client.GetMessage(1L);

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);

        Assert.IsNotNull(message);
        Assert.AreEqual(1L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);

        // get again
        Assert.AreSame(message, client.GetMessage(1L));
      });
    }

    [Test]
    public void TestGetFirstMessage()
    {
      GetFirstMessage(false);
    }

    [Test]
    public void TestGetFirstMessage_GetUniqueId()
    {
      GetFirstMessage(true);
    }

    private void GetFirstMessage(bool getUniqueId)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");

        if (getUniqueId)
          server.EnqueueResponse("UIDL 1\r\n",
                                 "+OK 1 whqtswO00WBw418f9t5JxYwZ\r\n");

        var message = getUniqueId
          ? client.GetFirstMessage(true)
          : client.GetFirstMessage();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(5L, client.MessageCount);
        Assert.AreEqual(1200L, client.TotalSize);

        Assert.IsNotNull(message);
        Assert.AreEqual(1L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);

        if (getUniqueId)
          Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);

        // get again
        Assert.AreSame(message, getUniqueId ? client.GetFirstMessage(true) : client.GetFirstMessage());
      });
    }

    [Test]
    public void TestGetLastMessage()
    {
      GetLastMessage(false);
    }

    [Test]
    public void TestGetLastMessage_GetUniqueId()
    {
      GetLastMessage(true);
    }

    private void GetLastMessage(bool getUniqueId)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("LIST 5\r\n",
                               "+OK 5 200\r\n");

        if (getUniqueId)
          server.EnqueueResponse("UIDL 5\r\n",
                                 "+OK 5 whqtswO00WBw418f9t5JxYwZ\r\n");

        var message = getUniqueId
          ? client.GetLastMessage(true)
          : client.GetLastMessage();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(5L, client.MessageCount);
        Assert.AreEqual(1200L, client.TotalSize);

        Assert.IsNotNull(message);
        Assert.AreEqual(5L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);

        if (getUniqueId)
          Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);

        // get again
        Assert.AreSame(message, getUniqueId ? client.GetLastMessage(true) : client.GetLastMessage());
      });
    }

    [Test]
    [TestCase(false)]
    [TestCase(true)]
    public void TestGetLastMessage_StatFailed(bool getUniqueId)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "-ERR\r\n");

        Assert.Throws<PopErrorResponseException>(() => client.GetLastMessage(getUniqueId));

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetFirstMessage_NoMessagesExist()
    {
      GetFirstOrLastMessage(delegate(PopClient client) {
        var ex = Assert.Throws<PopMessageNotFoundException>(() => client.GetFirstMessage());
        Assert.AreEqual(1L, ex.MessageNumber);
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetLastMessage_NoMessagesExist()
    {
      GetFirstOrLastMessage(delegate(PopClient client) {
        var ex = Assert.Throws<PopMessageNotFoundException>(() => client.GetLastMessage());
        Assert.AreEqual(1L, ex.MessageNumber);
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    private void GetFirstOrLastMessage(Action<PopClient> action)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 0 0\r\n");

        action(client);
      });
    }

    [Test]
    public void TestGetMessageGetUniqueId()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");
        server.EnqueueResponse("UIDL 1\r\n",
                               "+OK 1 whqtswO00WBw418f9t5JxYwZ\r\n");

        var message = client.GetMessage(1L, true);

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);

        Assert.IsNotNull(message);
        Assert.AreEqual(1L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);
      });
    }

    [Test]
    public void TestGetMessageDeleted()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");

        var message = client.GetMessage(1L);

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);

        server.EnqueueResponse("DELE 1\r\n",
                               "+OK\r\n");

        message.MarkAsDeleted();

        Assert.IsTrue(message.IsMarkedAsDeleted);

        try {
          client.GetMessage(1L);
          Assert.Fail("PopMessageDeletedException not thrown");
        }
        catch (PopMessageDeletedException ex) {
          Assert.AreEqual(1L, ex.MessageNumber);
        }
      });
    }

    [Test]
    public void TestGetMessageNotFound()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 120\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "-ERR no such message\r\n");

        try {
          client.GetMessage(1L);
          Assert.Fail("PopMessageNotFoundException not thrown");
        }
        catch (PopMessageNotFoundException ex) {
          Assert.AreEqual(1L, ex.MessageNumber);
          Assert.IsNull(ex.UniqueId);
        }

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageZero()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentOutOfRangeException>(() => client.GetMessage(0L));
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageGreaterThanMessageCount()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 3 1024\r\n");

        var ex = Assert.Throws<PopMessageNotFoundException>(() => client.GetMessage(4L));
        Assert.AreEqual(4L, ex.MessageNumber);
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageNotConnected()
    {
      using (var client = new PopClient()) {
        Assert.Throws<InvalidOperationException>(() => client.GetMessage(1L));
        Assert.IsFalse(client.IsBusy, "IsBusy");
      }
    }

    [Test]
    public void TestGetMessageReconnect()
    {
      var sessionMessages = new PopMessageInfo[2];

      for (var i = 0; i < 2; i++) {
        TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
          server.EnqueueResponse("STAT\r\n",
                                 "+OK 1 200\r\n");
          server.EnqueueResponse("LIST 1\r\n",
                                 "+OK 1 200\r\n");

          var message = client.GetMessage(1L);

          Assert.IsFalse(client.IsBusy, "IsBusy");

          Assert.AreEqual(1L, client.MessageCount);
          Assert.AreEqual(200L, client.TotalSize);

          Assert.IsNotNull(message);
          Assert.AreEqual(1L, message.MessageNumber);
          Assert.AreEqual(200L, message.Length);

          sessionMessages[i] = message;
        });
      }

      Assert.AreNotEqual(sessionMessages[0], sessionMessages[1]);
    }

    [Test]
    public void TestGetMessageByUid()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");
        server.EnqueueResponse("LIST 2\r\n",
                               "+OK 2 240\r\n");

        var message = client.GetMessage("QhdPYR:00WBw1Ph7x7");

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(2L, message.MessageNumber);
        Assert.AreEqual(240L, message.Length);
        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", message.UniqueId);
      });
    }

    [Test]
    public void TestGetMessageByUidError()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("UIDL\r\n",
                               "-ERR\r\n");

        try {
          client.GetMessage("QhdPYR:00WBw1Ph7x7");
          Assert.Fail("PopErrorResponseException not thrown");
        }
        catch (PopErrorResponseException) {
        }

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageByUidDeleted()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");
        server.EnqueueResponse("LIST 2\r\n",
                               "+OK 2 240\r\n");

        var message = client.GetMessage("QhdPYR:00WBw1Ph7x7");

        server.EnqueueResponse("DELE 2\r\n",
                               "+OK\r\n");

        message.MarkAsDeleted();

        try {
          client.GetMessage(message.UniqueId);
          Assert.Fail("PopMessageDeletedException not thrown");
        }
        catch (PopMessageDeletedException ex) {
          Assert.AreEqual(message.MessageNumber, ex.MessageNumber);
        }

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageByUidAlreadyListed()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");

        var messages = client.GetMessages(true).ToArray();

        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", messages[1].UniqueId);

        var message = client.GetMessage("QhdPYR:00WBw1Ph7x7");

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.IsNotNull(message);
        Assert.AreEqual(2L, message.MessageNumber);
        Assert.AreSame(messages[1], message);
      });
    }

    [Test]
    public void TestGetMessageByUidNull()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentNullException>(() => client.GetMessage((string)null));
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageByUidEmpty()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentException>(() => client.GetMessage(string.Empty));
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageByUidNotFound()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");

        try {
          client.GetMessage("non-existent");
          Assert.Fail("PopMessageNotFoundException not thrown");
        }
        catch (PopMessageNotFoundException ex) {
          Assert.AreEqual(0L, ex.MessageNumber);
          Assert.IsNotNull(ex.UniqueId);
          Assert.AreEqual("non-existent", ex.UniqueId);
        }

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessages()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");

        var messages = client.GetMessages().ToArray();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(1L, messages[0].MessageNumber);
        Assert.AreEqual(120L, messages[0].Length);

        Assert.AreEqual(2L, messages[1].MessageNumber);
        Assert.AreEqual(240L, messages[1].Length);

        Assert.AreEqual(3L, messages[2].MessageNumber);
        Assert.AreEqual(360L, messages[2].Length);

        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");

        var messagesSecond = client.GetMessages().ToArray();

        Assert.AreSame(messages[0], messagesSecond[0]);
        Assert.AreSame(messages[1], messagesSecond[1]);
        Assert.AreSame(messages[2], messagesSecond[2]);
      });
    }

    [Test]
    public void TestGetMessagesGetUniqueId()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");

        var messages = client.GetMessages(true).ToArray();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(1L, messages[0].MessageNumber);
        Assert.AreEqual(120L, messages[0].Length);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", messages[0].UniqueId);

        Assert.AreEqual(2L, messages[1].MessageNumber);
        Assert.AreEqual(240L, messages[1].Length);
        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", messages[1].UniqueId);

        Assert.AreEqual(3L, messages[2].MessageNumber);
        Assert.AreEqual(360L, messages[2].Length);
        Assert.AreEqual("xxxxxxxxx", messages[2].UniqueId);
      });
    }

    [Test]
    public void TestGetMessagesEmpty()
    {
      GetMessagesEmpty(false);
    }

    [Test]
    public void TestGetMessagesEmptyGetUniqueId()
    {
      GetMessagesEmpty(true);
    }

    private void GetMessagesEmpty(bool getUniqueId)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               ".\r\n");

        var messages = (getUniqueId ? client.GetMessages(true) : client.GetMessages()).ToArray();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(0, messages.Length);
      });
    }

    [Test]
    public void TestGetMessagesGetUniqueIdExtraUniqueId()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               "4 yyyyyyyyyyy\r\n" + 
                               ".\r\n");

        var messages = client.GetMessages(true).ToArray();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(1L, messages[0].MessageNumber);
        Assert.AreEqual(120L, messages[0].Length);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", messages[0].UniqueId);

        Assert.AreEqual(2L, messages[1].MessageNumber);
        Assert.AreEqual(240L, messages[1].Length);
        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", messages[1].UniqueId);

        Assert.AreEqual(3L, messages[2].MessageNumber);
        Assert.AreEqual(360L, messages[2].Length);
        Assert.AreEqual("xxxxxxxxx", messages[2].UniqueId);
      });
    }

    [Test]
    public void TestGetMessagesGetUniqueIdMissingUniqueId()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");
        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");

        var messages = client.GetMessages(true).ToArray();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(3, messages.Length);

        Assert.AreEqual(1L, messages[0].MessageNumber);
        Assert.AreEqual(120L, messages[0].Length);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", messages[0].UniqueId);

        Assert.AreEqual(2L, messages[1].MessageNumber);
        Assert.AreEqual(240L, messages[1].Length);
        Assert.IsNull(messages[1].UniqueId);

        Assert.AreEqual(3L, messages[2].MessageNumber);
        Assert.AreEqual(360L, messages[2].Length);
        Assert.AreEqual("xxxxxxxxx", messages[2].UniqueId);
      });
    }

    [Test]
    public void TestUpdateStatus()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");

        client.UpdateStatus();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);

        // re-issue
        server.EnqueueResponse("STAT\r\n",
                               "+OK 2 400\r\n");

        client.UpdateStatus();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.AreEqual(2L, client.MessageCount);
        Assert.AreEqual(400L, client.TotalSize);
      });
    }

    [Test]
    public void TestUpdateStatus_LoggedOut()
    {
      TestUtils.TestLoggedOut(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<InvalidOperationException>(() => client.UpdateStatus());
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestUpdateStatus_Disconnected()
    {
      TestUtils.TestDisconnected(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<Smdn.Net.Pop3.Protocol.PopConnectionException>(() => client.UpdateStatus());
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestCancelDelete()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);

        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");

        var message = client.GetMessage(1L);

        server.EnqueueResponse("DELE 1\r\n",
                               "+OK\r\n");

        message.MarkAsDeleted();

        Assert.IsTrue(message.IsMarkedAsDeleted);

        server.EnqueueResponse("STAT\r\n",
                               "+OK 0 0\r\n");

        Assert.AreEqual(0L, client.MessageCount);
        Assert.AreEqual(0L, client.TotalSize);

        server.EnqueueResponse("RSET\r\n",
                               "+OK\r\n");

        client.CancelDelete();

        Assert.IsFalse(client.IsBusy, "IsBusy");

        Assert.IsFalse(message.IsMarkedAsDeleted);

        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);
      });
    }

    [Test]
    public void TestCancelDelete_MessageNotListed()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("RSET\r\n",
                               "+OK\r\n");

        client.CancelDelete();

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestCancelDelete_LoggedOut()
    {
      TestUtils.TestLoggedOut(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<InvalidOperationException>(() => client.CancelDelete());
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestCancelDelete_Disconnected()
    {
      TestUtils.TestDisconnected(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<Smdn.Net.Pop3.Protocol.PopConnectionException>(() => client.CancelDelete());
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestKeepAlive()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("NOOP\r\n",
                               "+OK\r\n");

        client.KeepAlive();

        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestKeepAlive_LoggedOut()
    {
      TestUtils.TestLoggedOut(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<InvalidOperationException>(() => client.KeepAlive());
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestKeepAlive_Disconnected()
    {
      TestUtils.TestDisconnected(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<Smdn.Net.Pop3.Protocol.PopConnectionException>(() => client.KeepAlive());
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestInternalImpl_RunExclusively()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        using (var waitHandle = new ManualResetEvent(false)) {
          ThreadPool.QueueUserWorkItem(delegate {
            client.KeepAlive();
          });
          ThreadPool.QueueUserWorkItem(delegate {
            server.WaitForClientRequest();

            waitHandle.Set();

            Thread.Sleep(500);

            server.EnqueueResponse("NOOP\r\n",
                                 "+OK\r\n");
          });

          waitHandle.WaitOne();

          for (;;) {
            if (!client.IsBusy)
              break;

            Assert.Throws<InvalidOperationException>(() => client.KeepAlive());

#if ASYNC_OPS
            if (!client.IsBusy)
              break;

            Assert.Throws<InvalidOperationException>(() => client.KeepAliveAsync());
#endif

            Thread.Sleep(5);
          }
        }
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestInternalImpl_StartTaskExclusively()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        ThreadPool.QueueUserWorkItem(delegate {
          server.WaitForClientRequest();

          Thread.Sleep(500);

          server.EnqueueResponse("NOOP\r\n",
                                 "+OK\r\n");
        });

        var t = client.KeepAliveAsync();

        for (;;) {
          if (!client.IsBusy)
            break;

          Assert.Throws<InvalidOperationException>(() => client.KeepAlive());

          if (!client.IsBusy)
            break;

          Assert.Throws<InvalidOperationException>(() => client.KeepAliveAsync());

          Thread.Sleep(5);
        }

        t.Wait();
      });
    }
#endif
  }
}
