#if NET_4_0
#define ASYNC_OPS
#endif

#if ASYNC_OPS
using System;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.IO;
using Smdn.Net.Pop3.Client.Session;

namespace Smdn.Net.Pop3.Client {
  [TestFixture]
  public class PopClientOperationsAsyncTests {
    [Test]
    [TestCase(true)]
    [TestCase(false)]
    public void TestGetMessageAsync_ByMessageNumber(bool getUniqueId)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        var t1 = getUniqueId
                 ? client.GetMessageAsync(3L, true)
                 : client.GetMessageAsync(3L);

        Assert.IsTrue(client.IsBusy, "IsBusy task started");

        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("LIST 3\r\n",
                               "+OK 3 200\r\n");
        if (getUniqueId)
          server.EnqueueResponse("UIDL 3\r\n",
                                 "+OK 3 whqtswO00WBw418f9t5JxYwZ\r\n");

        t1.Wait();

        Assert.IsFalse(client.IsBusy, "IsBusy task finished");

        var message = t1.Result;

        Assert.AreEqual(3L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);

        if (getUniqueId)
          Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);

        // get again
        var t2 = getUniqueId
                 ? client.GetMessageAsync(3L, true)
                 : client.GetMessageAsync(3L);

        Assert.AreSame(message, t2.Result);
      });
    }

    [Test]
    public void TestGetMessageAsync_ByMessageNumber_MessageNumberOutOfRange()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentOutOfRangeException>(() => client.GetMessageAsync(0L));
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageAsync_ByMessageNumber_CancellationRequestedBeforeStart()
    {
      TestUtils.Repeat(100, delegate {
        TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
          var cts = new CancellationTokenSource();

          cts.Cancel();

          var t = client.GetMessageAsync(1L, cts.Token);

          if (t.IsCanceled)
            Assert.IsFalse(client.IsBusy, "IsBusy task started");

          var ex = Assert.Throws<AggregateException>(() => t.Wait());

          Assert.IsFalse(client.IsBusy, "IsBusy task finished");

          Assert.IsInstanceOf<OperationCanceledException>(ex.InnerException);

          Assert.IsNull(server.DequeueRequest()); // nothing requested
        });
      });
    }

    [Test]
    public void TestGetMessageAsync_ByUniqueId()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        var t = client.GetMessageAsync("QhdPYR:00WBw1Ph7x7");

        Assert.IsTrue(client.IsBusy, "IsBusy task started");

        server.EnqueueResponse("UIDL\r\n",
                               "+OK\r\n" +
                               "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                               "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                               "3 xxxxxxxxx\r\n" + 
                               ".\r\n");
        server.EnqueueResponse("LIST 2\r\n",
                               "+OK 2 240\r\n");

        t.Wait();

        Assert.IsFalse(client.IsBusy, "IsBusy task finished");

        var message = t.Result;

        Assert.AreEqual(2L, message.MessageNumber);
        Assert.AreEqual(240L, message.Length);
        Assert.AreEqual("QhdPYR:00WBw1Ph7x7", message.UniqueId);
      });
    }

    [Test]
    public void TestGetMessageAsync_ByUniqueId_UniqueIdNull()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentNullException>(() => client.GetMessageAsync((string)null));
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageAsync_ByUniqueId_UniqueIdEmpty()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentException>(() => client.GetMessageAsync(string.Empty));
        Assert.IsFalse(client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestGetMessageAsync_ByUniqueId_CancellationRequestedBeforeStart()
    {
      TestUtils.Repeat(100, delegate {
        TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
          var cts = new CancellationTokenSource();

          cts.Cancel();

          var t = client.GetMessageAsync("uid", cts.Token);

          if (t.IsCanceled)
            Assert.IsFalse(client.IsBusy, "IsBusy task started");

          var ex = Assert.Throws<AggregateException>(() => t.Wait());

          Assert.IsFalse(client.IsBusy, "IsBusy task finished");

          Assert.IsInstanceOf<OperationCanceledException>(ex.InnerException);

          Assert.IsNull(server.DequeueRequest()); // nothing requested
        });
      });
    }

    [Test]
    [TestCase(true)]
    [TestCase(false)]
    public void TestGetFirstMessageAsync(bool getUniqueId)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        var t1 = getUniqueId
                 ? client.GetFirstMessageAsync(true)
                 : client.GetFirstMessageAsync();

        Assert.IsTrue(client.IsBusy, "IsBusy task started");

        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");
        if (getUniqueId)
          server.EnqueueResponse("UIDL 1\r\n",
                                 "+OK 1 whqtswO00WBw418f9t5JxYwZ\r\n");

        t1.Wait();

        Assert.IsFalse(client.IsBusy, "IsBusy task finished");

        var message = t1.Result;

        Assert.AreEqual(1L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);

        if (getUniqueId)
          Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);

        // get again
        var t2 = getUniqueId
                 ? client.GetFirstMessageAsync(true)
                 : client.GetFirstMessageAsync();

        Assert.AreSame(message, t2.Result);
      });
    }

    [Test]
    [TestCase(true)]
    [TestCase(false)]
    public void TestGetLastMessageAsync(bool getUniqueId)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        var t1 = getUniqueId
                 ? client.GetLastMessageAsync(true)
                 : client.GetLastMessageAsync();

        Assert.IsTrue(client.IsBusy, "IsBusy task started");

        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("LIST 5\r\n",
                               "+OK 5 200\r\n");
        if (getUniqueId)
          server.EnqueueResponse("UIDL 5\r\n",
                                 "+OK 5 whqtswO00WBw418f9t5JxYwZ\r\n");

        t1.Wait();

        Assert.IsFalse(client.IsBusy, "IsBusy task finished");

        var message = t1.Result;

        Assert.AreEqual(5L, message.MessageNumber);
        Assert.AreEqual(200L, message.Length);

        if (getUniqueId)
          Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);

        // get again
        var t2 = getUniqueId
                 ? client.GetLastMessageAsync(true)
                 : client.GetLastMessageAsync();

        Assert.AreSame(message, t2.Result);
      });
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestGetMessagesAsync(bool getUniqueId)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        var t1 = getUniqueId
                 ? client.GetMessagesAsync(true)
                 : client.GetMessagesAsync();

        Assert.IsTrue(client.IsBusy, "IsBusy task started");

        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");

        if (getUniqueId)
          server.EnqueueResponse("UIDL\r\n",
                                 "+OK\r\n" +
                                 "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                                 "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                                 "3 xxxxxxxxx\r\n" + 
                                 ".\r\n");

        t1.Wait();

        Assert.IsFalse(client.IsBusy, "IsBusy task finished");

        var messages = t1.Result;

        Assert.AreEqual(3, messages.Count);

        Assert.AreEqual(1L, messages[0].MessageNumber);
        Assert.AreEqual(120L, messages[0].Length);
        if (getUniqueId)
          Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", messages[0].UniqueId);

        Assert.AreEqual(2L, messages[1].MessageNumber);
        Assert.AreEqual(240L, messages[1].Length);
        if (getUniqueId)
          Assert.AreEqual("QhdPYR:00WBw1Ph7x7", messages[1].UniqueId);

        Assert.AreEqual(3L, messages[2].MessageNumber);
        Assert.AreEqual(360L, messages[2].Length);
        if (getUniqueId)
          Assert.AreEqual("xxxxxxxxx", messages[2].UniqueId);

#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => messages.Add(messages[0]));
#endif

        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 240\r\n" +
                               "3 360\r\n" +
                               ".\r\n");

        if (getUniqueId)
          server.EnqueueResponse("UIDL\r\n",
                                 "+OK\r\n" +
                                 "1 whqtswO00WBw418f9t5JxYwZ\r\n" + 
                                 "2 QhdPYR:00WBw1Ph7x7\r\n" + 
                                 "3 xxxxxxxxx\r\n" + 
                                 ".\r\n");

        var t2 = getUniqueId
                 ? client.GetMessagesAsync(true)
                 : client.GetMessagesAsync();

        var messagesSecond = t2.Result;

        Assert.AreSame(messages[0], messagesSecond[0]);
        Assert.AreSame(messages[1], messagesSecond[1]);
        Assert.AreSame(messages[2], messagesSecond[2]);
      });
    }

    [Test]
    public void TestCancelDeleteAsync()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);

        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");

        var message = client.GetMessage(1L);

        server.EnqueueResponse("DELE 1\r\n",
                               "+OK\r\n");

        message.MarkAsDeleted();

        Assert.IsTrue(message.IsMarkedAsDeleted);

        server.EnqueueResponse("STAT\r\n",
                               "+OK 0 0\r\n");

        Assert.AreEqual(0L, client.MessageCount);
        Assert.AreEqual(0L, client.TotalSize);

        var t = client.CancelDeleteAsync();

        Assert.IsTrue(client.IsBusy, "IsBusy task started");
        Assert.Throws<InvalidOperationException>(() => client.CancelDelete());

        server.EnqueueResponse("RSET\r\n",
                               "+OK\r\n");

        t.Wait();

        Assert.IsFalse(client.IsBusy, "IsBusy task finished");

        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);
      });
    }

    [Test]
    public void TestUpdateStatusAsync()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        var t = client.UpdateStatusAsync();

        Assert.IsTrue(client.IsBusy, "IsBusy task started");
        Assert.Throws<InvalidOperationException>(() => client.UpdateStatus());

        server.EnqueueResponse("STAT\r\n",
                               "+OK 1 200\r\n");

        t.Wait();

        Assert.IsFalse(client.IsBusy, "IsBusy task finished");

        Assert.AreEqual(1L, client.MessageCount);
        Assert.AreEqual(200L, client.TotalSize);
      });
    }

    [Test]
    public void TestUpdateStatusAsync_LoggedOut()
    {
      TestUtils.TestLoggedOut(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<InvalidOperationException>(() => client.UpdateStatusAsync());
      });
    }

    [Test]
    public void TestCancelDeleteAsync_MessageNotListed()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        var t = client.CancelDeleteAsync();

        Assert.IsTrue(client.IsBusy, "IsBusy task started");
        Assert.Throws<InvalidOperationException>(() => client.CancelDelete());

        server.EnqueueResponse("RSET\r\n",
                               "+OK\r\n");

        t.Wait();

        Assert.IsFalse(client.IsBusy, "IsBusy task finished");
      });
    }

    [Test]
    public void TestCancelDeleteAsync_LoggedOut()
    {
      TestUtils.TestLoggedOut(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<InvalidOperationException>(() => client.CancelDeleteAsync());
      });
    }

    [Test]
    public void TestKeepAliveAsync()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        var t = client.KeepAliveAsync();

        Assert.IsTrue(client.IsBusy, "IsBusy task started");
        Assert.Throws<InvalidOperationException>(() => client.KeepAlive());

        server.EnqueueResponse("NOOP\r\n",
                               "+OK\r\n");

        t.Wait();

        Assert.IsFalse(client.IsBusy, "IsBusy task finished");
      });
    }

    [Test]
    public void TestKeepAliveAsync_LoggedOut()
    {
      TestUtils.TestLoggedOut(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<InvalidOperationException>(() => client.KeepAliveAsync());
      });
    }
  }
}
#endif // ASYNC_OPS