#if NET_4_0
#define ASYNC_OPS
#endif

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using NUnit.Framework;

using Smdn.IO;
using Smdn.Net.Pop3.Client.Session;

namespace Smdn.Net.Pop3.Client {
  [TestFixture]
  public class PopClientOperationsDownloadTests {
    private static void GetMessageBody(out string messageBody, out string messageByteStuffedBody, out int octets)
    {
      GetMessageBody(null, out messageBody, out messageByteStuffedBody, out octets);
    }

    private static void GetMessageBody(string subject, out string messageBody, out string messageByteStuffedBody, out int octets)
    {
      messageBody = @"MIME-Version: 1.0
Content-Type: text/plain
Subject: " + (subject ?? string.Empty) +
@"

..byte-stuffed
+OK response style line
-ERR response style line2
.
end of message
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      messageByteStuffedBody = messageBody.Replace("\r\n.", "\r\n..");

      octets = NetworkTransferEncoding.Transfer8Bit.GetByteCount(messageBody);
    }

    private void TestMessage(Action<PopPseudoServer, PopClient, string> action)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("STAT\r\n",
                               "+OK 3 200\r\n");
        server.EnqueueResponse("LIST 3\r\n",
                               "+OK 3 200\r\n");
        server.EnqueueResponse("RETR 3\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        action(server, client, messageBody);

        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n");

        client.Disconnect(true);
      });
    }

    private void TestFirstMessage(Action<PopPseudoServer, PopClient, string> action)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");
        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        action(server, client, messageBody);

        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n");

        client.Disconnect(true);
      });
    }

    private void TestLastMessage(Action<PopPseudoServer, PopClient, string> action)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("LIST 5\r\n",
                               "+OK 5 200\r\n");
        server.EnqueueResponse("RETR 5\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        action(server, client, messageBody);

        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n");

        client.Disconnect(true);
      });
    }

    private void TestNoMessagesExist(Action<PopClient> action)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 0 0\r\n");

        action(client);
      });
    }

    private void TestAllMessages(Action<PopPseudoServer, PopClient, string[]> action)
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        const int messageCount = 5;
        var messageBodies = new string[messageCount];
        var byteStuffedMessageBodies = new string[messageCount];

        for (var i = 0; i < messageCount; i++) {
          int octets;
          string messageBody, byteStuffedMessageBody;

          GetMessageBody("#" + (i + 1).ToString(), out messageBody, out byteStuffedMessageBody, out octets);

          messageBodies[i] = messageBody;
          byteStuffedMessageBodies[i] = byteStuffedMessageBody;
        }
        
        // LIST
        var listResponse = "+OK\r\n";

        for (var i = 0; i < messageCount; i++) {
          listResponse += string.Format("{0} 200\r\n", i + 1);
        }

        listResponse += ".\r\n";

        server.EnqueueResponse("LIST\r\n",
                               listResponse);

        // RETR
        for (var i = 0; i < messageCount; i++) {
          server.EnqueueResponse(string.Format("RETR {0}\r\n", i + 1),
                                 string.Format("+OK\r\n{0}.\r\n", byteStuffedMessageBodies[i]));
        }

        action(server, client, messageBodies);

        // QUIT
        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n");

        client.Disconnect(true);
      });
    }

    [Test]
    public void TestDownloadMessageAsText()
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        Assert.AreEqual(expectedResult,
                        client.DownloadMessageAsText(3L, Encoding.UTF8));
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadMessageAsTextAsync()
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        Assert.AreEqual(expectedResult,
                        client.DownloadMessageAsTextAsync(3L, Encoding.UTF8).Result);
      });
    }
#endif

    [Test]
    public void TestDownloadFirstMessageAsText()
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        Assert.AreEqual(expectedResult,
                        client.DownloadFirstMessageAsText(Encoding.UTF8));
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadFirstMessageAsTextAsync()
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        Assert.AreEqual(expectedResult,
                        client.DownloadFirstMessageAsTextAsync(Encoding.UTF8).Result);
      });
    }
#endif

    [Test]
    public void TestDownloadFirstMessageAsText_NoMessagesExist()
    {
      TestNoMessagesExist(delegate(PopClient client) {
        Assert.Throws<PopMessageNotFoundException>(() => client.DownloadFirstMessageAsText(Encoding.UTF8));
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadFirstMessageAsTextAsync_NoMessagesExist()
    {
      TestNoMessagesExist(delegate(PopClient client) {
        var ex = Assert.Throws<AggregateException>(() => client.DownloadFirstMessageAsTextAsync(Encoding.UTF8).Wait());

        Assert.IsInstanceOf<PopMessageNotFoundException>(ex.Flatten().InnerException);
      });
    }
#endif

    [Test]
    public void TestDownloadLastMessageAsText()
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        Assert.AreEqual(expectedResult,
                        client.DownloadLastMessageAsText(Encoding.UTF8));
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadLastMessageAsTextAsync()
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        Assert.AreEqual(expectedResult,
                        client.DownloadLastMessageAsTextAsync(Encoding.UTF8).Result);
      });
    }
#endif

    [Test]
    public void TestDownloadLastMessageAsText_NoMessagesExist()
    {
      TestNoMessagesExist(delegate(PopClient client) {
        Assert.Throws<PopMessageNotFoundException>(() => client.DownloadLastMessageAsText(Encoding.UTF8));
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadLastMessageAsTextAsync_NoMessagesExist()
    {
      TestNoMessagesExist(delegate(PopClient client) {
        var ex = Assert.Throws<AggregateException>(() => client.DownloadLastMessageAsTextAsync(Encoding.UTF8).Wait());

        Assert.IsInstanceOf<PopMessageNotFoundException>(ex.Flatten().InnerException);
      });
    }
#endif

    [Test]
    public void TestDownloadAllMessagesAsText()
    {
      TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
        var downloadedMessages = client.DownloadAllMessagesAsText(Encoding.UTF8).ToArray();

        Assert.AreEqual(expectedResults.Length, downloadedMessages.Length, "downloaded message count");

        for (var i = 0; i < expectedResults.Length; i++) {
          Assert.AreEqual(expectedResults[i], downloadedMessages[i]);
        }
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadAllMessagesAsTextAsync()
    {
      TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
        var downloadedMessages = client.DownloadAllMessagesAsTextAsync(Encoding.UTF8).Result;

        Assert.AreEqual(expectedResults.Length, downloadedMessages.Count, "downloaded message count");

#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => downloadedMessages.Clear());
#endif

        for (var i = 0; i < expectedResults.Length; i++) {
          Assert.AreEqual(expectedResults[i], downloadedMessages[i]);
        }
      });
    }
#endif

    [Test]
    public void TestDownloadAllMessagesAsText_NoMessagesExist()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               ".\r\n");

        CollectionAssert.IsEmpty(client.DownloadAllMessagesAsText(Encoding.UTF8));
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadAllMessagesAsTextAsync_NoMessagesExist()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               ".\r\n");

        CollectionAssert.IsEmpty(client.DownloadAllMessagesAsTextAsync(Encoding.UTF8).Result);
      });
    }
#endif

    [Test]
    public void TestDownloadMessageAsByteArray()
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        CollectionAssert.AreEqual(Encoding.UTF8.GetBytes(expectedResult),
                                  client.DownloadMessageAsByteArray(3L));
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadMessageAsByteArrayAsync()
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        CollectionAssert.AreEqual(Encoding.UTF8.GetBytes(expectedResult),
                                  client.DownloadMessageAsByteArrayAsync(3L).Result);
      });
    }
#endif

    [Test]
    public void TestDownloadFirstMessageAsByteArray()
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        CollectionAssert.AreEqual(Encoding.UTF8.GetBytes(expectedResult),
                                  client.DownloadFirstMessageAsByteArray());
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadFirstMessageAsByteArrayAsync()
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        CollectionAssert.AreEqual(Encoding.UTF8.GetBytes(expectedResult),
                                  client.DownloadFirstMessageAsByteArrayAsync().Result);
      });
    }
#endif

    [Test]
    public void TestDownloadLastMessageAsByteArray()
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        CollectionAssert.AreEqual(Encoding.UTF8.GetBytes(expectedResult),
                                  client.DownloadLastMessageAsByteArray());
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadLastMessageAsByteArrayAsync()
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        CollectionAssert.AreEqual(Encoding.UTF8.GetBytes(expectedResult),
                                  client.DownloadLastMessageAsByteArrayAsync().Result);
      });
    }
#endif

    [Test]
    public void TestDownloadMessageAsStream()
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        using (var stream = client.DownloadMessageAsStream(3L)) {
          FileAssert.AreEqual(new MemoryStream(Encoding.UTF8.GetBytes(expectedResult)),
                              stream);
        }
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadMessageAsStreamAsync()
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        using (var stream = client.DownloadMessageAsStreamAsync(3L).Result) {
          FileAssert.AreEqual(new MemoryStream(Encoding.UTF8.GetBytes(expectedResult)),
                              stream);
        }
      });
    }
#endif

    [Test]
    public void TestDownloadFirstMessageAsStream()
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        using (var stream = client.DownloadFirstMessageAsStream()) {
          FileAssert.AreEqual(new MemoryStream(Encoding.UTF8.GetBytes(expectedResult)),
                              stream);
        }
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadFirstMessageAsStreamAsync()
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        using (var stream = client.DownloadFirstMessageAsStreamAsync().Result) {
          FileAssert.AreEqual(new MemoryStream(Encoding.UTF8.GetBytes(expectedResult)),
                              stream);
        }
      });
    }
#endif

    [Test]
    public void TestDownloadLastMessageAsStream()
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        using (var stream = client.DownloadLastMessageAsStream()) {
          FileAssert.AreEqual(new MemoryStream(Encoding.UTF8.GetBytes(expectedResult)),
                              stream);
        }
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadLastMessageAsStreamAsync()
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        using (var stream = client.DownloadLastMessageAsStreamAsync().Result) {
          FileAssert.AreEqual(new MemoryStream(Encoding.UTF8.GetBytes(expectedResult)),
                              stream);
        }
      });
    }
#endif

    [Test]
    public void TestDownloadMessageAs_Converter()
    {
      DownloadMessageAs_Converter((client, number, converter) => client.DownloadMessageAs(number, converter));
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadMessageAsAsync_Converter()
    {
      DownloadMessageAs_Converter((client, number, converter) => client.DownloadMessageAsAsync(number, converter).Result);
    }
#endif

    private void DownloadMessageAs_Converter(Func<PopClient, long, Converter<Stream, long>, long> testFunc)
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        var ret = testFunc(client, 3L, delegate(Stream stream) {
          return stream.ReadToEnd().Length;
        });

        Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResult),
                        ret);
      });
    }

    [Test]
    public void TestDownloadFirstMessageAs_Converter()
    {
      DownloadFirstMessageAs_Converter((client, converter) => client.DownloadFirstMessageAs(converter));
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadFirstMessageAsAsync_Converter()
    {
      DownloadFirstMessageAs_Converter((client, converter) => client.DownloadFirstMessageAsAsync(converter).Result);
    }
#endif

    private void DownloadFirstMessageAs_Converter(Func<PopClient, Converter<Stream, long>, long> testFunc)
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        var ret = testFunc(client, delegate(Stream stream) {
          return stream.ReadToEnd().Length;
        });

        Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResult),
                        ret);
      });
    }

    [Test]
    public void TestDownloadLastMessageAs_Converter()
    {
      DownloadLastMessageAs_Converter((client, converter) => client.DownloadLastMessageAs(converter));
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadLastMessageAsAsync_Converter()
    {
      DownloadLastMessageAs_Converter((client, converter) => client.DownloadLastMessageAsAsync(converter).Result);
    }
#endif

    private void DownloadLastMessageAs_Converter(Func<PopClient, Converter<Stream, long>, long> testFunc)
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        var ret = testFunc(client, delegate(Stream stream) {
          return stream.ReadToEnd().Length;
        });

        Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResult),
                        ret);
      });
    }

    [Test]
    public void TestDownloadAllMessagesAs_Converter()
    {
      TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
        var downloadedResults = client.DownloadAllMessagesAs(delegate(Stream stream) {
          return stream.ReadToEnd().Length;
        }).ToArray();
        
        Assert.AreEqual(expectedResults.Length, downloadedResults.Length, "downloaded message count");
        
        for (var i = 0; i < expectedResults.Length; i++) {
          Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResults[i]), downloadedResults[i]);
        }
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadAllMessagesAsAsync_Converter()
    {
      TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
        var downloadedResults = client.DownloadAllMessagesAsAsync(delegate(Stream stream) {
          return stream.ReadToEnd().Length;
        }).Result;

        Assert.AreEqual(expectedResults.Length, downloadedResults.Count, "downloaded message count");

#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => downloadedResults.Clear());
#endif

        for (var i = 0; i < expectedResults.Length; i++) {
          Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResults[i]), downloadedResults[i]);
        }
      });
    }
#endif

    [Test]
    public void TestDownloadMessageAs_Func()
    {
      DownloadMessageAs_Func((client, number, func, arg) => client.DownloadMessageAs(number, func, arg));
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadMessageAsAsync_Func()
    {
      DownloadMessageAs_Func((client, number, func, arg) => client.DownloadMessageAsAsync(number, func, arg).Result);
    }
#endif

    private void DownloadMessageAs_Func(Func<PopClient, long, Func<Stream, string, int>, string, int> testFunc)
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        var ret = testFunc(client, 3L, delegate(Stream stream, string arg) {
          Assert.AreEqual(arg, "arg");

          return stream.ReadToEnd().Length;
        }, "arg");

        Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResult),
                        ret);
      });
    }

    [Test]
    public void TestDownloadFirstMessageAs_Func()
    {
      DownloadFirstMessageAs_Func((client, func, arg) => client.DownloadFirstMessageAs(func, arg));
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadFirstMessageAsAsync_Func()
    {
      DownloadFirstMessageAs_Func((client, func, arg) => client.DownloadFirstMessageAsAsync(func, arg).Result);
    }
#endif

    private void DownloadFirstMessageAs_Func(Func<PopClient, Func<Stream, string, int>, string, int> testFunc)
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        var ret = testFunc(client, delegate(Stream stream, string arg) {
          Assert.AreEqual(arg, "arg");

          return stream.ReadToEnd().Length;
        }, "arg");

        Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResult),
                        ret);
      });
    }

    [Test]
    public void TestDownloadLastMessageAs_Func()
    {
      DownloadLastMessageAs_Func((client, func, arg) => client.DownloadLastMessageAs(func, arg));
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadLastMessageAsAsync_Func()
    {
      DownloadLastMessageAs_Func((client, func, arg) => client.DownloadLastMessageAsAsync(func, arg).Result);
    }
#endif

    private void DownloadLastMessageAs_Func(Func<PopClient, Func<Stream, string, int>, string, int> testFunc)
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        var ret = testFunc(client, delegate(Stream stream, string arg) {
          Assert.AreEqual(arg, "arg");

          return stream.ReadToEnd().Length;
        }, "arg");

        Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResult),
                        ret);
      });
    }

    [Test]
    public void TestDownloadAllMessagesAs_Func()
    {
      TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
        var downloadedResults = client.DownloadAllMessagesAs(delegate(Stream stream, string arg) {
          Assert.AreEqual(arg, "arg");

          return stream.ReadToEnd().Length;
        }, "arg").ToArray();
        
        Assert.AreEqual(expectedResults.Length, downloadedResults.Length, "downloaded message count");
        
        for (var i = 0; i < expectedResults.Length; i++) {
          Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResults[i]), downloadedResults[i]);
        }
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadAllMessagesAsAsync_Func()
    {
      TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
        var downloadedResults = client.DownloadAllMessagesAsAsync(delegate(Stream stream, string arg) {
          Assert.AreEqual(arg, "arg");

          return stream.ReadToEnd().Length;
        }, "arg").Result;

        Assert.AreEqual(expectedResults.Length, downloadedResults.Count, "downloaded message count");

#if !NET_4_5
        Assert.Throws<NotSupportedException>(() => downloadedResults.Clear());
#endif

        for (var i = 0; i < expectedResults.Length; i++) {
          Assert.AreEqual(Encoding.UTF8.GetByteCount(expectedResults[i]), downloadedResults[i]);
        }
      });
    }
#endif

    private void TestFile(Action<string, string> action)
    {
      const string expectedFile = "expected.eml";
      const string actualFile = "actual.eml";

      try {
        if (File.Exists(expectedFile))
          File.Delete(expectedFile);
        if (File.Exists(actualFile))
          File.Delete(actualFile);

        action(expectedFile, actualFile);

        FileAssert.AreEqual(expectedFile, actualFile);
      }
      finally {
        if (File.Exists(expectedFile))
          File.Delete(expectedFile);
        if (File.Exists(actualFile))
          File.Delete(actualFile);
      }
    }

    [Test]
    public void TestDownloadMessageToFile()
    {
      DownloadMessageToFile((client, number, path) => client.DownloadMessageToFile(number, path));
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadMessageToFileAsync()
    {
      DownloadMessageToFile((client, number, path) => client.DownloadMessageToFileAsync(number, path).Result);
    }
#endif

    private void DownloadMessageToFile(Func<PopClient, long, string, FileInfo> testFunc)
    {
      TestMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        TestFile(delegate(string expectedFilePath, string actualFilePath) {
          File.WriteAllText(expectedFilePath, expectedResult, new UTF8Encoding(false));

          var file = testFunc(client, 3L, actualFilePath);

          Assert.IsTrue(file.Exists, "downloaded file exists");
          StringAssert.EndsWith(actualFilePath, file.FullName);
        });
      });
    }

    [Test]
    public void TestDownloadFirstMessageToFile()
    {
      DownloadFirstMessageToFile((client, path) => client.DownloadFirstMessageToFile(path));
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadFirstMessageToFileAsync()
    {
      DownloadFirstMessageToFile((client, path) => client.DownloadFirstMessageToFileAsync(path).Result);
    }
#endif

    private void DownloadFirstMessageToFile(Func<PopClient, string, FileInfo> testFunc)
    {
      TestFirstMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        TestFile(delegate(string expectedFilePath, string actualFilePath) {
          File.WriteAllText(expectedFilePath, expectedResult, new UTF8Encoding(false));

          var file = testFunc(client, actualFilePath);

          Assert.IsTrue(file.Exists, "downloaded file exists");
          StringAssert.EndsWith(actualFilePath, file.FullName);
        });
      });
    }

    [Test]
    public void TestDownloadLastMessageToFile()
    {
      DownloadLastMessageToFile((client, path) => client.DownloadLastMessageToFile(path));
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadLastMessageToFileAsync()
    {
      DownloadLastMessageToFile((client, path) => client.DownloadLastMessageToFileAsync(path).Result);
    }
#endif

    private void DownloadLastMessageToFile(Func<PopClient, string, FileInfo> testFunc)
    {
      TestLastMessage(delegate(PopPseudoServer server, PopClient client, string expectedResult) {
        TestFile(delegate(string expectedFilePath, string actualFilePath) {
          File.WriteAllText(expectedFilePath, expectedResult, new UTF8Encoding(false));

          var file = testFunc(client, actualFilePath);

          Assert.IsTrue(file.Exists, "downloaded file exists");
          StringAssert.EndsWith(actualFilePath, file.FullName);
        });
      });
    }

    [Test]
    public void TestDownloadAllMessagesToDirectory()
    {
      DownloadAllMessagesToDirectory((client, directory) => client.DownloadAllMessagesToDirectory(directory));
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadAllMessagesToDirectoryAsync()
    {
      DownloadAllMessagesToDirectory((client, directory) => client.DownloadAllMessagesToDirectoryAsync(directory).Result);
    }
#endif

#if NET_4_5
    private void DownloadAllMessagesToDirectory(Func<PopClient, string, IReadOnlyList<FileInfo>> testFunc)
#else
    private void DownloadAllMessagesToDirectory(Func<PopClient, string, IList<FileInfo>> testFunc)
#endif
    {
      const string dir = "downloaded";

      try {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);

        Directory.CreateDirectory(dir);

        TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
          var files = testFunc(client, dir);

          Assert.AreEqual(expectedResults.Length, files.Count, "downloaded file count");

          for (var i = 0; i < expectedResults.Length; i++) {
            var expectedFileName = string.Format("{0}{1}{2}.eml", dir, Path.DirectorySeparatorChar, i + 1);

            Assert.IsTrue(files[i].Exists, "downloaded file exists #{0}", i);
            StringAssert.EndsWith(expectedFileName, files[i].FullName, "file name #{0}", i);

            Assert.AreEqual(expectedResults[i],
                            File.ReadAllText(files[i].FullName),
                            "downloaded content #{0}", i);
          }
        });
      }
      finally {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);
      }
    }

    [Test]
    public void TestDownloadAllMessagesToDirectory_TargetDirectoryNotExist()
    {
      DownloadAllMessagesToDirectoryAsync_TargetDirectoryNotExist((client, directory) => client.DownloadAllMessagesToDirectory(directory));
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadAllMessagesToDirectoryAsync_TargetDirectoryNotExist()
    {
      DownloadAllMessagesToDirectoryAsync_TargetDirectoryNotExist((client, directory) => client.DownloadAllMessagesToDirectoryAsync(directory).Result);
    }
#endif

#if NET_4_5
    private void DownloadAllMessagesToDirectoryAsync_TargetDirectoryNotExist(Func<PopClient, string, IReadOnlyList<FileInfo>> testFunc)
#else
    private void DownloadAllMessagesToDirectoryAsync_TargetDirectoryNotExist(Func<PopClient, string, IList<FileInfo>> testFunc)
#endif
    {
      const string dir = "downloaded";
      
      try {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);

        TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
          var files = testFunc(client, dir);

          Assert.IsTrue(Directory.Exists(dir));
          Assert.AreEqual(expectedResults.Length, files.Count, "downloaded file count");

          for (var i = 0; i < expectedResults.Length; i++) {
            var expectedFileName = string.Format("{0}{1}{2}.eml", dir, Path.DirectorySeparatorChar, i + 1);

            Assert.IsTrue(files[i].Exists, "downloaded file exists #{0}", i);
            StringAssert.EndsWith(expectedFileName, files[i].FullName, "file name #{0}", i);

            Assert.AreEqual(expectedResults[i],
                            File.ReadAllText(files[i].FullName),
                            "downloaded content #{0}", i);
          }
        });
      }
      finally {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);
      }
    }

    [Test]
    public void TestDownloadAllMessagesToDirectory_ArgumentNull()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentNullException>(delegate {
          client.DownloadAllMessagesToDirectory(null);
        });
      });
    }

    [Test]
    public void TestDownloadAllMessagesToDirectory_ArgumentEmpty()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentException>(delegate {
          client.DownloadAllMessagesToDirectory(string.Empty);
        });
      });
    }

    [Test]
    public void TestDownloadAllMessagesToFiles()
    {
      DownloadAllMessagesToFiles((client, selector) => client.DownloadAllMessagesToFiles(selector));
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadAllMessagesToFilesAsync()
    {
      DownloadAllMessagesToFiles((client, selector) => client.DownloadAllMessagesToFilesAsync(selector).Result);
    }
#endif

#if NET_4_5
    private void DownloadAllMessagesToFiles(Func<PopClient, Func<PopMessageInfo, string>, IReadOnlyList<FileInfo>> testFunc)
#else
    private void DownloadAllMessagesToFiles(Func<PopClient, Func<PopMessageInfo, string>, IList<FileInfo>> testFunc)
#endif
    {
      const string dir = "downloaded";

      try {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);

        Directory.CreateDirectory(dir);

        TestAllMessages(delegate(PopPseudoServer server, PopClient client, string[] expectedResults) {
          var expectedFiles = new List<string>();

          var files = testFunc(client, delegate(PopMessageInfo msg) {
            var file = Path.Combine(dir, msg.MessageNumber.ToString("D4"));

            expectedFiles.Add(file);

            return file;
          });

          for (var i = 0; i < expectedResults.Length; i++) {
            Assert.IsTrue(files[i].Exists, "downloaded file exists #{0}", i);
            StringAssert.EndsWith(expectedFiles[i], files[i].FullName, "file name #{0}", i);

            Assert.AreEqual(expectedResults[i],
                            File.ReadAllText(files[i].FullName),
                            "downloaded content #{0}", i);
          }
        });
      }
      finally {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);
      }
    }

    [Test]
    public void TestDownloadAllMessagesToFiles_TargetDirectoryNotExist()
    {
      const string dir = "downloaded";

      try {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);

        TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
          server.EnqueueResponse("LIST\r\n",
                                 "+OK\r\n" +
                                 "1 200\r\n" +
                                 "2 200\r\n" +
                                 "3 200\r\n" +
                                 ".\r\n");

          Assert.Throws<DirectoryNotFoundException>(delegate {
            client.DownloadAllMessagesToFiles(delegate(PopMessageInfo msg) {
              return Path.Combine(dir, msg.MessageNumber.ToString("D4"));
            });
          });
        });
      }
      finally {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);
      }
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadAllMessagesToFilesAsync_TargetDirectoryNotExist()
    {
      const string dir = "downloaded";

      try {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);

        TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
          server.EnqueueResponse("LIST\r\n",
                                 "+OK\r\n" +
                                 "1 200\r\n" +
                                 "2 200\r\n" +
                                 "3 200\r\n" +
                                 ".\r\n");

          var ex = Assert.Throws<AggregateException>(delegate {
            client.DownloadAllMessagesToFilesAsync(delegate(PopMessageInfo msg) {
              return Path.Combine(dir, msg.MessageNumber.ToString("D4"));
            }).Wait();
          });

          Assert.IsInstanceOf<DirectoryNotFoundException>(ex.Flatten().InnerException);
        });
      }
      finally {
        if (Directory.Exists(dir))
          Directory.Delete(dir, true);
      }
    }
#endif

    [Test]
    public void TestDownloadAllMessagesToFiles_ArgumentNull()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentNullException>(delegate {
          client.DownloadAllMessagesToFiles(null);
        });
      });
    }

#if ASYNC_OPS
    [Test]
    public void TestDownloadAllMessagesToFilesAsync_ArgumentNull()
    {
      TestUtils.TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        Assert.Throws<ArgumentNullException>(delegate {
          client.DownloadAllMessagesToFilesAsync(null);
        });
      });
    }
#endif
  }
}