using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using NUnit.Framework;

using Smdn.Net.Pop3.Client.Session;
using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.Client;

namespace Smdn.Net.Pop3.Client {
  [TestFixture]
  public class PopClientOperationsOneShotTests {
    [Test]
    public void TestInternalDownloadAppropriateAuthMechanismNotFound()
    {
      using (var server = PopPseudoServer.StartNew()) {
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");

        Assert.Throws<PopNoAppropriateAuthMechanismException>(() => PopClient.GetMessageCount(new Uri("pop://user@" + server.HostPort), "pass"));
      }
    }

    [Test]
    public void TestInternalDownloadAuthenticateWithSpecifiedMechanism()
    {
      using (var server = PopPseudoServer.StartNew()) {
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL PLAIN\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH PLAIN ",
                               "+OK\r\n");
        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n",
                               true);

        PopClient.GetMessageCount(new Uri("pop://user;AUTH=PLAIN@" + server.HostPort), "pass");
      }
    }

    [Test]
    public void TestInternalDownloadUpgradeStream()
    {
      using (var server = PopPseudoServer.StartNew()) {
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "STLS\r\n" +
                               ".\r\n");
        server.EnqueueResponse("STLS\r\n",
                               "-ERR\r\n");

        Assert.Throws<PopUpgradeConnectionException>(delegate {
          PopClient.GetMessageCount(new Uri("pop://user@" + server.HostPort), "pass");
        });
      }
    }

    [Test]
    public void TestInternalDownloadSecurePort()
    {
      try {
        using (var server = PopPseudoServer.StartNew()) {
          // CAPA
          server.EnqueueResponse("+OK\r\n" +
                                 ".\r\n");

          Assert.Throws<PopUpgradeConnectionException>(delegate {
            PopClient.GetMessageCount(new Uri("pops://user@" + server.HostPort), "pass");
          });
        }
      }
      catch (ApplicationException ex) {
        if (Runtime.IsRunningOnMono)
          Assert.IsInstanceOf<System.Net.Sockets.SocketException>(ex.InnerException);
        else
          throw; // unexpected exception
      }

    }

    private void TestOneShot(Action<PopPseudoServer> testAction)
    {
      using (var server = PopPseudoServer.StartNew()) {
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               "SASL CRAM-MD5\r\n" +
                               ".\r\n");
        server.EnqueueResponse("AUTH CRAM-MD5\r\n",
                               "+ PDQwMDEzNDQxMTIxNDM1OTQuMTI3MjQ5OTU1MEBsb2NhbGhvc3Q+\r\n");
        server.EnqueueResponse("+OK\r\n");

        testAction(server);
      }
    }

    [Test]
    public void TestGetMessageCount()
    {
      TestOneShot(delegate(PopPseudoServer server) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n",
                               true);

        var ret = PopClient.GetMessageCount(new Uri("pop://user@" + server.HostPort), "pass");

        Assert.AreEqual(5, ret);
      });
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestDownloadFirstMessageAsText(bool deleteAfterDownload)
    {
      TestOneShot(delegate(PopPseudoServer server) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("LIST 1\r\n",
                               "+OK 1 200\r\n");
        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               "message#1\r\n" +
                               ".\r\n");

        if (deleteAfterDownload)
          server.EnqueueResponse("DELE 1\r\n",
                                 "+OK\r\n");

        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n",
                               true);

        var ret = PopClient.DownloadFirstMessageAsText(new Uri("pop://user@" + server.HostPort),
                                                       "pass",
                                                       deleteAfterDownload,
                                                       Encoding.UTF8);

        Assert.AreEqual("message#1\r\n", ret);
      });
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestDownloadLastMessageAsText(bool deleteAfterDownload)
    {
      TestOneShot(delegate(PopPseudoServer server) {
        server.EnqueueResponse("STAT\r\n",
                               "+OK 5 1200\r\n");
        server.EnqueueResponse("LIST 5\r\n",
                               "+OK 5 200\r\n");
        server.EnqueueResponse("RETR 5\r\n",
                               "+OK\r\n" +
                               "message#5\r\n" +
                               ".\r\n");

        if (deleteAfterDownload)
          server.EnqueueResponse("DELE 5\r\n",
                                 "+OK\r\n");

        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n",
                               true);

        var ret = PopClient.DownloadLastMessageAsText(new Uri("pop://user@" + server.HostPort),
                                                      "pass",
                                                      deleteAfterDownload,
                                                      Encoding.UTF8);

        Assert.AreEqual("message#5\r\n", ret);
      });
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestDownloadAllMessagesAsText(bool deleteAfterDownload)
    {
      TestOneShot(delegate(PopPseudoServer server) {
        server.EnqueueResponse("LIST\r\n",
                               "+OK\r\n" +
                               "1 120\r\n" +
                               "2 120\r\n" +
                               "3 120\r\n" +
                               ".\r\n");
        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               "message#1\r\n" +
                               ".\r\n");
        if (deleteAfterDownload)
          server.EnqueueResponse("DELE 1\r\n",
                                 "+OK\r\n");

        server.EnqueueResponse("RETR 2\r\n",
                               "+OK\r\n" +
                               "message#2\r\n" +
                               ".\r\n");
        if (deleteAfterDownload)
          server.EnqueueResponse("DELE 2\r\n",
                                 "+OK\r\n");

        server.EnqueueResponse("RETR 3\r\n",
                               "+OK\r\n" +
                               "message#3\r\n" +
                               ".\r\n");
        if (deleteAfterDownload)
          server.EnqueueResponse("DELE 3\r\n",
                                 "+OK\r\n");

        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n",
                               true);

        var ret = PopClient.DownloadAllMessagesAsText(new Uri("pop://user@" + server.HostPort),
                                                      "pass",
                                                      deleteAfterDownload,
                                                      Encoding.UTF8);

        Assert.IsNotEmpty(ret);
        Assert.AreEqual(3, ret.Length);
        Assert.AreEqual("message#1\r\n", ret[0]);
        Assert.AreEqual("message#2\r\n", ret[1]);
        Assert.AreEqual("message#3\r\n", ret[2]);
      });
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestDownloadFirstMessageToFile(bool deleteAfterDownload)
    {
      const string dest = "download.eml";

      try {
        if (File.Exists(dest))
          File.Delete(dest);

        TestOneShot(delegate(PopPseudoServer server) {
          server.EnqueueResponse("STAT\r\n",
                                 "+OK 5 1200\r\n");
          server.EnqueueResponse("LIST 1\r\n",
                                 "+OK 1 200\r\n");
          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 "message#1\r\n" +
                                 ".\r\n");

          if (deleteAfterDownload)
            server.EnqueueResponse("DELE 1\r\n",
                                   "+OK\r\n");

          server.EnqueueResponse("QUIT\r\n",
                                 "+OK\r\n",
                                 true);

          var file = PopClient.DownloadFirstMessageToFile(new Uri("pop://user@" + server.HostPort),
                                                          "pass",
                                                          deleteAfterDownload,
                                                          dest);

          Assert.IsTrue(file.Exists, "downloaded file exists");
          StringAssert.EndsWith(dest, file.FullName, "downloaded file name");
          Assert.AreEqual("message#1\r\n", File.ReadAllText(file.FullName, Encoding.UTF8));
        });
      }
      finally {
        if (File.Exists(dest))
          File.Delete(dest);
      }
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestDownloadLastMessageToFile(bool deleteAfterDownload)
    {
      const string dest = "download.eml";

      try {
        if (File.Exists(dest))
          File.Delete(dest);

        TestOneShot(delegate(PopPseudoServer server) {
          server.EnqueueResponse("STAT\r\n",
                                 "+OK 5 1200\r\n");
          server.EnqueueResponse("LIST 5\r\n",
                                 "+OK 5 200\r\n");
          server.EnqueueResponse("RETR 5\r\n",
                                 "+OK\r\n" +
                                 "message#5\r\n" +
                                 ".\r\n");

          if (deleteAfterDownload)
            server.EnqueueResponse("DELE 5\r\n",
                                   "+OK\r\n");

          server.EnqueueResponse("QUIT\r\n",
                                 "+OK\r\n",
                                 true);

          var file = PopClient.DownloadLastMessageToFile(new Uri("pop://user@" + server.HostPort),
                                                         "pass",
                                                         deleteAfterDownload,
                                                         dest);

          Assert.IsTrue(file.Exists, "downloaded file exists");
          StringAssert.EndsWith(dest, file.FullName, "downloaded file name");
          Assert.AreEqual("message#5\r\n", File.ReadAllText(file.FullName, Encoding.UTF8));
        });
      }
      finally {
        if (File.Exists(dest))
          File.Delete(dest);
      }
    }

    [TestCase(true)]
    [TestCase(false)]
    public void TestDownloadAllMessagesToDirectory(bool deleteAfterDownload)
    {
      const string destdir = "download";

      try {
        if (Directory.Exists(destdir))
          Directory.Delete(destdir, true);

        TestOneShot(delegate(PopPseudoServer server) {
          server.EnqueueResponse("LIST\r\n",
                                 "+OK\r\n" +
                                 "1 120\r\n" +
                                 "2 120\r\n" +
                                 "3 120\r\n" +
                                 ".\r\n");
          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 "message#1\r\n" +
                                 ".\r\n");
          if (deleteAfterDownload)
            server.EnqueueResponse("DELE 1\r\n",
                                   "+OK\r\n");

          server.EnqueueResponse("RETR 2\r\n",
                                 "+OK\r\n" +
                                 "message#2\r\n" +
                                 ".\r\n");
          if (deleteAfterDownload)
            server.EnqueueResponse("DELE 2\r\n",
                                   "+OK\r\n");

          server.EnqueueResponse("RETR 3\r\n",
                                 "+OK\r\n" +
                                 "message#3\r\n" +
                                 ".\r\n");
          if (deleteAfterDownload)
            server.EnqueueResponse("DELE 3\r\n",
                                   "+OK\r\n");

          server.EnqueueResponse("QUIT\r\n",
                                 "+OK\r\n");

          var files = PopClient.DownloadAllMessagesToDirectory(new Uri("pop://user@" + server.HostPort),
                                                               "pass",
                                                               deleteAfterDownload,
                                                               destdir);

          Assert.IsTrue(Directory.Exists(destdir));

          Assert.IsTrue(files[0].Exists, "downloaded file exists #0");
          StringAssert.EndsWith(Path.Combine(destdir, "1.eml"), files[0].FullName, "downloaded file name #0");
          Assert.AreEqual("message#1\r\n", File.ReadAllText(files[0].FullName, Encoding.UTF8));

          Assert.IsTrue(files[1].Exists, "downloaded file exists #1");
          StringAssert.EndsWith(Path.Combine(destdir, "2.eml"), files[1].FullName, "downloaded file name #1");
          Assert.AreEqual("message#2\r\n", File.ReadAllText(files[1].FullName, Encoding.UTF8));

          Assert.IsTrue(files[2].Exists, "downloaded file exists #2");
          StringAssert.EndsWith(Path.Combine(destdir, "3.eml"), files[2].FullName, "downloaded file name #2");
          Assert.AreEqual("message#3\r\n", File.ReadAllText(files[2].FullName, Encoding.UTF8));
        });
      }
      finally {
        if (Directory.Exists(destdir))
          Directory.Delete(destdir, true);
      }
    }
  }
}