using System;
using System.IO;
using System.Text;
using NUnit.Framework;

using Smdn.IO;
using Smdn.Net.Pop3.Client.Session;

namespace Smdn.Net.Pop3.Client {
  [TestFixture]
  public class PopMessageInfoTests {
    private static void GetTestMessageBody(out string messageBody, out string messageByteStuffedBody, out int octets)
    {
      messageBody = @"MIME-Version: 1.0
Content-Type: text/plain

..byte-stuffed
+OK response style line
-ERR response style line2
.
end of message
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      messageByteStuffedBody = TestUtils.GetByteStuffed(messageBody);

      octets = NetworkTransferEncoding.Transfer8Bit.GetByteCount(messageBody);
    }

    [Test]
    public void TestPropertyGetClientSessionClosed()
    {
      TestUtils.TestClosedSessionMessage(delegate(PopMessageInfo message) {
        Assert.Throws<InvalidOperationException>(() => Assert.IsNull(message.Client));
      });
    }

    [Test]
    public void TestPropertyGetMessageNumberSessionClosed()
    {
      TestUtils.TestClosedSessionMessage(delegate(PopMessageInfo message) {
        Assert.Throws<InvalidOperationException>(() => Assert.AreEqual(0L, message.MessageNumber));
      });
    }

    [Test]
    public void TestPropertyGetLengthSessionClosed()
    {
      TestUtils.TestClosedSessionMessage(delegate(PopMessageInfo message) {
        Assert.Throws<InvalidOperationException>(() => Assert.AreEqual(0L, message.Length));
      });
    }

    [Test]
    public void TestPropertyGetIsMarkedAsDeletedSessionClosed()
    {
      TestUtils.TestClosedSessionMessage(delegate(PopMessageInfo message) {
        Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(message.IsMarkedAsDeleted));
      });
    }

    [Test]
    public void TestPropertyGetUniqueIdServerSupportsUidl()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        server.EnqueueResponse("UIDL 1\r\n",
                               "+OK 1 whqtswO00WBw418f9t5JxYwZ\r\n");

        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);

        // retrieve again
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);
      });
    }

    [Test]
    public void TestPropertyGetUniqueIdServerNotSupportsUidl()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        server.EnqueueResponse("UIDL 1\r\n",
                               "-ERR\r\n");

        Assert.IsNull(message.UniqueId);

        // retrieve again
        Assert.IsNull(message.UniqueId);
      });
    }

    [Test]
    public void TestPropertyGetUniqueIdDeleted()
    {
      TestUtils.TestDeletedMessage(delegate(PopMessageInfo message) {
        try {
          Assert.IsNotNull(message.UniqueId);
          Assert.Fail("PopMessageDeletedException not thrown");
        }
        catch (PopMessageDeletedException ex) {
          Assert.AreEqual(message.MessageNumber, ex.MessageNumber);
        }
      });
    }

    [Test]
    public void TestPropertyGetUniqueIdSessionClosed()
    {
      TestUtils.TestClosedSessionMessage(delegate(PopMessageInfo message) {
        Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(message.UniqueId));
      });
    }

    [Test]
    public void TestMarkAsDeleted()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        server.EnqueueResponse("DELE 1\r\n",
                               "+OK\r\n");

        message.MarkAsDeleted();

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        Assert.IsTrue(message.IsMarkedAsDeleted);

        server.EnqueueResponse("STAT\r\n",
                               "+OK 0 0\r\n");

        Assert.AreEqual(0L, message.Client.MessageCount);
        Assert.AreEqual(0L, message.Client.TotalSize);

        // delete again
        message.MarkAsDeleted();

        Assert.IsTrue(message.IsMarkedAsDeleted);
      });
    }

    [Test]
    public void TestMarkAsDeletedSessionClosed()
    {
      TestUtils.TestClosedSessionMessage(delegate(PopMessageInfo message) {
        Assert.Throws<InvalidOperationException>(() => message.MarkAsDeleted());
      });
    }

    [Test]
    public void TestGetUniqueIdServerSupportsUidl()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        server.EnqueueResponse("UIDL 1\r\n",
                               "+OK 1 whqtswO00WBw418f9t5JxYwZ\r\n");

        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.GetUniqueId());

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        // retrieve again
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.GetUniqueId());
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);
      });
    }

    [Test]
    public void TestGetUniqueIdServerNotSupportsUidl()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        server.EnqueueResponse("UIDL 1\r\n",
                               "-ERR\r\n");

        var ex = Assert.Throws<PopIncapableException>(() => message.GetUniqueId());
        Assert.AreEqual(PopCapability.Uidl, ex.RequiredCapability);

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        // retrieve again
        ex = Assert.Throws<PopIncapableException>(() => message.GetUniqueId());
        Assert.AreEqual(PopCapability.Uidl, ex.RequiredCapability);

        Assert.IsNull(message.UniqueId);
      });
    }

    [Test]
    public void TestTryGetUniqueIdServerSupportsUidl()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        server.EnqueueResponse("UIDL 1\r\n",
                               "+OK 1 whqtswO00WBw418f9t5JxYwZ\r\n");

        string uniqueId;

        Assert.IsTrue(message.TryGetUniqueId(out uniqueId), "method return value 1");
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", uniqueId);

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        // retrieve again
        Assert.IsTrue(message.TryGetUniqueId(out uniqueId), "method return value 2");
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", uniqueId);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);
      });
    }

    [Test]
    public void TestTryGetUniqueIdServerNotSupportsUidl()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        // UIDL
        server.EnqueueResponse("UIDL 1\r\n",
                               "-ERR\r\n");

        string uniqueId;

        Assert.IsFalse(message.TryGetUniqueId(out uniqueId), "method return value 1");
        Assert.IsNull(uniqueId);

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        // retrieve again
        Assert.IsFalse(message.TryGetUniqueId(out uniqueId), "method return value 1");
        Assert.IsNull(uniqueId);
        Assert.IsNull(message.UniqueId);
      });
    }

    [Test]
    public void TestTryGetUniqueIdSessionClosed()
    {
      TestUtils.TestClosedSessionMessage(delegate(PopMessageInfo message) {
        string discard;

        Assert.Throws<InvalidOperationException>(() => message.TryGetUniqueId(out discard));
      });
    }

    [Test]
    public void TestTryGetUniqueIdDeleted()
    {
      TestUtils.TestDeletedMessage(delegate(PopMessageInfo message) {
        string discard;

        Assert.Throws<PopMessageDeletedException>(() => message.TryGetUniqueId(out discard));
      });
    }

    [Test]
    public void TestOpenReadRetr()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        message.Client.DeleteAfterRetrieve = false;

        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        using (var stream = message.OpenRead()) {
          Assert.IsFalse(message.Client.IsBusy, "IsBusy");

          Assert.IsFalse(message.IsMarkedAsDeleted);

          Assert.IsNotNull(stream);
          Assert.AreEqual(octets, stream.Length);
          Assert.AreEqual(messageBody, (new StreamReader(stream, NetworkTransferEncoding.Transfer7Bit)).ReadToEnd());
        }
      });
    }

    [Test]
    public void TestOpenReadTop()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        message.Client.DeleteAfterRetrieve = false;

        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("TOP 1 3\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        using (var stream = message.OpenRead(3)) {
          Assert.IsFalse(message.Client.IsBusy, "IsBusy");

          Assert.IsFalse(message.IsMarkedAsDeleted);

          Assert.IsNotNull(stream);
          Assert.AreEqual(octets, stream.Length);
          Assert.AreEqual(messageBody, (new StreamReader(stream, NetworkTransferEncoding.Transfer7Bit)).ReadToEnd());
        }
      });
    }

    [Test]
    public void TestOpenReadSessionClosed()
    {
      TestUtils.TestClosedSessionMessage(delegate(PopMessageInfo message) {
        Assert.Throws<InvalidOperationException>(() => message.OpenRead());
      });
    }

    [Test]
    public void TestOpenReadDeleteAfterRetrieve()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        message.Client.DeleteAfterRetrieve = true;

        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");
        server.EnqueueResponse("DELE 1\r\n",
                               "+OK\r\n");

        using (var stream = message.OpenRead()) {
          Assert.IsFalse(message.Client.IsBusy, "IsBusy");

          Assert.IsTrue(message.IsMarkedAsDeleted);

          Assert.IsNotNull(stream);
          Assert.AreEqual(octets, stream.Length);
          Assert.AreEqual(messageBody, (new StreamReader(stream, NetworkTransferEncoding.Transfer7Bit)).ReadToEnd());

          server.EnqueueResponse("STAT\r\n",
                                 "+OK 0 0\r\n");

          Assert.AreEqual(0L, message.Client.MessageCount);
          Assert.AreEqual(0L, message.Client.TotalSize);
        }
      });
    }

    private void TestReadAs(Action<PopMessageInfo, int, string> test)
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        test(message, octets, messageBody);
      });
    }

    [Test]
    public void TestReadAsStreamConverter()
    {
      TestReadAs(delegate(PopMessageInfo message, int octets, string expectedMessageBody) {
        var ret = message.ReadAs(delegate(Stream stream) {
          Assert.IsNotNull(stream);
          Assert.AreEqual(octets, stream.Length);

          return stream.ReadToEnd();
        });

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        Assert.AreEqual(NetworkTransferEncoding.Transfer7Bit.GetBytes(expectedMessageBody),
                        ret);
      });
    }

    [Test]
    public void TestReadAsStreamFunc1()
    {
      TestReadAs(delegate(PopMessageInfo message, int octets, string expectedMessageBody) {
        var ret = message.ReadAs(delegate(Stream stream, string arg1) {
          Assert.IsNotNull(stream);
          Assert.AreEqual(octets, stream.Length);

          Assert.AreEqual("arg1", arg1);

          return stream.ReadToEnd();
        }, "arg1");

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        Assert.AreEqual(NetworkTransferEncoding.Transfer7Bit.GetBytes(expectedMessageBody),
                        ret);
      });
    }

    [Test]
    public void TestReadAsStreamFunc2()
    {
      TestReadAs(delegate(PopMessageInfo message, int octets, string expectedMessageBody) {
        var ret = message.ReadAs(delegate(Stream stream, string arg1, string arg2) {
          Assert.IsNotNull(stream);
          Assert.AreEqual(octets, stream.Length);

          Assert.AreEqual("arg1", arg1);
          Assert.AreEqual("arg2", arg2);

          return stream.ReadToEnd();
        }, "arg1", "arg2");

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        Assert.AreEqual(NetworkTransferEncoding.Transfer7Bit.GetBytes(expectedMessageBody),
                        ret);
      });
    }

    [Test]
    public void TestReadAsStreamFunc3()
    {
      TestReadAs(delegate(PopMessageInfo message, int octets, string expectedMessageBody) {
        var ret = message.ReadAs(delegate(Stream stream, string arg1, string arg2, string arg3) {
          Assert.IsNotNull(stream);
          Assert.AreEqual(octets, stream.Length);

          Assert.AreEqual("arg1", arg1);
          Assert.AreEqual("arg2", arg2);
          Assert.AreEqual("arg3", arg3);

          return stream.ReadToEnd();
        }, "arg1", "arg2", "arg3");

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        Assert.AreEqual(NetworkTransferEncoding.Transfer7Bit.GetBytes(expectedMessageBody),
                        ret);
      });
    }

    [Test]
    public void TestReadAsStreamFunc4()
    {
      TestReadAs(delegate(PopMessageInfo message, int octets, string expectedMessageBody) {
        var ret = message.ReadAs(delegate(Stream stream, string arg1, string arg2, string arg3, string arg4) {
          Assert.IsNotNull(stream);
          Assert.AreEqual(octets, stream.Length);

          Assert.AreEqual("arg1", arg1);
          Assert.AreEqual("arg2", arg2);
          Assert.AreEqual("arg3", arg3);
          Assert.AreEqual("arg4", arg4);

          return stream.ReadToEnd();
        }, "arg1", "arg2", "arg3", "arg4");

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        Assert.AreEqual(NetworkTransferEncoding.Transfer7Bit.GetBytes(expectedMessageBody),
                        ret);
      });
    }

    [Test]
    public void TestReadAllBytes()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        Assert.AreEqual(NetworkTransferEncoding.Transfer7Bit.GetBytes(messageBody),
                        message.ReadAllBytes());
      });
    }

    [Test]
    public void TestSave()
    {
      const string outfile = "retrieved.eml";

      try {
        File.Delete(outfile);

        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          int octets;
          string messageBody, byteStuffedMessageBody;

          GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          message.Save(outfile);

          Assert.IsFalse(message.Client.IsBusy, "IsBusy");

          Assert.IsTrue(File.Exists(outfile));
          Assert.AreEqual(octets, (new FileInfo(outfile)).Length);
          Assert.AreEqual(NetworkTransferEncoding.Transfer7Bit.GetBytes(messageBody),
                          File.ReadAllBytes(outfile));
        });
      }
      finally {
        File.Delete(outfile);
      }
    }

    [Test]
    public void TestSave_OverwriteExistingFile()
    {
      const string outfile = "retrieved.eml";
      const int existingFileLength = 10 * 1024;

      try {
        File.Delete(outfile);

        File.WriteAllText(outfile, new string('x', existingFileLength));

        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          int octets;
          string messageBody, byteStuffedMessageBody;

          GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

          Assert.Less(octets, (new FileInfo(outfile)).Length, "less than existing file length");

          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          message.Save(outfile);

          Assert.IsFalse(message.Client.IsBusy, "IsBusy");

          Assert.IsTrue(File.Exists(outfile), "file existence");
          Assert.AreEqual(octets, (new FileInfo(outfile)).Length, "file length");
          Assert.AreEqual(NetworkTransferEncoding.Transfer7Bit.GetBytes(messageBody),
                          File.ReadAllBytes(outfile));
        });
      }
      finally {
        File.Delete(outfile);
      }
    }

    [Test]
    public void TestSave_PathNull()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.Save(null));
        Assert.IsFalse(message.Client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestSave_PathEmpty()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentException>(() => message.Save(string.Empty));
        Assert.IsFalse(message.Client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestSave_CanNotOpenRetrieveStream()
    {
      const string outfile = "retrieved.eml";

      try {
        File.Delete(outfile);

        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          server.EnqueueResponse("RETR 1\r\n",
                                 "-ERR no such message\r\n");

          Assert.Throws<PopErrorResponseException>(() => message.Save(outfile));

          Assert.IsTrue(File.Exists(outfile), "file existence");
        });
      }
      finally {
        File.Delete(outfile);
      }
    }

    [Test]
    public void TestSave_FileNotCreatedOnException_ArgumentOutOfRange()
    {
      const string outfile = "retrieved.eml";

      try {
        File.Delete(outfile);

        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          Assert.Throws<ArgumentOutOfRangeException>(() => message.Save(outfile, -2));

          Assert.IsFalse(message.Client.IsBusy, "IsBusy");

          Assert.IsFalse(File.Exists(outfile));
        });
      }
      finally {
        File.Delete(outfile);
      }
    }

    [Test]
    public void TestSave_FileNotCreatedOnException_MessageDeleted()
    {
      const string outfile = "retrieved.eml";

      try {
        File.Delete(outfile);

        TestUtils.TestDeletedMessage(delegate(PopMessageInfo message) {
          var ex = Assert.Throws<PopMessageDeletedException>(() => message.Save(outfile));

          Assert.IsFalse(message.Client.IsBusy, "IsBusy");

          Assert.AreEqual(message.MessageNumber, ex.MessageNumber);

          Assert.IsFalse(File.Exists(outfile));
        });
      }
      finally {
        File.Delete(outfile);
      }
    }

    [Test]
    public void TestSave_FileNotCreatedOnException_SessionClosed()
    {
      const string outfile = "retrieved.eml";

      try {
        File.Delete(outfile);

        TestUtils.TestClosedSessionMessage(delegate(PopMessageInfo message) {
          Assert.Throws<InvalidOperationException>(() => message.Save(outfile));

          Assert.IsFalse(File.Exists(outfile));
        });
      }
      finally {
        File.Delete(outfile);
      }
    }

    [Test]
    public void TestWriteToStream()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        var outputStream = new MemoryStream();

        message.WriteTo(outputStream);

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        Assert.AreEqual(octets, outputStream.Length);

        outputStream.Close();

        Assert.AreEqual(NetworkTransferEncoding.Transfer7Bit.GetBytes(messageBody),
                        outputStream.ToArray());
      });
    }

    [Test]
    public void TestWriteToStream_DestinationNull()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.WriteTo((Stream)null));
        Assert.IsFalse(message.Client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestWriteToStream_DestinationNotWritable()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        using (var stream = new MemoryStream(new byte[0], false)) {
          Assert.Throws<ArgumentException>(() => message.WriteTo(stream));
          Assert.IsFalse(message.Client.IsBusy, "IsBusy");
        }
      });
    }

    [Test]
    public void TestWriteToBinaryWriter()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        var outputStream = new MemoryStream();
        var writer = new BinaryWriter(outputStream);

        message.WriteTo(writer);

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        Assert.AreEqual(octets, outputStream.Length);

        outputStream.Close();

        Assert.AreEqual(NetworkTransferEncoding.Transfer7Bit.GetBytes(messageBody),
                        outputStream.ToArray());
      });
    }

    [Test]
    public void TestWriteToBinaryWriter_DestinationNull()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.WriteTo((BinaryWriter)null));
        Assert.IsFalse(message.Client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestOpenText()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        using (var reader = message.OpenText(Encoding.UTF8)) {
          Assert.IsFalse(message.Client.IsBusy, "IsBusy");

          Assert.IsNotNull(reader);
          Assert.AreEqual(Encoding.UTF8, reader.CurrentEncoding);
          Assert.AreEqual(messageBody, reader.ReadToEnd());
        }
      });
    }

    [Test]
    public void TestOpenText_DefaultCharset()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        using (var reader = message.OpenText()) {
          Assert.IsFalse(message.Client.IsBusy, "IsBusy");

          Assert.IsNotNull(reader);
          Assert.AreEqual(Encoding.GetEncoding("latin1"), reader.CurrentEncoding);
          Assert.AreEqual(messageBody, reader.ReadToEnd());
        }
      });
    }

    [Test]
    public void TestOpenText_CharsetNull()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.OpenText((Encoding)null));
        Assert.IsFalse(message.Client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestReadAllLines()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        Assert.AreEqual(messageBody.TrimEnd(),
                        string.Join("\r\n", message.ReadAllLines()));

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestReadAllText()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        Assert.AreEqual(messageBody,
                        message.ReadAllText());

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");
      });
    }

    [Test]
    public void TestOpenReadDeleted()
    {
      TestUtils.TestDeletedMessage(delegate(PopMessageInfo message) {
        try {
          message.OpenRead();
          Assert.Fail("PopMessageDeletedException not thrown");
        }
        catch (PopMessageDeletedException ex) {
          Assert.AreEqual(message.MessageNumber, ex.MessageNumber);
        }

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");
      });
    }
  }
}
