#if NET_4_0
#define ASYNC_OPS
#endif

#if ASYNC_OPS
using System;
using System.IO;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using NUnit.Framework;

using Smdn.IO;
using Smdn.Net.Pop3.Client.Session;

namespace Smdn.Net.Pop3.Client {
  [TestFixture]
  public class PopMessageInfoOperationsAsyncTests {
    [Test]
    public void TestMarkAsDeletedAsync()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        var t = message.MarkAsDeletedAsync();

        Assert.IsTrue(message.Client.IsBusy, "IsBusy task started");
        Assert.IsFalse(message.IsMarkedAsDeleted, "IsMarkedAsDeleted task started");

        server.EnqueueResponse("DELE 1\r\n",
                               "+OK\r\n");

        t.Wait();

        Assert.IsFalse(message.Client.IsBusy, "IsBusy task finished");
        Assert.IsTrue(message.IsMarkedAsDeleted, "IsMarkedAsDeleted task finished");

        server.EnqueueResponse("STAT\r\n",
                               "+OK 0 0\r\n");

        Assert.AreEqual(0L, message.Client.MessageCount);
        Assert.AreEqual(0L, message.Client.TotalSize);

        // delete again
        message.MarkAsDeletedAsync().Wait();

        Assert.IsTrue(message.IsMarkedAsDeleted);
      });
    }

    [Test]
    public void TestGetUniqueIdAsync_ServerSupportsUidl()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        var t = message.GetUniqueIdAsync();

        Assert.IsTrue(message.Client.IsBusy, "IsBusy task started #1");

        server.EnqueueResponse("UIDL 1\r\n",
                               "+OK 1 whqtswO00WBw418f9t5JxYwZ\r\n");

        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", t.Result);
        Assert.IsFalse(message.Client.IsBusy, "IsBusy task finished #1");

        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);

        // retrieve again
        t = message.GetUniqueIdAsync();

        Assert.IsFalse(message.Client.IsBusy, "IsBusy task started #1");

        Assert.IsTrue(t.IsCompleted);
        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", t.Result);
        Assert.IsFalse(message.Client.IsBusy, "IsBusy task finished #2");

        Assert.AreEqual("whqtswO00WBw418f9t5JxYwZ", message.UniqueId);
      });
    }

    [Test]
    public void TestGetUniqueIdAsync_ServerNotSupportsUidl()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        var t = message.GetUniqueIdAsync();

        Assert.IsTrue(message.Client.IsBusy, "IsBusy task started #1");

        server.EnqueueResponse("UIDL 1\r\n",
                               "-ERR\r\n");

        var ex = Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsInstanceOf<PopIncapableException>(ex.InnerException);
        Assert.AreEqual(PopCapability.Uidl, (ex.InnerException as PopIncapableException).RequiredCapability);
        Assert.IsFalse(message.Client.IsBusy, "IsBusy task finished #1");

        // retrieve again
        var ex2 = Assert.Throws<PopIncapableException>(() => message.GetUniqueIdAsync());

        Assert.IsFalse(message.Client.IsBusy, "IsBusy task not started");

        Assert.AreEqual(PopCapability.Uidl, ex2.RequiredCapability);

        Assert.IsNull(message.UniqueId);
      });
    }

    [Test]
    public void TestGetUniqueIdAsync_SessionClosed()
    {
      TestUtils.TestClosedSessionMessage(delegate(PopMessageInfo message) {
        Assert.Throws<InvalidOperationException>(() => message.GetUniqueIdAsync());
      });
    }

    [Test]
    public void TestGetUniqueIdAsync_Deleted()
    {
      TestUtils.TestDeletedMessage(delegate(PopMessageInfo message) {
        Assert.Throws<PopMessageDeletedException>(() => message.GetUniqueIdAsync());
      });
    }

    private static void GetTestMessageBody(out string messageBody, out string messageByteStuffedBody, out int octets)
    {
      messageBody = @"MIME-Version: 1.0
Content-Type: text/plain

..byte-stuffed
+OK response style line
-ERR response style line2
.
end of message
".Replace("\r\n", "\n").Replace("\n", "\r\n");

      messageByteStuffedBody = TestUtils.GetByteStuffed(messageBody);

      octets = NetworkTransferEncoding.Transfer8Bit.GetByteCount(messageBody);
    }

    [Test]
    public void TestOpenReadAsync()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        message.Client.DeleteAfterRetrieve = false;

        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        var t = message.OpenReadAsync();

        Assert.IsFalse(t.IsCanceled, "task.IsCanceled #1");
        Assert.IsFalse(t.IsFaulted, "task.IsFaulted #1");
        Assert.IsFalse(t.IsCompleted, "task.IsCompleted #1");

        Assert.IsTrue(message.Client.IsBusy, "IsBusy task started");

        Assert.Throws<InvalidOperationException>(() => message.Client.KeepAlive());
        Assert.Throws<InvalidOperationException>(() => message.Client.KeepAliveAsync());

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");

        using (var stream = t.Result) {
          Assert.IsFalse(message.Client.IsBusy, "IsBusy task finished");

          Assert.IsFalse(t.IsCanceled, "task.IsCanceled #2");
          Assert.IsFalse(t.IsFaulted, "task.IsFaulted #2");
          Assert.IsTrue(t.IsCompleted, "task.IsCompleted #2");

          Assert.IsNotNull(stream);
          Assert.AreEqual(octets, stream.Length);
          Assert.AreEqual(messageBody,
                          (new StreamReader(stream, NetworkTransferEncoding.Transfer8Bit)).ReadToEnd());
        }
      });
    }

    [Test]
    public void TestOpenReadAsync_DeleteAfterRetrieve()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        message.Client.DeleteAfterRetrieve = true;

        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        var t = message.OpenReadAsync();

        Assert.IsFalse(t.IsCanceled, "task.IsCanceled #1");
        Assert.IsFalse(t.IsFaulted, "task.IsFaulted #1");
        Assert.IsFalse(t.IsCompleted, "task.IsCompleted #1");

        Assert.IsTrue(message.Client.IsBusy, "IsBusy task started");

        Assert.Throws<InvalidOperationException>(() => message.Client.KeepAlive());
        Assert.Throws<InvalidOperationException>(() => message.Client.KeepAliveAsync());

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n");
        server.EnqueueResponse("DELE 1\r\n",
                               "+OK\r\n");

        using (var stream = t.Result) {
          Assert.IsFalse(message.Client.IsBusy, "IsBusy task finished");

          Assert.IsTrue(message.IsMarkedAsDeleted);

          Assert.IsNotNull(stream);
          Assert.AreEqual(octets, stream.Length);
          Assert.AreEqual(messageBody, (new StreamReader(stream, NetworkTransferEncoding.Transfer7Bit)).ReadToEnd());

          server.EnqueueResponse("STAT\r\n",
                                 "+OK 0 0\r\n");

          Assert.AreEqual(0L, message.Client.MessageCount);
          Assert.AreEqual(0L, message.Client.TotalSize);
        }
      });
    }

    [Test]
    public void TestOpenReadAsync_Overloads1()
    {
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var openReadAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync()),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(progress)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(CancellationToken.None)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(CancellationToken.None, progress)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(-1)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(-1, progress)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(-1, CancellationToken.None)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(-1, CancellationToken.None, progress)),
      };
#else
      var openReadAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync()),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(CancellationToken.None)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(-1)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(-1, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < openReadAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          message.Client.DeleteAfterRetrieve = false;

          int octets;
          string messageBody, byteStuffedMessageBody;

          GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          try {
            var ret = openReadAsyncMethods[i](message)
              .ContinueWith(t => (new StreamReader(t.Result, NetworkTransferEncoding.Transfer8Bit)).ReadToEnd())
              .Result;

            Assert.IsFalse(message.Client.IsBusy, "IsBusy");

            Assert.AreEqual(messageBody, ret);
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }
        });
      }
    }

    [Test]
    public void TestOpenReadAsync_Overloads2()
    {
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var openReadAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(0)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(0, progress)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(0, CancellationToken.None)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(0, CancellationToken.None, progress)),
      };
#else
      var openReadAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(0)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(0, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < openReadAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          message.Client.DeleteAfterRetrieve = false;

          int octets;
          string messageBody, byteStuffedMessageBody;

          GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

          server.EnqueueResponse("TOP 1 0\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          try {
            var ret = openReadAsyncMethods[i](message)
              .ContinueWith(t => (new StreamReader(t.Result, NetworkTransferEncoding.Transfer8Bit)).ReadToEnd())
              .Result;

            Assert.IsFalse(message.Client.IsBusy, "IsBusy");

            Assert.AreEqual(messageBody, ret);
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }
        });
      }
    }

    [Test]
    public void TestOpenReadAsync_CancellationRequestedBeforeStart()
    {
      TestUtils.Repeat(100, delegate {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          var cts = new CancellationTokenSource();

          cts.Cancel();

          var t = message.OpenReadAsync(cts.Token);

          if (t.IsCanceled)
            Assert.IsFalse(message.Client.IsBusy, "IsBusy task started");

          var ex = Assert.Throws<AggregateException>(() => t.Wait());

          Assert.IsFalse(message.Client.IsBusy, "IsBusy task finished");

          Assert.IsInstanceOf<OperationCanceledException>(ex.InnerException);

          Assert.IsNull(server.DequeueRequest()); // nothing requested
        });
      });
    }

    [Test]
    public void TestOpenReadAsync_CancellationRequestedBeforeYieldResult()
    {
      OpenReadAsync_CancellationRequestedAfterRetrieve(false);
    }

    [Test]
    public void TestOpenReadAsync_CancellationRequestedBeforeMarkAsDelete()
    {
      OpenReadAsync_CancellationRequestedAfterRetrieve(true);
    }

    private void OpenReadAsync_CancellationRequestedAfterRetrieve(bool deleteAfterRetrieve)
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        message.Client.DeleteAfterRetrieve = deleteAfterRetrieve;

        int octets;
        string messageBody, byteStuffedMessageBody;

        GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

        var cts = new CancellationTokenSource();

        var t = message.OpenReadAsync(cts.Token);

        Assert.IsTrue(message.Client.IsBusy, "IsBusy task started");

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               byteStuffedMessageBody +
                               ".\r\n",
                               () => cts.Cancel(),
                               null);

        var ex = Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsFalse(message.Client.IsBusy, "IsBusy task finished");

        Assert.IsInstanceOf<OperationCanceledException>(ex.InnerException);

        server.DequeueRequest(); // RETR
        Assert.IsNull(server.DequeueRequest()); // DELE not requested

        Assert.IsFalse(message.IsMarkedAsDeleted);
      });
    }

#if NET_4_5
    [Test]
    public void TestOpenReadAsync_ReportProgress()
    {
      var messageBody = new StringBuilder();

      messageBody.Append("MIME-Version: 1.0").Append("\r\n");
      messageBody.Append("Content-Type: text/plain").Append("\r\n");
      messageBody.Append("\r\n");

      var messageLine = new string('x', 254); // 254+2 bytes per line

      // 160 lines = 40 * 4 * 256 bytes = 40k bytes
      for (var i = 0; i < 160; i++) {
        // 256 bytes per line
        messageBody.Append(messageLine).Append("\r\n");
      }

      TestUtils.TestMessage(1L, messageBody.Length, delegate(PopPseudoServer server, PopMessageInfo message) {
        message.Client.DeleteAfterRetrieve = false;

        var cts = new CancellationTokenSource();
        var lastProgressPercentage = int.MinValue;
        var lastProgressLineCount = long.MinValue;
        var reportedCount = 0;
        var reportedFinish = false;
        var progress = new Progress<PopMessageRetrieveProgressInfo>(p => {
          Interlocked.Increment(ref reportedCount);

          if (p.RetrievedOctetCount == p.TotalOctetCount)
            reportedFinish = true;

          Assert.AreEqual(cts.Token, p.CancellationToken, "CancellationToken #{0}", reportedCount);
          Assert.AreEqual(messageBody.Length, p.TotalOctetCount, "TotalOctetCount #{0}", reportedCount);
          Assert.GreaterOrEqual(p.ProgressPercentage, 0, "0 <= ProgressPercentage #{0}", reportedCount);
          Assert.LessOrEqual(p.ProgressPercentage, 100, "ProgressPercentage <= 100 #{0}", reportedCount);

          if (SynchronizationContext.Current != null && lastProgressPercentage < 100) {
            Assert.Greater(p.ProgressPercentage, lastProgressPercentage, "lastProgressPercentage < ProgressPercentage #{0}", reportedCount);
            Assert.Greater(p.RetrievedLineCount, lastProgressLineCount, "lastProgressLineCount < RetrievedLineCount #{0}", reportedCount);
          }

          Interlocked.Exchange(ref lastProgressPercentage, p.ProgressPercentage);
          Interlocked.Exchange(ref lastProgressLineCount, p.RetrievedLineCount);
        });

        var task = message.OpenReadAsync(cts.Token, progress);

        server.EnqueueResponse("RETR 1\r\n",
                               "+OK\r\n" +
                               TestUtils.GetByteStuffed(messageBody.ToString()) + 
                               ".\r\n");

        Assert.DoesNotThrow(() => task.Wait());

        Thread.Sleep(100); // wait for progress task

        Assert.Greater(reportedCount, 0, "reproted count");
        Assert.IsTrue(reportedFinish, "reported finish");
      });
    }
#endif

    [Test]
    public void TestOpenReadAsync_SessionClosed()
    {
      TestUtils.TestClosedSessionMessage(delegate(PopMessageInfo message) {
        Assert.Throws<InvalidOperationException>(() => message.OpenReadAsync());
      });
    }

    [Test]
    public void TestOpenReadAsync_MessageDeleted()
    {
      TestUtils.TestDeletedMessage(delegate(PopMessageInfo message) {
        var ex = Assert.Throws<PopMessageDeletedException>(() => message.OpenReadAsync());

        Assert.IsFalse(message.Client.IsBusy, "IsBusy");

        Assert.AreEqual(message.MessageNumber, ex.MessageNumber);
      });
    }

    [Test]
    public void TestOpenReadAsync_MaxLinesOutOfRange()
    {
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var openReadAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(-2)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(-2, progress)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(-2, CancellationToken.None)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(-2, CancellationToken.None, progress)),
      };
#else
      var openReadAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(-2)),
        new Func<PopMessageInfo, Task<Stream>>(m => m.OpenReadAsync(-2, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < openReadAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          Assert.Throws<ArgumentOutOfRangeException>(() => openReadAsyncMethods[i](message), "#{0}", i);
          Assert.IsFalse(message.Client.IsBusy, "IsBusy");
        });
      }
    }

    [Test]
    public void TestReadAsAsync_StreamConverter()
    {
      var converter = new Converter<Stream, byte[]>(s => s.ReadToEnd());
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var readAsAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(converter)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(converter, progress)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(converter, CancellationToken.None)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(converter, CancellationToken.None, progress)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(-1, converter)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(-1, converter, progress)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(-1, converter, CancellationToken.None)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(-1, converter, CancellationToken.None, progress)),
      };
#else
      var readAsAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(converter)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(converter, CancellationToken.None)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(-1, converter)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(-1, converter, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAsAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          int expectedOctets;
          string expectedMessageBody, byteStuffedMessageBody;

          GetTestMessageBody(out expectedMessageBody, out byteStuffedMessageBody, out expectedOctets);

          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          try {
            var ret = readAsAsyncMethods[i](message).Result;

            Assert.IsNotNull(ret);
            Assert.AreEqual(expectedOctets, ret.Length);
            Assert.AreEqual(expectedMessageBody,
                            NetworkTransferEncoding.Transfer8Bit.GetString(ret));
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }
        });
      }
    }

    [Test]
    public void TestReadAsAsync_StreamConverter_ConverterNull()
    {
      var converter = (Converter<Stream, byte[]>)null;

      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.ReadAsAsync(converter));
      });
    }

    [Test]
    public void TestReadAsAsync_StreamConverter_MaxLinesOutOfRange()
    {
      var converter = new Converter<Stream, byte[]>(s => s.ReadToEnd());
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var readAsAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(-2, converter)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(-2, converter, progress)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(-2, converter, CancellationToken.None)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(-2, converter, CancellationToken.None, progress)),
      };
#else
      var readAsAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(-2, converter)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAsAsync(-2, converter, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAsAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          Assert.Throws<ArgumentOutOfRangeException>(() => readAsAsyncMethods[i](message), "#{0}", i);
        });
      }
    }

    [Test]
    public void TestReadAllBytesAsync()
    {
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var readAllBytesAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync()),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(progress)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(CancellationToken.None)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(CancellationToken.None, progress)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(-1)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(-1, progress)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(-1, CancellationToken.None)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(-1, CancellationToken.None, progress)),
      };
#else
      var readAllBytesAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync()),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(CancellationToken.None)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(-1)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(-1, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAllBytesAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          int expectedOctets;
          string expectedMessageBody, byteStuffedMessageBody;

          GetTestMessageBody(out expectedMessageBody, out byteStuffedMessageBody, out expectedOctets);

          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          try {
            var ret = readAllBytesAsyncMethods[i](message).Result;

            Assert.IsNotNull(ret);
            Assert.AreEqual(expectedOctets, ret.Length);
            Assert.AreEqual(expectedMessageBody,
                            NetworkTransferEncoding.Transfer8Bit.GetString(ret));
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }
        });
      }
    }

    [Test]
    public void TestReadAllBytesAsync_MaxLinesOutOfRange()
    {
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var readAllBytesAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(-2)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(-2, progress)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(-2, CancellationToken.None)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(-2, CancellationToken.None, progress)),
      };
#else
      var readAllBytesAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(-2)),
        new Func<PopMessageInfo, Task<byte[]>>(m => m.ReadAllBytesAsync(-2, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAllBytesAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          Assert.Throws<ArgumentOutOfRangeException>(() => readAllBytesAsyncMethods[i](message), "#{0}", i);
        });
      }
    }

    [Test]
    public void TestWriteToAsync_Stream()
    {
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var writeToAsyncMethods = new[] {
        new Func<PopMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s)),
        new Func<PopMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, progress)),
        new Func<PopMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, CancellationToken.None)),
        new Func<PopMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, CancellationToken.None, progress)),
        new Func<PopMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, -1)),
        new Func<PopMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, -1, progress)),
        new Func<PopMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, -1, CancellationToken.None)),
        new Func<PopMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, -1, CancellationToken.None, progress)),
      };
#else
      var writeToAsyncMethods = new[] {
        new Func<PopMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s)),
        new Func<PopMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, CancellationToken.None)),
        new Func<PopMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, -1)),
        new Func<PopMessageInfo, Stream, Task>((m, s) => m.WriteToAsync(s, -1, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < writeToAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          int expectedOctets;
          string expectedMessageBody, byteStuffedMessageBody;

          GetTestMessageBody(out expectedMessageBody, out byteStuffedMessageBody, out expectedOctets);

          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          try {
            using (var outputStream = new MemoryStream()) {
              writeToAsyncMethods[i](message, outputStream).Wait();

              Assert.AreEqual(expectedOctets, outputStream.Length);

              outputStream.Close();

              Assert.AreEqual(NetworkTransferEncoding.Transfer8Bit.GetBytes(expectedMessageBody),
                              outputStream.ToArray());
            }
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }
        });
      }
    }

    [Test]
    public void TestWriteToAsync_Stream_DestinationNull()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.WriteToAsync((Stream)null));
      });
    }

    [Test]
    public void TestWriteToAsync_Stream_DestinationNotWritable()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        using (var stream = new MemoryStream(new byte[0], false)) {
          Assert.Throws<ArgumentException>(() => message.WriteToAsync(stream));
        }
      });
    }

    [Test]
    public void TestWriteToAsync_Stream_MaxLinesOutOfRange()
    {
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var writeToAsyncMethods = new[] {
        new Func<PopMessageInfo, Task>(m => m.WriteToAsync(Stream.Null, -2)),
        new Func<PopMessageInfo, Task>(m => m.WriteToAsync(Stream.Null, -2, progress)),
        new Func<PopMessageInfo, Task>(m => m.WriteToAsync(Stream.Null, -2, CancellationToken.None)),
        new Func<PopMessageInfo, Task>(m => m.WriteToAsync(Stream.Null, -2, CancellationToken.None, progress)),
      };
#else
      var writeToAsyncMethods = new[] {
        new Func<PopMessageInfo, Task>(m => m.WriteToAsync(Stream.Null, -2)),
        new Func<PopMessageInfo, Task>(m => m.WriteToAsync(Stream.Null, -2, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < writeToAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          Assert.Throws<ArgumentOutOfRangeException>(() => writeToAsyncMethods[i](message), "#{0}", i);
        });
      }
    }

    [Test]
    public void TestWriteToAsync_BinaryWriter()
    {
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var writeToAsyncMethods = new[] {
        new Func<PopMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w)),
        new Func<PopMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, progress)),
        new Func<PopMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, CancellationToken.None)),
        new Func<PopMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, CancellationToken.None, progress)),
        new Func<PopMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, -1)),
        new Func<PopMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, -1, progress)),
        new Func<PopMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, -1, CancellationToken.None)),
        new Func<PopMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, -1, CancellationToken.None, progress)),
      };
#else
      var writeToAsyncMethods = new[] {
        new Func<PopMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w)),
        new Func<PopMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, CancellationToken.None)),
        new Func<PopMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, -1)),
        new Func<PopMessageInfo, BinaryWriter, Task>((m, w) => m.WriteToAsync(w, -1, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < writeToAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          int expectedOctets;
          string expectedMessageBody, byteStuffedMessageBody;

          GetTestMessageBody(out expectedMessageBody, out byteStuffedMessageBody, out expectedOctets);

          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          try {
            using (var outputStream = new MemoryStream()) {
              var writer = new BinaryWriter(outputStream);

              writeToAsyncMethods[i](message, writer).Wait();

              Assert.AreEqual(expectedOctets, outputStream.Length);

              outputStream.Close();

              Assert.AreEqual(NetworkTransferEncoding.Transfer8Bit.GetBytes(expectedMessageBody),
                              outputStream.ToArray());
            }
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }
        });
      }
    }

    [Test]
    public void TestWriteToAsync_BinaryWriter_DestinationNull()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.WriteToAsync((BinaryWriter)null));
      });
    }

    [Test]
    public void TestWriteToAsync_BinaryWriter_MaxLinesOutOfRange()
    {
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var writeToAsyncMethods = new[] {
        new Func<PopMessageInfo, Task>(m => m.WriteToAsync(BinaryWriter.Null, -2)),
        new Func<PopMessageInfo, Task>(m => m.WriteToAsync(BinaryWriter.Null, -2, progress)),
        new Func<PopMessageInfo, Task>(m => m.WriteToAsync(BinaryWriter.Null, -2, CancellationToken.None)),
        new Func<PopMessageInfo, Task>(m => m.WriteToAsync(BinaryWriter.Null, -2, CancellationToken.None, progress)),
      };
#else
      var writeToAsyncMethods = new[] {
        new Func<PopMessageInfo, Task>(m => m.WriteToAsync(BinaryWriter.Null, -2)),
        new Func<PopMessageInfo, Task>(m => m.WriteToAsync(BinaryWriter.Null, -2, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < writeToAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          Assert.Throws<ArgumentOutOfRangeException>(() => writeToAsyncMethods[i](message), "#{0}", i);
        });
      }
    }

    [Test]
    public void TestSaveAsync()
    {
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var saveAsyncMethods = new[] {
        new Func<PopMessageInfo, string, Task>((m, f) => m.SaveAsync(f)),
        new Func<PopMessageInfo, string, Task>((m, f) => m.SaveAsync(f, progress)),
        new Func<PopMessageInfo, string, Task>((m, f) => m.SaveAsync(f, CancellationToken.None)),
        new Func<PopMessageInfo, string, Task>((m, f) => m.SaveAsync(f, CancellationToken.None, progress)),
        new Func<PopMessageInfo, string, Task>((m, f) => m.SaveAsync(f, -1)),
        new Func<PopMessageInfo, string, Task>((m, f) => m.SaveAsync(f, -1, progress)),
        new Func<PopMessageInfo, string, Task>((m, f) => m.SaveAsync(f, -1, CancellationToken.None)),
        new Func<PopMessageInfo, string, Task>((m, f) => m.SaveAsync(f, -1, CancellationToken.None, progress)),
      };
#else
      var saveAsyncMethods = new[] {
        new Func<PopMessageInfo, string, Task>((m, f) => m.SaveAsync(f)),
        new Func<PopMessageInfo, string, Task>((m, f) => m.SaveAsync(f, CancellationToken.None)),
        new Func<PopMessageInfo, string, Task>((m, f) => m.SaveAsync(f, -1)),
        new Func<PopMessageInfo, string, Task>((m, f) => m.SaveAsync(f, -1, CancellationToken.None)),
      };
#endif
      const string outfile = "retrieved.eml";

      for (var i = 0; i < saveAsyncMethods.Length; i++) {
        try {
          File.Delete(outfile);

          TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
            int expectedOctets;
            string expectedMessageBody, byteStuffedMessageBody;

            GetTestMessageBody(out expectedMessageBody, out byteStuffedMessageBody, out expectedOctets);

            server.EnqueueResponse("RETR 1\r\n",
                                   "+OK\r\n" +
                                   byteStuffedMessageBody +
                                   ".\r\n");

            try {
              saveAsyncMethods[i](message, outfile).Wait();

              Assert.IsTrue(File.Exists(outfile));
              Assert.AreEqual(expectedOctets, (new FileInfo(outfile)).Length);
              Assert.AreEqual(NetworkTransferEncoding.Transfer8Bit.GetBytes(expectedMessageBody),
                              File.ReadAllBytes(outfile));
            }
            catch (Exception ex) {
              Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
            }
          });
        }
        finally {
          File.Delete(outfile);
        }
      }
    }

    [Test]
    public void TestSaveAsync_CancellationRequestedBeforeStart()
    {
      const string outfile = "retrieved.eml";

      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        var cts = new CancellationTokenSource();

        cts.Cancel();

        var t = message.SaveAsync(outfile, cts.Token);

        var ex = Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsInstanceOf<OperationCanceledException>(ex.Flatten().InnerException);

        Assert.IsFalse(File.Exists(outfile), "file existence");
      });
    }

    [Test]
    public void TestSaveAsync_OverwriteExistingFile()
    {
      const string outfile = "retrieved.eml";
      const int existingFileLength = 10 * 1024;

      try {
        File.Delete(outfile);

        File.WriteAllText(outfile, new string('x', existingFileLength));

        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          int octets;
          string messageBody, byteStuffedMessageBody;

          GetTestMessageBody(out messageBody, out byteStuffedMessageBody, out octets);

          Assert.Less(octets, (new FileInfo(outfile)).Length, "less than existing file length");

          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          message.SaveAsync(outfile).Wait();

          Assert.IsTrue(File.Exists(outfile), "file existence");
          Assert.AreEqual(octets, (new FileInfo(outfile)).Length, "file length");
          Assert.AreEqual(NetworkTransferEncoding.Transfer7Bit.GetBytes(messageBody),
                          File.ReadAllBytes(outfile));
        });
      }
      finally {
        File.Delete(outfile);
      }
    }

    [Test]
    public void TestSaveAsync_PathNull()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.SaveAsync(null));
      });
    }

    [Test]
    public void TestSaveAsync_PathEmpty()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentException>(() => message.SaveAsync(string.Empty));
      });
    }

    [Test]
    public void TestSaveAsync_MaxLinesOutOfRange()
    {
      const string outfile = "retrieved.eml";
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var saveAsyncMethods = new[] {
        new Func<PopMessageInfo, Task>(m => m.SaveAsync(outfile, -2)),
        new Func<PopMessageInfo, Task>(m => m.SaveAsync(outfile, -2, progress)),
        new Func<PopMessageInfo, Task>(m => m.SaveAsync(outfile, -2, CancellationToken.None)),
        new Func<PopMessageInfo, Task>(m => m.SaveAsync(outfile, -2, CancellationToken.None, progress)),
      };
#else
      var saveAsyncMethods = new[] {
        new Func<PopMessageInfo, Task>(m => m.SaveAsync(outfile, -2)),
        new Func<PopMessageInfo, Task>(m => m.SaveAsync(outfile, -2, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < saveAsyncMethods.Length; i++) {
        try {
          File.Delete(outfile);

          TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
            Assert.Throws<ArgumentOutOfRangeException>(() => saveAsyncMethods[i](message), "#{0}", i);
          });
        }
        finally {
          File.Delete(outfile);
        }
      }
    }

    [Test]
    public void TestSaveAsync_CanNotOpenRetrieveStream()
    {
      const string outfile = "retrieved.eml";

      try {
        File.Delete(outfile);

        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          server.EnqueueResponse("RETR 1\r\n",
                                 "-ERR no such message\r\n");

          var t = message.SaveAsync(outfile);

          var ex = Assert.Throws<AggregateException>(() => t.Wait());

          Assert.IsInstanceOf<PopErrorResponseException>(ex.Flatten().InnerException);

          Assert.IsTrue(File.Exists(outfile), "file existence");
        });
      }
      finally {
        File.Delete(outfile);
      }
    }

    [Test]
    public void TestOpenTextAsync()
    {
      var encoding = Encoding.GetEncoding("ISO-8859-1");
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var openTextAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync()),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(-1)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, -1)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, -1, CancellationToken.None)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, -1, progress)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, -1, CancellationToken.None, progress)),
      };
#else
      var openTextAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync()),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(-1)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, -1)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, -1, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < openTextAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          int expectedOctets;
          string expectedMessageBody, byteStuffedMessageBody;

          GetTestMessageBody(out expectedMessageBody, out byteStuffedMessageBody, out expectedOctets);

          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          try {
            using (var reader = openTextAsyncMethods[i](message).Result) {
              Assert.IsNotNull(reader);
              Assert.AreEqual(encoding, reader.CurrentEncoding);
              Assert.AreEqual(expectedMessageBody, reader.ReadToEnd());
            }
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }
        });
      }
    }

    [Test]
    public void TestOpenTextAsync_CharsetNull()
    {
      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.OpenTextAsync((Encoding)null));
      });
    }

    [Test]
    public void TestOpenTextAsync_MaxLinesOutOfRange()
    {
      var encoding = Encoding.GetEncoding("ISO-8859-1");
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var openTextAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(-2)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, -2)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, -2, CancellationToken.None)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, -2, progress)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, -2, CancellationToken.None, progress)),
      };
#else
      var openTextAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(-2)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, -2)),
        new Func<PopMessageInfo, Task<StreamReader>>(m => m.OpenTextAsync(encoding, -2, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < openTextAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          Assert.Throws<ArgumentOutOfRangeException>(() => openTextAsyncMethods[i](message), "#{0}", i);
        });
      }
    }

    [Test]
    public void TestReadAsAsync_StreamReaderConverter()
    {
      var encoding = Encoding.GetEncoding("ISO-8859-1");
      var converter = new Converter<StreamReader, string>(s => s.ReadToEnd());
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var readAsAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(converter)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(-1, converter)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, converter)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, -1, converter)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, -1, converter, CancellationToken.None)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, -1, converter, progress)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, -1, converter, CancellationToken.None, progress)),
      };
#else
      var readAsAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(converter)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(-1, converter)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, converter)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, -1, converter)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, -1, converter, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAsAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          int expectedOctets;
          string expectedMessageBody, byteStuffedMessageBody;

          GetTestMessageBody(out expectedMessageBody, out byteStuffedMessageBody, out expectedOctets);

          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          try {
            Assert.AreEqual(expectedMessageBody,
                            readAsAsyncMethods[i](message).Result);
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }
        });
      }
    }

    [Test]
    public void TestReadAsAsync_StreamReaderConverter_ConverterNull()
    {
      var converter = (Converter<StreamReader, string>)null;

      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.ReadAsAsync(converter));
      });
    }

    [Test]
    public void TestReadAsAsync_StreamReaderConverter_CharsetNull()
    {
      var converter = (Converter<StreamReader, string>)(r => r.ReadToEnd());

      TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        Assert.Throws<ArgumentNullException>(() => message.ReadAsAsync((Encoding)null, converter));
      });
    }

    [Test]
    public void TestReadAsAsync_StreamReaderConverter_MaxLinesOutOfRange()
    {
      var encoding = Encoding.GetEncoding("ISO-8859-1");
      var converter = new Converter<StreamReader, string>(s => s.ReadToEnd());
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var readAsAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(-2, converter)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, -2, converter)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, -2, converter, CancellationToken.None)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, -2, converter, progress)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, -2, converter, CancellationToken.None, progress)),
      };
#else
      var readAsAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(-2, converter)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, -2, converter)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAsAsync(encoding, -2, converter, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAsAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          Assert.Throws<ArgumentOutOfRangeException>(() => readAsAsyncMethods[i](message), "#{0}", i);
        });
      }
    }

    [Test]
    public void TestReadAllLinesAsync()
    {
      var encoding = Encoding.GetEncoding("ISO-8859-1");
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var readAsAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync()),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(-1)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, -1)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, -1, CancellationToken.None)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, -1, progress)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, -1, CancellationToken.None, progress)),
      };
#else
      var readAsAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync()),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(-1)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, -1)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, -1, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAsAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          int expectedOctets;
          string expectedMessageBody, byteStuffedMessageBody;

          GetTestMessageBody(out expectedMessageBody, out byteStuffedMessageBody, out expectedOctets);

          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          try {
            Assert.AreEqual(expectedMessageBody.Replace("\r\n", "\n").Replace("\n", "\r\n").TrimEnd(),
                            string.Join("\r\n", message.ReadAllLinesAsync().Result));
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }
        });
      }
    }

    [Test]
    public void TestReadAllLinesAsync_MaxLinesOutOfRange()
    {
      var encoding = Encoding.GetEncoding("ISO-8859-1");
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var readAsAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(-2)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, -2)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, -2, CancellationToken.None)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, -2, progress)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, -2, CancellationToken.None, progress)),
      };
#else
      var readAsAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(-2)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, -2)),
        new Func<PopMessageInfo, Task<string[]>>(m => m.ReadAllLinesAsync(encoding, -2, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAsAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          Assert.Throws<ArgumentOutOfRangeException>(() => readAsAsyncMethods[i](message), "#{0}", i);
        });
      }
    }

    [Test]
    public void TestReadAllTextAsync()
    {
      var encoding = Encoding.GetEncoding("ISO-8859-1");
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var readAllTextAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync()),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(-1)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, -1)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, -1, CancellationToken.None)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, -1, progress)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, -1, CancellationToken.None, progress)),
      };
#else
      var readAllTextAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync()),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(-1)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, -1)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, -1, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAllTextAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          int expectedOctets;
          string expectedMessageBody, byteStuffedMessageBody;

          GetTestMessageBody(out expectedMessageBody, out byteStuffedMessageBody, out expectedOctets);

          server.EnqueueResponse("RETR 1\r\n",
                                 "+OK\r\n" +
                                 byteStuffedMessageBody +
                                 ".\r\n");

          try {
            Assert.AreEqual(expectedMessageBody,
                            readAllTextAsyncMethods[i](message).Result);
          }
          catch (Exception ex) {
            Assert.Fail("unexpected exception (overload #{0}) : {1}", i, ex);
          }
        });
      }
    }

    [Test]
    public void TestReadAllTextAsync_MaxLinesOutOfRange()
    {
      var encoding = Encoding.GetEncoding("ISO-8859-1");
#if NET_4_5
      var progress = (IProgress<PopMessageRetrieveProgressInfo>)null;
      var readAllTextAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(-2)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, -2)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, -2, CancellationToken.None)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, -2, progress)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, -2, CancellationToken.None, progress)),
      };
#else
      var readAllTextAsyncMethods = new[] {
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(-2)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, -2)),
        new Func<PopMessageInfo, Task<string>>(m => m.ReadAllTextAsync(encoding, -2, CancellationToken.None)),
      };
#endif

      for (var i = 0; i < readAllTextAsyncMethods.Length; i++) {
        TestUtils.TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
          Assert.Throws<ArgumentOutOfRangeException>(() => readAllTextAsyncMethods[i](message), "#{0}", i);
        });
      }
    }
  }
}
#endif // ASYNC_OPS

