using System;
using NUnit.Framework;

using Smdn.Net.Pop3.Client.Session;

namespace Smdn.Net.Pop3.Client {
  internal static class TestUtils {
    public static void Repeat(int count, Action action)
    {
      for (var i = 0; i < count; i++) {
        action();
      }
    }

    public static string GetByteStuffed(string input)
    {
      return input.Replace("\r\n.", "\r\n..");
    }

    public static void TestAuthenticated(Action<PopPseudoServer, PopClient> testAction)
    {
      using (var server =PopPseudoServer.StartNew()) {
        server.EnqueueResponse("CAPA\r\n",
                               "+OK\r\n" +
                               ".\r\n");
        server.EnqueueResponse("USER user\r\n",
                               "+OK\r\n");
        server.EnqueueResponse("PASS pass\r\n",
                               "+OK\r\n");

        using (var client = new PopClient(new Uri(string.Format("pop://user@{0}/", server.HostPort)))) {
          client.Timeout = 5000;
          client.Profile.UsingSaslMechanisms = null;
          client.Profile.UseTlsIfAvailable = false;
          client.Profile.AllowInsecureLogin = true;

          client.Connect("pass");

          server.DequeueRequest(); // CAPA
          server.DequeueRequest(); // USER
          server.DequeueRequest(); // PASS

          testAction(server, client);
        }
      }
    }

    public static void TestDisconnected(Action<PopPseudoServer, PopClient> testAction)
    {
      TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("NOOP\r\n",
                               "+OK\r\n",
                               true); // stop server

        client.KeepAlive();

        testAction(server, client);
      });
    }

    public static void TestLoggedOut(Action<PopPseudoServer, PopClient> testAction)
    {
      TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("QUIT\r\n",
                               "+OK\r\n");

        client.Logout();

        testAction(server, client);
      });
    }

    public static void TestMessage(Action<PopPseudoServer, PopMessageInfo> action)
    {
      TestMessage(1L, 120L, action);
    }

    public static void TestMessage(long messageNumber, long messageLength, Action<PopPseudoServer, PopMessageInfo> action)
    {
      TestAuthenticated(delegate(PopPseudoServer server, PopClient client) {
        server.EnqueueResponse("STAT\r\n",
                               string.Format("+OK {0} {1}\r\n", messageNumber, messageLength));
        server.EnqueueResponse("LIST ",
                               string.Format("+OK {0} {1}\r\n", messageNumber, messageLength));

        var message = client.GetMessage(messageNumber);

        Assert.IsNotNull(message);
        Assert.AreEqual(messageNumber, message.MessageNumber);
        Assert.AreEqual(messageLength, message.Length);
        Assert.IsFalse(message.IsMarkedAsDeleted);

        server.DequeueRequest(); // STAT
        server.DequeueRequest(); // LIST

        action(server, message);
      });
    }

    public static void TestDeletedMessage(Action<PopMessageInfo> action)
    {
      TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        server.EnqueueResponse("DELE ",
                               "+OK\r\n");

        message.MarkAsDeleted();

        server.DequeueRequest(); // DELE

        Assert.IsTrue(message.IsMarkedAsDeleted);

        server.EnqueueResponse("STAT\r\n",
                               "+OK 0 0\r\n");

        Assert.AreEqual(0L, message.Client.MessageCount);
        Assert.AreEqual(0L, message.Client.TotalSize);

        server.DequeueRequest(); // STAT

        action(message);
      });
    }

    public static void TestClosedSessionMessage(Action<PopMessageInfo> action)
    {
      TestMessage(delegate(PopPseudoServer server, PopMessageInfo message) {
        message.Client.Disconnect();

        action(message);
      });
    }
  }
}
