using System;
using System.Text;
using NUnit.Framework;

namespace Smdn.Net.Pop3 {
  [TestFixture]
  public class PopUriBuilderTests {
    [Test]
    public void TestConstructDefault()
    {
      var b = new PopUriBuilder();

      Assert.AreEqual("pop://localhost/", b.Uri.AbsoluteUri);
    }

    [Test]
    public void TestConstructFromUri()
    {
      var uri = "POPS://user;auth=cram-md5@localhost:10110";
      var b = new PopUriBuilder(new Uri(uri));

      Assert.AreEqual("pops://user;AUTH=CRAM-MD5@localhost:10110/", b.Uri.AbsoluteUri);
      Assert.AreEqual("pops", b.Scheme);
      Assert.AreEqual("user", b.UserName);
      Assert.AreEqual(PopAuthenticationMechanism.CRAMMD5, b.AuthType);
      Assert.AreEqual("localhost", b.Host);
      Assert.AreEqual(10110, b.Port);
    }

    [Test]
    public void TestClone()
    {
      var origin = new PopUriBuilder("pops://user;auth=cram-md5@localhost:10110/");
      var cloned = origin.Clone();

      Assert.AreEqual(cloned, origin);
      Assert.AreEqual("pops", cloned.Scheme);
      Assert.AreEqual("user", cloned.UserName);
      Assert.AreEqual(PopAuthenticationMechanism.CRAMMD5, cloned.AuthType);
      Assert.AreEqual("localhost", cloned.Host);
      Assert.AreEqual(10110, cloned.Port);
    }

    [Test]
    public void TestBuildUri()
    {
      var b = new PopUriBuilder();

      b.Host = "pop.example.net";

      Assert.AreEqual("pop://pop.example.net/", b.Uri.AbsoluteUri);

      b.Port = 10110;

      Assert.AreEqual("pop://pop.example.net:10110/", b.Uri.AbsoluteUri);

      b.UserName = "user";

      Assert.AreEqual("pop://user@pop.example.net:10110/", b.Uri.AbsoluteUri);

      b.AuthType = PopAuthenticationMechanism.CRAMMD5;

      Assert.AreEqual("pop://user;AUTH=CRAM-MD5@pop.example.net:10110/", b.Uri.AbsoluteUri);

      b.Scheme = "pops";

      Assert.AreEqual("pops://user;AUTH=CRAM-MD5@pop.example.net:10110/", b.Uri.AbsoluteUri);

      b.Scheme = "POP";

      Assert.AreEqual("pop://user;AUTH=CRAM-MD5@pop.example.net:10110/", b.Uri.AbsoluteUri);

      b.UserName = null;

      Assert.AreEqual("pop://;AUTH=CRAM-MD5@pop.example.net:10110/", b.Uri.AbsoluteUri);

      b.Port = 110;

      Assert.AreEqual("pop://;AUTH=CRAM-MD5@pop.example.net:110/", b.Uri.AbsoluteUri);

      b.Port = 0;

      if (Runtime.IsRunningOnMono)
        Assert.AreEqual("pop://;AUTH=CRAM-MD5@pop.example.net:0/", b.Uri.AbsoluteUri);
      else
        Assert.AreEqual("pop://;AUTH=CRAM-MD5@pop.example.net/", b.Uri.AbsoluteUri); // ???

      b.Port = -1;

      Assert.AreEqual("pop://;AUTH=CRAM-MD5@pop.example.net/", b.Uri.AbsoluteUri);
    }


    [Test]
    public void TestConstructFromUriUserNameContainsAtmark()
    {
      var uri = "pops://user%40pop.example.net@pop.example.net/";
      var b = new PopUriBuilder(new Uri(uri));

      Assert.AreEqual(uri.ToString(), b.ToString());
      Assert.AreEqual("pops", b.Scheme);
      Assert.AreEqual("user@pop.example.net", b.UserName);
      Assert.AreEqual("pop.example.net", b.Host);
    }

    [Test]
    public void TestBuildUserNameContainsAtmark()
    {
      var b = new PopUriBuilder();

      b.Scheme = "pops";
      b.Host = "pop.example.net";
      b.UserName = "user@pop.example.net";

      Assert.AreEqual("pops://user%40pop.example.net@pop.example.net/",
                      b.Uri.AbsoluteUri);
    }

    [Test]
    public void TestConstructFromUriUserNameContainsColon()
    {
      var uri = "pops://recent%3Auser@pop.example.net/";
      var b = new PopUriBuilder(new Uri(uri));

      Assert.AreEqual(uri.ToString(), b.ToString());
      Assert.AreEqual("pops", b.Scheme);
      Assert.AreEqual("recent:user", b.UserName);
      Assert.AreEqual("pop.example.net", b.Host);
    }

    [Test]
    public void TestBuildUserNameContainsColon()
    {
      var b = new PopUriBuilder();

      b.Scheme = "pops";
      b.Host = "pop.example.net";
      b.UserName = "recent:user";

      Assert.AreEqual("pops://recent%3Auser@pop.example.net/",
                      b.Uri.AbsoluteUri);
    }

    [Test]
    public void TestBuildExample1()
    {
      var b = new PopUriBuilder();

      b.Scheme = "pop";
      b.Host = "mailsrv.qualcomm.com";
      b.UserName = "rg";

      Assert.AreEqual("pop://rg@mailsrv.qualcomm.com/",
                      b.Uri.AbsoluteUri);
    }

    [Test]
    public void TestBuildExample2()
    {
      var b = new PopUriBuilder();

      b.Scheme = "pop";
      b.Host = "mail.eudora.com";
      b.Port = 8110;
      b.UserName = "rg";
      b.AuthType = PopAuthenticationMechanism.Apop;

      Assert.AreEqual("pop://rg;AUTH=+APOP@mail.eudora.com:8110/",
                      b.Uri.AbsoluteUri);
    }

    [Test]
    public void TestBuildExample3()
    {
      var b = new PopUriBuilder();

      b.Scheme = "pop";
      b.Host = "foo.bar";
      b.UserName = "baz";
      b.AuthType = new PopAuthenticationMechanism("SCRAM-MD5");

      Assert.AreEqual("pop://baz;AUTH=SCRAM-MD5@foo.bar/",
                      b.Uri.AbsoluteUri);
    }
  }
}
