using System;
using System.Net;
using NUnit.Framework;

namespace Smdn.Security.Authentication.Sasl.Client {
  [TestFixture]
  public class XOAuth2MechanismTests {
    [Test]
    public void TestCreate()
    {
      foreach (var name in new[] {
        "XOAUTH2",
        "xoauth2",
      }) {
        Assert.Throws<SaslMechanismNotSupportedException>(() => SaslClientMechanism.Create(name));
      }
    }

    [Test]
    public void TestConstruct()
    {
      Assert.DoesNotThrow(() => new XOAuth2Mechanism(null, "accessToken"));
      Assert.DoesNotThrow(() => new XOAuth2Mechanism("user", null));
      Assert.DoesNotThrow(() => new XOAuth2Mechanism(null, null));
      Assert.DoesNotThrow(() => new XOAuth2Mechanism("user", "accessToken"));
    }

    [Test]
    public void TestIsPlainText()
    {
      using (var client = new XOAuth2Mechanism("user", "accessToken")) {
        Assert.IsFalse(client.IsPlainText);
      }
    }

    [Test]
    public void TestUser()
    {
      using (var client = new XOAuth2Mechanism("user", "accessToken")) {
        Assert.AreEqual("user", client.User);

        client.User = "newuser";

        Assert.AreEqual("newuser", client.User);

        Assert.DoesNotThrow(() => client.User = null);

        client.Dispose();

        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(client.User));
        Assert.Throws<ObjectDisposedException>(() => client.User = "userDisposed");
      }
    }

    [Test]
    public void TestAccessToken()
    {
      using (var client = new XOAuth2Mechanism("user", "accessToken")) {
        Assert.AreEqual("accessToken", client.AccessToken);

        client.AccessToken = "newAccessToken";

        Assert.AreEqual("newAccessToken", client.AccessToken);

        Assert.DoesNotThrow(() => client.AccessToken = null);

        client.Dispose();

        Assert.Throws<ObjectDisposedException>(() => Assert.IsNotNull(client.AccessToken));
        Assert.Throws<ObjectDisposedException>(() => client.AccessToken = "accessTokenDisposed");
      }
    }

    [Test]
    public void TestName()
    {
      using (var client = new XOAuth2Mechanism("user", "accessToken")) {
        Assert.AreEqual("XOAUTH2", client.Name);
      }
    }

    private class XOAuth2MechanismEx : XOAuth2Mechanism {
    }

    [Test]
    public void TestNameInherited()
    {
      using (var client = new XOAuth2MechanismEx()) {
        Assert.AreEqual("XOAUTH2", client.Name);
      }
    }

    [Test]
    public void TestInitialize()
    {
      using (var client = new XOAuth2Mechanism("user", "accessToken")) {
        Assert.AreEqual(SaslExchangeStatus.None, client.ExchangeStatus);
        Assert.IsNull(client.LastErrorResults);

        byte[] initialResponseFirst;

        Assert.AreEqual(SaslExchangeStatus.Continuing,
                        client.GetInitialResponse(out initialResponseFirst));
        Assert.AreEqual(SaslExchangeStatus.Continuing, client.ExchangeStatus);
        Assert.IsNull(client.LastErrorResults);

        byte[] clientResponse;

        client.Exchange(ByteString.CreateImmutable("error results").ToArray(),
                        out clientResponse);

        Assert.IsNotNull(client.LastErrorResults);

        client.Initialize();

        Assert.AreEqual(SaslExchangeStatus.None, client.ExchangeStatus);
        Assert.IsNull(client.LastErrorResults);

        byte[] initialResponseSecond;

        Assert.AreEqual(SaslExchangeStatus.Continuing,
                        client.GetInitialResponse(out initialResponseSecond));
        Assert.AreEqual(SaslExchangeStatus.Continuing, client.ExchangeStatus);
        Assert.IsNull(client.LastErrorResults);

        Assert.AreEqual(initialResponseFirst, initialResponseSecond);
      }
    }

    [Test]
    public void TestGetInitialResponse()
    {
      using (var client = new XOAuth2Mechanism("someuser@example.com", "vF9dft4qmTc2Nvb3RlckBhdHRhdmlzdGEuY29tCg==")) {
        Assert.AreEqual(SaslExchangeStatus.None, client.ExchangeStatus);

        Assert.IsTrue(client.ClientFirst);

        byte[] initialResponse;

        Assert.AreEqual(SaslExchangeStatus.Continuing,
                        client.GetInitialResponse(out initialResponse));
        Assert.AreEqual(SaslExchangeStatus.Continuing, client.ExchangeStatus);

        BytesAssert.AreEqual("user=someuser@example.com\x0001auth=Bearer vF9dft4qmTc2Nvb3RlckBhdHRhdmlzdGEuY29tCg==\x0001\x0001",
                             initialResponse);
      }
    }

    [Test]
    public void TestExchangeUserNotSet()
    {
      using (var client = new XOAuth2Mechanism(null, "accessToken")) {
        Assert.IsNull(client.User);

        Assert.AreEqual(SaslExchangeStatus.None, client.ExchangeStatus);

        byte[] clientResponse;

        Assert.AreEqual(SaslExchangeStatus.Failed, client.Exchange(EmptyArray<byte>.Instance, out clientResponse));
        Assert.IsNull(clientResponse);

        Assert.Throws<InvalidOperationException>(() => client.Exchange(EmptyArray<byte>.Instance, out clientResponse));
      }
    }

    [Test]
    public void TestExchangeAccessTokenSet()
    {
      using (var client = new XOAuth2Mechanism("user", null)) {
        Assert.IsNull(client.AccessToken);

        Assert.AreEqual(SaslExchangeStatus.None, client.ExchangeStatus);

        byte[] clientResponse;

        Assert.AreEqual(SaslExchangeStatus.Failed, client.Exchange(EmptyArray<byte>.Instance, out clientResponse));
        Assert.IsNull(clientResponse);

        Assert.Throws<InvalidOperationException>(() => client.Exchange(EmptyArray<byte>.Instance, out clientResponse));
      }
    }

    [Test]
    public void TestExchangeError()
    {
      using (var client = new XOAuth2Mechanism("someuser@example.com", "vF9dft4qmTc2Nvb3RlckBhdHRhdmlzdGEuY29tCg==")) {
        Assert.AreEqual(SaslExchangeStatus.None, client.ExchangeStatus);
        Assert.IsNull(client.LastErrorResults);

        Assert.IsTrue(client.ClientFirst);

        // step1
        byte[] clientResponse;

        Assert.AreEqual(SaslExchangeStatus.Continuing,
                        client.Exchange(null, out clientResponse));
        Assert.AreEqual(SaslExchangeStatus.Continuing, client.ExchangeStatus);
        Assert.IsNull(client.LastErrorResults);

        BytesAssert.AreEqual("user=someuser@example.com\x0001auth=Bearer vF9dft4qmTc2Nvb3RlckBhdHRhdmlzdGEuY29tCg==\x0001\x0001",
                             clientResponse);

        // step2
        Assert.AreEqual(SaslExchangeStatus.Failed,
                        client.Exchange(ByteString.CreateImmutable("{\"status\":\"401\",\"schemes\":\"bearer mac\",\"scope\":\"https://mail.google.com/\"}").ToArray(),
                        out clientResponse));
        Assert.AreEqual(SaslExchangeStatus.Failed, client.ExchangeStatus);
        Assert.AreEqual("{\"status\":\"401\",\"schemes\":\"bearer mac\",\"scope\":\"https://mail.google.com/\"}",
                        client.LastErrorResults);
        Assert.IsNull(clientResponse);
      }
    }
  }
}
