using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Windows.Forms;

using Smdn.Formats.Ini;

using Smdn.Net.Imap4;
using Smdn.Net.Imap4.Client;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

using Smdn.Net.Pop3;
using Smdn.Net.Pop3.Client;
using Smdn.Net.Pop3.Protocol;
using Smdn.Net.Pop3.Protocol.Client;

using Smdn.Security.Authentication.Sasl;
using Smdn.Security.Authentication.Sasl.Client;

using Smdn.Windows.Forms;

namespace Smdn.Applications.MRAConnector {
  public partial class MainDialog : Form {
    private readonly string[] defaultUsingSaslMechanisms = new[] {
      "DIGEST-MD5", "CRAM-MD5", "NTLM", "PLAIN", "LOGIN",
    };

    private static class Fonts {
      //public static Font BoldFont;
      //public static Font StrikeoutFont;
      public static Font Monospace;

      static Fonts()
      {
        //BoldFont = new Font(Control.DefaultFont, FontStyle.Bold);
        //StrikeoutFont = new Font(Control.DefaultFont, FontStyle.Strikeout);
        Monospace = new Font(FontFamily.GenericMonospace, Control.DefaultFont.SizeInPoints, FontStyle.Regular, GraphicsUnit.Point);
      }
    }

    public MainDialog()
    {
      InitializeComponent();
      Initialize();
    }

    private void Initialize()
    {
      ImapSslConnection.ServerCertificateValidationCallback += ValidateServerCertificate;
      PopSslConnection.ServerCertificateValidationCallback += ValidateServerCertificate;

      ImapConnection.TraceSource.Switch.Level = SourceLevels.All;
      PopConnection.TraceSource.Switch.Level = SourceLevels.All;

      protocolLogListener = new TextBoxTraceListener("textbox", textBoxProtocolLog);

      ImapConnection.TraceSource.Listeners.Add(protocolLogListener);
      PopConnection.TraceSource.Listeners.Add(protocolLogListener);

      connectBackground = new ConnectBackgroundWorker();
      connectBackground.RunWorkerCompleted += ConnectCompleted;

      //components.Add(connectBackground);

      numericUpDownPort.Minimum = IPEndPoint.MinPort;
      numericUpDownPort.Maximum = IPEndPoint.MaxPort;

      numericUpDownPort.ValueChanged += delegate {
        if (!checkBoxConnectToDefaultPort.Checked)
          currentProfile.Port = (int)numericUpDownPort.Value;
      };

      radioButtonProtocolImap.CheckedChanged += ProtocolChanged;
      radioButtonProtocolPop.CheckedChanged += ProtocolChanged;

      checkBoxConnectToDefaultPort.CheckedChanged += ConnectToDefaultPortChanged;

      textBoxHost.Validating += HostAndUserNameValidating;
      textBoxHost.Validated += delegate {
        currentProfile.Host = textBoxHost.Text;
      };

      textBoxUserName.Validating += HostAndUserNameValidating;
      textBoxUserName.Validated += delegate {
        currentProfile.UserName = textBoxUserName.Text;
      };

      checkBoxAuthenticateAnonymous.CheckedChanged += AuthenticateAnonymousChanged;

      radioButtonSslUsageAlways.CheckedChanged += SslUsageChanged;
      radioButtonSslUsageIfAvailable.CheckedChanged += SslUsageChanged;
      radioButtonSslUsageNever.CheckedChanged += SslUsageChanged;

      radioButtonAuthMechanismUseSupplied.CheckedChanged += AuthMechanismChanged;
      radioButtonAuthMechanismSelectAppropriate.CheckedChanged += AuthMechanismChanged;

      foreach (var mechanismName in SaslClientMechanism.GetAvailableMechanisms()) {
        if (!string.Equals(mechanismName, SaslMechanisms.Anonymous, StringComparison.OrdinalIgnoreCase))
          comboBoxAuthType.Items.Add(mechanismName);
      }

      comboBoxAuthType.SelectedValueChanged += AuthTypeChanged;

      checkBoxForceSecureLogin.CheckedChanged += delegate {
        currentProfile.AllowInsecureLogin = !checkBoxForceSecureLogin.Checked;

        UpdatePopLoginEnability();
      };

      checkBoxAllowPopStandardLogin.CheckedChanged += delegate {
        currentProfile.AllowPopStandardLogin = checkBoxAllowPopStandardLogin.Checked;
      };

      checkBoxAllowApopLogin.CheckedChanged += delegate {
        currentProfile.AllowApopLogin = checkBoxAllowApopLogin.Checked;
      };

      buttonConnect.Click += delegate {
        Connect(currentProfile);
      };

      buttonSaveProfile.Click += delegate { SaveCurrentProfile(); };
      buttonRemoveProfile.Click += delegate { RemoveProfile(); };

      defaultProfile = null;

      profiles = new Dictionary<string, Profile>(ProfileStore.LoadProfiles(out defaultProfile));

      comboBoxProfiles.Items.Clear();

      foreach (var key in profiles.Keys) {
        comboBoxProfiles.Items.Add(key);
      }

      comboBoxProfiles.SelectedIndexChanged += delegate {
        Profile prof;

        if (profiles.TryGetValue((string)comboBoxProfiles.SelectedItem, out prof))
          SetProfile(prof.Clone());
      };

      textBoxProtocolLog.Font = Fonts.Monospace;
    }

    protected override void OnLoad(EventArgs e)
    {
      base.OnLoad(e);

      if (defaultProfile != null)
        SetProfile(defaultProfile);
    }

    protected override void OnClosing (CancelEventArgs e)
    {
      base.OnClosing(e);

      if (!e.Cancel)
        ProfileStore.SaveProfiles(currentProfile, profiles);
    }

    private bool ValidateServerCertificate(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
    {
      if (sslPolicyErrors == SslPolicyErrors.None)
        return true;

      var message = new StringBuilder();

      message.AppendFormat("SslPolicyErrors: {0}", sslPolicyErrors);
      message.AppendLine();

      if (chain != null) {
        foreach (var element in chain.ChainElements) {
          message.AppendFormat("発行先: {0}", element.Certificate.Subject);
          message.AppendLine();
          message.AppendFormat("発行者: {0}", element.Certificate.Issuer);
          message.AppendLine();
          message.AppendFormat("有効期間: {0} から {1}", element.Certificate.GetEffectiveDateString(), element.Certificate.GetExpirationDateString());
          message.AppendLine();

          foreach (var status in element.ChainElementStatus) {
            message.AppendFormat("Status: {0}", status.Status);
            message.AppendLine();
            message.AppendFormat("StatusInformation: {0}", status.StatusInformation);
            message.AppendLine();
            message.AppendLine();
          }
        }
      }

      textBoxValidationError.Invoke((Action<string>)SetValidationErrorText, message.ToString());

#if false
      var result = MessageBox.Show(this,
                                   message.ToString(),
                                   "ValidateServerCertificate",
                                   MessageBoxButtons.YesNo);

      return (result == DialogResult.Yes);
#endif

      return true;
    }

    private void SetValidationErrorText(string message)
    {
      textBoxValidationError.Text = message;
    }

    private void ProtocolChanged(object sender, EventArgs e)
    {
      if (checkBoxConnectToDefaultPort.Checked)
        SetDefaultPort();

      var authTypeApop = (string)PopAuthenticationMechanism.Apop;
      var indexOfAuthTypeApop = comboBoxAuthType.Items.IndexOf(authTypeApop);

      if (radioButtonProtocolImap.Checked) {
        currentProfile.ProtocolType = ProtocolType.Imap;

        if (indexOfAuthTypeApop != -1)
          comboBoxAuthType.Items.RemoveAt(indexOfAuthTypeApop);
      }
      else if (radioButtonProtocolPop.Checked) {
        currentProfile.ProtocolType = ProtocolType.Pop;

        if (indexOfAuthTypeApop == -1)
          comboBoxAuthType.Items.Add(authTypeApop);
      }

      if (checkBoxAuthenticateAnonymous.Checked)
        currentProfile.AuthType = "ANONYMOUS";
      else
        AuthTypeChanged(null, EventArgs.Empty);

      UpdatePopLoginEnability();
    }

    private void UpdatePopLoginEnability()
    {
      var enablePopLogin = currentProfile.ProtocolType == ProtocolType.Pop
        && checkBoxForceSecureLogin.Enabled
        && !checkBoxForceSecureLogin.Checked;

      checkBoxAllowApopLogin.Enabled = enablePopLogin;
      checkBoxAllowPopStandardLogin.Enabled = enablePopLogin;
    }

    private void HostAndUserNameValidating(object sender, CancelEventArgs e)
    {
      try {
#pragma warning disable 168
        var uri = currentProfile.IsImap
          ? (new ImapUriBuilder(textBoxHost.Text, textBoxUserName.Text, (ImapAuthenticationMechanism)null)).Uri
          : (new PopUriBuilder(textBoxHost.Text, textBoxUserName.Text, (PopAuthenticationMechanism)null)).Uri;
#pragma warning restore 168

        e.Cancel = false;
      }
      catch (Exception) {
        MessageBox.Show(this, string.Format("invalid host or username: {0}@{1}", textBoxUserName.Text, textBoxHost.Text));
        e.Cancel = true;
      }
    }

    private void ConnectToDefaultPortChanged(object sender, EventArgs e)
    {
      if (checkBoxConnectToDefaultPort.Checked) {
        numericUpDownPort.Enabled = false;
        SetDefaultPort();
      }
      else {
        numericUpDownPort.Enabled = true;
      }
    }

    private void AuthenticateAnonymousChanged(object sender, EventArgs e)
    {
      if (checkBoxAuthenticateAnonymous.Checked) {
        panelAuthMechanism.Enabled = false;
        textBoxPassword.Enabled = false;
        labelUserName.Text = "ログイントークン";
        currentProfile.AuthType = "ANONYMOUS";
      }
      else {
        panelAuthMechanism.Enabled = true;
        textBoxPassword.Enabled = true;
        labelUserName.Text = "ユーザ名";
        AuthMechanismChanged(null, EventArgs.Empty);
      }
    }

    private void SslUsageChanged(object sender, EventArgs e)
    {
      if (radioButtonSslUsageAlways.Checked)
        currentProfile.SslUsage = SslUsage.Always;
      else if (radioButtonSslUsageIfAvailable.Checked)
        currentProfile.SslUsage = SslUsage.IfAvailable;
      else if (radioButtonSslUsageNever.Checked)
        currentProfile.SslUsage = SslUsage.Never;
    }

    private void AuthTypeChanged(object sender, EventArgs e)
    {
      checkBoxForceSecureLogin.Enabled = string.IsNullOrEmpty(currentProfile.AuthType) || currentProfile.AuthType == "*";

      currentProfile.AuthType = (string)comboBoxAuthType.SelectedItem;

      UpdatePopLoginEnability();
    }

    private void AuthMechanismChanged(object sender, EventArgs e)
    {
      comboBoxAuthType.Enabled = radioButtonAuthMechanismUseSupplied.Checked;
      checkBoxForceSecureLogin.Enabled = !radioButtonAuthMechanismUseSupplied.Checked;

      if (radioButtonAuthMechanismUseSupplied.Checked)
        currentProfile.AuthType = (string)comboBoxAuthType.SelectedItem;
      else
        currentProfile.AuthType = null;

      UpdatePopLoginEnability();
    }

    private void SetProfile(Profile profile)
    {
      if (currentProfile != null)
        currentProfile.AuthorityChanged -= ProfileAuthorityChanged;

      currentProfile = profile;

      currentProfile.AuthorityChanged += ProfileAuthorityChanged;

      var authType = profile.AuthType;
      var port = profile.Port;
      var actualPort = profile.GetPort();

      if (profile.IsImap)
        radioButtonProtocolImap.Checked = true;
      else
        radioButtonProtocolPop.Checked = true;

      textBoxHost.Text = profile.Host;

      if (port == -1) {
        checkBoxConnectToDefaultPort.Checked = true;
        numericUpDownPort.Enabled = false;
      }
      else {
        checkBoxConnectToDefaultPort.Checked = false;
        numericUpDownPort.Enabled = true;
      }

      numericUpDownPort.Value = actualPort;

      textBoxUserName.Text = profile.UserName;

      switch (profile.SslUsage) {
        case SslUsage.Always:
          radioButtonSslUsageAlways.Checked = true;
          break;
        case SslUsage.IfAvailable:
          radioButtonSslUsageIfAvailable.Checked = true;
          break;
        case SslUsage.Never:
          radioButtonSslUsageNever.Checked = true;
          break;
      }

      if (string.IsNullOrEmpty(authType) || authType == "*") {
        checkBoxAuthenticateAnonymous.Checked = false;
        radioButtonAuthMechanismSelectAppropriate.Checked = true;
      }
      else if (string.Equals(authType, "ANONYMOUS", StringComparison.OrdinalIgnoreCase)) {
        checkBoxAuthenticateAnonymous.Checked = true;
      }
      else {
        checkBoxAuthenticateAnonymous.Checked = false;
        radioButtonAuthMechanismUseSupplied.Checked = true;
        comboBoxAuthType.SelectedIndex = comboBoxAuthType.Items.IndexOf(authType);
      }

      checkBoxForceSecureLogin.Checked = !profile.AllowInsecureLogin;
      checkBoxAllowPopStandardLogin.Checked = profile.AllowPopStandardLogin;
      checkBoxAllowApopLogin.Checked = profile.AllowApopLogin;

      textBoxPassword.Text = string.Empty;

      ProfileAuthorityChanged(currentProfile, EventArgs.Empty);
    }

    private void ProfileAuthorityChanged(object sender, EventArgs e)
    {
      var p = sender as Profile;

      textBoxAuthority.Text = p.Authority.AbsoluteUri;
    }

    private void SetDefaultPort()
    {
      currentProfile.Port = -1;

      numericUpDownPort.Value = currentProfile.GetPort();
    }

    private void SaveCurrentProfile()
    {
      string profileName = null;

      using (var dialog = new InputDialog()) {
        dialog.Description = "保存する設定に付ける名前を入力してください。";

        if (dialog.ShowDialog(this) != DialogResult.OK)
          return;

        profileName = dialog.InputText;
      }

      if (string.IsNullOrEmpty(profileName)) {
        MessageBox.Show(this,
                        "1文字以上の名前を付けてください。",
                        "不正な名前",
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Error);
        return;
      }

      var overwrite = false;

      if (profiles.ContainsKey(profileName)) {
        var result = MessageBox.Show(this,
                                     string.Format("'{0}'はすでに存在します。　上書きしますか?", profileName),
                                     "設定の上書き",
                                     MessageBoxButtons.YesNo,
                                     MessageBoxIcon.Question,
                                     MessageBoxDefaultButton.Button2);

        if (result != DialogResult.Yes)
          return;
      
        overwrite = true;
      }

      if (!overwrite)
        comboBoxProfiles.Items.Add(profileName);

      profiles[profileName] = currentProfile.Clone();

      MessageBox.Show(this,
                      string.Format("'{0}'を保存しました。", profileName),
                      "設定の保存",
                      MessageBoxButtons.OK,
                      MessageBoxIcon.Information);
    }

    private void RemoveProfile()
    {
      var profileName = (string)comboBoxProfiles.SelectedItem;

      if (string.IsNullOrEmpty(profileName))
        return;
      if (!profiles.ContainsKey(profileName))
        return;

      var result = MessageBox.Show(this,
                                   string.Format("'{0}'の設定を削除します。", profileName),
                                   "設定の削除",
                                   MessageBoxButtons.OKCancel,
                                   MessageBoxIcon.Question,
                                   MessageBoxDefaultButton.Button2);

      if (result != DialogResult.OK)
        return;

      comboBoxProfiles.Items.Remove(profileName);
      profiles.Remove(profileName);
    }

    private void Connect(Profile profile)
    {
      if (connectBackground.IsBusy) {
        var result = MessageBox.Show(this,
                                     "接続を中止しますか?",
                                     "接続の中止",
                                     MessageBoxButtons.YesNo,
                                     MessageBoxIcon.Question,
                                     MessageBoxDefaultButton.Button1);

        if (result == DialogResult.Yes)
          connectBackground.Cancel();
      }
      else {
        protocolLogListener.Clear();

        if (profile.IsImap)
          connectBackground.RunWorkerAsync(Tuple.Create(profile.GetImapProfile(),
                                                        new NetworkCredential(profile.UserName, textBoxPassword.Text),
                                                        defaultUsingSaslMechanisms));
        else
          connectBackground.RunWorkerAsync(Tuple.Create(profile.GetPopProfile(),
                                                        new NetworkCredential(profile.UserName, textBoxPassword.Text),
                                                        defaultUsingSaslMechanisms));

        labelStatus.BackColor = Color.LightGray;
        labelStatus.ForeColor = Color.Black;
        labelStatus.Text = "接続中...";

        buttonConnect.Text = "中止";
      }
    }

    private void ConnectCompleted(object sender, RunWorkerCompletedEventArgs e)
    {
      buttonConnect.Text = "接続";

      if (e.Cancelled || e.Error is OperationCanceledException) {
        labelStatus.BackColor = Color.Yellow;
        labelStatus.ForeColor = Color.Black;
        labelStatus.Text = "接続をキャンセルしました。";
      }
      else {
        if (e.Error == null) {
          labelStatus.BackColor = Color.Lime;
          labelStatus.ForeColor = Color.Black;
          labelStatus.Text = "ログインに成功しました。";
        }
        else {
          string errorMessage = null;

          if (e.Error is SaslException) {
            errorMessage = "SASL認証エラー";
          }
          else if (e.Error is ImapException) {
            if (e.Error is ImapSecureConnectionException)
              errorMessage = "SSL/TLSエラー";
            else if (e.Error is ImapConnectionException)
              errorMessage = "接続エラー";
            else if (e.Error is ImapAuthenticationException)
              errorMessage = "認証エラー";
            else if (e.Error is ImapNoAppropriateAuthMechanismException)
              errorMessage = "認証エラー (試行できる認証機構なし)";
            else if (e.Error is ImapProtocolViolationException)
              errorMessage = "プロトコルエラー";
            else if (e.Error is ImapIncapableException)
              errorMessage = "サポートされていない機能";
            else if (e.Error is ImapErrorResponseException)
              errorMessage = "エラー応答";
          }
          else if (e.Error is PopException) {
            if (e.Error is PopUpgradeConnectionException)
              errorMessage = "SSL/TLSエラー";
            else if (e.Error is PopConnectionException)
              errorMessage = "接続エラー";
            else if (e.Error is PopAuthenticationException)
              errorMessage = "認証エラー";
            else if (e.Error is PopNoAppropriateAuthMechanismException)
              errorMessage = "認証エラー (試行できる認証機構なし)";
            else if (e.Error is PopProtocolViolationException)
              errorMessage = "プロトコルエラー";
            else if (e.Error is PopIncapableException)
              errorMessage = "サポートされていない機能";
            else if (e.Error is PopErrorResponseException)
              errorMessage = "エラー応答";
          }

          if (errorMessage == null)
            errorMessage = "予期せぬエラー";

          textBoxProtocolLog.Text += e.Error.ToString();

          labelStatus.BackColor = Color.OrangeRed;
          labelStatus.ForeColor = Color.Black;
          labelStatus.Text = errorMessage;
        }
      }
    }

    private Profile currentProfile;
    private Profile defaultProfile;
    private Dictionary<string, Profile> profiles;
    private TextBoxTraceListener protocolLogListener;
    private ConnectBackgroundWorker connectBackground;
  }
}
