using System;


using Smdn.Net.Imap4;
using Smdn.Net.Imap4.Client;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Pop3;
using Smdn.Net.Pop3.Client;
using Smdn.Net.Pop3.Protocol;

namespace Smdn.Applications.MRAConnector {
  public enum ProtocolType {
    Imap,
    Pop,
  }

  public enum SslUsage {
    Always,
    IfAvailable,
    Never,
  }

  public class Profile : ICloneable {
    public event EventHandler AuthorityChanged;

    public bool IsImap {
      get { return protocolType == ProtocolType.Imap; }
    }

    public Uri Authority {
      get
      {
        if (IsImap)
          return imapProfile.Authority;
        else
          return popProfile.Authority;
      }
    }

    public ProtocolType ProtocolType {
      get { return protocolType; }
      set
      {
        if (protocolType == value)
          return;

        protocolType = value;

        OnAuthorityChanged();
      }
    }

    public string Host {
      get { return host; }
      set
      {
        if (string.Equals(host, value))
          return;

        host = value;
        imapProfile.Host = host;
        popProfile.Host = host;

        OnAuthorityChanged();
      }
    }

    public int Port {
      get { return port; }
      set
      {
        if (port == value)
          return;

        port = value;
        imapProfile.Port = port;
        popProfile.Port = port;

        OnAuthorityChanged();
      }
    }

    public SslUsage SslUsage {
      get { return sslUsage; }
      set
      {
        if (sslUsage == value)
          return;

        sslUsage = value;

        switch (sslUsage) {
          case SslUsage.Always:
            imapProfile.SecurePort = true;
            popProfile.SecurePort = true;
            break;
          case SslUsage.IfAvailable:
            imapProfile.SecurePort = false;
            imapProfile.UseTlsIfAvailable = true;
            popProfile.SecurePort = false;
            popProfile.UseTlsIfAvailable = true;
            break;
          case SslUsage.Never:
            imapProfile.SecurePort = false;
            imapProfile.UseTlsIfAvailable = false;
            popProfile.SecurePort = false;
            popProfile.UseTlsIfAvailable = false;
            break;
        }

        OnAuthorityChanged();
      }
    }

    public string UserName {
      get { return userName; }
      set
      {
        if (string.Equals(userName, value))
          return;

        userName = value;
        imapProfile.UserName = userName;
        popProfile.UserName = userName;

        OnAuthorityChanged();
      }
    }

    public string AuthType {
      get { return authType; }
      set
      {
        if (string.Equals(authType, value))
          return;

        authType = value;
        imapProfile.AuthType = authType;
        popProfile.AuthType = authType;

        OnAuthorityChanged();
      }
    }

    public bool AllowInsecureLogin {
      get { return allowInsecureLogin; }
      set
      {
        allowInsecureLogin = value;
        imapProfile.AllowInsecureLogin = allowInsecureLogin;
        popProfile.AllowInsecureLogin = allowInsecureLogin;
      }
    }

    public bool AllowPopStandardLogin {
      get { return allowPopStandardLogin; }
      set
      {
        allowPopStandardLogin = value;
        popProfile.AllowStandardLogin = allowPopStandardLogin;
      }
    }

    public bool AllowApopLogin {
      get { return allowApopLogin; }
      set
      {
        allowApopLogin = value;
        popProfile.AllowApopLogin = allowApopLogin;
      }
    }

    public int Timeout {
      get { return timeout; }
      set
      {
        timeout = value;
        imapProfile.Timeout = timeout;
        popProfile.Timeout = timeout;
      }
    }

    public Profile()
    {
      Host = "localhost";
      Port = -1;
      UserName = "user";
      AuthType = null;
    }

    public Profile(Uri authority)
    {
      if (ImapUri.IsImap(authority)) {
        ProtocolType = ProtocolType.Imap;
        UserName = ImapStyleUriParser.GetUser(authority);

        var authType = ImapStyleUriParser.GetAuthType(authority);

        if (authType == null)
          AuthType = null;
        else
          AuthType = authType.Value;
      }
      else if (PopUri.IsPop(authority)) {
        ProtocolType = ProtocolType.Pop;
        UserName = PopStyleUriParser.GetUser(authority);

        var authType = PopStyleUriParser.GetAuthType(authority);

        if (authType == null)
          AuthType = null;
        else
          AuthType = authType.Value;
      }
      else {
        throw new ArgumentException("uri is not IMAP URL or POP URL", "authority");
      }

      Host = authority.Host;
      Port = authority.Port;
    }

    public int GetPort()
    {
      if (port == -1) {
        if (SslUsage == SslUsage.Always) {
          if (IsImap)
            return ImapDefaultPorts.Imaps;
          else
            return PopDefaultPorts.Pops;
        }
        else {
          if (IsImap)
            return ImapDefaultPorts.Imap;
          else
            return PopDefaultPorts.Pop;
        }
      }
      else {
        return port;
      }
    }

    public ImapClientProfile GetImapProfile()
    {
      return imapProfile.Clone();
    }

    public PopClientProfile GetPopProfile()
    {
      return popProfile.Clone();
    }

    public Profile Clone()
    {
      var cloned = (Profile)MemberwiseClone();

      cloned.AuthorityChanged = null;
      cloned.imapProfile = this.imapProfile.Clone();
      cloned.popProfile = this.popProfile.Clone();

      return cloned;
    }

    object ICloneable.Clone()
    {
      return Clone();
    }

    private void OnAuthorityChanged()
    {
      var ev = this.AuthorityChanged;

      if (ev != null)
        ev(this, EventArgs.Empty);
    }

    private ProtocolType protocolType = ProtocolType.Imap;
    private string host;
    private int port;
    private SslUsage sslUsage = SslUsage.IfAvailable;
    private string userName;
    private string authType;
    private bool allowInsecureLogin = false;
    private bool allowPopStandardLogin = true;
    private bool allowApopLogin = true;
    private int timeout = -1;

    private ImapClientProfile imapProfile = new ImapClientProfile();
    private PopClientProfile popProfile = new PopClientProfile();
  }
}
